// **********************************************************************
//  PicImageFormat class definition
//  Part of : PicAlg
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef PICIMAGEFORMAT_HXX
#define PICIMAGEFORMAT_HXX

#include <bstrarr.hxx>
#include <pstream.hxx>
#include <pobject.hxx>

#include <PicAlgDefs.hxx>

class PicRawPicture;

/**
  * PicImageFormat represents an image file format (for example JPEG, GIF).
  * This base class has methods for reading images from files.
  * The actual implementations need to be able to create a PicRawPicture
  * from stream of data and write a PicRawPicture into stream. Each 
  * image format has a name and an unique id (define in file PicAlgDefs.hxx). 
  * Image formats should also have at least one filename extension 
  * defined (e.g. "jpg" for *.jpg or "gif" for *.gif). This is used when
  * guessing the format of a file opened for loading. It is also used to
  * determine if this is the desired format when saving to a user-defined
  * filename.
  **/

class PicImageFormat : public PicObject
{
 public:
  PICABSOBJECTCLASS(PicImageFormat)

  /**
    * The default constructor.
    **/
  PicImageFormat();
  /**
    * The destructor.
    **/
  virtual ~PicImageFormat();

  /**
    * Opens a file stream for reading and when ok, calls
    * readPicture(PicStream &from, PICULONG size).
    *
    * @param filename File name of the picture to be loaded.
    * @return Pointer to a PicRawPicture object generated from the file.
    **/
  virtual PicRawPicture *readPicture(BaseString filename);
  /**
    * Opens a file stream for writing and when ok, calls
    * writePicture(PicStream &to, PicRawPicture * ).
    * 
    * @param filename File name of the picture to be saved.
    * @return True if successful.
    **/
  virtual PICBOOL writePicture(BaseString filename, PicRawPicture *);
  /**
    * This function should create (allocated with new) a PicRawPicture out 
    * of a stream of bytes. The length of the stream is equal to argument 
    * variable "size". If the data is not valid according to the image 
    * format, return NULLPTR.
    * 
    * @param from Reference to a PicStream object where to read the data from.
    * @param size Number of bytes to read.
    * @return Pointer to a PicRawPicture object generated from the stream.
    **/
  virtual PicRawPicture *readPicture(PicStream &from, PICULONG size) = 0;
  /**
    * This function transform a PicRawPicture into stream of bytes. 
    * Return boolTrue on success, otherwise (raise a descriptive error)
    * and return boolFalse.
    *
    * @param to Reference to a PicStream object where to write the data to.
    * @return True if successful.
    **/
  virtual PICBOOL writePicture(PicStream &to, PicRawPicture *) = 0;

  /// A descriptive name for the image format.
  BaseString      name;
  /// Unique ID (define in file PicAlgDefs.hxx).
  PICULONG        id;
  /**
    * Array of file extension strings, case is not significant, "jpg"
    * matches (*.jpg).
    **/
  BaseStringArray fileExtensions;
  /**
    * Should be boolTrue, if part of image data may be lost (for example
    * in compression) when using the file format; otherwise boolFalse.
    **/
  PICBOOL         isLossy;

  /// Assignment operator.
  virtual PicObject& operator = (const PicObject&);
  /// "Equal compare" operator.
  virtual PICBOOL operator == (const PicObject&) const;
  /// "Less than" operator.
  virtual PICBOOL operator <  (const PicObject&) const;
  /**
    * Tests if this object is empty.
    *
    * @param True is successful.
    **/
  virtual PICBOOL isNull() const;
  /**
    * Sets this object as empty.
    **/
  virtual void    setNull();
  /**
    * Tests is this object is default.
    *
    * @param True is successful.
    **/
  virtual PICBOOL isDefault() const;
  /**
    * Sets this object as default.
    **/
  virtual void    setDefault();
  /**
    * Exports this object to a BaseString presentation.
    *
    * @return BaseString object.
    **/
  virtual BaseString getImage() const;
  /**
    * Exports the specified object to a BaseString presentation.
    *
    * @param format Reference to an object to be converted.
    * @return BaseString object.
    **/
  virtual BaseString getImage(const PicObject& format) const;
  /**
    * Imports a BaseString presentation of this object.
    *
    * @param image Reference to a BaseString object holding the presentation.
    **/
  virtual void setImage(const BaseString& image);
  /**
    * Imports a BaseString presentation of an object.
    *
    * @param format Reference to an object to import the new presentation.
    * @param image Reference to a BaseString object holding the presentation.
    **/
  virtual void setImage(const PicObject& format, const BaseString& image);

};

/**
  * PicPPM implements PPM image file format "*.ppm".
  * PPM format includes a simple header plus RGB-values
  * of each pixel.
  **/

class PicPPM : public PicImageFormat
{
 public:
  PICOBJECTCLASS(PicPPM)

  /**
    * The default constructor.
    **/
  PicPPM();
  /**
    * Opens a file stream for reading and when ok, calls
    * readPicture(PicStream &from, PICULONG size).
    *
    * @param filename File name of the picture to be loaded.
    * @return Pointer to a PicRawPicture object generated from the file.
    **/
  virtual PicRawPicture *readPicture(PicStream &from, PICULONG size);
  /**
    * Opens a file stream for writing and when ok, calls
    * writePicture(PicStream &to, PicRawPicture * ).
    * 
    * @param filename File name of the picture to be saved.
    * @return True if successful.
    **/
  virtual PICBOOL writePicture(PicStream &to, PicRawPicture *);
};

/**
  * PicJpeg implements Jpeg image file format "*.jpg", "*.jpeg".
  * Jpeg has a very good compression rate but some image data is lost.
  **/

class PicJpeg : public PicImageFormat
{
 public:
  PICOBJECTCLASS(PicJpeg)

  /**
    * The default constructor.
    **/
  PicJpeg();
  /**
    * Opens a file stream for reading and when ok, calls
    * readPicture(PicStream &from, PICULONG size).
    *
    * @param filename File name of the picture to be loaded.
    * @return Pointer to a PicRawPicture object generated from the file.
    **/
  virtual PicRawPicture *readPicture(PicStream &from, PICULONG size);
  /**
    * Opens a file stream for writing and when ok, calls
    * writePicture(PicStream &to, PicRawPicture * ).
    * 
    * @param filename File name of the picture to be saved.
    * @return True if successful.
    **/
  virtual PICBOOL writePicture(PicStream &to, PicRawPicture *);
};

/**
  * PicSunRaster implements SunRaster image file format "*.ras".
  * Not all subtypes are included.
  **/

class PicSunRaster : public PicImageFormat
{
 public:
  PICOBJECTCLASS(PicSunRaster)

  /**
    * The default constructor.
    **/
  PicSunRaster();
  /**
    * Opens a file stream for reading and when ok, calls
    * readPicture(PicStream &from, PICULONG size).
    *
    * @param filename File name of the picture to be loaded.
    * @return Pointer to a PicRawPicture object generated from the file.
    **/
  virtual PicRawPicture *readPicture(PicStream &from, PICULONG size);
  /**
    * Opens a file stream for writing and when ok, calls
    * writePicture(PicStream &to, PicRawPicture * ).
    * 
    * @param filename File name of the picture to be saved.
    * @return True if successful.
    **/
  virtual PICBOOL writePicture(PicStream &to, PicRawPicture *);
};

// **********************************************************************

inline PICBOOL PicImageFormat::isDefault() const 
  { return isNull(); }
inline void PicImageFormat::setDefault()
  { setNull(); }
inline BaseString PicImageFormat::getImage(const PicObject& format) const
  { if (&format) return getImage(); else return getImage(); }
inline void PicImageFormat::setImage(const PicObject& format, 
                                    const BaseString& image)
  { if (&format) setImage(image); else setImage(image); }


#endif


