// **********************************************************************
//  PicSearchAlgorithm class definition
//  Part of : PicAlg
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef PICSEARCHALGORITHM_HXX
#define PICSEARCHALGORITHM_HXX

#include <PicAlgDefs.hxx>
#include <PicSignature.hxx>
#include <PicPicture.hxx>

/**
  * PicSearchAlgorithm is the base class for all search algorithms.
  * Search algorithms are used to create and compare concise 
  * representations describing some (ideally all interesting!) aspects 
  * of an image. Search algorithms are the main method for image
  * queries based on pictore contents.
  **/

class PicSearchAlgorithm
{
 public:

  /**
    * The default constructor.
    **/
  PicSearchAlgorithm(const char *libPath, int id);
  ~PicSearchAlgorithm();

  /**
    * This function is meant for accessing the (possible) parameters of this
    * algorithm. Note that you must separately update the GUI to create a
    * suitable window for the parameters.
    *
    * NOTE: The paramaters of a search algorithm should have no 
    *       influence on the signatures that are computed for images
    *       (since those are already in the database and would have
    *       to be recomputed when changing parameters!). 
    *       The paramaters are only meant for giving you at run-time 
    *       a way of controlling how to compare signatures.
    *
    * @param params Reference to a collection object holding the parameters.
    * @return True if successful.
    **/
	PICBOOL (*setParams)(PicIxCollect &params);

  /**
    * The implementation of this function should determine the 
    * 'closeness' of the two argument signatures. Return 1000 for perfect
    * match, 0 for worst match. The first argument signature is the one 
    * computed for the query image (does not matter if your algorithm is 
    * symmetric).
    *
    * @param firstSig Reference to the first signature object.
    * @param secSig Reference to the second signature object.
    * @return Match rate (0-1000).
    **/
	PICUSHORT (*compare)(PicSignature &firstSig, PicSignature &secSig);

  /**
    * The implementation of this function should compute a signature for 
    * the argument image. The return value (allocated by operator new)
    * should be an instance of the PicSignature-based signature class
    * created for your search algorithm. On error, return a 'null' 
    * signature instead of NULLPTR if possible.
    *
    * @param pic Reference to a picture object from which to compute the
    *            signature.
    * @return Pointer to the generated signature object.
    **/
	PicSignature *(*computeSignature)(PicPicture &Pic);

	char *(*algName)(void);
	char *(*algInfo)(void);

	int algID(void)
	{
		return(alg_id);
	}

 private:
  void *libhandle;
  int alg_id;
  
  // No longer true: fix 
  /** ID should be a unique (within search algorithms) ID identifying
    * this algorithm (define in PicAlgDefs.hxx).
    **/

};

// **********************************************************************

#endif


