// **********************************************************************
//  This file contains PicSearchType implementation
//  Part of : PicAlg
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#include <string.h>

#include <picerror.hxx>
#include <PicImageFormat.hxx>
#include <PicPicture.hxx>
#include <mmain.hxx>

MAKE_PICOBJECTCLASS(PicPicture, PicObject)

ERROR_MESSAGES(PicPicture)
PARAMSTR(rawNull, "Couldn't read image data")
PARAMSTR(noPixel, "No pixel data")
END_PARAM_STRINGS

PicPicture::PicPicture()
  : matchRate(0), matchRanking(0), id(0), DBid(0),
	identicalToFilename(boolFalse), format(NULLPTR), raw(NULLPTR)
{

}

PicPicture::~PicPicture()
{
  setNull();
}

PicPicture::PicPicture(const PicPicture &other)
  : matchRate(0), matchRanking(0), id(0), DBid(0), 
	identicalToFilename(boolFalse), format(NULLPTR), raw(NULLPTR)
{
  *this = (const PicObject &)other;
}

PicObject& PicPicture::operator = (const PicObject& obj)
{
  if(&obj == this)
    return *this;

  if(obj.IS_A(PicPicture)){
    const PicPicture &other = AS(PicPicture, obj);

    setNull();
    id = other.id;
    DBid = other.DBid;
	name = other.name;
    par1 = other.par1;
    par2 = other.par2;

    filename = other.filename;
    format = other.format;

    signatures = other.signatures;
    segmentations = other.segmentations;

    matchRate = other.matchRate;
    matchRanking = other.matchRanking;

    identicalToFilename = other.identicalToFilename;
    origData = (PicObject &)(other.origData);

    if(other.raw){
      raw = (PicRawPicture *)other.raw->copy();
    }
  }
  return *this;
}

PICBOOL PicPicture::operator == (const PicObject& other) const
{
  return (this == &other);
}

PICBOOL PicPicture::operator <  (const PicObject&) const
{
  undefined("PicPicture::operator <");
  return boolFalse;
}

PICBOOL PicPicture::isNull() const
{
  return(raw==NULLPTR);
}

void PicPicture::setNull()
{
  filename.setNull();
  name.setNull();
  par1.setNull();
  par2.setNull();
  matchRate=0;
  matchRanking=0;
  id=0;
  DBid=0;
  segmentations.setNull();
  identicalToFilename = boolFalse;
  origData.setNull();
  setRaw(NULLPTR);
  format = NULLPTR;
}

BaseString PicPicture::getImage() const
{
  return name;
}

void PicPicture::setImage(const BaseString& image)
{
  undefined("PicPicture::setImage");
}

void PicPicture::setRaw(PicRawPicture *newRaw)
{
  if(raw)
    delete raw;
  raw = newRaw;
}

PICBOOL PicPicture::loadFromFile(BaseString fileName)
{
  DEBUG_ALG("loadFromFile: " & fileName & "\n");

  PicFile       file(fileName);
  BaseString    ext = file.suffix();
  PicIterator   iter;
  PicCollection &coll = PicMain::globData.imageFormats;
  PICUSHORT     rounds;
  PICBOOL       useExtension;
  PicRawPicture *newRaw = NULLPTR;

  // First try to use file extension
  ext.makeLowercase();
  for(rounds=0; rounds<=1; rounds++){
    useExtension = (rounds==0) ? boolTrue : boolFalse;
    ITERATE(coll, iter)
    {
      PicImageFormat &curr = AS(PicImageFormat, coll.at(iter));
      if(curr.fileExtensions.contains(ext) == useExtension){
        newRaw = curr.readPicture(fileName);
        if(newRaw != NULLPTR) {
          format = &curr;
          setRaw(newRaw);
          identicalToFilename = boolTrue;
          origData.setNull();
          if(format->isLossy){
            if(!origData.readFromFile(fileName)){
              PicError::clearCurrent();   // never mind
            }
          }
          return boolTrue;
        }
      }
    }
  }
  EHOLD(PicPicture, rawNull);
  return boolFalse;
}

PICBOOL PicPicture::writeToFile(BaseString fileName)
{
  DEBUG_ALG("writeToFile: " & fileName & "\n");

  PicFile        file(fileName);
  PicIterator    iter;
  PicCollection  &coll = PicMain::globData.imageFormats;
  PicImageFormat *formatUsed = NULLPTR;
  PICBOOL        ret;

  if (raw==NULLPTR){
    EHOLD(PicPicture, noPixel);
    return boolFalse;
  }

  ITERATE(coll, iter)
  {
    PicImageFormat &curr = AS(PicImageFormat, coll.at(iter));
    PICUSHORT pos = curr.fileExtensions.next(0);
    while(pos)
    {
      BaseString extension = curr.fileExtensions[pos];
      if(extension.equal(file.suffix())){
        formatUsed = &curr;
        break;
      }
      pos = curr.fileExtensions.next(pos);
    }
    if (formatUsed!=NULLPTR)
      break;
  }

  if (formatUsed==NULLPTR){
    if(format != NULLPTR)
      formatUsed = format;  // Use the old one 
    else
      formatUsed = PicMain::globData.getImageFormat(PIC_PPM_IMAGEFORMAT_ID);
  }
  if ((ret=formatUsed->writePicture(fileName, raw))){
    format=formatUsed;
    identicalToFilename = boolTrue;
    origData.setNull();
    if(format->isLossy){
      if(!origData.readFromFile(fileName)){
        PicError::clearCurrent();   // never mind
      }
    }
  }
  return ret;
}

PICBOOL PicPicture::getBlock(Tk_PhotoImageBlock *block)
{
  if(!block)
    return boolFalse;

  DEBUG_ALG("GetBlock: width: " & asString(block->width) &\
            " height: " & asString(block->height) & "\n");

  setRaw(new PicRGBPicture(block));
  return boolTrue;
}

