// **********************************************************************
//  This file contains PicSunRaster implementation
//  Part of : PicAlg
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#include <mmain.hxx>
#include <PicImageFormat.hxx>
#include <pfilestm.hxx>
#include <string.h>

MAKE_PICOBJECTCLASS(PicSunRaster, PicImageFormat)

PicSunRaster::PicSunRaster()
{
  name = "SunRaster";
  id = PIC_SUNRASTER_ID;
  fileExtensions.add("ras");
  isLossy=boolFalse;
}

PICBOOL changeByteOrder(PICINT *endianInt)
{
  char *p=(char *)endianInt;
  PICINT temp;
  char *tempP=(char *)&temp;

  *(tempP+3)=*p;
  *(tempP+2)=*(p+1);
  *(tempP+1)=*(p+2);
  *tempP=*(p+3);
  memcpy(p, tempP, 4);
  return boolTrue;
}

PicRawPicture *PicSunRaster::readPicture(PicStream &from, PICULONG size)
{
  PicRawPicture *ans = NULLPTR;
  unsigned char *dataBuf=NULL;
  unsigned char *rawBuf=NULL;
  unsigned char *colorMap=NULL;
  PICULONG magicNumber,i;
  PICINT width, height, depth, length, type, colorMapType, colorMapLength;
  PICSIZE bytesToRead;
  PICSIZE dword=4;
  PICBOOL littleFlag=boolFalse;

  PICINT endianTest=0x12345678;
  char *byteInInt=(char *)&endianTest;
  if (*byteInInt==0x78)
    littleFlag=boolTrue;

  DEBUG_ALG("Reading SunRaster...\n");

  from.read(&magicNumber, dword);

  if (magicNumber!=0x59a66a95) {
    if (magicNumber==0x956aa659) {
      DEBUG_ALG("Integers in little endian byte order.\n");
    }
    else return NULLPTR; //It wasn't a SunRaster image.
  }

  from.read(&width, dword);
  from.read(&height, dword);
  from.read(&depth, dword);
  from.read(&length, dword);
  from.read(&type, dword);
  from.read(&colorMapType, dword);
  from.read(&colorMapLength, dword);

  if (littleFlag) {
    changeByteOrder(&width);
    changeByteOrder(&height);
    changeByteOrder(&depth);
    changeByteOrder(&length);
    changeByteOrder(&type);
    changeByteOrder(&colorMapType);
    changeByteOrder(&colorMapLength);
  }

  switch(type) {
  case 0x0:
  case 0x1:
  case 0x3:
    break;
  case 0x2:
    DEBUG_ALG("Byte encoding not implemented yet.\n");
    return NULLPTR;
  case 0x4:
    DEBUG_ALG("TIFF format reading not implemented.\n");
    return NULLPTR;
  case 0x5:
    DEBUG_ALG("IFF format reading not implemented.\n");
    return NULLPTR;
  case 0xffff:
    DEBUG_ALG("SunRaster file is of type 'Experimental'.\n");
    return NULLPTR;
  default:
    DEBUG_ALG("Unknown SunRaster bitmap type " & asString(type) & "\n");
    return NULLPTR;
  }

  if (colorMapType>0x1 || (colorMapType && depth!=8)) {
    DEBUG_ALG("ColorMaps (other than depth-8 type-1) not implemented yet.\n");
    return NULLPTR;
  }

  if (colorMapType) {
    colorMap=(unsigned char *)picAllocate(colorMapLength);
    from.read(colorMap, colorMapLength);
  }

  dataBuf=(unsigned char *)picAllocate(width*height*3+1);

  if (length==0)
    bytesToRead=width*height*depth/8;
  else
    bytesToRead=length;

  rawBuf=(unsigned char *)picAllocate(bytesToRead+1);
  from.read(rawBuf, bytesToRead);

  unsigned char *byteFromRaw=rawBuf;
  unsigned char *byteToData=dataBuf;

  if (depth==1) { //Black and white
    PICINT b;
    unsigned char c=0;
    PICSHORT j;
    for (i=0, b=-1; (PICULONG)i<bytesToRead; i++) {
      if (b < 0) {
	b=7;
	c=~*byteFromRaw++;
      }
      while(b>=0) {
	for (j=0; j<3; j++)
	  *byteToData++ = ((c >> b) & 1)*255;
	b--;
      }
    }
  }

  if (depth==8) { //Usually grayscale (may be color if map present)
    PICSHORT j;
    if (colorMapType==0x1) {
      if(!colorMap)
      {
        picFree(rawBuf);
	picFree(dataBuf);
      	return(NULLPTR);
      }
      PICINT RGBBoundary=colorMapLength/3;
      for (i=0; i<bytesToRead; i++) {
	for (j=0; j<3; j++) {
	  *byteToData++ = *(colorMap+j*RGBBoundary+(int)(*byteFromRaw));
	}
	byteFromRaw++;
      }
    } else {
      for (i=0; i<bytesToRead; i++) {
	for (j=0; j<3; j++)
	  *byteToData++ = *byteFromRaw;
	byteFromRaw++;
      }
    }
  }

  if (depth==24) {
    if (type==0x3) //RGB
      memcpy(dataBuf, rawBuf, length);
    else { //BGR
      for (i=0; i<bytesToRead/3; i++) {
	*(byteToData+3*i+2)=*(byteFromRaw+3*i);   //Blue
	*(byteToData+3*i+1)=*(byteFromRaw+3*i+1); //Green
	*(byteToData+3*i)  =*(byteFromRaw+3*i+2); //Red
      }
    }
  }

  if (depth==32) { //Pad byte precedes the pixel values.
    if (type==0x3) { //RGB
      for (i=0; i<bytesToRead; i++) {
	if (i%4 != 0) {
	  *byteToData++=*(byteFromRaw+i);
	}
      }
    }
    else { //BGR
      for (i=0; i<bytesToRead/4; i++) {
	*(byteToData+3*i+2)=*(byteFromRaw+4*i+1); //Blue
	*(byteToData+3*i+1)=*(byteFromRaw+4*i+2); //Green
	*(byteToData+3*i)  =*(byteFromRaw+4*i+3); //Red
      }
    }
  }

  ans=new PicRGBPicture(width, height, dataBuf);
  DEBUG_ALG("Read SunRaster ok.\n");

  picFree(dataBuf);
  picFree(rawBuf);

  return ans;
}

// **********************************************************************

PICBOOL PicSunRaster::writePicture(PicStream &out, PicRawPicture *raw)
{
  if (!raw || raw->isNull())
    return boolFalse;

  PICINT dword=4;
  PICINT size;
  PICBOOL littleFlag=boolFalse;

  PICINT endianTest=0x12345678;
  char *byteInInt=(char *)&endianTest;
  if (*byteInInt==0x78)
    littleFlag=boolTrue;

  PICINT magicNumber=0x59a66a95;
  PICINT width=raw->getWidth();
  PICINT height=raw->getHeight();
  PICINT depth=24;
  PICINT length=size=width*height*3;
  PICINT type=3; //RGB
  PICINT colorMapType=0;
  PICINT colorMapLength=0;

  if (littleFlag) {
    changeByteOrder(&magicNumber);
    changeByteOrder(&width);
    changeByteOrder(&height);
    changeByteOrder(&depth);
    changeByteOrder(&length);
    changeByteOrder(&type);
  }

  out.write(&magicNumber, dword);
  out.write(&width, dword);
  out.write(&height, dword);
  out.write(&depth, dword);
  out.write(&length, dword);
  out.write(&type, dword);
  out.write(&colorMapType, dword);
  out.write(&colorMapLength, dword);

  out.write(raw->getDataPtr(), size);

  return boolTrue;
}

