// **********************************************************************
//
//      PICServerDataBase.h
//
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//
// General Public License for more details.
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// ********************************************************************** 

#ifndef PICSERVERDATABASE_H_
#define PICSERVERDATABASE_H_  

#include <parray.hxx>
#include <pixmap.hxx>
#include <pstring.hxx>

/** PICServerDataBase.
  * Structure where client stores informations of the  server's databases
  * and algorithms.
  **/

// ********************************************************************** 

// The server's database information.

class PICServerData : public PicObject
{
 public:
  PICOBJECTCLASS(PICServerData)

  PICServerData();
  PICServerData(const PICServerData &other);
  virtual ~PICServerData();

//:var

  /// Address of the server.
  BaseString ipAddress;

  /// Portnumber of the servers socket.
  int portNbr;     // TCP-port

  /// Name of the server.
  BaseString name;

  /// URL of servers homepage.   
  BaseString httpAddress;

  /// Commercial text of the server.
  BaseString commercial;

  /// Users name.
  BaseString userName;

  /// Users authorization level.
  int        authLevel;

  /// Can user change password or not...
  bool       canChangePassword;

  /// Servers databases.
  PicArray databases;
 
  /// Servers algorithms
  PicArray algorithms;

  /// Servers user information array.
  PicArray users;

  /// Servers databases (ALL)
  PicArray allDatabases; 

  virtual PicObject& operator = (const PicObject&);
  virtual PICBOOL operator == (const PicObject&) const;
  virtual PICBOOL operator <  (const PicObject&) const;
  virtual PICBOOL isNull() const;
  virtual void    setNull();
  virtual PICBOOL isDefault() const;
  virtual void    setDefault();
  virtual BaseString getImage() const;
  virtual BaseString getImage(const PicObject& format) const;
  virtual void setImage(const BaseString& image);
  virtual void setImage(const PicObject& format, const BaseString& image);

};

// ********************************************************************** 

/** Servers database structure.
  **/
class PICServerDbData : public PicObject
{
 public:
  PICOBJECTCLASS(PICServerDbData)

  PICServerDbData();
  virtual ~PICServerDbData();
  PICServerDbData(const PICServerDbData &other);

//:var
  /// Name of the database.
  BaseString name;
  /// id of the database.
  long       id;
  /// Supported algorithms (ids only).
  PicArray   dbAlgIds;
  /// Users right level.
  int        userRightsLevel;

  virtual PicObject& operator = (const PicObject&);
  virtual PICBOOL operator == (const PicObject&) const;
  virtual PICBOOL operator <  (const PicObject&) const;
  virtual PICBOOL isNull() const;
  virtual void    setNull();
  virtual PICBOOL isDefault() const;
  virtual void    setDefault();
  virtual BaseString getImage() const;
  virtual BaseString getImage(const PicObject& format) const;
  virtual void setImage(const BaseString& image);
  virtual void setImage(const PicObject& format, const BaseString& image);

};

// ********************************************************************** 

/** Servers algorithm structure.
  **/
class PICDbAlgorithm : public PicObject
{
 public:
  PICOBJECTCLASS(PICDbAlgorithm)

  PICDbAlgorithm();
  virtual ~PICDbAlgorithm();
  PICDbAlgorithm(const PICDbAlgorithm &other);

//:var
  /// Name of the algorithm.
  BaseString name;
  /// Id of the algorithm.
  long       id;
  /// Description of the algorithm.
  BaseString description;
  /// Parameters window of the algorithm.
  BaseString tk;
  /// Parameters set for the algorithm
  PicArray params;

  virtual PicObject& operator = (const PicObject&);
  virtual PICBOOL operator == (const PicObject&) const;
  virtual PICBOOL operator <  (const PicObject&) const;
  virtual PICBOOL isNull() const;
  virtual void    setNull();
  virtual PICBOOL isDefault() const;
  virtual void    setDefault();
  virtual BaseString getImage() const;
  virtual BaseString getImage(const PicObject& format) const;
  virtual void setImage(const BaseString& image);
  virtual void setImage(const PicObject& format, const BaseString& image);

};

// ********************************************************************** 
// ======================================================================


/// User Information table

class PICUserInfo : public PicObject
{
 public:
  PICOBJECTCLASS(PICUserInfo)

  PICUserInfo();
  virtual ~PICUserInfo();
  PICUserInfo(const PICUserInfo &other);

//:var
  /// Username.
  PicString name;
  /// Server authorization level.
  long       authLevel;
  char       canChangePassword;

  ///  Databases where the user have an access.
  PicArray   dbRights;

  virtual PicObject& operator = (const PicObject&);
  virtual PICBOOL operator == (const PicObject&) const;
  virtual PICBOOL operator <  (const PicObject&) const;
  virtual PICBOOL isNull() const;
  virtual void    setNull();
  virtual PICBOOL isDefault() const;
  virtual void    setDefault();
  virtual BaseString getImage() const;
  virtual BaseString getImage(const PicObject& format) const;
  virtual void setImage(const BaseString& image);
  virtual void setImage(const PicObject& format, const BaseString& image);

};

// ======================================================================

/// Database authorization structure

class PICDbRight : public PicObject
{
 public:
  PICOBJECTCLASS(PICDbRight)

  PICDbRight();
  virtual ~PICDbRight();
  PICDbRight(const PICDbRight &other);

//:var
  /// Id of the database.
  long id;
  /// Database authorization level.
  long authLevel;

  virtual PicObject& operator = (const PicObject&);
  virtual PICBOOL operator == (const PicObject&) const;
  virtual PICBOOL operator <  (const PicObject&) const;
  virtual PICBOOL isNull() const;
  virtual void    setNull();
  virtual PICBOOL isDefault() const;
  virtual void    setDefault();
  virtual BaseString getImage() const;
  virtual BaseString getImage(const PicObject& format) const;
  virtual void setImage(const BaseString& image);
  virtual void setImage(const PicObject& format, const BaseString& image);

};


// ======================================================================
// **********************************************************************


/** PICServerData
  **/
inline PICBOOL PICServerData::isDefault() const 
  { return isNull(); }
inline void PICServerData::setDefault()
  { setNull(); }
inline BaseString PICServerData::getImage(const PicObject& format) const
  { return getImage(); }
inline void PICServerData::setImage(const PicObject& format, 
                                    const BaseString& image)
  { setImage(image); }

// **********************************************************************

/** PICServerDbdata
  **/
inline PICBOOL PICServerDbData::isDefault() const 
  { return isNull(); }
inline void PICServerDbData::setDefault()
  { setNull(); }
inline BaseString PICServerDbData::getImage(const PicObject& format) const
  { return getImage(); }
inline void PICServerDbData::setImage(const PicObject& format, 
                                      const BaseString& image)
  { setImage(image); }

// **********************************************************************

/** PICDbAlgorithm
  **/
inline PICBOOL PICDbAlgorithm::isDefault() const 
  { return isNull(); }
inline void PICDbAlgorithm::setDefault()
  { setNull(); }
inline BaseString PICDbAlgorithm::getImage(const PicObject& format) const
  { return getImage(); }
inline void PICDbAlgorithm::setImage(const PicObject& format, 
                                      const BaseString& image)
  { setImage(image); }

// **********************************************************************
// ======================================================================

/// PICUserInfo

inline PICBOOL PICUserInfo::isDefault() const 
  { return isNull(); }
inline void PICUserInfo ::setDefault()
  { setNull(); }
inline BaseString PICUserInfo ::getImage(const PicObject& format) const
  { return getImage(); }
inline void PICUserInfo ::setImage(const PicObject& format, 
                                      const BaseString& image)
  { setImage(image); }

// ======================================================================

/// PICDbRight

inline PICBOOL PICDbRight::isDefault() const 
  { return isNull(); }
inline void PICDbRight::setDefault()
  { setNull(); }
inline BaseString PICDbRight::getImage(const PicObject& format) const
  { return getImage(); }
inline void PICDbRight::setImage(const PicObject& format, 
                                      const BaseString& image)
  { setImage(image); }

// ======================================================================


#endif 

