// **********************************************************************
//  PicSearchType class definition
//  Part of : PicAlg
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef PICSEARCHTYPE_HXX
#define PICSEARCHTYPE_HXX

#include <pobject.hxx>
#include <bstrarr.hxx>

class PICDbAlgorithm;

/**
  * PicSearchType is a base class for different search types. Each 
  * search type includes names of the search algorithms belonging to
  * the search type. 
  **/

class PicSearchType : public PicObject
{
 public:
  PICOBJECTCLASS(PicSearchType)

  /**
    * The default constructor.
    **/
  PicSearchType();
  /**
    * Constructor setting the name for this search type.
    *
    * @param stypeName BaseString object holding the name string.
    **/
  PicSearchType(BaseString stypeName);

  /**
    * After creating search algorithms, you may register them to
    * (possibly many) search types using this function. Typically
    * the call is made from inside a search algorithm's constructor.
    *
    * @param alg Pointer to a search algorithm object.
    * @return True if successful.
    **/
  PICBOOL addAlgorithm(PICDbAlgorithm *alg);

  /// The name for this search type.
  BaseString name;
  ///Contains the names of the search algorithms belonging to this search type.
  BaseStringArray algs;

  /// Assignment operator.
  virtual PicObject& operator = (const PicObject&);
  /// "Equal compare" operator.
  virtual PICBOOL operator == (const PicObject&) const;
  /// "Less than" operator.
  virtual PICBOOL operator <  (const PicObject&) const;
  /**
    * Tests if this object is empty.
    *
    * @param True is successful.
    **/
  virtual PICBOOL isNull() const;
  /**
    * Sets this object as empty.
    **/
  virtual void    setNull();
  /**
    * Tests is this object is default.
    *
    * @param True is successful.
    **/
  virtual PICBOOL isDefault() const;
  /**
    * Sets this object as default.
    **/
  virtual void    setDefault();
  /**
    * Exports this object to a BaseString presentation.
    *
    * @return BaseString object.
    **/
  virtual BaseString getImage() const;
  /**
    * Exports the specified object to a BaseString presentation.
    *
    * @param format Reference to an object to be converted.
    * @return BaseString object.
    **/
  virtual BaseString getImage(const PicObject& format) const;
  /**
    * Imports a BaseString presentation of this object.
    *
    * @param image Reference to a BaseString object holding the presentation.
    **/
  virtual void setImage(const BaseString& image);
  /**
    * Imports a BaseString presentation of an object.
    *
    * @param format Reference to an object to import the new presentation.
    * @param image Reference to a BaseString object holding the presentation.
    **/
  virtual void setImage(const PicObject& format, const BaseString& image);

};

// **********************************************************************

inline PICBOOL PicSearchType::isDefault() const 
  { return isNull(); }
inline void PicSearchType::setDefault()
  { setNull(); }
inline BaseString PicSearchType::getImage(const PicObject& format) const
  { if (&format) return getImage(); else return getImage(); }
inline void PicSearchType::setImage(const PicObject& format, 
                                    const BaseString& image)
  { if (&format) setImage(image); else setImage(image); }

#endif
