// **********************************************************************
//  This file contains the initialization routines of PicMain
//  Part of : PicCore
// **********************************************************************
// Copyright (C) 1997-1999 PICSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#include <mmain.hxx>
#include <picerror.hxx>
#include <pinteger.hxx>
#include <pstring.hxx>
#include <penviron.hxx>
#include <bstrarr.hxx>
#include <PICClient.h>

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

MAKE_PICCLASS(PicMain)

ERROR_MESSAGES(PicMain)
PARAMSTR(initFailed, "Initialization failed")
PARAMSTR(unKnownImageFormat, "Unknown image format")
PARAMSTR(unKnownSearchType, "Unknown search type \"@*\"")
PARAMSTR(unKnownSearchAlg, "Unknown search algorithm \"@*\"")
PARAMSTR(unKnownSegmAlg, "Unknown segmentation algorithm \"@*\"")
PARAMSTR(unKnownIPAlg, "Unknown image processing algorithm \"@*\"")
PARAMSTR(unKnownExampleImage, "Unknown query image (index @*)")
PARAMSTR(threadSuspendFailure, "Thread did not suspend within @* seconds")
PARAMSTR(threadAbortFailure, "Thread did not abort within @* seconds")
PARAMSTR(invalidParams, "Invalid parameters")
PARAMSTR(imageInsertFailed, "Image insertion failed")
PARAMSTR(imageReadFailed, "Image reading failed")
PARAMSTR(deleteImageFailed, "Image deletion failed")
PARAMSTR(exImageCreateFailed, "Could not create new query image")
PARAMSTR(exImageCloseFailed, "Could not close query image")
PARAMSTR(exImageClearFailed, "Could not clear query image")
END_PARAM_STRINGS

static char *copyRightMesg = 
  "PICSearch version 3.1, Copyright (C) 1997-1999 PICSearch team." \
  "PICSearch comes with ABSOLUTELY NO WARRANTY; " \
  "This is free software, and you are welcome to redistribute it " \
  "under certain conditions; see General Public Licence for details.\n";


// **********************************************************************
//  This function creates an object identified by class id and
//  adds it into given collection. Also descendants are inserted
//  recursively. Of course, all abstract classes are skipped.

static void addToArray(const PicObjectClass *id, PicSeqIxCollect *coll)
{
  if(!id || !coll)
    return;

  PicObject *obj = id->createNew();
  if(obj){
    // It's a non-abstract class, so add to collection
    coll->add(*obj);
    delete obj;
  }
  // Add children also:
  const PicVirtClass *curr = id->firstDerived();
  while(curr){
    DEBUG_REQUIRE(curr->hasAncestor(PicObject::classId()));
    addToArray((PicObjectClass *)curr, coll);
    curr = curr->nextSibling(); // carefully, now
  }
}

// **********************************************************************
// Construct globData first so others may use it

PicMainSharedData   PicMain::globData;
PicGUI              PicMain::gui;
PicArray            PicMain::examplePics;
PicArray            PicMain::exampleListIndexes;
PicMassInsertThread PicMain::batchThread;

// Function definition for libc qsort(), this really should be coded in C++
int sortCompareFunc(long *, long *);

PicMain::PicMain() 
{
  PicErrorContext ecxt;
  PicIterator     iter;
  // First create some search types

  globData.searchTypes.add(PicSearchType("All"));
  globData.searchTypes.add(PicSearchType("Basic"));
  globData.searchTypes.add(PicSearchType("Sketch"));
  globData.searchTypes.add(PicSearchType("Drawing"));
  globData.searchTypes.add(PicSearchType("Texture"));
  globData.searchTypes.add(PicSearchType("Color"));
  globData.searchTypes.add(PicSearchType("Combination"));
  globData.currentSearchType = "All";
  PicSearchType *allStype = globData.getSearchType("All");

  // Then iterate all known image formats and algorithms
  // to create one of each into suitable collection

  const PicVirtClass *classId = &PicImageFormat::classId();
  DEBUG_REQUIRE(classId && classId->hasAncestor(PicObject::classId()));
  classId = classId->firstDerived();
  while(classId){
    DEBUG_REQUIRE(classId->hasAncestor(PicObject::classId()));
    addToArray((PicObjectClass *)classId, &globData.imageFormats);
    classId = classId->nextSibling();
  }
  globData.picPars->formatForScaledImages = globData.getImageFormat(PIC_JPEG_IMAGEFORMAT_ID);

  // PicSearchTypes are already constructed so the 
  // PicSearchAlgorithm constructors are able to find them:

  DEBUG_CORE("JL: Algpart if PicMain() constructor disabled\n");
/*
	classId = &PicSearchAlgorithm::classId();
  DEBUG_REQUIRE(classId && classId->hasAncestor(PicObject::classId()));
  classId = classId->firstDerived();
  while(classId){
    DEBUG_REQUIRE(classId->hasAncestor(PicObject::classId()));
    addToArray((PicObjectClass *)classId, &globData.searchAlgs);
    classId = classId->nextSibling();
  }
  if(allStype){
     ITERATE(globData.searchAlgs, iter)
     {
       PicSearchAlgorithm &curr = 
         AS(PicSearchAlgorithm, globData.searchAlgs.at(iter));
       if(!(allStype->algs.contains(curr.name)))
         allStype->addAlgorithm(&curr);
     }
  }
  ITERATE_BACKWARDS(globData.searchTypes, iter)
  {
    PicSearchType &curr = 
      AS(PicSearchType, globData.searchTypes.at(iter));
    if(curr.algs.isEmpty())
      globData.searchTypes.deleteAt(iter); // Remove empty search types
  }
*/
  if(PicError::isDetected()){
    EHOLD(PicMain, initFailed);
    return;
  }
}

// **********************************************************************

PicMain::~PicMain() {
}

// **********************************************************************

static PicSemaphore guiSemaph;
PICBOOL picSearchIsExiting; // set in PicMain::exitProgram

#include <picerror.hxx>
#include <sys/time.h>

bool PicMain::threadRunning()
{
	if(batchThread.state() == PicThread::running)
 		return(true);
	else
		return(false);
}

void PicMain::guiLeavingC()
{
  if(picSearchIsExiting)
    return;
  guiSemaph.set();
}

void PicMain::guiEnteringC()
{
  if(picSearchIsExiting)
    return;
  guiSemaph.clear();
}

// **********************************************************************

PICBOOL PicMain::startGUI(int argc, char *argv[])
{
  guiSemaph.set();
  return gui.start(argc, argv);
}

// **********************************************************************

PICBOOL PicMain::guiHasStarted(void)
{
  PicErrorContext errCxt;
  PicIterator     iter;
  BaseStringArray searchTypes;
  PicCollection   *coll;

  INFO_PRINT("Starting up.\n");  
  INFO_PRINT(copyRightMesg);

  searchTypes.setNull();
  coll = &globData.searchTypes;
  ITERATE(*coll, iter)
  {
    PicSearchType &curr = AS(PicSearchType, coll->at(iter));
    searchTypes.add(curr.name);
  }
  gui.setSearchTypes(searchTypes);

//  changeSearchType(globData.currentSearchType, searchTypes);
  
  // at start, first example image is null:
  FLUSH_ERRORS();

//  newExampleImage(1);		in v3 no initial example image is required!

  if(PicError::isDetected()){
    EHOLD(PicMain, initFailed);
    FLUSH_ERRORS();
    return boolFalse;
  }
  // Obsoleted in v3, all parameters are stored in the
  // client, more precisely in picsearch.cfg
  //readParamsFromDb();

  FLUSH_ERRORS();

  return boolTrue;
}

// **********************************************************************

/* BEGIN: Methods for PicParams follow, enhanced by v3 */

PicParams::PicParams() {
  globalErrorFlag = -1;
  PicParams(DEFAULT_CONFIG);
}

PicParams::PicParams(char *configFile) {
  int offset;
  int dirOffset;

  BaseString fileName(configFile);
  BaseString dirName;
  BaseString completeName[3];

  for(offset = 0; offset < SERVER_LIST_SIZE; offset++)
    serverList[offset].setNull();
  // Prefill with values that are known to be correct
  defaultSettings();
  // Try to find the config file in $PICSEARCHDIR, $HOME and .
  dirName = PicEnvironment::value("PICSEARCHDIR");
  if(dirName.isEmpty()) {
    completeName[0].setNull();
  }
  else {
    completeName[0] = dirName.append("/");
    completeName[0] = dirName.append(fileName);
  }
  dirName = PicEnvironment::value("HOME");
  if(dirName.isEmpty()) {
    completeName[1].setNull();
  }
  else {
    completeName[1] = dirName.append("/");
    completeName[1] = dirName.append(fileName);
  }
  completeName[2] = fileName;
  dirOffset = 0;
  configFp = NULL;
  do {
    if(!completeName[dirOffset].isEmpty())
      configFp = fopen(completeName[dirOffset].chars(), "r+");
    if(configFp != NULL)
      break;
    else
      dirOffset++;
  } while(dirOffset < 3);
  if(configFp) {
    readSettings();
  }
  else {
    dirOffset = 0;
    do {
      if(!completeName[dirOffset].isEmpty()) {
	configFp = fopen(completeName[dirOffset].chars(), "w+");
	writeSettings();
	break;
      }
      else
	dirOffset++;
    } while(dirOffset < 3);
  }
}

PicParams::PicParams(const PicParams& a) {
  // initialize Parameters from another PicParams
  int offset;
  windowSize = a.windowSize;
  posX = a.posX;
  posY = a.posY;
  platform = a.platform;
  perfIndex = a.perfIndex;
  networkPerformance = a.networkPerformance;
  for(offset = 0; offset < SERVER_LIST_SIZE; offset++)
    serverList[offset] = a.serverList[offset];
  logonUser = a.logonUser;
  cacheSize = a.cacheSize;
  resultImages = a.resultImages;
  hitPercentage = a.hitPercentage;
  defWidth = a.defWidth;
  defHeight = a.defHeight;
  stampWidth = a.stampWidth;
  stampHeight = a.stampHeight;
  querySetSize = a.querySetSize;
  dbImageFramesOnScreen = a.dbImageFramesOnScreen;
  scaleQueryImage = a.scaleQueryImage;
  askAlgParameters = a.askAlgParameters;
  picGUIDebug = a.picGUIDebug;
  picCoreDebug = a.picCoreDebug;
  picAlgDebug = a.picAlgDebug;
  picDbDebug = a.picDbDebug;
  picCacheDebug = a.picCacheDebug;
  showAbortedQuery = a.showAbortedQuery;
  askChangeConfirm = a.askChangeConfirm;
  wwwBrowser = a.wwwBrowser;
  showAdvert = a.showAdvert;
  imageEditor = a.imageEditor;
  threadAbortTimeout = a.threadAbortTimeout;
  formatForScaledImages = a.formatForScaledImages;
}

PicParams::~PicParams() {
}

void PicParams::closeConfig() {
  fclose(configFp);
}
void PicParams::restoreSettings(const PicParams& a) {
  windowSize = a.windowSize;
  cacheSize = a.cacheSize;
}

PicParams &PicParams::operator=(const PicParams& a) {
  // assign Parameters from another PicParams
  int offset;

  windowSize = a.windowSize;
  posX = a.posX;
  posY = a.posY;
  platform = a.platform;
  perfIndex = a.perfIndex;
  networkPerformance = a.networkPerformance;
  for(offset = 0; offset < SERVER_LIST_SIZE; offset++)
    serverList[offset] = a.serverList[offset];
  logonUser = a.logonUser;
  cacheSize = a.cacheSize;
  resultImages = a.resultImages;
  hitPercentage = a.hitPercentage;
  defWidth = a.defWidth;
  defHeight = a.defHeight;
  stampWidth = a.stampWidth;
  stampHeight = a.stampHeight;
  querySetSize = a.querySetSize;
  dbImageFramesOnScreen = a.dbImageFramesOnScreen;
  scaleQueryImage = a.scaleQueryImage;
  askAlgParameters = a.askAlgParameters;
  picGUIDebug = a.picGUIDebug;
  picCoreDebug = a.picCoreDebug;
  picAlgDebug = a.picAlgDebug;
  picDbDebug = a.picDbDebug;
  showAbortedQuery = a.showAbortedQuery;
  askChangeConfirm = a.askChangeConfirm;
  wwwBrowser = a.wwwBrowser;
  showAdvert = a.showAdvert;
  imageEditor = a.imageEditor;
  threadAbortTimeout = a.threadAbortTimeout;
  formatForScaledImages = a.formatForScaledImages;
  return *this;
}

PICBOOL PicParams::defaultSettings() {
  windowSize = 1;
  posX = 10;
  posY = 10;
  platform = detectPlatform();
  perfIndex = 10;
  networkPerformance = 10;
  serverList[0] = "localhost";
  logonUser = "SuperUser";
  cacheSize = 100;
  resultImages = 200;
  hitPercentage = 0;
  defWidth = 300;
  defHeight = 300;
  stampWidth = 128;
  stampHeight = 128;
  querySetSize = 200;
  dbImageFramesOnScreen = 12;
  scaleQueryImage = boolTrue;
  askAlgParameters = boolTrue;
  picGUIDebug = boolFalse;
  picCoreDebug = boolFalse;
  picAlgDebug = boolFalse;
  picDbDebug = boolFalse;
  picCacheDebug = boolFalse;
  showAbortedQuery = boolFalse;
  askChangeConfirm = boolTrue;
  wwwBrowser = "/usr/local/netscape/netscape -display :0";
  showAdvert = boolTrue;
  imageEditor = "/usr/local/netscape/xpaint -display :0";
  threadAbortTimeout = 15;
  formatForScaledImages = NULL;
  return boolTrue;
}

PICBOOL PicParams::readSettings() {
  char linebuf[256];
  char option[128];
  char optionValue[128];
  int section;
  int userServerOffset;
  section = -1;
  userServerOffset = 0;
  while(fgets(linebuf, 256, configFp) != NULL) {
    if(isOptSec(linebuf)) {
      // find out which section is in question
      section = getOptSec(linebuf);
    }
    else {
      // it's not a section line, find out what it is
      if(isOpt(linebuf) == true) {
        // it's a option line, get the option
        getOpt(linebuf, option);
        getOptVal(linebuf, optionValue);
        switch(section) {
        case GUI_PARAMS:
	  if(!strcmp(option, "windowsize"))
            windowSize = atol(optionValue);
          if(!strcmp(option, "windowposx"))
            posX = atol(optionValue);
          if(!strcmp(option, "windowposy"))
            posY = atol(optionValue);
          break;
        case RUNTIME_PARAMS:
          if(!strcmp(option, "platform")) {
            platform = atol(optionValue);
            if(detectPlatform() != platform)
              platform = detectPlatform();
          }
          if(!strcmp(option, "performance"))
            perfIndex = atol(optionValue);
          if(!strcmp(option, "networkperformance"))
            networkPerformance = atol(optionValue);
          break;
        case USED_SERVERS_PARAMS:
          if(userServerOffset < SERVER_LIST_SIZE) {
            if(!strcmp(option, "server"))
              serverList[userServerOffset++] = optionValue;
          }
          break;
        case USER_PARAMS:
	  if(!strcmp(option, "logonuser"))
            logonUser = optionValue;
          break;
        case CACHE_PARAMS:
          if(!strcmp(option, "size"))
            cacheSize = atol(optionValue);
          break;
        case QUERY_PARAMS:
          if(!strcmp(option, "resultimages"))
            resultImages = atol(optionValue);
          if(!strcmp(option, "hitpercentage"))
            hitPercentage = atol(optionValue);
          if(!strcmp(option, "defwidth"))
            defWidth = atol(optionValue);
          if(!strcmp(option, "defheight"))
            defHeight = atol(optionValue);
          if(!strcmp(option, "stampwidth"))
            stampWidth = atol(optionValue);
          if(!strcmp(option, "stampheight"))
	    stampHeight = atol(optionValue);
          if(!strcmp(option, "querysetsize"))
            querySetSize = atol(optionValue);
          if(!strcmp(option, "dbimageframesonscreen"))
            dbImageFramesOnScreen = atol(optionValue);
          if(!strcmp(option, "scalequeryimage")) {
            if(atol(optionValue) == 0)
              scaleQueryImage = boolFalse;
            else
              scaleQueryImage = boolTrue;
          }
          if(!strcmp(option, "askalgparameters")) {
            if(atol(optionValue) == 0)
              askAlgParameters = boolFalse;
            else
              askAlgParameters = boolTrue;
          }
          break;
        case DEBUG_PARAMS:
	  if(!strcmp(option, "picguidebug")) {
            if(atol(optionValue) == 0)
              picGUIDebug = boolFalse;
            else
              picGUIDebug = boolTrue;
          }
          if(!strcmp(option, "piccoredebug")) {
            if(atol(optionValue) == 0)
              picCoreDebug = boolFalse;
            else
              picCoreDebug = boolTrue;
          }
          if(!strcmp(option, "picalgdebug")) {
            if(atol(optionValue) == 0)
              picAlgDebug = boolFalse;
            else
              picAlgDebug = boolTrue;
          }
	  if(!strcmp(option, "picdbdebug")) {
            if(atol(optionValue) == 0)
              picDbDebug = boolFalse;
            else
              picDbDebug = boolTrue;
          }
	  if(!strcmp(option, "piccachedebug")) {
            if(atol(optionValue) == 0)
              picCacheDebug = boolFalse;
            else
              picCacheDebug = boolTrue;
	  }
          break;
        case MISC_PARAMS:
          if(!strcmp(option, "showabortedquery")) {
            if(atol(optionValue) == 0)
              showAbortedQuery = boolFalse;
            else
              showAbortedQuery = boolTrue;
          }
          if(!strcmp(option, "askchangeconfirm")) {
            if(atol(optionValue) == 0)
              askChangeConfirm = boolFalse;
            else
              askChangeConfirm = boolTrue;
          }
          if(!strcmp(option, "wwwbrowser")) {
            getOptValString(linebuf, optionValue);
            wwwBrowser = optionValue;
          }
          if(!strcmp(option, "showadvert")) {
            if(atol(optionValue) == 0)
              showAdvert = boolFalse;
            else
              showAdvert = boolTrue;
          }
          if(!strcmp(option, "imageeditor")) {
            getOptValString(linebuf, optionValue);
            imageEditor = optionValue;
          }
	  if(!strcmp(option, "threadaborttimeout"))
            threadAbortTimeout = atol(optionValue);
          break;
        default:
          break;
        }
      }
    }
  }
  return boolTrue;
}

PICBOOL PicParams::isOptSec(char *line) {
  bool foundLeft = false;
  bool foundRight = false;
  int lenght = strlen(line);
  int offset;

  for(offset = 0; (offset < lenght) && (foundLeft == false || foundRight == false); offset++) {
    if(line[offset] == '[')
      foundLeft = true;
    if(line[offset] == ']')
      foundRight = true;
  }
  if(foundLeft == true && foundRight == true)
    return boolTrue;
  else
    return boolFalse;
}

PICBOOL PicParams::isOpt(char *line) {
  PICBOOL isOption = boolFalse;
  int lenght = strlen(line);
  int offset;

  for(offset = 0; (offset < lenght) && (isOption == boolFalse); offset++)
    if(line[offset] == '=')
      isOption = boolTrue;
  return isOption;
}

PICSHORT PicParams::getOptSec(char *line) {
  char section[32];
  int counter;
  int offset;

  for(counter = 0; line[counter] != '['; counter++);
  for(offset = 0, counter++; line[counter] != ']' && offset < 32; counter++)
    section[offset++] = line[counter];
  section[offset] = '\0';
  if(!strcmp(section, "gui"))
    return GUI_PARAMS;
  if(!strcmp(section, "runtime"))
    return RUNTIME_PARAMS;
  if(!strcmp(section, "usedservers"))
    return USED_SERVERS_PARAMS;
  if(!strcmp(section, "user"))
    return USER_PARAMS;
  if(!strcmp(section, "cache"))
    return CACHE_PARAMS;
  if(!strcmp(section, "query"))
    return QUERY_PARAMS;
  if(!strcmp(section, "debug"))
    return DEBUG_PARAMS;
  if(!strcmp(section, "misc"))
    return MISC_PARAMS;
  return -1;
}

PICBOOL PicParams::getOpt(char *line, char *result) {
  int counter;
  int offset;
  int lenght = strlen(line);

  for(offset = 0, counter = 0; (counter < lenght) && (line[counter] != '='); counter++)
    if(isalpha(line[counter]) || isalnum(line[counter]) || isgraph(line[counter]))
      result[offset++] = line[counter];
  result[offset] = '\0';
  return boolTrue;
}

PICBOOL PicParams::getOptVal(char *line, char *result) {
  int counter;
  int offset;
  int lenght = strlen(line);

  for(counter = 0; (counter < lenght) && (line[counter] != '='); counter++);
  for(counter++, offset = 0; (line[counter] != '\0') && (counter < lenght); counter++)
    if(isalpha(line[counter]) || isalnum(line[counter]) || isgraph(line[counter]))
      result[offset++] = line[counter];
  result[offset] = '\0';
  return boolTrue;
}

PICBOOL PicParams::getOptValString(char *line, char *result) {
  int counter;
  int offset;
  int lenght = strlen(line);

  for(counter = 0; (counter < lenght) && (line[counter] != '='); counter++);
  for(counter++, offset = 0; (line[counter] != '\0') && (counter < lenght); counter++)
    if(isalpha(line[counter]) || isalnum(line[counter]) || isgraph(line[counter]) || line[counter] \
== ' ')
      result[offset++] = line[counter];
  result[offset] = '\0';
  return boolTrue;
}

void PicParams::showSettings() {
  cout << "Window size: " << windowSize << endl;
  cout << "Window X position: " << posX << endl;
  cout << "Window Y position: " << posY << endl;
  cout << "Platform : " << platform << endl;
  cout << "Performance index: " << perfIndex << endl;
  cout << "Network performance index: " << networkPerformance << endl;
  cout << "Server list" << endl;
  for(int offset = 0; (offset < SERVER_LIST_SIZE) && (serverList[offset].isNull() == boolFalse); offset++)

    cout << "\tServer " << offset + 1 << ": " << serverList[offset].chars() << endl;
  cout << "Last logon user: " << logonUser.chars() << endl;
  cout << "Amount of pictures stored in cache: " << cacheSize << endl;
  cout << "Result images: " << resultImages << endl;
  cout << "Minimun match percentage: " << hitPercentage << endl;
  cout << "New image width: " << defWidth << endl;
  cout << "New image height: " << defHeight << endl;
  cout << "Thumbnail width: " << stampWidth << endl;
  cout << "Thumbnail height: " << stampHeight << endl;
  cout << "Size of the query result set: " << querySetSize << endl;
  cout << "Picture frames on screen: " << dbImageFramesOnScreen << endl;
  cout << "Scale query image: ";
  if(scaleQueryImage == boolTrue)
    cout << "Yes" << endl;
  else
    cout << "No" << endl;
  cout << "Ask algorithm parameters: ";
  if(askAlgParameters == boolTrue)
    cout << "Yes" << endl;
  else
    cout << "No" << endl;
  cout << "GUI debug messages: ";
  if(picGUIDebug == boolTrue)
    cout << "Yes" << endl;
  else
    cout << "No" << endl;
  cout << "Core debug messages: ";
  if(picCoreDebug == boolTrue)
    cout << "Yes" << endl;
  else
    cout << "No" << endl;
  cout << "Algorithm debug messages: ";
  if(picAlgDebug == boolTrue)
    cout << "Yes" << endl;
  else
    cout << "No" << endl;
  cout << "Database debug messages: ";
  if(picDbDebug == boolTrue)
    cout << "Yes" << endl;
  else
    cout << "No" << endl;
  cout << "Cache debug messages: ";
  if(picCacheDebug == boolTrue)
    cout << "Yes" << endl;
  else
    cout << "No" << endl;
  cout << "Show results of aborted query: ";
  if(showAbortedQuery == boolTrue)
    cout << "Yes" << endl;
  else
    cout << "No" << endl;
  cout << "Confirm editor changes: ";
  if(askChangeConfirm == boolTrue)
    cout << "Yes" << endl;
  else
    cout << "No" << endl;
  cout << "WWW-browser: " << wwwBrowser.chars() << endl;
  cout << "Show server advertisement: ";
  if(showAdvert == boolTrue)
    cout << "Yes" << endl;
  else
    cout << "No" << endl;
  cout << "Image Editor: " << imageEditor.chars() << endl;
  cout << "Thread time-out: " << threadAbortTimeout << endl;
}

PICBOOL PicParams::writeSettings() {
  // Our rutine uses standard ANSI-C filehandling, C++ bites
  char linebuf[256];

  char *lineFeed = "\n";

  // rewind to begin of file
  rewind(configFp);
  // write [gui] section
  sprintf(linebuf, "[gui]%s", lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "windowsize=%u%s", windowSize, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "windowposx=%u%s", posX, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "windowposy=%u%s%s", posY, lineFeed, lineFeed);
  fputs(linebuf, configFp);
  // write [runtime] section
  sprintf(linebuf, "[runtime]%s", lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "platform=%d%s", platform, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "performance=%d%s", perfIndex, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "networkperformance=%d%s%s", networkPerformance, lineFeed, lineFeed);
  fputs(linebuf, configFp);

  // write [usedservers] section
  sprintf(linebuf, "[usedservers]%s", lineFeed);
  fputs(linebuf, configFp);
  for(int offset = 0; (offset < SERVER_LIST_SIZE) && (serverList[offset].isNull() == boolFalse); offset++) {
    sprintf(linebuf, "server=%s%s", serverList[offset].chars(), lineFeed);
    fputs(linebuf, configFp);
  }

  // write [user] section, additional newline...
  sprintf(linebuf, "%s[user]%s", lineFeed, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "logonuser=%s%s%s", logonUser.chars(), lineFeed, lineFeed);
  fputs(linebuf, configFp);

  // write [cache] section
  sprintf(linebuf, "[cache]%s", lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "size=%d%s%s", cacheSize, lineFeed, lineFeed);
  fputs(linebuf, configFp);

  // write [query] section
  sprintf(linebuf, "[query]%s", lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "resultimages=%lu%s", resultImages, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "hitpercentage=%u%s", hitPercentage, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "defwidth=%u%s", defWidth, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "defheight=%u%s", defHeight, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "stampwidth=%u%s", stampWidth, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "stampheight=%u%s", stampHeight, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "querysetsize=%u%s", querySetSize, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "dbimageframesonscreen=%u%s", dbImageFramesOnScreen, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "scalequeryimage=%d%s", (scaleQueryImage == boolTrue) ? 1 : 0, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "askalgparameters=%d%s%s", (askAlgParameters == boolTrue) ? 1 : 0, lineFeed, lineFeed);
  fputs(linebuf, configFp);

  // write [debug] section
  sprintf(linebuf, "[debug]%s", lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "picguidebug=%d%s", (picGUIDebug == boolTrue) ? 1 : 0, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "piccoredebug=%d%s", (picCoreDebug == boolTrue) ? 1 : 0, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "picalgdebug=%d%s", (picAlgDebug == boolTrue) ? 1 : 0, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "picdbdebug=%d%s", (picDbDebug == boolTrue) ? 1 : 0, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "piccachedebug=%d%s%s", (picCacheDebug == boolTrue) ? 1 : 0, lineFeed, lineFeed);
  fputs(linebuf, configFp);

  // write [misc] section
  sprintf(linebuf, "[misc]%s", lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "showabortedquery=%d%s", (showAbortedQuery == boolTrue) ? 1 : 0, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "askchangeconfirm=%d%s", (askChangeConfirm == boolTrue) ? 1 : 0, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "wwwbrowser=%s%s", wwwBrowser.chars(), lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "showadvert=%d%s", (showAdvert == boolTrue) ? 1 : 0, lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "imageeditor=%s%s", imageEditor.chars(), lineFeed);
  fputs(linebuf, configFp);
  sprintf(linebuf, "threadaborttimeout=%u%s", threadAbortTimeout, lineFeed);
  fputs(linebuf, configFp);
  return boolTrue;
}

PICBOOL PicParams::addServer(char *line) {
  int offset;
  PICBOOL inList;

  inList = boolFalse;

  for (offset = 1; (offset < SERVER_LIST_SIZE) && (serverList[offset].isNull() == boolFalse); offset++) {
    INFO_PRINT(serverList[offset]);
    if (serverList[offset] == line)
      inList = boolTrue; 
  }

/*
if (inList == boolTrue)
  INFO_PRINT("oli listalla\n")
else
  INFO_PRINT("EI ollut listalla\n");
*/

  if (inList == boolFalse) {
    for(offset = 1; (offset < SERVER_LIST_SIZE) && (serverList[offset].isNull() == boolFalse); offset++);
    if(offset == SERVER_LIST_SIZE)
      offset--;
    for(; offset > 1; offset--)
      serverList[offset] = serverList[offset - 1];
    serverList[1] = line;
  }
  
  return boolTrue;
}

PICBOOL PicParams::addUser(BaseString user) {
  logonUser = user;
}

PICSHORT PicParams::detectPlatform() {
  char *envPtr;
  if((envPtr = getenv("OSTYPE")) == NULL)
     return 0;
  if(!strcmp(envPtr, "osf1"))
    return 1;
  if(!strcmp(envPtr, "hpux"))
    return 1;
  if(!strcmp(envPtr, "solaris"))
    return 1;
  if(!strcmp(envPtr, "linux-gnu"))
    return 1;
  if(!strcmp(envPtr, "Linux"))
    return 1;
  return -1;
}

/* END: Methods for PicParams follow, enhanced by v3 */

/* BEGIN: Methods for PICCache follow, introduced by v3 */

PICCache::PICCache() {
  PICCache(100);
}

PICCache::PICCache(long size) {
  cacheSize = size;
  arraySize = size >> PER_CHAIN_SHIFT;
  cacheHits = 0;
  cacheMisses = 0;
  array = (CacheObj **) malloc(sizeof(CacheObj *) * arraySize);
  globalErrorFlag = 0;
  PICULONG i = 0;
  
  while(i < arraySize && globalErrorFlag == 0) {
    array[i] = (CacheObj *) malloc(sizeof(CacheObj));
    if(array[i] == NULL) {
      globalErrorFlag = -1;
      break;
    }
    else {
      array[i]->next = NULL;
      array[i]->prev = NULL;
      array[i]->databaseId = -1;
      array[i]->pictureId = -1;
      array[i]->inserted = -1;
      array[i]->lastUsed = -1;
      array[i]->hitCount = 0;
      array[i]->thisPic = NULL;
      array[i]->inThisChain = 1;
      array[i]->collisions = 0;
      i++;
    }
  }
}

PICCache::~PICCache() {
}

void PICCache::purgeCache() {
  CacheObj *p;
  CacheObj *q;
  for(PICULONG i = 0; i < arraySize; i++) {
    p = array[i];
    q = p;
    while(q->next != NULL)
      q = q->next;
    if(q->prev == NULL) {
      // there is only one element in chain
      if(q->thisPic != NULL) {
	q->thisPic->setNull();
	delete q->thisPic;
      }
      free(q);
    }
    else {
      // q points to last in chain, go prev and free
      q = q->prev;
      if(q->next->thisPic != NULL) {
        q->next->thisPic->setNull();
	delete q->next->thisPic;
      }
      free(q->next);
      // free next in chain until in first object, q = p
      while(q->prev != NULL) {
        q = q->prev;
	if(q->next->thisPic != NULL) {
	  q->next->thisPic->setNull();
	  delete q->next->thisPic;
	}
        free(q->next);
      }
      if(q == p) {
	if(q->thisPic != NULL) {
	  q->thisPic->setNull();
	  delete q->thisPic;
	}
        free(q);
      }
    }
  }
  free(array);
}

PICULONG PICCache::getHashKey(PICULONG databaseId, PICULONG pictureId) {
  // limited testing showed that the collision rate is somewhat 10-20%
  float base = (pictureId * HASH_CONST) + (databaseId * HASH_CONST);
  long medRes = (long)base;
  base = (base - medRes) * arraySize;
  
  return ((PICULONG)base);
}

bool PICCache::putThumb(long databaseId, long pictureId, const PicPicture &thisPic) {
  long toChain;
  long candHitCount;
  time_t candLastUsed;
  CacheObj *p;
  CacheObj *q;
  CacheObj *r;
  CacheObj *candidate;
  // Find out which chain this picture will be stored into
  toChain = getHashKey(databaseId, pictureId);
  p = array[toChain];
  // The first in chain is a special case
  if(p->inThisChain == 1 && p->inserted == -1) {
    p->databaseId = databaseId;
    p->pictureId = pictureId;
    p->inserted = time(NULL);
    p->lastUsed = -1;
    p->hitCount = 0;
    p->thisPic = new PicPicture(thisPic);
    return true;
  }
  // Next we will see if the chain is full or not
  if(p->inThisChain < PER_CHAIN) {
    // Chain isn't full, we can extend it
    q = p;
    while(q->next != NULL)
      q = q->next;
    q->next = (CacheObj *) malloc(sizeof(CacheObj));
    if(q->next != NULL) {
      r = q->next;
      r->prev = q;
      r->next = NULL;
      r->databaseId = databaseId;
      r->pictureId = pictureId;
      r->inserted = time(NULL);
      r->lastUsed = -1;
      r->hitCount = 0;
      r->thisPic = new PicPicture(thisPic);
      r->inThisChain = -1; // not used here
      r->collisions = -1; // --""--
      p->inThisChain++;
    }
    return true;
  }
  // Chain full, find which object to replace
  p->collisions++;
  candLastUsed = -1;
  candHitCount = 0;
  candidate = NULL;
  q = p;
  while(q->next != NULL) {
    if(q->lastUsed >= candLastUsed || q->hitCount >= candHitCount) {
      candidate = q;
      candLastUsed = q->lastUsed;
      candHitCount = q->hitCount;
    }
    q = q->next;
  }
  if(candidate != NULL) {
    candidate->databaseId = databaseId;
    candidate->pictureId = pictureId;
    candidate->inserted = time(NULL);
    candidate->lastUsed = -1;
    candidate->hitCount = 0;
    if(candidate->thisPic != NULL) {
      // Maybe a bit over top, but...
      candidate->thisPic->setNull();
      delete candidate->thisPic;
    }
    candidate->thisPic = new PicPicture(thisPic);
    return true;
  }
  return false;
}

PicPicture *PICCache::getThumb(long databaseId, long pictureId) {
  long inChain;
  CacheObj *p;
  // Find out where the thumbnail should be found
  inChain = getHashKey(databaseId, pictureId);
  p = array[inChain];
  while(p != NULL) {
    if((p->pictureId == pictureId) && (p->databaseId == databaseId)) {
      p->lastUsed = time(NULL);
      p->hitCount++;
      cacheHits++;
      return p->thisPic;
    }
    p = p->next;
  }
  cacheMisses++;
  return NULL;
}

bool PICCache::isThumbInCache(long databaseId, long pictureId) {
  // this method is necessary of how PICClient::QueryResults() works,
  // it could also be if'd in getThumb() but this is cleaner
  long inChain;
  CacheObj *p;
  // Find out where the thumbnail should be found
  inChain = getHashKey(databaseId, pictureId);
  p = array[inChain];
  while(p != NULL) {
    if((p->pictureId == pictureId) && (p->databaseId == databaseId))
      return true;
    p = p->next;
  }
  return false;
}

PICULONG PICCache::chainSize(PICULONG offset) {
  if(offset < arraySize)
    return array[offset]->inThisChain;
  else
    return 0;
}

PICULONG PICCache::elementsInChain(PICULONG offset) {
  // Method is required because the first element in
  // a chain is *special*
  PICULONG inChain;
  CacheObj *p;

  if(offset < arraySize) {
    inChain = 0;
    p = array[offset];
    while(p != NULL) {
      if(p->thisPic != NULL)
	inChain++;
      p = p->next;
    }
    return inChain;
  }
  else
    return 0;
}

PICULONG PICCache::chainCollisions(PICULONG offset) {
  if(offset < arraySize)
    return array[offset]->collisions;
  else
    return 0;
}

cacheArrayDescr *PICCache::getArrayDescr() {
  CacheObj *p;
  PICULONG chainOffset;
  PICULONG arrayOffset;
  cacheArrayDescr *thisDesc;

  if(!(thisDesc = (cacheArrayDescr *) malloc(sizeof(cacheArrayDescr))))
    return NULL;
  if(!(thisDesc->descArray = (PICULONG *) malloc(sizeof(PICULONG) * cacheSize * 2))) {
    free(thisDesc);
    return NULL;
  }
  thisDesc->nonNull = 0;
  thisDesc->elements = (cacheSize * 2);
  for(arrayOffset = 0; arrayOffset < thisDesc->elements; arrayOffset++)
    // If you change this, make according changes to sortCompareFunc() also
    thisDesc->descArray[arrayOffset] = 0;
  for(chainOffset = 0, arrayOffset = 0; chainOffset < arraySize; chainOffset++) {
    p = array[chainOffset];
    while(p != NULL) {
      if(p->thisPic != NULL) {
        thisDesc->descArray[arrayOffset] = p->databaseId;
        thisDesc->descArray[arrayOffset + 1] = p->pictureId;
        arrayOffset++;
        arrayOffset++;
	thisDesc->nonNull++;
	thisDesc->nonNull++;
      }
      p = p->next;
    }
  }
  // Sort is based on pictureId, server knows this
  qsort(thisDesc->descArray, cacheSize, sizeof(long) * 2, sortCompareFunc);
  return(thisDesc);
}

// This really should be coded in C++, libc power!
int sortCompareFunc(long *a, long *b) {
  long *ptr1 = a;
  long *ptr2 = b;

  ptr1++;
  ptr2++;
  // This fails miserably if neither id is negative
  return(*ptr1 - *ptr2);
}

CacheObj::CacheObj() {
}

CacheObj::~CacheObj() {
}

/* END: Methods for PICCache follow, introduced by v3 */


PICBOOL PicParams::fromColl(PicIxCollect &coll)
{
  PicErrorContext ecxt;
  static PicString trueVal("1");

  if(coll.contains(PICPARAMS_ASK_CHANGE_CONF)){
    askChangeConfirm = 
      (AS(PicString, coll.at(PICPARAMS_ASK_CHANGE_CONF)) == trueVal) ?
      boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_GUI_DEBUG)){
    picGUIDebug = 
      (AS(PicString, coll.at(PICPARAMS_GUI_DEBUG)) == trueVal) ?
      boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_CORE_DEBUG)){
    picCoreDebug = 
      (AS(PicString, coll.at(PICPARAMS_CORE_DEBUG)) == trueVal) ?
      boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_ALG_DEBUG)){
    picAlgDebug = 
      (AS(PicString, coll.at(PICPARAMS_ALG_DEBUG)) == trueVal) ?
      boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_DB_DEBUG)){
    picDbDebug = 
      (AS(PicString, coll.at(PICPARAMS_DB_DEBUG)) == trueVal) ?
      boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_CACHE_DEBUG)){
    picCacheDebug = 
      (AS(PicString, coll.at(PICPARAMS_CACHE_DEBUG)) == trueVal) ?
      boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_DEF_WIDTH)){
    PicString &ref = AS(PicString, coll.at(PICPARAMS_DEF_WIDTH));
    PicInteger newWidth(ref);
    if(!PicError::isDetected()){
      PICUSHORT trueWidth = setDefWidth((PICUSHORT)(newWidth.asInt()));
      ref = asString(trueWidth);
    }
  }
  if(coll.contains(PICPARAMS_WWWBROWSER)){
    wwwBrowser = AS(PicString, coll.at(PICPARAMS_WWWBROWSER));
  }
  if(coll.contains(PICPARAMS_NUM_OF_RESULT_IMAGES)){
    PicString &ref = AS(PicString, coll.at(PICPARAMS_NUM_OF_RESULT_IMAGES));
    PicInteger newVal(ref);
    if(!PicError::isDetected()){
      resultImages = newVal.asInt();
    }
  }
  if(coll.contains(PICPARAMS_HIT_PERC)){
    PicString &ref = AS(PicString, coll.at(PICPARAMS_HIT_PERC));
    PicInteger newVal(ref);
    if(!PicError::isDetected()){
      hitPercentage = newVal.asInt();
    }
  }

  /* BEGIN: New (or not used until now) parameters introduced by v3 */

  if(coll.contains(PICPARAMS_STAMP_WIDTH)) {
    PicString &ref = AS(PicString, coll.at(PICPARAMS_STAMP_WIDTH));
    PicInteger newVal(ref);
    if(!PicError::isDetected()){
      stampWidth = newVal.asInt();
    }
  }
  if(coll.contains(PICPARAMS_STAMP_HEIGHT)) {
    PicString &ref = AS(PicString, coll.at(PICPARAMS_STAMP_HEIGHT));
    PicInteger newVal(ref);
    if(!PicError::isDetected()){
      stampHeight = newVal.asInt();
    }
  }
  if(coll.contains(PICPARAMS_QUERY_SIZE)) {
    PicString &ref = AS(PicString, coll.at(PICPARAMS_QUERY_SIZE));
    PicInteger newVal(ref);
    if(!PicError::isDetected()){
      querySetSize = newVal.asInt();
    }
  }
  if(coll.contains(PICPARAMS_DB_IMAGE_FRAMES_ON_SCREEN)) {
    PicString &ref = AS(PicString, coll.at(PICPARAMS_DB_IMAGE_FRAMES_ON_SCREEN));
    PicInteger newVal(ref);
    if(!PicError::isDetected()){
      dbImageFramesOnScreen = newVal.asInt();
    }
  }
  if(coll.contains(PICPARAMS_SHOW_ABORTED_QUERY)) {
    showAbortedQuery = (AS(PicString, coll.at(PICPARAMS_SHOW_ABORTED_QUERY)) == trueVal) \
      ? boolTrue : boolFalse;
  }
  // We don't like segmenatation faults
  if(PicMain::globData.cacheIsDisabled == false) {
    if(coll.contains(PICPARAMS_CACHE_SIZE)) {
      PicString &ref = AS(PicString, coll.at(PICPARAMS_CACHE_SIZE));
      PicInteger newVal(ref);
      if(!PicError::isDetected()){
	cacheSize = newVal.asInt();
      }
    }
  }
  if(coll.contains(PICPARAMS_WINDOW_SIZE)) {
    PicString &ref = AS(PicString, coll.at(PICPARAMS_WINDOW_SIZE));
    PicInteger newVal(ref);
    if(!PicError::isDetected()){
      windowSize = newVal.asInt();
    }
  }
  if(coll.contains(PICPARAMS_WINDOW_POSITION_X)) {
    PicString &ref = AS(PicString, coll.at(PICPARAMS_WINDOW_POSITION_X));
    PicInteger newVal(ref);
    if(!PicError::isDetected()){
      posX = newVal.asInt();
    }
  }
  if(coll.contains(PICPARAMS_WINDOW_POSITION_Y)) {
    PicString &ref = AS(PicString, coll.at(PICPARAMS_WINDOW_POSITION_Y));
    PicInteger newVal(ref);
    if(!PicError::isDetected()){
      posY = newVal.asInt();
    }
  }
  if(coll.contains(PICPARAMS_PLATFORM)) {
    PicString &ref = AS(PicString, coll.at(PICPARAMS_PLATFORM));
    PicInteger newVal(ref);
    if(!PicError::isDetected()){
      platform = newVal.asInt();
    }
  }
  if(coll.contains(PICPARAMS_PERFORMANCE_INDEX)) {
    PicString &ref = AS(PicString, coll.at(PICPARAMS_PERFORMANCE_INDEX));
    PicInteger newVal(ref);
    if(!PicError::isDetected()){
      perfIndex = newVal.asInt();
    }
  }
  if(coll.contains(PICPARAMS_NETWORK_PERFORMANCE_INDEX)) {
    PicString &ref = AS(PicString, coll.at(PICPARAMS_NETWORK_PERFORMANCE_INDEX));
    PicInteger newVal(ref);
    if(!PicError::isDetected()){
      networkPerformance = newVal.asInt();
    }
  }
  if(coll.contains(PICPARAMS_SCALE_QUERY_IMAGE)) {
    scaleQueryImage = (AS(PicString, coll.at(PICPARAMS_SCALE_QUERY_IMAGE)) == trueVal) \
      ? boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_ASK_ALGORITHM_PARAMETERS)) {
    askAlgParameters = (AS(PicString, coll.at(PICPARAMS_ASK_ALGORITHM_PARAMETERS)) == trueVal) \
      ? boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_SHOW_SERVER_ADVERT)) {
    showAdvert = (AS(PicString, coll.at(PICPARAMS_SHOW_SERVER_ADVERT)) == trueVal) \
      ? boolTrue : boolFalse;
  }
  if(coll.contains(PICPARAMS_IMAGE_EDITOR)) {
    imageEditor = AS(PicString, coll.at(PICPARAMS_IMAGE_EDITOR));
  }
  if(coll.contains(PICPARAMS_LOGON_USER)) {
    logonUser = AS(PicString, coll.at(PICPARAMS_LOGON_USER));
  }
  if(coll.contains(PICPARAMS_THREAD_TIMEOUT)) {
    PicString &ref = AS(PicString, coll.at(PICPARAMS_THREAD_TIMEOUT));
    PicInteger newVal(ref);
    if(!PicError::isDetected()){
      threadAbortTimeout = newVal.asInt();
    }
  }
  for(int offset = 0; offset < SERVER_LIST_SIZE; offset++) {
    if(coll.contains(PICPARAMS_SERVER_LIST + offset)) {
      serverList[offset] = AS(PicString, coll.at(PICPARAMS_SERVER_LIST + offset));
    }
  }

  /* END: New (or not used until now) parameters introduced by v3 */

  return boolTrue;
}

// **********************************************************************

PICBOOL PicParams::toColl(PicIxCollect &coll)
{
  PicErrorContext ecxt;
  static PicString falseVal("0");
  static PicString trueVal("1");

  coll.deleteAll();

  coll.putAt(PICPARAMS_ASK_CHANGE_CONF, askChangeConfirm ? trueVal : falseVal);
  coll.putAt(PICPARAMS_GUI_DEBUG, picGUIDebug ? trueVal : falseVal);
  coll.putAt(PICPARAMS_CORE_DEBUG, picCoreDebug ? trueVal : falseVal);
  coll.putAt(PICPARAMS_ALG_DEBUG, picAlgDebug ? trueVal : falseVal);
  coll.putAt(PICPARAMS_DB_DEBUG, picDbDebug ? trueVal : falseVal);
  coll.putAt(PICPARAMS_CACHE_DEBUG, picCacheDebug ? trueVal : falseVal);
  coll.putAt(PICPARAMS_DEF_WIDTH, PicString(asString(getDefWidth())));
  coll.putAt(PICPARAMS_DEF_HEIGHT, PicString(asString(getDefHeight())));
  coll.putAt(PICPARAMS_WWWBROWSER, PicString(wwwBrowser));
  coll.putAt(PICPARAMS_NUM_OF_RESULT_IMAGES, PicString(asString(resultImages)));
  coll.putAt(PICPARAMS_HIT_PERC, PicString(asString(hitPercentage)));
  
  /* BEGIN: New (or not used until now) parameters introduced by v3 */

  coll.putAt(PICPARAMS_STAMP_WIDTH, PicString(asString(stampWidth)));
  coll.putAt(PICPARAMS_STAMP_HEIGHT, PicString(asString(stampHeight)));
  coll.putAt(PICPARAMS_QUERY_SIZE, PicString(asString(querySetSize)));
  coll.putAt(PICPARAMS_DB_IMAGE_FRAMES_ON_SCREEN, PicString(asString(dbImageFramesOnScreen)));
  coll.putAt(PICPARAMS_SHOW_ABORTED_QUERY, showAbortedQuery ? trueVal : falseVal);
  // We don't like segmenatation faults
  if(PicMain::globData.cacheIsDisabled == false)
    coll.putAt(PICPARAMS_CACHE_SIZE, PicString(asString(cacheSize)));
  else
    coll.putAt(PICPARAMS_CACHE_SIZE, PicString(asString(0)));
  coll.putAt(PICPARAMS_WINDOW_SIZE, PicString(asString(windowSize)));
  coll.putAt(PICPARAMS_WINDOW_POSITION_X, PicString(asString(posX)));
  coll.putAt(PICPARAMS_WINDOW_POSITION_Y, PicString(asString(posY)));
  coll.putAt(PICPARAMS_PLATFORM, PicString(asString(platform)));
  coll.putAt(PICPARAMS_PERFORMANCE_INDEX, PicString(asString(perfIndex)));
  coll.putAt(PICPARAMS_NETWORK_PERFORMANCE_INDEX, PicString(asString(networkPerformance)));
  coll.putAt(PICPARAMS_SCALE_QUERY_IMAGE, scaleQueryImage ? trueVal : falseVal);
  coll.putAt(PICPARAMS_ASK_ALGORITHM_PARAMETERS, askAlgParameters ? trueVal : falseVal);
  coll.putAt(PICPARAMS_SHOW_SERVER_ADVERT, showAdvert ? trueVal : falseVal);
  coll.putAt(PICPARAMS_IMAGE_EDITOR, PicString(imageEditor));
  coll.putAt(PICPARAMS_LOGON_USER, PicString(logonUser));
  coll.putAt(PICPARAMS_THREAD_TIMEOUT, PicString(asString(threadAbortTimeout)));
  for(int offset = 0; offset < SERVER_LIST_SIZE; offset++)
    coll.putAt(PICPARAMS_SERVER_LIST + offset, PicString(serverList[offset]));

  /* END: New (or not used until now) parameters introduced by v3 */

  return !PicError::isDetected();
}

// **********************************************************************

PICUSHORT PicParams::getDefWidth()
{
  return defWidth;
}

PICUSHORT PicParams::setDefWidth(PICUSHORT newWidth)
{
  if(newWidth > 20 && newWidth < 1025){
    defWidth = newWidth;
  }
  return defWidth;
}

PICUSHORT PicParams::getDefHeight()
{
  return defHeight;
}

PICUSHORT PicParams::setDefHeight(PICUSHORT newHeight)
{
  if(newHeight > 20 && newHeight < 769){
    defHeight = newHeight;
  }
  return defHeight;
}









