// --------------------------------------------------------------------------
//  'GENERIC' ARRAY BASE CLASS
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef BCOLLECT_HXX
#define BCOLLECT_HXX

#include <picbase.hxx>
class BaseString;

//:var ----------------------------------------------------------------------

extern "C" {
typedef void      PROC_SETITEM(void*);
typedef void      PROC_COPYITEM(void*, const void*);
typedef PICUSHORT PROC_TESTITEM(const void*);
typedef int       PROC_COMPAREITEMS(const void*, const void*);
}

/*
These types of function pointers are used in the instantiation of generic
array classes.

:list Function type
:item PROC_SETITEM:       Modifies the array component pointed to by the
                          argument.
:item PROC_COPYITEM:      Copies the array component pointed to by the
                          second argument.
:item PROC_TESTITEM:      Tests the state of the array component pointed
                          to by the argument.
:item PROC_COMPAREITEMS:  Compares the two array components and returns zero
                          if they are equal; -1 if the first one sorts before
                          the second one; or 1 otherwise.
:end
*/

/*:class *** ARRAY TYPE INFORMATION *****************************************

PicGenArrayInfo structure provides instantiation-specific type information
that is used by BaseCollection to implement the array operations. 

:list Field
:item itemSize:        Size of an array component in bytes.
:item createItem:      A function to construct a new component.
:item initItem:        Same as createItem, but if a null component is same
                       as a pure zero object, this can be set NULLPTR.
:item createItemCopy:  A function to construct a copy of another component.
:item setItemNull:     A function to set a component in null state. A null
                       component is regarded as non-existent, and it must
                       not allocate dynamic memory.
:item isItemNull:      A function for nullity test. It should return 1 if
                       the component is null or zero otherwise.
:item ascending:       A function for comparison of two components. Sorting
                       with this function should lead to ascending order.
                       If comparison is not a meaningful operation, you
                       should still provide a dummy implementation; possibly
                       one to raise an error.
:item descending:      Another comparison function for descending order.
:end  
*/

struct PicGenArrayInfo
{
    PICSIZE            itemSize;
    PROC_SETITEM*      createItem;
    PROC_SETITEM*      initItem;
    PROC_COPYITEM*     createItemCopy;
    PROC_SETITEM*      setItemNull;
    PROC_TESTITEM*     isItemNull;
    PROC_COMPAREITEMS* ascending;
    PROC_COMPAREITEMS* descending;
};


/*:class *** BASE ARRAY *****************************************************

BaseCollection is the base class of arrays generated from the array
template. Also it contains general error codes of collections. This
is an abstract class in practice, although not virtual.

****************************************************************************/

class BaseCollection
{

    public:
// --------------------------------------------------------------------------
//      DESTRUCTOR
//:fun ----------------------------------------------------------------------

   ~BaseCollection ();

/*
The desctructor sets the array in null state.
*/

// --------------------------------------------------------------------------
//      CLASS ID & ERROR CODES
// --------------------------------------------------------------------------

    PICCLASS()

/*
The class id of BaseCollection is used with most errors of collection
classes.
*/

//:mac -----------------------------------------------------------

    PICERRORCODES { ok,
        full              = ERR_MEMORY_ALLOCATION,
        illegalIndex      = ERR_COLLECTION_ILLEGAL_INDEX,
        illegalLowerBound = ERR_COLLECTION_ILLEGAL_LOWER_BOUND,
        putNullObject     = ERR_COLLECTION_PUT_NULLOBJECT,
        unsortable        = ERR_COLLECTION_UNSORTABLE
    };

/*
The error status codes of most collection classes.
*/

// --------------------------------------------------------------------------
//:fun  COMPARISON

    PICBOOL operator == (const BaseCollection&) const;
    PICBOOL operator != (const BaseCollection&) const;
    PICBOOL is          (const BaseCollection&) const;

/*
Arrays are equal if they contain equal components at the same locations.
Equality is defined by the function PicGenArrayInfo::ascending defined
for each array instantiation.

The is-method tests whether two arrays are the same object. 
*/

// --------------------------------------------------------------------------
//:fun

    void    setNull ();
    PICBOOL isNull  () const;

/*
A null array is considered empty and it allocates no dynamic memory.
Setting an array null will set all the components null. 
*/

// --------------------------------------------------------------------------
//:fun

    PICPOSITION next     (PICPOSITION) const;
    PICPOSITION previous (PICPOSITION) const;

    void    forAll (PROC_SETITEM*);
    PICSIZE count  (PROC_TESTITEM*, PICUSHORT result) const;

    PICPOSITION findNext (PICPOSITION, PROC_TESTITEM*, PICUSHORT) const;
    PICPOSITION findPrev (PICPOSITION, PROC_TESTITEM*, PICUSHORT) const;

/*
:list Method
:item next:      Returns the position of the next non-null component
                 starting from the given position. If such component is
                 not found, returns zero.
:item previous:  Same as next but searches for the previous component.
:item forAll:    Applies the given function to each component.
:item count:     Counts the number of components for whom the given test
                 function returns the given result.
:item findNext:  Returns the position of the next component for whom the
                 given test function returns the given result.
:item findPrev:  Same as findNext but searches for the previous component.
:end
*/

// --------------------------------------------------------------------------
//:fun

    PICBOOL     put    (PICPOSITION);
    PICPOSITION insert (PICPOSITION = 0);

/*
:list Method
:item put:      Ensures that the given location exists. It may or may
                not contain a component.
:item insert:   Inserts an empty location at the given position. If the
                location is non-empty, components are moved to make room.
                The default argument zero instructs to insert at the end;
                you should fill that location before doing any other
                operations, because empty locations at the end are purged
                automatically.
:end

:ret boolTrue:  The given location exists and can be safely used.
:ret boolFalse: The position was invalid (zero) or the array cannot be
                resized sufficiently.
*/

// --------------------------------------------------------------------------
//:fun

    void deleteAll ();
    void removeAll ();

    void purge (PICPOSITION, PICBOOL inOrder);
    void purge (PICBOOL inOrder);

/*
:list Method
:item deleteAll: Sets all components in null state.
:item removeAll: Deletes the array buffer without setting components null.
                 The array is null after this operation. Beware of memory
                 leaks when using this method.
:item purge:     Discards a given location after setting it null or discards
                 all null locations in the array buffer. Successive
                 locations are moved in memory. If it is not necessary to
                 retain the order of the components, the operation may be
                 faster.
:end
*/

// --------------------------------------------------------------------------
//:fun

    PICBOOL contains (PICPOSITION) const;
    PICBOOL contains (const BaseCollection&) const;

/*
The first variant returns boolTrue, if there is a non-null component in
the given location. The second variant returns boolTrue, if each non-null
component in the argument has an equal counterpart in the receiver, i.e.
the argument is a subset of the receiver.
*/

//:fun -----------------------------------------------------------------------

    PICPOSITION findEqual (
        const void*        obj,
        PROC_COMPAREITEMS* compare) const;

    PICPOSITION findSorted (
        const void*        obj,
        PROC_COMPAREITEMS* compare) const;

    PICPOSITION findMin (PROC_COMPAREITEMS* compare) const;
    PICBOOL     sort    (PROC_COMPAREITEMS* compare);

    PICPOSITION smallest () const;
    PICPOSITION greatest () const;

    PICBOOL sortAscending  ();
    PICBOOL sortDescending ();

/*
:list Method
:item findEqual:   Searches the first component that is equal to the first
                   argument (i.e. compare-function returns zero). This is
                   a linear search.
:item findSorted:  Same as findEqual, but uses a binary search algorithm,
                   so the array must be in a sorted order according to the
                   given comparison function.
:item findMin:     Searches the first component according to the given sort
                   order (for ascending order the smallest one; for descending
                   order the greatest one).
:item sort:        Sorts the array according to the given comparison function.
:item smallest:    Calls findMin with PicGenArrayInfo::ascending.
:item greatest:    Calls findMin with PicGenArrayInfo::descending.
:item sortAscending: Calls sort with PicGenArrayInfo::ascending.
:item sortDescending: Calls sort with PicGenArrayInfo::descending.
:end
*/

// --------------------------------------------------------------------------
//:fun

    PICSIZE size     () const;
    PICBOOL isEmpty  () const;

    PICPOSITION firstPos () const;
    PICPOSITION lastPos  () const;
    PICPOSITION maxPos   () const;

    PICPOSITION firstFree () const;

    PICBOOL inRange       (PICPOSITION) const;
    void    positionError (PICPOSITION) const;

/*
:list Method
:item size:      Returns the number of non-null components.
:item isEmpty:   Returns boolTrue if size is zero.
:item firstPos:  Returns the position of the first non-null component or
                 zero if the array is empty.
:item lastPos:   Returns the position of the last non-null component or
                 zero if the array is empty. Also sets the upper bound
                 at that location.
:item maxPos:    Returns the current maximum position (== allocation size).
:item firstFree: Returns the position of the first empty location.
:item inRange:   Tests whether the given position is nonzero and below the
                 upper bound.
:item positionError: Raises BaseCollection::illegalIndex. You can use this
                     when the inRange-check fails.
:end
*/

    protected:
// ==========================================================================
//:var

    PicGenArrayInfo* info;     // Type information
    char*            content;  // Store to hold elements (0..alloc).
    PICSIZE          alloc;    // Allocated store size (in item size units).
    PICPOSITION      ub;       // Upper bound for array.

/*
Instance variables are all protected, so the instantiated array classes
may access them. 

:list Instance variable
:item info:    Type information for the implementation of BaseCollection.
:item content: Buffer to hold the components (0..alloc - 1). The first
               location (#0) is always unused and may be used to return
               a reference to a null component.
:item alloc:   Size of the allocated buffer (in units of component size).
:item ub:      Cached position of the upper bound. This is usually the
               position of the last non-null component, but inserting at
               the end will temporarily move it. The method lastPos
               resets the cached position.
:end                
*/

// --------------------------------------------------------------------------
//:fun

    BaseCollection (PicGenArrayInfo*);
    void copyFrom (const BaseCollection&);

/*
The constructor sets the type information. The method copyFrom empties
the array and builds it anew by copying the argument array.
*/

// --------------------------------------------------------------------------
//:fun

    const void* operator [] (PICPOSITION) const;
          void* operator [] (PICPOSITION);

/*
These operators can access locations for which inRange is true. Note that 
the instantiated array classes should define their own access operators, 
because they can be implemented more efficiently (component size is known 
at compile time). 
*/

// --------------------------------------------------------------------------
//  ALLOCATION

    PICBOOL addLoc    (PICPOSITION);
    PICBOOL insertLoc (PICPOSITION);

    PICBOOL allocate (PICSIZE, PICBOOL clear);
    PICBOOL allocate (PICSIZE);
    PICBOOL resize   (PICSIZE, PICBOOL clear);
    PICBOOL resize   (PICSIZE);
};


// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline BaseCollection::BaseCollection (PicGenArrayInfo* typeInfo) :
       info(typeInfo), content(NULLPTR), alloc(0), ub(0) {}
inline BaseCollection::~BaseCollection ()
       { setNull(); }

inline void* BaseCollection::operator [] (PICPOSITION ix)
       { return content + ix * info->itemSize; }
inline const void* BaseCollection::operator [] (PICPOSITION ix) const
       { return content + ix * info->itemSize; }

inline PICBOOL BaseCollection::operator != (const BaseCollection& array) const
       { return !(*this == array); }
inline PICBOOL BaseCollection::is (const BaseCollection& array) const
       { return this == &array; }

inline PICBOOL BaseCollection::isNull () const
       { return !content; }

inline PICBOOL BaseCollection::inRange (PICPOSITION ix) const
       { return ix > 0 && ix <= ub; }

inline PICPOSITION BaseCollection::next (PICPOSITION ix) const
       { return findNext(ix, info->isItemNull, boolFalse); }
inline PICPOSITION BaseCollection::previous (PICPOSITION ix) const
       { return findPrev(ix, info->isItemNull, boolFalse); }

inline PICBOOL BaseCollection::contains (PICPOSITION ix) const
       { return inRange(ix) && !info->isItemNull(THIS[ix]); }

inline PICSIZE BaseCollection::size () const
       { return count(info->isItemNull, boolFalse); }
inline PICBOOL BaseCollection::isEmpty () const
       { return next(0) == 0; }

inline PICPOSITION BaseCollection::firstPos () const
       { return next(0); }
inline PICPOSITION BaseCollection::lastPos () const
       { return ((BaseCollection*)this)->ub = previous(ub + 1); }
inline PICPOSITION BaseCollection::maxPos () const
       { return alloc ? alloc - 1 : 0; }

inline PICPOSITION BaseCollection::smallest () const
       { return findMin(info->ascending); }
inline PICPOSITION BaseCollection::greatest () const
       { return findMin(info->descending); }

inline PICBOOL BaseCollection::sortAscending ()
       { return sort(info->ascending); }
inline PICBOOL BaseCollection::sortDescending ()
       { return sort(info->descending); }

#endif
