// --------------------------------------------------------------------------
//  STRING
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef BSTRING_HXX
#define BSTRING_HXX

#include <pchar.hxx>
#include <picerror.erh>

struct PicStringInfo
{
    PICSIZE alloc;  // Current maximum size of the string.
    PICSIZE size;   // Current string size, excluding NUL terminator.
    PICSIZE count;  // Number of references to this string.
                    // The character buffer follows.

    char* chars () { return (char*)(this + 1); }
};

const PICSIZE maxStringSize = maxNewSize - sizeof(PicStringInfo) - 1; // Maximum size of BaseString.

/*:class *** STRING *********************************************************

BaseString is a dynamic collection of ASCII characters. Size of a string is
limited only by the maximum allocation size. Character components are
referred to with positions within 1..size(). An empty string contains no
characters. It is not to be confused with null string, whose value is
considered unknown. A NUL character is stored at the end of every string,
but it is not included to the size.

BaseString uses a private class StringRep to store the character buffer.
The buffer contains a reference count that is incremented when strings are
copied. The buffer is copied only when one of the referents changes it.
Buffer is deleted when its reference count falls to zero. Empty string
reserves a buffer whereas null string doesn't.

****************************************************************************/

class BaseString
{
    public:

    PICCLASS()

    PICERRORCODES { ok,
        nullInstance  = ERR_NULL_INSTANCE,
        positionError = ERR_STRING_POSITION,
        sizeError     = ERR_STRING_SIZE
    };

// --------------------------------------------------------------------------
//:enum 

    enum CaseSensitivity { caseSensitive, caseInsensitive };
    enum Justification   { toRight, toLeft, center };

/*
These enumerated constants are used in certain string operations. Case
sensitivity means that upper- and lowercase characters are considered
unequal. Justification is used when aligning strings to a fixed-size field.
*/

// --------------------------------------------------------------------------
//:fun

    BaseString ();
    BaseString (const char*);
    BaseString (char, PICSIZE count);
    BaseString (const BaseString&);

    ~BaseString ();

/*
Strings can be constructed in the following ways:
    
    1. BaseString x1;
    2. BaseString x2 ("cccc...cccc");
    3. BaseString x4 ('c',1);
    4. BaseString x5 (aBaseString);

1. Default constructor creates a null BaseString.
2. A zero-terminated string is copied into the newly created BaseString.
   If the argument is NULLPTR a null BaseString is created.
3. The given character is duplicated to fill the string. If count is 0 an
   empty BaseString is created.
4. Copy constructor increments reference count. Character buffer is copied
   only when needed.

Destructor deletes the character buffer when its reference count falls to
zero.

:err outOfMemory: The character buffer is allocated dynamically which may
                  cause memory allocation failures.
:err sizeError:   The given character count or initialSize exceeded string
                  limits. The string becomes null.
*/

//:fun ----------------------------------------------------------------------

    BaseString& operator = (const char*);
    BaseString& operator = (const BaseString&);

/*
When a zero-terminated string is assigned, the characters are always copied.
If the argument is NULLPTR the receiver is set null. Assigning a BaseString
causes the reference count of the argument to be incremented.
*/

// --------------------------------------------------------------------------
//:fun

    BaseString copy () const;

/*
This is the force copy method of BaseString. It returns a BaseString
that is a copy of the receiver and has a distinct character buffer. This
is needed in multi-thread programs to avoid concurrency problems.
*/

// --------------------------------------------------------------------------
//:fun

    static const BaseString& null  ();
    static const BaseString& empty ();

/*
These functions return a global null and empty (non-null) BaseString.
*/

// --------------------------------------------------------------------------
//:fun

    PICBOOL isNull  () const;
    void    setNull ();

/*
The method setNull makes the receiver a null BaseString. It deletes the
character buffer if such exists.

:ret isNull: boolTrue if the receiver is a null BaseString.
*/

// --------------------------------------------------------------------------
//:fun

    PICPOSITION positionOf (
        const char* search,
        PICPOSITION    startPos = 1,
        CaseSensitivity         = caseSensitive) const;

    PICPOSITION positionOf (
        char     search,
        PICPOSITION startPos = 1,
        CaseSensitivity      = caseSensitive) const;

    PICPOSITION positionOf (
        const BaseString& search,
        PICPOSITION       startPos = 1,
        CaseSensitivity            = caseSensitive) const;

    static const char* find (
        const char* string,
        const char* search,
        CaseSensitivity = caseSensitive);

    static const char* find (
        const char* string,
        char        search,
        CaseSensitivity = caseSensitive);

/*
These methods search for a substring in the receiver or in the given string.

:arg string:    String from where to search a substring.
:arg search:    The substring to be found. A null string (NULLPTR) is never
                a substring. An empty string is a substring of another empty
                string, but not of non-empty string.
:arg startPos:  Starting position of the search. It must be a valid position.
                By default the search starts from the beginning.
:arg PicCaseSensitivity: Searching is case-sensitive by default. That mode
                         is more efficient.

:ret PICPOSITION: Beginning position of the substring or outOfRange if
                  the substring is not found.
:ret const char*: Beginning of the found substring or NULLPTR is not found.
*/

// --------------------------------------------------------------------------
//:fun  COMPARISON

    PICBOOL equal (
        const BaseString&,
        CaseSensitivity = caseInsensitive) const;
    PICBOOL notEqual (
        const BaseString&,
        CaseSensitivity = caseInsensitive) const;
    PICBOOL lessEqualThan (
        const BaseString&,
        CaseSensitivity = caseInsensitive) const;
    PICBOOL lessThan (
        const BaseString&,
        CaseSensitivity = caseInsensitive) const;
    PICBOOL greaterEqualThan (
        const BaseString&,
        CaseSensitivity = caseInsensitive) const;
    PICBOOL greaterThan (
        const BaseString&,
        CaseSensitivity = caseInsensitive) const;

    PICBOOL operator == (const BaseString&) const;
    PICBOOL operator != (const BaseString&) const;
    PICBOOL operator <= (const BaseString&) const;
    PICBOOL operator <  (const BaseString&) const;
    PICBOOL operator >= (const BaseString&) const;
    PICBOOL operator >  (const BaseString&) const;

    static PicCompareResult compare (
        const char*, const char*,
        CaseSensitivity = caseSensitive);

/*
These methods compare the argument and the receiver.
Comparison operators are case-sensitive;
other methods are case-insensitive by default, except compare. Null strings
are considered equal and smaller than any non-null string.

The method compare performs comparison for the given null-terminated
strings.
*/

//:fun ----------------------------------------------------------------------

    PICBOOL is (const BaseString&) const;

/*
Two strings are the same string if they occupy the same memory location.
Note that two strings referring to the same StringRep are not necessarily
the same string.
*/

// --------------------------------------------------------------------------
//:fun

    void setChars (const char*, PICSIZE size);
    void setChars (char, PICSIZE times);

/*
Set the string from a non-zero-terminated buffer by copying the given
amount of characters or by replicating a single character.

:arg const char*: Pointer to a non-zero-terminated character buffer.
                     If this is NULLPTR the BaseString is set null.
:arg const char:  Character to be replicated.
:arg size:           Size of the character buffer.
:arg times:          Count of characters to copy.
:err sizeError:      Given size is too big.
*/

//:fun ----------------------------------------------------------------------


    const char* chars () const;
          char* chars (PICSIZE setSize);

/*
These methods return a pointer to the beginning of the string.

The first variant returns a pointer to the beginning of the existing data
in the BaseString. It is always a zero-terminated string (but note that a
BaseString may contain embedded NUL characters). Also note that the address
of the character buffer may change when the string is resized, so don't
save the pointer over a longer period of time. For null string a pointer
to an empty string is returned.

The second variant sets the size of the BaseString to the argument and
ensures that the buffer is large enough. An error is raised and NULLPTR
returned if the allocation fails. Because the buffer is not initialized,
you must copy characters into the buffer.

These methods are typically used along with lower-level libraries, e.g. the
C run-time library. Examples:

  printf("%s", aString.chars());

  PICSIZE len = GetTextLength();
  GetText(aString.chars(len), len + 1);

:err sizeError
*/

// --------------------------------------------------------------------------
//:fun

    char at (PICPOSITION) const;

    char first () const;
    char last  () const;

/*
These methods access the component characters of the receiver string.

:arg PICPOSITION: A valid position inside the receiver string. The validity
                  is checked.

:ret at:    The character at the given position or NUL in case of error.
:ret first: The first character of the receiver, or NUL if the receiver is
            null or empty.
:ret last:  The last character of the receiver, or NUL if the receiver is
            null or empty.

:err positionError: The position argument of at is not valid. For
                    first and last no errors are raised.
*/

// --------------------------------------------------------------------------
//:fun

    BaseString at (PICPOSITION startPos, PICSIZE length) const;

/*
This method extracts a substring from the receiver:

    [startPos .. startPos + length - 1]

The following constraints are applied to the arguments:

    startPos == 0             --> startPos = 1
    startPos > size           --> empty string is returned
    length > size - startPos  --> length = size - startPos
    length == 0               --> empty string is returned

If the receiver is empty or null an empty string is returned.
*/

//:fun ----------------------------------------------------------------------

    BaseString between (PICPOSITION startPos, PICPOSITION endPos) const;

/*
This method extracts a substring from the receiver:

    [startPos .. endPos]

The following constraints are applied to the arguments:

    startPos == 0       --> startPos = 1
    startPos > size     --> positionError is raised
    endPos > size       --> endPos = size
    endPos < startPos   --> empty string is returned

If the receiver is empty or null an empty string is returned.

*/

//:fun ----------------------------------------------------------------------

    BaseString before (PICPOSITION beforePos) const;
    BaseString after  (PICPOSITION afterPos)  const;

/*
These methods extract a substring from the receiver:

    before: [1 .. beforePos - 1]
    after:  [afterPos + 1 .. size]

The following constraints are applied to the arguments:

    beforePos > size   --> whole string is returned
    beforePos <= 1     --> empty string is returned
    afterPos >= size   --> empty string is returned
    afterPos < 1       --> whole string is returned

If the receiver is empty or null an empty string is returned.

*/

// --------------------------------------------------------------------------
//:fun  DELETING CHARACTERS

    void deleteAt (PICPOSITION startPos, PICSIZE length);

/*
This method deletes a substring from the receiver:

    [startPos .. startPos + length - 1]

The following constraints are applied to the arguments:

    startPos == 0             --> startPos = 1
    startPos > size           --> positionError is raised
    length > size - startPos  --> length = size - startPos
    length == 0               --> no operation

If the receiver is empty or null nothing is done. If the deletion range
covers the whole string, the receiver is set empty.

*/

//:fun ----------------------------------------------------------------------

    void deleteBetween (PICPOSITION startPos, PICPOSITION endPos);

/*
This method deletes a substring from the receiver:

    [startPos .. endPos]

The following constraints are applied to the arguments:

    startPos == 0       --> startPos = 1
    startPos > size     --> positionError is raised
    endPos > size       --> endPos = size
    endPos < startPos   --> no operation

If the receiver is empty or null nothing is done. If the deletion range
covers the whole string, the receiver is set empty.

*/

//:fun ----------------------------------------------------------------------

    void deleteBefore (PICPOSITION beforePos);
    void deleteAfter  (PICPOSITION afterPos);

/*
These methods delete a substring from the receiver:

    deleteBefore: [1 .. beforePos - 1]
    deleteAfter:  [afterPos + 1 .. size]

The following constraints are applied to the arguments:

    beforePos > size   --> receiver is set empty
    beforePos <= 1     --> no operation
    afterPos >= size   --> no operation
    afterPos < 1       --> receiver is set empty

If the receiver is empty or null nothing is done.

*/

// --------------------------------------------------------------------------
//      INSERT METHODS
//:fun ----------------------------------------------------------------------

    void insert (PICPOSITION, const BaseString&);
    void insert (PICPOSITION, const char*);
    void insert (PICPOSITION, char, PICSIZE times = 1);

/*
These methods insert characters beginning from the given position. The
receiver will grow.

:arg PICPOSITION:       A valid position inside the receiver.
:arg const BaseString&: String to be inserted. If it is empty or null nothing
                        is done.
:arg const char*:       Zero-terminated string to be inserted. If the argument
                        is NULLPTR or points to NUL character, nothing is done.
:arg char:              Character to be inserted.
:arg times:             Number of characters to be inserted; 1 by default.
                        If this is zero, nothing is done.

:err positionError: Starting position is not valid.

*/

// --------------------------------------------------------------------------
//:fun

    void replace (PICPOSITION, const BaseString&);
    void replace (PICPOSITION, const char*);
    void replace (PICPOSITION, char, PICSIZE times = 1);

/*
These methods replace characters starting at the given position. If the
replacement exceeds the size of the receiver, the receiver will grow.

:arg PICPOSITION: A valid position inside the receiver.
:arg BaseString&: Replacement string. If it is empty or null, nothing is done.
:arg const char*: Replacement string. If it is NULLPTR or points to NUL
                     character, nothing is done.
:arg char:        Replacement character.
:arg times:          Specifies how many original characters the replacement
                     character replaces. If this is zero, nothing is done.

:err positionError: Starting position is not valid.

*/

//:fun ----------------------------------------------------------------------

    void replace (
        const BaseString& keyword,
        const BaseString& replacement,
        CaseSensitivity = caseSensitive);

    void replaceKeychar (
        char keyChar,
        char replacement,
        CaseSensitivity = caseSensitive);

/*
These methods replace all occurences of a keyword with a replacement.

:arg keyword: Keyword string. If this is empty or null, nothing is done.
:arg keyChar: Key character to be replaced with another character.
:arg BaseString& replacement: Replacement string. If this is empty or null,
                              the keywords will be deleted.
:arg char replacement:        Replacement character.
:arg PicCaseSensitivity:      Keyword searching is case-sensitive by default.

*/

// --------------------------------------------------------------------------
//:fun

    BaseString& append (const BaseString&);
    BaseString& append (const char*);
    BaseString& append (char, PICSIZE times = 1);

    BaseString& prepend (const BaseString&);
    BaseString& prepend (const char*);
    BaseString& prepend (char, PICSIZE times = 1);

/*
These methods insert characters at the end (append) or beginning (prepend)
of the receiver.

:arg const BaseString&: String to be inserted. If this is empty or null,
                        nothing is done.
:arg const char*:       String to be inserted. If this is NULLPTR or points
                        to a NUL character, nothing is done.
:arg char:              Character to be inserted.
:arg times:             Count of characters to be inserted. Default is one.
                        If this is zero, nothing is done.

:ret BaseString&:       Reference to the receiver.

*/

//:fun ----------------------------------------------------------------------

    friend BaseString operator & (const BaseString&, const BaseString&);
    friend BaseString operator & (const BaseString&, const char*);
    friend BaseString operator & (const BaseString&, char);

/*
These operators catenate two strings together. They are friend functions so
that they can be applied via implicit conversions, e.g:

    aBaseString & "an ASCIIZ string";
    "an ASCIIZ string" & aBaseString;
    aBaseString & 'c';

But because operators cannot be overloaded for char*, the following is
illegal:

    "an ASCIIZ string" & "another ASCIIZ string";

And because constructing a BaseString from char requires an additional
PICSIZE argument, the following is also illegal:

    'c' & aBaseString;

Note: compiler will not detect this error: it casts char into PICSIZE
and constructs an empty BaseString out of the 'c'-argument.

:ret BaseString: A newly created BaseString that contains the arguments
                 as catenated.

:err nullInstance: A BaseString-argument is null. If an char*-argument is
                   NULLPTR, nothing is done.

*/

// --------------------------------------------------------------------------
//:fun

           PICSIZE size () const;
    static PICSIZE size (const char*);

    PICBOOL isEmpty  () const;
    void    setEmpty ();

/*
Empty string reserves a character buffer that contains no characters
(except the ending NUL). The static method calculates the size of a
zero-terminated string.

:ret size:    Count of characters in the receiver. Zero for empty and null
              string.
:ret isEmpty: boolTrue if the receiver is empty or null (size == 0).
*/

//:fun ----------------------------------------------------------------------

    PICBOOL inRange (PICPOSITION) const;

/*
:arg PICPOSITION: Any PICPOSITION.
:ret boolTrue:  The argument position resides inside the receiver string.
:ret boolFalse: The argument position falls outside the receiver string.
*/

//:fun ----------------------------------------------------------------------

    BaseString& trim ();

/*
This method removes leading and trailing whitespace from the receiver.
If the receiver is null, nothing is done. For example:

    "  string  "  -->  "string"

:ret BaseString&: Reference to the receiver.
*/

//:fun ----------------------------------------------------------------------

    BaseString& reverse ();

/*
This method reverses the characters in the receiver. If the receiver is
null, nothing is done. For example:

    "string"  -->  "gnirts"

:ret BaseString&: Reference to the receiver.
*/

//:fun ----------------------------------------------------------------------

    BaseString& format (
        PICSIZE fieldWidth,
        Justification  = toRight,
        char filler = ' ');

    BaseString getImage () const;
    BaseString getImage (
        PICSIZE fieldWidth,
        Justification  = toRight,
        char filler = ' ') const;

/*
String formatting means justification inside a fixed-length field. Three
justification options are provided:

    "string"; field length = 8, filler blank

    BaseString::toLeft   -->  "string  "
    BaseString::toRight  -->  "  string"
    BaseString::center   -->  " string "

If the string is longer than the field, it is cut to fit. If the string is
empty, filler is added to make its length equal to fieldWidth.

The getImage-methods are provided for symmetry.

:arg fieldWidth:       Width of the field in characters.
:arg PicJustification: Justification style. Default is right-aligned.
:arg filler:           Character used to pad the empty positions. Default is
                       blank.

:ret BaseString&: Reference to the receiver.
:ret BaseString:  A copy of the receiver as formatted (not a force copy).
*/

// --------------------------------------------------------------------------
//:fun

    BaseString& makeUppercase ();
    BaseString& makeLowercase ();
    BaseString& makeProper    ();

/*
These methods change the case of characters (if any) in the receiver
string. 'Proper' capitalization is appropriate for names. It is designed
mainly for Finnish language, but the rules should work quite well for
other languages too.

:ret BaseString&: Reference to the receiver.
*/

// --------------------------------------------------------------------------
//      CLASSIFICATION
//:fun ----------------------------------------------------------------------

    PICBOOL isNumeric      () const;
    PICBOOL isAlphabetic   () const;
    PICBOOL isAlphanumeric () const;
    PICBOOL isSpace        () const;

/*
These methods return boolTrue if all the characters in the receiver string
fulfill the constraint (or the string is empty). Character classification
is defined by the methods of PicChar.
*/

//:fun ----------------------------------------------------------------------

    PICBOOL containsChars (const BaseString& charSet) const;

/*
:ret PICBOOL: boolTrue if the receiver string contains only characters in
              the argument charSet. For null or empty string boolTrue is
              returned, because such string doesn't contain illegal
              characters.
*/

    protected:
// ==========================================================================
//:fun

    char  operator [] (PICPOSITION) const;
    char& operator [] (PICPOSITION);

/*
This operator allows access to character buffer without range checking and
isNull-test. Because a reference is returned, characters can be modified
directly.
*/

// --------------------------------------------------------------------------
//:fun

    void insert (PICPOSITION, const char*, PICSIZE);

/*
This method implements the insert operation.

:arg PICPOSITION:   Position where to insert the argument. It must be
                    within [1 .. size + 1].
:arg const char*: Pointer to the string to be inserted. If this is
                     NULLPTR, nothing is done.
:arg PICSIZE:       Size of the string to be inserted. If this is zero,
                    the string must be zero-terminated (or NULLPTR).

:err positionError: Starting position is out of range.
:err sizeError:     The string would grow too much.

*/

// --------------------------------------------------------------------------
//:fun

    void replace (PICPOSITION, const char*, PICSIZE);

/*
This method implements the replace operation.

:arg PICPOSITION:   Position at which to replace characters with the
                    argument. It must be within [1 .. size + 1].
:arg const char*: Pointer to the replacement string. If this is
                     NULLPTR, nothing is done.
:arg PICSIZE:       Size of the replacement string. If this is zero,
                    the string must be zero-terminated (or NULLPTR).

:err positionError: Starting position is out of range.
:err sizeError:     The string would grow too much.

*/

// --------------------------------------------------------------------------
//:fun

    BaseString& append (const char*, PICSIZE);

/*
This method implements the append operation.

:arg const char*:    The string to be appended. If this is NULLPTR,
                        nothing is done.
:arg PICSIZE:           Size of the argument string. If this is zero,
                        the string must be zero-terminated (or NULLPTR).

:err sizeError:     The string would grow too much.

*/

// --------------------------------------------------------------------------
//:fun

    BaseString (const char*, PICSIZE);

/*
This protected constructor is used by substring methods. It copies
characters from the given ASCIIZ string.

:arg PICSIZE: The number of characters to be copied.
*/

    private:
// ==========================================================================
//      INSTANCE VARIABLES, PRIVATE METHODS AND FRIENDS
//
//      use makes sure that character buffer can be used (it is non-null
//      and the reference count is 1) and there is space for at least
//      requestSize characters (requestSize <= maxStringSize). The second
//      argument specifies whether to copy the old contents of the string.
//
//      If there are no other references to the character buffer,
//      deleteStringRep frees it. It always sets BaseString::rep to NULLPTR.
//
//      Errors : - MEMORY_ALLOCATION_ERROR
// ----------------------------------------------------------------------

    char* rep;

          PicStringInfo& info ()       { return *(((PicStringInfo*)rep) - 1); }
    const PicStringInfo& info () const { return *(((PicStringInfo*)rep) - 1); }

    PICBOOL use (PICSIZE requestedSize, PICBOOL saveOld = boolTrue);
    void deleteStringRep ();

    friend class PicText;
};

typedef BaseString::CaseSensitivity PicCaseSensitivity;
typedef BaseString::Justification PicJustification;

// ----------------------------------------------------------------------

    BaseString& operator << (BaseString&, char);
    BaseString& operator << (BaseString&, short);
    BaseString& operator << (BaseString&, unsigned short);
    BaseString& operator << (BaseString&, int);
    BaseString& operator << (BaseString&, unsigned int);
    BaseString& operator << (BaseString&, long);
    BaseString& operator << (BaseString&, unsigned long);
    BaseString& operator << (BaseString&, PICFLOAT);
    BaseString& operator << (BaseString&, const char*);
    BaseString& operator << (BaseString&, const BaseString&);

// --------------------------------------------------------------------------
//:fun

    BaseString asString (short);
    BaseString asString (unsigned short);
    BaseString asString (int);
    BaseString asString (unsigned int);
    BaseString asString (long);
    BaseString asString (unsigned long);
    BaseString asString (PICFLOAT);
//      BaseString asString (const void*);
 
/*
These functions return the textual representation of the argument.
Representations are self-explanatory, except for void* the pointer
format segment:offset is used.
*/

// --------------------------------------------------------------------------
//  INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline BaseString::BaseString () : rep(NULLPTR) {}

inline BaseString::~BaseString () { deleteStringRep(); }

inline PICBOOL BaseString::isNull  () const { return rep == NULLPTR; }
inline void    BaseString::setNull ()       { deleteStringRep(); }

inline char& BaseString::operator [] (PICPOSITION pos)
       { return rep[pos - 1]; }
inline char BaseString::operator [] (PICPOSITION pos) const
       { return rep[pos - 1]; }

inline PICBOOL BaseString::inRange (PICPOSITION pos) const
       { return rep && pos >= 1 && pos <= info().size; }
inline PICBOOL BaseString::isEmpty () const
       { return !rep || info().size == 0; }

inline PICBOOL BaseString::operator != (const BaseString& str) const
       { return !(*this == str); }
inline PICBOOL BaseString::operator <= (const BaseString& str) const
       { return str >= *this; }
inline PICBOOL BaseString::operator < (const BaseString& str) const
       { return str > *this; }

inline PICBOOL BaseString::notEqual (const BaseString& str, PicCaseSensitivity aCase) const
       { return !equal(str,aCase); }
inline PICBOOL BaseString::lessEqualThan (const BaseString& str, PicCaseSensitivity aCase) const
       { return str.greaterEqualThan(*this, aCase); }
inline PICBOOL BaseString::lessThan (const BaseString& str, PicCaseSensitivity aCase) const
       { return str.greaterThan(*this, aCase); }

inline PICBOOL BaseString::is (const BaseString& right) const
       { return this == &right; }

inline char BaseString::first () const { return rep ? rep[0] : '\0'; }

inline BaseString BaseString::before (PICPOSITION pos) const
       { return at(1, pos ? pos - 1 : 0); }
inline BaseString BaseString::after (PICPOSITION pos) const
       { return at(pos + 1, maxStringSize); }

inline void BaseString::deleteBefore (PICPOSITION pos)
       { if (pos > 1) deleteAt(firstPos, pos - 1); }
inline void BaseString::deleteAfter (PICPOSITION pos)
       { if (pos < size()) deleteAt(pos + 1, size() - pos); }

inline void BaseString::insert (PICPOSITION to, const BaseString& str)
       { insert(to, str.rep, str.size()); }
inline void BaseString::insert (PICPOSITION to, const char* azStr)
       { insert(to, azStr, 0); }

inline void BaseString::replace (PICPOSITION at, const BaseString& str)
       { replace(at, str.rep, str.size()); }
inline void BaseString::replace (PICPOSITION at, const char* azStr)
       { replace(at, azStr, 0); }

inline BaseString& BaseString::append (const BaseString& str)
       { append(str.rep, str.size()); return *this; }
inline BaseString& BaseString::append (const char* azStr)
       { append(azStr, 0); return *this; }

inline BaseString operator & (const BaseString& s1, const BaseString& s2)
       { return s1.copy().append(s2); }
inline BaseString operator & (const BaseString& s1, const char* s2)
       { return s1.copy().append(s2); }
inline BaseString operator & (const BaseString& str, char ch)
       { return str.copy().append(ch); }

inline BaseString& BaseString::prepend (const BaseString& str)
       { insert(1, str.rep, str.size()); return *this; }
inline BaseString& BaseString::prepend (const char* azStr)
       { insert(1, azStr, 0); return *this; }
inline BaseString& BaseString::prepend (char ch, PICSIZE times)
       { insert(1, ch, times); return *this; }

inline BaseString BaseString::getImage () const { return *this; }

inline BaseString BaseString::getImage (
    PICSIZE fieldWidth, PicJustification just, char filler) const
    { return copy().format(fieldWidth, just, filler); }

#endif
