// --------------------------------------------------------------------------
//  BASE TIME
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef BTIME_HXX
#define BTIME_HXX

#include <bstring.hxx>
#include <picerror.erh>
#include <picbase.hxx>
#include <picbase.erh>
class PicFmtText;

typedef char  PICHOUR;            // Hours in day (0..23).
typedef char  PICMINUTE;          // Minutes in hour (0..59).
typedef char  PICSECOND;          // Seconds in minute (0..59)

typedef PICLONG  PICHOURCOUNT;       // Time interval in hours.
typedef PICLONG  PICMINUTECOUNT;     // Time interval in minutes.
typedef PICLONG  PICSECONDCOUNT;     // Time interval in seconds.

const PICHOUR      minHour     = 0;  // Minimum value of PICHOUR.
const PICHOUR      maxHour     = 23; // Maximum value of PICHOUR.
const PICMINUTE    minMinute   = 0;  // Minimum value of PICMINUTE.
const PICMINUTE    maxMinute   = 59; // Maximum value of PICMINUTE.
const PICSECOND    minSecond   = 0;  // Minimum value of PICSECOND.
const PICSECOND    maxSecond   = 59; // Maximum value of PICSECOND.

const PICHOURCOUNT now         = 0;  // Current time.

const PICHOUR nullHour = 25;

/*:class ====================================================================
BaseTime implements the basic time handling operations. Time values include
an hour, minute and second component. Null time is used to represent unknown
or missing value. At any moment after creation BaseTime instance is either
null or represents a valid time of a day (i.e. within 00:00:00 and 23:59:59)

Time format can be selected with the PicTimeIniter class.

BaseTime belongs to the kernel level 1 library. It has a counterpart PicTime
in the level 2.
*/

class BaseTime
{

    public:
// --------------------------------------------------------------------------
//:enum TIME FORMATTING TYPES

    enum TResolution
    {
        hoursOnly      = FORMAT_HOURS_ONLY,
        hoursToMinutes = FORMAT_HOURS_TO_MINUTES,
        hoursToSeconds = FORMAT_HOURS_TO_SECONDS
    };

    enum TStyle
    {
        hours12        = FORMAT_HOURS_12,
        hours24        = FORMAT_HOURS_24
    };

/*
Time format is defined as a combination of these two parameters.
Resolution means the number of components included in the textual
representation:

    hoursOnly       12
    hoursToMinutes  12:35
    hoursToSeconds  12:35:45

Style means the way the components are presented:

    hours12         02:35:45 PM
    hours24         14:35:45
*/

// --------------------------------------------------------------------------
//:fun

    static TStyle      defaultStyle      ();
    static TResolution defaultResolution ();
    static char        defaultDelimiter  ();
    static char        defaultLeader     ();

/*
These methods return the components of the currently selected default
time format.
*/

// --------------------------------------------------------------------------
//:fun

    BaseTime ();
    BaseTime (PICHOURCOUNT plusInterval);
    BaseTime (PICHOUR, PICMINUTE, PICSECOND);
    BaseTime (const BaseString&);
    BaseTime (const BaseString&, char delimiter);
    BaseTime (const BaseTime&);

/*
BaseTimes can be created in the following ways:

    1. BaseTime a;
    2. BaseTime b (now + 2);
    3. BaseTime c (13, 30, 0);
    4. BaseTime d ("13.30");
    5. BaseTime e ("1:30:00 PM", ':');
    6. BaseTime f (aBaseTime);

1. Creates a null BaseTime.
2. Creates a BaseTime with the value 'after two hours' (now + 2 hours).
   This constructor uses the current system time.
3. Creates a BaseTime with the value 13h 30m given as components.
4. Creates a BaseTime with the value 13h 30m in default format.
5. Creates a BaseTime with the value 13h 30m in a specified format.
6. Copies another BaseTime.

:err deformatError: The given string is not of the given format.
:err formatError:   The given string is not a valid time.

*/

// ----------------------------------------------------------------------

    BaseTime& operator = (const BaseTime& right);

// --------------------------------------------------------------------------

    PICCLASS()

    PICERRORCODES { ok,
        nullInstance      = ERR_NULL_INSTANCE,
        deformatError     = ERR_TIME_DEFORMAT,
        formatError       = ERR_TIME_FORMAT,
        timeIllegal       = ERR_TIME_ILLEGAL
    };

// --------------------------------------------------------------------------

    PICBOOL isNull  () const;
    void    setNull ();

// --------------------------------------------------------------------------
//:fun

    BaseTime operator - (const BaseTime&) const;

/*
Minus-operator returns an interval between the receiver and the argument
normalized to be a valid BaseTime (within 0..23:59:59).

:err nullInstance: Receiver or argument is null.
*/

//:fun ----------------------------------------------------------------------

    void addHours   (PICHOURCOUNT);
    void addMinutes (PICMINUTECOUNT);
    void addSeconds (PICSECONDCOUNT);

/*
These methods add an interval to the receiver. The result is normalized
to be a valid BaseTime (within 0..23:59:59).

:err nullInstance: Receiver is null.
*/

//:fun ----------------------------------------------------------------------

    PICHOURCOUNT   hoursBetween   (const BaseTime&) const;
    PICMINUTECOUNT minutesBetween (const BaseTime&) const;
    PICSECONDCOUNT secondsBetween (const BaseTime&) const;

/*
These methods return the interval between the receiver and the argument
(in full units). If the argument is greater than the receiver the returned
value is negative. For example:

    hours between 12:23 and 13:23 --> -1
    hours between 12:23 and 13:22 -->  0  (not quite full hour)

:err nullInstance: Receiver or argument is null.
*/

// --------------------------------------------------------------------------
//:fun

    PICBOOL operator ==  (const BaseTime&) const;
    PICBOOL operator !=  (const BaseTime&) const;
    PICBOOL operator >   (const BaseTime&) const;
    PICBOOL operator >=  (const BaseTime&) const;
    PICBOOL operator <   (const BaseTime&) const;
    PICBOOL operator <=  (const BaseTime&) const;

/*
BaseTimes are compared along the time axis. Later times are considered
greater. Two null time instance are equal and null is smaller than any
other valid time instance.
*/

// --------------------------------------------------------------------------
//:fun

    BaseString getImage() const;
    BaseString getImage (
        TResolution, TStyle,
        char delimiter,
        char leader) const;

/*
These methods return the textual representation of the receiver. Default
format is used unless a specific format is given. For null time a null
string is returned.

Delimiter separates the components of the time value. Common delimiters
are point (13.30) and colon (13:30).

Leader defines the character which is used to fill time image when hours
are less than 10. If it is '\0' filling is not done.

    leader = '0'    -> (4h 3m 1s) -> "04:03:01"
    leader = '\0' -> (4h 3m 1s) -> "4:03:01"

*/

//:fun ----------------------------------------------------------------------

    void setImage (const BaseString&);
    void setImage (const BaseString&, char delimiter);

/*
These methods set the value of the receiver according to the given
textual time image. Time delimiter can be used as deformatting aid.
If no specific delimiter is given, the default is used.

See BaseDateTime::deformat for more details.

:err deformatError: The given string is not of the given format.
:err formatError:   The given string is not a valid time.

*/

//:fun ----------------------------------------------------------------------

    friend PicFmtText& operator << (PicFmtText&, const BaseTime&);

/*
Output operator to write a BaseTime to PicFmtText.
*/

// --------------------------------------------------------------------------
//:fun

    void setTime (PICHOUR, PICMINUTE, PICSECOND);

/*
This method sets the value of the receiver according to the given time
components.

:err timeIllegal: A time component is out of range. The receiver is
                  left unchanged.
*/

//:fun ----------------------------------------------------------------------

    void setRelativeTime (
        PICHOURCOUNT   hourInterval,
        PICMINUTECOUNT minuteInterval = 0,
        PICSECONDCOUNT secondInterval = 0);

/*
This method sets the value of the receiver based on the current system
time. Interval of hours, minutes and seconds can be added. The resulting
time value is normalized to be a valid BaseTime.
*/

// --------------------------------------------------------------------------
//:fun

    PICHOUR   hour   () const;
    PICMINUTE minute () const;
    PICSECOND second () const;

    PICSECONDCOUNT asSeconds () const;

/*
These methods return the components of the time value or the time as
seconds from 00:00:00. For null time all methods return zero.
*/

    private:
// ==========================================================================
//  INSTANCE VARIABLES, PRIVATE METHODS

    PICHOUR   theHour;
    PICMINUTE theMinute;
    PICSECOND theSecond;
    char   notUsed;      // To make BaseTime 4 bytes 

    friend class BaseDateTime;
};

typedef BaseTime::TResolution PicTimeResolution;
typedef BaseTime::TStyle      PicTimeStyle;

// --------------------------------------------------------------------------
//      TIME CLASS INITIALIZATION
/*:class --------------------------------------------------------------------

BaseTime/PicTime uses default parameters until an PicTimeIniter is created
(either as an automatic or static object). Given values are used until
another PicTimeIniter is created. When PicTimeIniter is destroyed the
previous values are restored.
*/

class PicTimeIniter
{
    public:

    PicTimeIniter (
        PicTimeStyle,
        PicTimeResolution,
        char delimiter,
        char leader);
   ~PicTimeIniter ();

    private:

    PicTimeStyle      tStyle;
    PicTimeResolution tResol;
    char           tDelim;
    char           tLeader;

};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PICBOOL BaseTime::isNull() const { return theHour == nullHour; }
inline void    BaseTime::setNull()      { theHour = nullHour; }

inline PICBOOL BaseTime::operator != (const BaseTime &right) const
       { return !(*this == right); }
inline PICBOOL BaseTime::operator >= (const BaseTime &right) const
       { return (*this == right || *this > right); }
inline PICBOOL BaseTime::operator < (const BaseTime &right) const
       { return !(*this >= right); }
inline PICBOOL BaseTime::operator <= (const BaseTime &right) const
       { return !(*this > right); }

inline PICHOUR   BaseTime::hour   () const { return isNull() ? 0 : theHour; }
inline PICMINUTE BaseTime::minute () const { return isNull() ? 0 : theMinute; }
inline PICSECOND BaseTime::second () const { return isNull() ? 0 : theSecond; }

#endif
