// --------------------------------------------------------------------------
//  ERROR LOG
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PERRLOG_HXX
#define PERRLOG_HXX

#include <picbase.hxx>
#include <perrmgr.hxx>
class BaseString;
class PicError;
class PicErrorManager;

/*:class ********************************************************************

PicErrorLog is able to log errors of multiple threads and multiple instances
of the same application. 

Instances of PicErrorLog are never created, since there is only one error
log in the application (even if there are multiple error managers). Error
log is manipulated through static methods of PicErrorLog.

****************************************************************************/

class PicErrorLog
{

    public:
// --------------------------------------------------------------------------
//      PARAMETERIZATION
//:enum ---------------------------------------------------------------------

    enum Stamp
    {
        baseStamp,
        clearStamp,
        discardStamp,
        holdStamp,
        raiseStamp,
        resetStamp,
        fatalStamp,
        internalStamp
    };

/*
A number of stamps are used to mark various error log entries.
*/

//:enum ---------------------------------------------------------------------

    enum
    {
        none      = 0,
        raised    = 1,
        held      = 2,
        discarded = 4,
        cleared   = 8,
        all       = 15
    };

/*
Error log filter determines what messages are logged. Log filter constants
are to be or'ed together.
*/

// --------------------------------------------------------------------------
//:fun  REGISTER USERS OF ERROR LOG

    static void login  ();
    static void logout ();

/*
Every user of error log should login before using the log, and logout when
not using any more. Error log is opened, when written the first time. File
is closed after all the users have logged out. These methods are called by
the constructor of PicL1ErrorManager, so they are usually not needed in
application programs.
*/

// --------------------------------------------------------------------------
//      LOGGING
//:fun ----------------------------------------------------------------------

    static void write (const PicError* message, Stamp);
    static void write (BaseString& logText);

    static void writeMemoryError (
        const char* logText,
        PICUSHORT      lineNumber = 0,
        const char* fileName   = NULLPTR);

/*
The first write-method creates an error message out of the given stamp.
The second method writes the text directly after prepending information
about process, thread, date and time (note that the argument itself is
modified).

Memory errors are logged in a special manner to avoid further memory
allocation.
*/

// --------------------------------------------------------------------------
//:fun  OPEN FLAG

    static PICBOOL open ();

/*
This flag is boolTrue when the error log is open, i.e. something has
been written to the log.
*/

    private:
// ==========================================================================
//      INSTANCE VARIABLES, PRIVATE METHODS AND FRIENDS
//
//      No instance variables since there is only one global error
//      log. PicErrorLog class is used to access the log from
//      multiple processes.

    DENY_CREATION_FOR(PicErrorLog)

    static PICBOOL opened;

    friend void PicError::log (BaseString& logText);

};

typedef PicErrorLog::Stamp PicErrorLogStamp;

class PicCircularLog
{
    public:

    PicCircularLog ();
    static void init ();
};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline void PicErrorLog::write (BaseString& logText)
{
    PicError::log(logText);
}

inline PicCircularLog::PicCircularLog () { init(); }

inline PICBOOL PicErrorLog::open () { return opened; }

#endif
