// --------------------------------------------------------------------------
//  ERROR MANAGER LEVEL 1
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PERRMGR1_HXX
#define PERRMGR1_HXX

#include <perrmgr.hxx>

/*:class ********************************************************************

PicL1ErrorManager is the basic implementation of Error Manager class. It
contains most of the functionality, but doesn't use the PicObject hierarchy.
All the features requiring PicObjects are no-operations, 
e.g. delayed error message texts are not displayed.

****************************************************************************/

class PicL1ErrorManager : public PicErrorManager
{
    public:
// --------------------------------------------------------------------------
//      CONSTRUCTOR AND DESTRUCTOR
//:fun ----------------------------------------------------------------------

    PicL1ErrorManager ();
   ~PicL1ErrorManager ();

/*
Constructor creates and initializes error manager for the
current thread. It creates the first error context as part
of the initialization. Destructor finishes the error manager.
Normally error manager is created at the beginning of the
program as an automatic object, for example:

    main ()
    {
        PicL1ErrorManager emgr;
        ...
        // Error manager is finished
    }
*/

// --------------------------------------------------------------------------
//      METHODS
// --------------------------------------------------------------------------

    PICERRORMGRCLASS()

    protected:
// ==========================================================================
//
//      Most methods are protected -- they are intended only for
//      error manager's internal use.
//
// --------------------------------------------------------------------------
//      CLEARING ERRORS
//
//      clearCurrent      Clears the errors of the current context.
//      reset             Resets the error manager.
//

    virtual void clearCurrent ();
    virtual void reset        ();

// --------------------------------------------------------------------------
//      RAISING ERRORS
//
//      See PERRMGR.HXX

    virtual void raise (const PicErrorText& message, PICBOOL hold);

    virtual void cumulate (const BaseString& cumulationText);

    virtual void raiseMemoryError (const char* explanation);

    virtual PICBOOL checkMemoryError (
        PICBOOL     wantToContinue,
        const char* cumulationText,
        PICUSHORT   lineNumber,
        const char* fileName);
                    
// --------------------------------------------------------------------------
//      CONTEXT HANDLING

    virtual PicError* removeError ();

// --------------------------------------------------------------------------
//      READING ERROR INFO FROM OBJECTS/RESOURCES
//
//      getErrorText    Handles direct message & keyword ($KEYWORD$).
//                      Resources are not used (in case of delayed
//                      message status code is shown).
//      replaceKeywords Does nothing.

    virtual BaseString getObjErrorText (
        const PicObject& messages,
        const PicClass&  errorClass,
        PICSTATUS        errorStatus) const;

// --------------------------------------------------------------------------
//      INTERNAL ERRORS
//
//      checkInternalError  Checks the current context for errors. Returns
//                          boolTrue if internal memory error was detected.

    PICBOOL checkInternalError (
        PICUSHORT      lineNumber,
        const char* fileName);

    #define CHECK_INTERNAL_ERROR() checkInternalError(__LINE__, __FILE__)

    private:
// ==========================================================================
//      PRIVATE METHODS
//
//      logErrorHook        Returns boolFalse.
//      *ExitHook           Writes to error log on any filter.

    virtual void transfer (PicErrorContext* current);

    DENY_COPYING_FOR(PicL1ErrorManager)

};

#endif
