// --------------------------------------------------------------------------
//  FILE
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PFILE_HXX
#define PFILE_HXX

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <bstring.hxx>
#include <picerror.erh>
class BaseStringArray;
class BaseDateTime;

/*:class ********************************************************************

PicFile represents a file having some attributes such as name,
attribute, existence and so on. File is identified by name that is given
when the PicFile instance is created. The name can be later changed. File
name has the following components: (drive,) path, name, root name, and suffix.
Application programs should not assume anything about file name.

The most common operations on directories, drives and file maintenance
are also provided. To manipulate the data in the file use PicFileStream.

****************************************************************************/

class PicFile
{
    public:
// --------------------------------------------------------------------------
//      CONSTRUCTORS & ASSIGNMENT
//:fun ----------------------------------------------------------------------

    PicFile ();
    PicFile (const char* name);
    PicFile (const BaseString& name);
    PicFile (const PicFile&);

/*
Default constructor creates null file. File name specifies file, directory
or group of files (pattern). Some methods also accept wildcards in file
name. See comments with each method for wildcard handling.

:arg const char*: File name.
:arg const BaseString&: File name.
:arg const PicFile&: File name.
*/

//:fun ----------------------------------------------------------------------

    PicFile& operator = (const PicFile&);

/*
Assignment changes name of the file.

:arg const PicFile&: New File name.
:ret PicFile&: Reference to the receiver.
*/

// --------------------------------------------------------------------------
//      METHODS
// --------------------------------------------------------------------------

    PICCLASS()

    PICERRORCODES { ok,
        error            = ERR_FILE_ERROR,
        fileNotOpened    = ERR_FILE_NOT_OPEN,
        fileStillOpen    = ERR_FILE_STILL_OPEN,
        fileAlreadyOpen  = ERR_FILE_ALREADY_OPEN,
        illegalDrive     = ERR_FILE_DRIVE,
        openModeError    = ERR_FILE_OPEN_MODE,
        fileNotFound     = ERR_FILE_NOT_FOUND,
        nameError        = ERR_FILE_NAME,
        sharingError     = ERR_FILE_SHARE,
        accessError      = ERR_FILE_ACCESS,
        driveNotReady    = ERR_FILE_NOT_READY,
        pathError        = ERR_FILE_PATH,
        eIsDir,
        eBadFileNbr,
        eInvalidArg
    };

// --------------------------------------------------------------------------
//      FILE CONSTANTS
//      Enumeration values are copied from bsedos.h/windows.h
//      (except for OpenMode).
//:enum ---------------------------------------------------------------------

    enum Attributes
    {
        normalFile     = 0100000,
        ownerRead      = 0400,
        ownerWrite     = 0200,
        ownerExec      = 0100,
        groupRead      = 0040,
        groupWrite     = 0020,
        groupExec      = 0010,
        otherRead      = 0004,
        otherWrite     = 0002,
        otherExec      = 0001,
        directory      = 040000
    };

/*
File attributes are specified when the file is created. Attributes can be
combined with bitwise-or. However, not all combinations are meaningful,
e.g. normalFile | readOnlyFile.

:list File attributes
:item normalFile: File is normal file.
:item readOnlyFile: File is read-only.
:item hiddenFile: File is hidden file.
:item systemFile: File is system file.
:item directory: File specifies directory, not file.
:item archived: File has been archived, e.g. with some backup program.
:end
*/

//:enum ---------------------------------------------------------------------

    enum Search
    {
        allFiles = normalFile,
        all = allFiles | directory,
        subdirectories = directory | 020000
    };

/*
PicFileSearch enumerates the most used file search attributes.

:list File search attributes
:item allFiles: Matches all all files but not directories.
:item all: Matches all files and directories.
:item subdirectories: Matches all subdirectories but not '.' (self) or
                      '..' (parent).
:end
*/

//:enum ---------------------------------------------------------------------

    enum OpenMode
    {
        forReading   = 0,
        forWriting   = 1,
        forAppending = 2
    };

/*
Open mode defines possible operations on file. Only one mode can be
specified at a time.

:list Open modes
:item forReading: The file is opened for reading.
:item forWriting: The file is created for writing.
:item forAppending: The file opened for writing at the end of file.
:end
*/

//:enum ---------------------------------------------------------------------

    enum ShareMode
    {
        shared = 0    // not used with *IX
    };

/*
Sharing defines access rights for other users.
:list Sharing mode
:item shared: Other users can open the file for any access.
:item denyRead: Other users can open the file for writing but not reading.
:item denyWrite: Other users can open the file for reading but not writing.
:item denyReadWrite: Other users can't open the file for any access.
:end
*/

// --------------------------------------------------------------------------
//      NULL
//:fun ----------------------------------------------------------------------

    PICBOOL isNull  () const;
    void    setNull ();

/*
Null file has null file name. Non-null file does not necessarily exist.

:ret boolTrue: File is null.
:ret boolFalse: File is not null.
*/

// --------------------------------------------------------------------------
//      FILE NAME COMPONENTS
//:fun ----------------------------------------------------------------------

    const BaseString& fullName () const;

    BaseString path     () const;
    BaseString name     () const;
    BaseString root     () const;
    BaseString suffix   () const;

/*
These methods query specific file name component. Components are specified
as follows:
- Drive consists of one character followed by colon (colon is not allowed
  in path), e.g. C:
- Path consists of all characters before the last path separator, not
  including it.
- Name consists of file name without any path, e.g. FILE.HXX.
- Root is file name without extension, e.g. FILE
- Suffix is file name extension (characters after the last point),
  e.g. HXX. Note that suffix can be of any length.

Currently '.' (current directory) and '..' (parent directory) are not
handled very intelligently, so avoid using them in the path.

:ret BaseString: Requested file name component. Null string is returned if
                 file is null, empty string if the component doesn't apply.
*/
//:fun ----------------------------------------------------------------------

    void setFullName (const BaseString&);
    void setPath     (const BaseString&);
    void setName     (const BaseString&);
    void setRoot     (const BaseString&);
    void setSuffix   (const BaseString&);

/*
These methods set the requested file name component. Resulting file
name is not validated. 

:arg const BaseString&: New value for file name component. If null or empty
                        component is removed.
*/

//:fun ----------------------------------------------------------------------

    PICBOOL exists () const;

/*
This method tests if the given file really exists. This doesn't mean that
the file can be opened because other users might have locked it. Wildcards
are not allowed in file name.

:ret boolTrue: The file exists.
:ret boolFalse: The file doesn't exist.
*/

//:fun ----------------------------------------------------------------------

    PICBOOL findPath () const;

//:fun ----------------------------------------------------------------------

    static PicFile makeUniqueName (const BaseString& baseName);

/*
This method forms a unique file name from the given base name. Name is
guaranteed to be unique only at the time of function call.

:arg const BaseString&: Template for unique name. First three characters
                        in the name and suffix are preserved if possible.
:ret PicFile: Unique file name, null if unique name is not formaed.
*/

//:fun ----------------------------------------------------------------------

    static PicFile find (
        const BaseString& pattern,
        const BaseString& dirs,
              PICBOOL     recurse    = boolFalse,
              PICUINT     attributes = normalFile);

    static PICBOOL find (
        BaseStringArray&  files,
        const BaseString& pattern,
        const BaseString& dirs, 
              PICBOOL     recurse    = boolFalse,
              PICUINT     attributes = normalFile);

/*
These are the most powerful search methods. They search matching file(s)
using directories specified by the given directory list. Semicolon is used
to separate different directories in the list. Find has two versions: the
first one returns full file name of the first matching file and the second
returns full file names of all matching files as an string array.

:arg files:   Array which receives absolute path names of all matching files.
:arg pattern: File pattern which is searched.
:arg dirs:    Directories which are used to search for pattern.
:arg recurse: If this is boolTrue dirs specify starting directories for
              subdirectory scan, if boolFalse only specified directories
              are searched.
:attributes:  File attributes specify type of the files to search.
:ret PicFile: First matching file.
:ret PICBOOL: boolTrue if at least one file was found, boolFalse if no file
              was found.
*/

// --------------------------------------------------------------------------
//      FILE OPERATIONS
//:fun ----------------------------------------------------------------------

    PICBOOL copyTo (const PicFile& newFile) const;

//:fun ----------------------------------------------------------------------

    PICBOOL rename (const BaseString& newName);

/*
This method renames existing file to new name (name of the PicFile
instance does not change). Both files must be on the same drive and
can not contain wild-cards.

:arg const BaseString&: New name for the file.
:err error: General operating system error.
:err accessError: Renamed file can not be read or the destination can not
                  be written.
:err sharingError: The renamed file is locked by some other process.
*/

//:fun ----------------------------------------------------------------------

    PICBOOL deleteFile ();

/*
This method deletes file. File name can not contain wild-cards.

:err error: General operating system error.
:err sharingError: The deleted file is locked by some other process.
*/
//:fun ----------------------------------------------------------------------

    static PICBOOL changeDirectory (const BaseString&);
    static PICBOOL makeDirectory   (const BaseString&);
    static PICBOOL removeDirectory (const BaseString&);

/*
Directory maintenace operations. Note that directory change has effect
only to current process. When process exits current directory is restored
to previous one.

:arg const BaseString&: Name specifying directory. It can not contain
                        wild-cards.
:err error: General operating system error.
:err accessError: The directory doesn't exist or is used by some other
                  process.
*/

//:fun ----------------------------------------------------------------------

    static BaseString currentDirectory ();

/*
Current directory contains drive and path information. Ending path
separator is not included.

:ret BaseString: Current directory name.
:err illegalDrive: The current drive doesn't contain readable storage media.
*/

// --------------------------------------------------------------------------
//      SIZE & ATTRIBUTES
//:fun ----------------------------------------------------------------------

    PICSIZE itemCount   (PICSIZE itemSize = sizeof(char)) const;
    PICUINT attributes  () const;
    PICBOOL isDirectory () const;

/*
These methods return file size and attributes. You can test for specific
attribute using bitwise-and :

    if (file.attributes() & PicFile::readOnlyFile)
    ...

:arg const PICSIZE: Specifies size of items to count.
:ret PICSIZE: Number of items in the file. If itemSize == 1 then file size
              is returned (default). Size is 0 if file has not been saved
              or the file is open for writing (for any user).
:ret PICUSHORT: Combination of PicFileAttributes describing file attributes.
:ret isDirectory: True if the file specifies directory.
*/

    private:
// ==========================================================================
//  INSTANCE VARIABLES, PRIVATE METHODS AND FRIENDS

    BaseString fName;

    PICPOSITION skipDrive () const;
    PICPOSITION skipPath  () const;
    PICPOSITION skipRoot  () const;

    static PICBOOL fileFind (
              PicFile&    match,
        const BaseString& pattern,
        const BaseString& startDir,
              PICBOOL     recurse,
              PICUINT     attr);

    static void fileFind (
        BaseStringArray&  files,
        const BaseString& pattern,
        const BaseString& startDir,
              PICBOOL     recurse,
              PICUINT     attr);

    friend class PicFileIterator;

};

/*:class ********************************************************************

PicFileIterator implements iteration of files that match a given pattern:

    PicFileIterator iter ("*.*", PicFile::normalFile);
    while (iter.next())
    {
        use(iter.file);
        ...
    }

Several file iterators can be active at the same time. 

****************************************************************************/

class PicFileIterator
{
  public:

    PicFileIterator (const PicFile&, PICUINT attr = PicFile::normalFile);
   ~PicFileIterator ();

    PICBOOL next ();

    PicFile file;

  private:

    PICUINT   attr;
    DIR *handle;
    PicFile pattern;
};

typedef PicFile::Attributes PicFileAttributes;
typedef PicFile::Search     PicFileSearch;
typedef PicFile::OpenMode   PicOpenMode;
typedef PicFile::ShareMode  PicShareMode;

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicFile::PicFile () {}
inline PicFile::PicFile (const char* name) : fName(name) {}
inline PicFile::PicFile (const BaseString& name) : fName (name) {}
inline PicFile::PicFile (const PicFile& file) : fName(file.fName) {}

inline PicFile& PicFile::operator = (const PicFile& file)
       { fName = file.fName; return *this; }

inline PICBOOL PicFile::isNull () const { return fName.isNull(); }
inline void    PicFile::setNull()       { fName.setNull(); }

inline PICBOOL PicFile::isDirectory() const
       { return attributes() & directory; }

inline const BaseString& PicFile::fullName () const { return fName; }

inline PicFileIterator::PicFileIterator (const PicFile& pattern, PICUINT a) :
    file(pattern), attr(a), handle(0) {}

#endif
