// --------------------------------------------------------------------------
//  DYNAMIC INTEGER ARRAY
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PINTARR_HXX
#define PINTARR_HXX

#include <picbase.hxx>
#include <picerror.erh>
class PicObject;
class BaseString;
class PicIterator;


/*:class ====================================================================
IntArray is a dynamic array of integers. Array has dynamic upper bound.
At any moment the array is either empty (upper bound < lower bound) or
has size() elements (size == upper bound - lower bound + 1). Size of the
array can vary between 0..maxNewWordSize as defined in picbase.hxx.
Because of the dynamic nature of the array every non-const operation might
resize the array (e.g. add, fillBetween). When the array is resized new
positions are filled with zeros.
Access outside current array bounds is illegal and raises an error (only
exception is current which raises error).

While PicIntArray provides efficient handling it restricts components
to short integers. PicArray can be used to store PicObjects.
*/

class PicIntArray
{
    public:

    PICCLASS()

// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT
//:fun ----------------------------------------------------------------------

    PicIntArray ();
    PicIntArray (
        PICINDEX lowerBound,
        PICSIZE  initialSize = defaultSize);
    PicIntArray (const PicIntArray&);
    PicIntArray (const PicObject&);
    PicIntArray (PICINDEX lowerBound, const PicObject&);

/*
PicIntArray can be created as :
    1. PicIntArray array;
    2. PicIntArray array (-1,10);
    3. PicIntArray array (anIntArray);
    4. PicIntArray array (collection);
    5. PicIntArray array (-1, collection);

1. Default constructor creates null array with lower bound 1.
2. Create empty array with -1 as lower bound and initially space
   for 10 short integers.
3. Copy constructor.
4. Create array from collection. It is iterated through and all
   the short integer components are added into the PicIntArray 
   (add(current().asInt()).
5. Similar to 4 but with given lower bound.

:err PicObject::typeMismatch: Argument object is not of correct type
                              or components are not short integers.
:err PicIntArray::illegalLowerBound: Given lower bound was outOfIndex.
:err MEMORY_ALLOCATION_ERROR: Can't allocate enough memory for 
                              array.
*/

   ~PicIntArray ();
  
//:fun ----------------------------------------------------------------------

    PicIntArray& operator = (const PicIntArray&);

/*
Assignment copies the argument array. After the assignment arrays have
same values and attributes (lower bound etc.).
*/

// --------------------------------------------------------------------------
//      NULL
//:fun ----------------------------------------------------------------------

    PICBOOL isNull  () const;
    void    setNull ();

/*
Null array is considered empty. It doesn't allocate memory.
*/

// --------------------------------------------------------------------------

    PICBOOL operator == (const PicIntArray&) const;
    PICBOOL is          (const PicIntArray&) const;

//:fun ----------------------------------------------------------------------

    PICINT at    (PICINDEX) const;
    PICINT first () const;
    PICINT last  () const;

/*
Indexed component access.

:ret at: Integer at given index.
:ret first: Integer at PicIntArray::firstIndex(), or 0 if the array
            is empty.
:ret last: Integer at PicIntArray::lastIndex(), or 0 if the array
           is empty.
:err PicIntArray::illegalIndex: Given index does not specify any component.
                                Return value is undefined and should not
                                be used.
*/

//:fun ----------------------------------------------------------------------

    const PICINT* ints () const;

/*
This method returns a pointer to the internal buffer of PicIntArray.
The first integer in the array is at the location 0 and the last one
at the location size() - 1. If the PicIntArray is null, a NULLPTR is
returned.
*/

// --------------------------------------------------------------------------
//      ADD, REPLACE, FILL, EMPTY
//:fun ----------------------------------------------------------------------

    void add   (PICINT);
    void putAt (PICINDEX, PICINT);

/*
add appends the value to the array at implicit index lastIndex()+1.
putAt assigns the value to given index. Previous value at that index
is lost. putAt will enlarge array if the given index is bigger than
current upper bound.

:arg const PICINT: New value.
:arg const PICINDEX: Array index between firstIndex()..maxIndex().
:err MEMORY_ALLOCATION_ERROR: Array is resized and there is not enough
                              memory to do it.
:err PicIntArray::illegalIndex: The given index is less than firstIndex(),
                                greater than maxIndex() or is outOfRange.
*/

//:fun ----------------------------------------------------------------------

    void fillBetween (PICINDEX from, PICINDEX to, PICINT value);
    void setEmpty    ();

/*
fillBetween fills the given positions (including limits) with the given
integer. If fillBetween causes the array to grow, upper bound is 'to'
after the operation.
setEmpty zeros the array and sets it in empty state.

:arg from: First location to fill. If 'from' is smaller than lower bound
           the array is filled beginning from lower bound.
:arg to: Last location to fill. If 'to' is smaller than 'from' nothing
         is done.
:arg PICINT: Value which is used to fill array.
:err MEMORY_ALLOCATION_ERROR: Can't allocate space for array components.
*/

// --------------------------------------------------------------------------
//      SEARCH
//:fun ----------------------------------------------------------------------

    PICBOOL  contains (PICINT) const;
    PICINDEX indexOf  (PICINT) const;

/*
:ret contains: boolTrue if the array contains given value.
:ret indexOf: Location of the value, outOfIndex if not found.
*/

// --------------------------------------------------------------------------
//      ATTRIBUTES
//:fun ----------------------------------------------------------------------

    PICSIZE  size       () const;
    PICBOOL  isEmpty    () const;

/*
:ret size: upper bound - lower bound + 1
:ret isEmpty: upper bound < lower bound
*/

//:fun ----------------------------------------------------------------------

    PICINDEX firstIndex () const;
    PICINDEX lastIndex  () const;
    PICINDEX minIndex   () const;
    PICINDEX maxIndex   () const;
    PICSIZE  maxSize    () const;
    PICBOOL  inRange    (PICINDEX) const;

/*
Array attributes.

:ret firstIndex: lower bound (or outOfIndex if empty)
:ret lastIndex: upper bound (or outOfIndex - 1 if empty)
:ret minIndex: lower bound
:ret maxIndex: lower bound + maxSize - 1
:ret maxSize: current maximum number of elements (allocation size)
:ret inRange: index >= lower bound && index <= upper bound
*/

//:fun ----------------------------------------------------------------------

    PICPOSITION position (PICINDEX) const;
    PICINDEX    index    (PICPOSITION) const;

/*
Conversion between position and index. Note that component at first 
index is at position 1. Validity of locations is not checked.
*/

// --------------------------------------------------------------------------
//      ITERATOR METHODS
//:fun ----------------------------------------------------------------------

    void     startIteration (PicIterator&) const;
    PICBOOL  next           (PicIterator&) const;
    PICBOOL  previous       (PicIterator&) const;
    PICINT   current        (PicIterator&) const;
    PICINDEX currentIndex   (PicIterator&) const;

/*
:err PicIntArray::illegalIndex: if current index is out of array bounds,
                                in which case outOfIndex is returned.
*/

// --------------------------------------------------------------------------
//      SORT
//:fun ----------------------------------------------------------------------

    void sort (PicSortOrder = ascending);

/*
Sort changes array indexes.
*/

    protected:
// ==========================================================================
//:fun PROTECTED METHODS

          PICINT& operator [] (PICINDEX);
    const PICINT& operator [] (PICINDEX) const;

/*
Component access without range checking.

:arg const PICINDEX: Valid array index.
*/

    private:
// ==========================================================================

    PICINT*   content;         // Store for the integers (0..alloc-1).
    PICSIZE   alloc;           // Current allocation size.
    PICINDEX  lb;              // Current lower bound for the array.
    PICINDEX  ub;              // Current upper bound for the array.

// --------------------------------------------------------------------------
//      COPYING
//
//      copy adds all values from given array to the receiver. Memory
//      for the receiver array must be allocated beforehand.
//      allocate allocates the requested amount of array cells.
//      resize resizes the array to the given size. If allocate fails,
//      the array is null afterwards. If resize fails, the array remains
//      unchanged and boolFalse is returned.
//
//      Errors : MEMORY_ALLOCATION_ERROR

    void    copy     (const PicIntArray&);
    void    allocate (PICSIZE amount);
    PICBOOL resize   (PICSIZE toSize);

};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicIntArray::PicIntArray ()
 : lb (firstPos), ub (firstPos - 1), alloc(0), content(NULLPTR) {}

inline PICBOOL PicIntArray::is (const PicIntArray& right) const
       { return this == &right; }

inline PICBOOL PicIntArray::isEmpty () const { return ub < lb; }
inline PICBOOL PicIntArray::isNull () const { return !content; }

inline PICINT& PicIntArray::operator [] (PICINDEX index)
       { return content[index - lb]; }
inline const PICINT& PicIntArray::operator [] (PICINDEX index) const
       { return content[index - lb]; }

inline PICINT PicIntArray::first () const
       { return isEmpty() ? 0 : content[0]; }
inline PICINT PicIntArray::last () const
       { return isEmpty() ? 0 : (*this)[ub]; }

inline const PICINT* PicIntArray::ints () const { return content; }

inline PICSIZE PicIntArray::size () const
       { return isEmpty() ? 0 : (ub - lb) + 1; }

inline PICINDEX PicIntArray::firstIndex () const
       { return isEmpty() ? outOfIndex : lb; }
inline PICINDEX PicIntArray::lastIndex () const
       { return isEmpty() ? outOfIndex - 1 : ub; }

inline PICINDEX PicIntArray::minIndex () const { return lb; }
inline PICSIZE  PicIntArray::maxSize  () const { return alloc; }

inline PICBOOL PicIntArray::inRange (PICINDEX anIndex) const
       { return (anIndex >= lb) && (anIndex <= ub); }
inline PICPOSITION PicIntArray::position (PICINDEX ind) const
       { return (ind - lb) + firstPos; }
inline PICINDEX PicIntArray::index (PICPOSITION pos) const
       { return (pos - firstPos) + lb; }

inline PICBOOL PicIntArray::contains (PICINT value) const
       { return indexOf(value) != outOfIndex; }

#endif
