// --------------------------------------------------------------------------
//  INTEGER SET
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PINTSET_HXX
#define PINTSET_HXX

#include <picbase.hxx>
class PicObject;
class BaseString;
class PicIterator;
struct PicIntSetData;

const PICUSHORT endOfList = USHRT_MAX;

/*:class ********************************************************************

Set is an unordered collection that doesn't allow duplicates. PicIntSet
is a dynamic set having integers as members, currently only positive
integers. Set has only associative and iterative access.

PicIntSet will grow as needed when new integers are added to it. This
means that every non-const method may raise MEMORY_ALLOCATION_ERROR.

PicIntSet provides efficient handling for integer members. Other
components (any PicObject) can be stored to PicSet.

****************************************************************************/

class PicIntSet
{

    public:
// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT
//:fun ----------------------------------------------------------------------

    PicIntSet ();
    PicIntSet (PICUSHORT, PICUSHORT, ...);
    PicIntSet (const PicIntSet&);
    PicIntSet (const PicObject&);

/*
PicIntSet can be created as :

    1. PicIntSet set;
    2. PicIntSet set (5,7,234,endOfList);
    3. PicIntSet set (anIntSet);
    4. PicIntSet set (collection);

1. Default constructor creates null set, which doesn't allocate memory.
2. Create set which contains list of integers. This list can contain
   any number of integers. The list ends with constant endOfList.
3. Copy constructor.
4. Create set from any collection which contains integers. The collection
   is iterated through and current element is added to the set.

:err MEMORY_ALLOCATION_ERROR: Can allocate space for set.
:err PicObject::typeMismatch: Collection component is not integer.
*/    

   ~PicIntSet ();
    PICCLASS()

//:fun ----------------------------------------------------------------------

    PicIntSet& operator = (const PicIntSet&);

/*
After copy both sets contain equal components.
*/

// --------------------------------------------------------------------------
//      COMPARISONS
//:fun ----------------------------------------------------------------------

    PICBOOL operator == (const PicIntSet&) const;
    PICBOOL is          (const PicIntSet&) const;

/*
Sets are equal if they contain equal members.
*/

// --------------------------------------------------------------------------
//      NULL SET
//:fun ----------------------------------------------------------------------

    PICBOOL isNull  () const;
    void    setNull ();

/*
Null set is considered empty. It does not allocate memory.
*/

// --------------------------------------------------------------------------
//      ADDING INTEGERS TO THE SET
//:fun ----------------------------------------------------------------------

    void add        (PICUSHORT);
    void add        (PICUSHORT, PICUSHORT, ...);
    void addBetween (PICUSHORT from, PICUSHORT to);

/*
Add the parameter(s) to the set. When using variable length argument
list the last argument must be endOfList.
addBetween adds all values between (and including) the limits to the
set.

:arg from: First member to add.
:arg to: If smaller than 'from' nothing is done.
:err MEMORY_ALLOCATION_ERROR: Can't resize the set.
*/

//:fun ----------------------------------------------------------------------

    void replace (PICUSHORT oldInt, PICUSHORT newInt);

/*
Replace the old integer value with new one if the old is found.
*/    

// --------------------------------------------------------------------------
//      REMOVING
//:fun ----------------------------------------------------------------------

    void remove        (PICUSHORT);
    void removeBetween (PICUSHORT from, PICUSHORT to);
    void setEmpty      ();

/*
Remove integer(s) from the set if they are members. 

:arg PICUSHORT: Member of the set. Should be positive integer.
:arg from: First member to remove. If greater than the greatest member
           nothing is done.
:arg to: If smaller than 'from' nothing is done.
*/

// --------------------------------------------------------------------------
//      ITERATOR
//:fun ----------------------------------------------------------------------

    PICBOOL next     (PICUSHORT&) const;
    PICBOOL previous (PICUSHORT&) const;

/*
You can iterate through the set as follows:

    PICUSHORT current = endOfList;
    while (intset.next(current))
        ...;

The elements appear in ascending or descending order.
*/

    void      startIteration (PicIterator&) const;
    PICBOOL   next           (PicIterator&) const;
    PICBOOL   previous       (PicIterator&) const;
    PICUSHORT current        (PicIterator&) const;

// --------------------------------------------------------------------------
//      SET OPERATIONS
//:fun ----------------------------------------------------------------------

    PicIntSet  operator +  (const PicIntSet&) const;
    PicIntSet  operator *  (const PicIntSet&) const;
    PicIntSet  operator -  (const PicIntSet&) const;

    PicIntSet& operator += (const PicIntSet&);
    PicIntSet& operator *= (const PicIntSet&);
    PicIntSet& operator -= (const PicIntSet&);

/*
Union (+) results in a set which contains components either in the
receiver or in the argument set.
Intersection (*) results in a set which contains components both in
the receiver and in the argument set.
Difference (-) results in a set which contains components in the
receiver but not in the argument set.

:err MEMORY_ALLOCATION_ERROR: Can't create/resize the set.
*/

// --------------------------------------------------------------------------
//      ATTRIBUTES
//:fun ----------------------------------------------------------------------

    PICBOOL   isEmpty  () const;    
    PICSIZE   size     () const;
    PICUSHORT smallest () const;
    PICUSHORT greatest () const;
    PICSIZE   maxSize  () const;

    PICBOOL   contains (PICUSHORT) const;

/*
Set attributes.

:ret isEmpty: boolTrue if the set is empty and has no members.
:ret size: Number of members in the set.
:ret smallest: Smallest integer in the set.
:ret greatest: Greatest integer in the set.
:ret maxSize: The current maximum number of members.
:ret contains: boolTrue if the argument is member of the set.
*/

    private:
// ==========================================================================
//   INSTANCE VARIABLES, PRIVATE METHODS AND FRIENDS    

    PicIntSetData* content;

};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicIntSet::PicIntSet () : content(NULLPTR) {}

inline PICBOOL PicIntSet::is (const PicIntSet& right) const
       { return this == &right; }

inline PICBOOL PicIntSet::isNull () const { return !content; }

inline PicIntSet PicIntSet::operator + (const PicIntSet& right) const
       { return (PicIntSet(*this) += right); }
inline PicIntSet PicIntSet::operator * (const PicIntSet& right) const
       { return (PicIntSet(*this) *= right); }
inline PicIntSet PicIntSet::operator - (const PicIntSet& right) const
       { return (PicIntSet(*this) -= right); }

#endif
