// --------------------------------------------------------------------------
//  ITERATOR
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PITERATE_HXX
#define PITERATE_HXX

#include <picbase.hxx>

// --------------------------------------------------------------------------
//      ITERATOR
/*:des ----------------------------------------------------------------------

    ITERATE           (Collection&, PicIterator&) iterateStatement
    ITERATE_BACKWARDS (Collection&, PicIterator&) iterateStatement

These are the standard iterator macros. They are followed by a statement
to be iterated (usually a compound statement in braces). For example:

    PicIterator i;
    ITERATE(anArray, i)
    {
        PicObject& obj = anArray.current(i);
        PICINDEX   ind = anArray.currentIndex(i);
        ...
    }

Iterating backwards is essential if you delete the traversed components
with purging: it ensures that components are not skipped.

:arg Collection:   Collection to be iterated. During iteration the
                   statement is executed once for every component in the
                   collection. If the collection is empty, the statement
                   is not executed. Inside the statement the current
                   component can be accessed with the method current
                   of the collection class.
:arg PicIterator&: Iterator object used in this iteration. In nested
                   iterations you must use different iterator objects.
                   Iterator object must be given as an argument to the
                   current-method.
*/

#define ITERATE(collection, iterator)            \
    for ((collection).startIteration (iterator); \
         (collection).next (iterator);)

#define ITERATE_PTR(collectionPtr,iterator)          \
    for ((collectionPtr)->startIteration (iterator); \
         (collectionPtr)->next (iterator);)

#define ITERATE_BACKWARDS(collection, iterator)  \
    for ((collection).startIteration (iterator); \
         (collection).previous (iterator);)

#define ITERATE_PTR_BACKWARDS(collectionPtr,iterator) \
    for ((collectionPtr)->startIteration (iterator);  \
         (collectionPtr)->previous (iterator);)

/*:class ====================================================================
PicIterator objects hold the current iterator index during iteration. In
practice the index is a long integer. The class whose object is being
iterated interprets the meaning of the index.

For the user of the iterator PicIterator has only one method: a constructor.

    class PicIterator
    {
        PicIterator ();
    };

Implementers of class-specific iterators use another class called
PicIteratorImp.
*/

class PicIterator
{

    public:

    PicIterator () : current (0) {}

    PICBOOL operator == (const PicIterator& i) { return current == i.current; }
    PICBOOL operator != (const PicIterator& i) { return current != i.current; }

    protected:

    PICLONG current;
};

#endif
