// --------------------------------------------------------------------------
//  INDEXED MAP OBJECTS
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PIXMAP_HXX
#define PIXMAP_HXX

#include <passarr.hxx>
#include <pcollect.hxx>

/*:class *** INDEXED MAP OBJECT *********************************************

PicIxMapObject is an abstract base class for two collections of index-value
pairs: PicIxMap and PicSortedIxMap. They are both implemented by using
array-like storage.

****************************************************************************/

class PicIxMapObject : public PicKeyIxCollect
{
    public:
// --------------------------------------------------------------------------
//:fun  ASSIGNMENT

    virtual PicObject&      operator = (const PicObject&);
            PicIxMapObject& operator = (const PicIxMapObject&);

/*
Assignment copies indexed collections.
*/

// --------------------------------------------------------------------------

    PICABSOBJECTCLASS(PicIxMapObject)

// --------------------------------------------------------------------------
//:fun  COMPARISONS

    virtual PICBOOL operator == (const PicObject&)      const;
            PICBOOL isSubsetOf  (const PicIxMapObject&) const;

/*
These methods perform a deep comparison of two maps. Maps are equal if they
contain equal components at equal indices in the same order. A subset map
contains equal components at equal indices in any order.
*/

// --------------------------------------------------------------------------
//:fun  NULL

    virtual PICBOOL isNull  () const;
    virtual void    setNull ();

/*
Null map is considered empty. It does not allocate dynamic memory.
*/

// --------------------------------------------------------------------------

    virtual PicObject& access  (const PicIterator&);
    virtual PicObject* replace (PicIterator&, PicObject*);

    virtual PICBOOL purge (PicIterator&, PICBOOL inOrder = boolTrue) =0;
    virtual PICBOOL allocate (PICSIZE forComponents);

    virtual PICINDEX indexAt   (const PicIterator&)     const;
    virtual PICBOOL  findIndex (PICINDEX, PicIterator&) const;
    virtual PICBOOL  insert    (PICINDEX, PicIterator&) =0;

    virtual void deleteAll ();

    virtual PICBOOL next     (PicIterator&) const;
    virtual PICBOOL previous (PicIterator&) const;

// --------------------------------------------------------------------------
//:fun  PURGE

    void purgeOnOrder ();

/*
Purges all empty locations. Because order is preserved, this method
doesn't modify the semantics of the map. It only saves space.
*/

// --------------------------------------------------------------------------
//:fun  ASKING ATTRIBUTES

    virtual PICSIZE size    () const;
            PICBOOL isEmpty () const;

    PICINDEX minIndex () const;
    PICINDEX maxIndex () const;

/*
:ret minIndex: The smallest index in use or outOfIndex if the map is empty.
:ret maxIndex: The greatest index in use or outOfIndex if the map is empty.
*/

    protected:
// ==========================================================================
//:fun  CONSTRUCTORS

    PicIxMapObject ();
    PicIxMapObject (PICSIZE initialSize);
    PicIxMapObject (const PicIxMapObject&);
    PicIxMapObject (const PicAssocArray&);

    const PicAssocArray& assocArray () const;
          PicAssocArray& assocArray ();

    PICPOSITION location (PICINDEX) const;
    PICPOSITION location (const PicObject&) const;

/*
PicIxMapObject has the following protected constructors:

    1. PicIxMapObject map;
    2. PicIxMapObject map(10);
    3. PicIxMapObject map(anIxMap);
    4. PicIxMapObject map(anAssocArray);

1. Creates a null map that doesn't allocate memory.
2. Creates an empty map with the given preallocated size.
3. Creates a map with components copied from the argument.
4. Creates a map out of an association array (the internal representation).
   The internal array can be accessed with the method assocArray. The
   location-methods search for an association by index or by value. If
   not found, outOfRange is returned.
*/

    PICBOOL equals (const PicIxMapObject&) const;

    private:
// ==========================================================================
//      INSTANCE VARIABLES

    PicAssocArray  assocs;

};


/*:class *** INDEXED MAP ****************************************************

PicIxMap is a collection of index-value pairs having a fixed order
independent of the numeric order of the indices.

****************************************************************************/

class PicIxMap : public PicIxMapObject
{
    public:
// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, COPY & ASSIGNMENT

    PicIxMap ();
    PicIxMap (PICSIZE initialSize);
    PicIxMap (const PicIxMap&);
    PicIxMap (const PicIxCollect& fromIxCollect);

/*
PicIxMaps can be created as follows:

    1. PicIxMap map;
    2. PicIxMap map(10);
    3. PicIxMap map(anIxMap);
    4. PicIxMap map(anIxCollect);

1. Creates a null map that doesn't allocate memory.
2. Creates an empty map with the given preallocated size.
3. Creates a map with components copied from the argument.
4. Creates a map with components copied from an indexed collection.
*/

//:fun ----------------------------------------------------------------------

   ~PicIxMap ();

/*
Destructor deletes the map and all its components.
*/

    PicObject& operator = (const PicObject&);
    PicIxMap&  operator = (const PicIxMap&);

    PICOBJECTCLASS(PicIxMap)

    PICBOOL operator == (const PicObject&) const;
    PICBOOL operator == (const PicIxMap&)  const;

// --------------------------------------------------------------------------

    virtual PICBOOL insert (PICINDEX, PicIterator&);
    virtual PICBOOL purge  (PicIterator&, PICBOOL inOrder = boolTrue);

//:fun ----------------------------------------------------------------------

    PICBOOL putPtrAt (PICINDEX index, PicObject*);

/*
Puts the given object at the given index. If there is a component with
the index, it is deleted. Returns true if the operation is successful.
*/

//:fun ----------------------------------------------------------------------

    void purge ();

/*
Purges all empty locations without preserving the order of components.
*/

// --------------------------------------------------------------------------
//:fun  SORT

    void sort (PICIXSORTER* = PicIxCollect::ascending);

/*
Map is sorted according to a compare function of the following form:

    PICSHORT compare (PICINDEX i1, PICINDEX i2)
    {
        // return 1 if i1 sorts after i2
        //        0 if i1 == i2
        //       -1 if i1 sorts before i2
    }
*/

// --------------------------------------------------------------------------

    PicIxMap& operator *= (const PicCollection&);
    PicIxMap& operator -= (const PicCollection&);
    PicIxMap& operator &= (const PicIxCollect&);
    PicIxMap& operator |= (const PicIxCollect&);
    PicIxMap& operator += (const PicIxCollect&);

    protected:
// ==========================================================================
//  PROTECTED CONSTRUCTOR

    PicIxMap (const PicAssocArray&);

};


/*:class *** SORTED INDEXED MAP *********************************************

PicSortedIxMap is a collection of index-value pairs that is kept in the
numeric order of the indices.

****************************************************************************/

class PicSortedIxMap : public PicIxMapObject
{

    public:
// --------------------------------------------------------------------------
//      CONSTRUCTORS, COPY & ASSIGNMENT
//:fun ----------------------------------------------------------------------

    PicSortedIxMap ();
    PicSortedIxMap (PICSIZE initialSize);
    PicSortedIxMap (const PicSortedIxMap&);
    PicSortedIxMap (const PicIxCollect& ixCollect);

/*
PicSortedIxMaps can be created as follows:

    1. PicSortedIxMap map;
    2. PicSortedIxMap map(10);
    3. PicSortedIxMap map(aSortedIxMap);
    4. PicSortedIxMap map(anIxCollect);

1. Creates a null map that doesn't allocate memory.
2. Creates an empty map with the given preallocated size.
3. Creates a map with components copied from the argument.
4. Creates a map with components copied from an indexed collection.

By default the sorting order is PicIxCollect::ascending.
*/

//:fun ----------------------------------------------------------------------

   ~PicSortedIxMap ();

/*
Destructor deletes the map and all its components.
*/

    PicObject&      operator = (const PicObject&);
    PicSortedIxMap& operator = (const PicSortedIxMap&);

    PICOBJECTCLASS(PicSortedIxMap)

    PICBOOL operator == (const PicObject&) const;
    PICBOOL operator == (const PicSortedIxMap&) const;

// --------------------------------------------------------------------------
//:fun  SEARCH

    virtual PICBOOL findIndex (PICINDEX, PicIterator&) const;
    virtual PICBOOL insert    (PICINDEX, PicIterator&);
    virtual PICBOOL purge     (PicIterator&, PICBOOL inOrder = boolTrue);

/*
Insert and purge are implemented to preserve the order of indices.
Searching is implemented as a binary search of ordered components.
*/

// --------------------------------------------------------------------------
//:fun  MODIFYING COMPONENTS

    PICBOOL putPtrAt (PICINDEX index, PicObject*);

/*
Puts the given object at the given index. If there is a component with
the index, it is deleted. Returns true if the operation is successful.
*/

// --------------------------------------------------------------------------
//:fun  SORT

    void sort (PICIXSORTER* = PicIxCollect::ascending);
    virtual PICBOOL isSorted () const;

/*
This method sets the sorting criterion for the map. The map is sorted
according to the new order and it retains the order from that point on.

:ret isSorted: boolTrue.
*/

// --------------------------------------------------------------------------

    PicSortedIxMap& operator *= (const PicCollection&);
    PicSortedIxMap& operator -= (const PicCollection&);
    PicSortedIxMap& operator &= (const PicIxCollect&);
    PicSortedIxMap& operator |= (const PicIxCollect&);
    PicSortedIxMap& operator += (const PicIxCollect&);

    protected:
// ==========================================================================
//  PROTECTED CONSTRUCTOR

    PicSortedIxMap (const PicAssocArray&);

    private:
// ==========================================================================
//      INSTANCE VARIABLES

    PICIXSORTER* order;

};


// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicIxMapObject::PicIxMapObject () : PicKeyIxCollect(), assocs() {}
inline PicIxMapObject::PicIxMapObject (PICSIZE initSize) :
    PicKeyIxCollect(), assocs(initSize) {}
inline PicIxMapObject::PicIxMapObject (const PicIxMapObject& map) :
    PicKeyIxCollect(), assocs(map.assocs) {}
inline PicIxMapObject::PicIxMapObject (const PicAssocArray& assocarr) :
    PicKeyIxCollect(), assocs(assocarr) {}

inline PicIxMapObject& PicIxMapObject::operator = (const PicIxMapObject& map)
       { assocs = map.assocs; return *this; }
inline PICBOOL PicIxMapObject::equals (const PicIxMapObject& map) const
       { return assocs.equals(map.assocs); }
inline PICBOOL PicIxMapObject::isSubsetOf (const PicIxMapObject& map) const
       { return map.assocs.contains(this->assocs); }

inline void PicIxMapObject::purgeOnOrder () { assocs.purge(boolTrue); }
inline PICBOOL PicIxMapObject::allocate (PICSIZE size)
       { return assocs.put(size); }

inline PICBOOL PicIxMapObject::isEmpty () const
       { return assocs.isEmpty(); }

inline const PicAssocArray& PicIxMapObject::assocArray () const
       { return assocs; }
inline PicAssocArray& PicIxMapObject::assocArray ()
       { return assocs; }

inline PICPOSITION PicIxMapObject::location (PICINDEX index) const
       { return assocs.locationOf(index); }
inline PICPOSITION PicIxMapObject::location (const PicObject& obj) const
       { return assocs.locationOf(obj); }

// --------------------------------------------------------------------------

inline PicObject& PicIxMap::operator = (const PicObject& obj)
       { return PicIxMapObject::operator = (obj); }
inline PicIxMap& PicIxMap::operator = (const PicIxMap& map)
       { PicIxMapObject::operator = (map); return *this; }

inline PICBOOL PicIxMap::operator == (const PicObject& obj) const
       { return PicIxMapObject::operator == (obj); }
inline PICBOOL PicIxMap::operator == (const PicIxMap& map) const
       { return PicIxMapObject::equals(map); }

inline void PicIxMap::purge () { assocArray().purge(boolFalse); }

inline PicIxMap& PicIxMap::operator *= (const PicCollection& cl)
       { return (PicIxMap&)PicCollection::operator *= (cl); }
inline PicIxMap& PicIxMap::operator -= (const PicCollection& cl)
       { return (PicIxMap&)PicCollection::operator -= (cl); }
inline PicIxMap& PicIxMap::operator &= (const PicIxCollect& ixcl)
       { return (PicIxMap&)PicKeyIxCollect::operator &= (ixcl); }
inline PicIxMap& PicIxMap::operator |= (const PicIxCollect& ixcl)
       { return (PicIxMap&)PicKeyIxCollect::operator |= (ixcl); }
inline PicIxMap& PicIxMap::operator += (const PicIxCollect& ixcl)
       { return (PicIxMap&)PicKeyIxCollect::operator += (ixcl); }

// --------------------------------------------------------------------------

inline PicObject& PicSortedIxMap::operator = (const PicObject& obj)
       { return PicIxMapObject::operator = (obj); }
inline PicSortedIxMap& PicSortedIxMap::operator = (const PicSortedIxMap& map)
       { PicKeyIxCollect::operator=(map); return *this; }

inline PICBOOL PicSortedIxMap::operator == (const PicObject& obj) const
       { return PicIxMapObject::operator == (obj); }
inline PICBOOL PicSortedIxMap::operator == (const PicSortedIxMap& map) const
       { return PicIxMapObject::equals(map); }

inline PicSortedIxMap& PicSortedIxMap::operator *= (const PicCollection& cl)
       { return (PicSortedIxMap&)PicCollection::operator *= (cl); }
inline PicSortedIxMap& PicSortedIxMap::operator -= (const PicCollection& cl)
       { return (PicSortedIxMap&)PicCollection::operator -= (cl); }
inline PicSortedIxMap& PicSortedIxMap::operator &= (const PicIxCollect& ixcl)
       { return (PicSortedIxMap&)PicKeyIxCollect::operator &= (ixcl); }
inline PicSortedIxMap& PicSortedIxMap::operator |= (const PicIxCollect& ixcl)
       { return (PicSortedIxMap&)PicKeyIxCollect::operator |= (ixcl); }
inline PicSortedIxMap& PicSortedIxMap::operator += (const PicIxCollect& ixcl)
       { return (PicSortedIxMap&)PicKeyIxCollect::operator += (ixcl); }

inline PICBOOL PicSortedIxMap::isSorted () const { return boolTrue; }

#endif
