// --------------------------------------------------------------------------
//  NAME-OBJECT ASSOCIATION ARRAY
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PICNAMASSOCARRAY_HXX
#define PICNAMASSOCARRAY_HXX

#include <passarr.hxx>
#include <pcollect.hxx>

struct NameBuffer;                // The name buffer data structure
typedef PICUSHORT NAMEENTRYSIZE;  // Size of an entry in the name buffer

/*:class *******************************************************

PicNamAssocArray is array of string + object associations.

****************************************************************************/

class PicNamAssocArray : public PicAssocArray
{
    public:
// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT

    PicNamAssocArray ();
    PicNamAssocArray (PICSIZE initSize);
    PicNamAssocArray (const PicNamAssocArray&);

   ~PicNamAssocArray ();

    PicNamAssocArray& operator = (const PicNamAssocArray&);

    PICCLASS()

// --------------------------------------------------------------------------
//      COMPARE
//
//      Returns boolTrue if the two arrays contain same objects with
//      same names in the same order.

    PICBOOL equals (const PicNamAssocArray&) const;

// --------------------------------------------------------------------------
//      ADD, REPLACE & REMOVE
//:fun

    PICPOSITION putName (const char* name);
    PICPOSITION putName (const char* name, PICNMSORTER* order);

    void deleteAll ();
    void removeAll ();

    void purge (PICPOSITION, PICBOOL inOrder);
    void purge (PICBOOL inOrder);

/*
putName         Finds/puts the given name in the name buffer.
                Location of the name is returned or outOfIndex
                if the operation failed. Second variant puts
                according to sorted order.
deleteAll       Empties the array and removes all the names.
*/

// --------------------------------------------------------------------------
//      SEARCH

    const char* nameOf (const PicObject&) const;
    PICBOOL contains (const PicNamAssocArray&) const;

// --------------------------------------------------------------------------
//      SORT

    void sort (PICNMSORTER*);

// --------------------------------------------------------------------------
//      NAME BUFFER
//:fun

    void purgeNameBuffer () const;

/*
This method removes fragmentation of the name buffer. The same
operation is performed implicitly, when the name buffer is resized.
Note that the purging is performed only, if the fragmentation of
the buffer is significant.
*/

// --------------------------------------------------------------------------
//      INTERNAL LOCATIONS
//:fun

    const char* nameAt         (PICPOSITION) const;
    PICPOSITION    locationOfName (const char* name) const;

    PICBOOL searchName (
        const char* name,
        PICPOSITION&   loc,
        PICNMSORTER*   order) const;

/*
nameAt          Returns the name at the given location. Location
                must be inside array bounds.
locationOfName  Returns the location of the object with the
                given name or outOfIndex if not found.
searchName      Searches a name assuming the array is sorted
                according to alphabetical order. Returns true
                if a match is found, otherwise stops at the
                correct place for argument name.
*/

// --------------------------------------------------------------------------

          PicAssoc& operator [] (PICPOSITION);
    const PicAssoc& operator [] (PICPOSITION) const;

    private:
// ==========================================================================
//      INSTANCE VARIABLES AND PRIVATE METHODS
//
//      copyNamesFrom       Copies the name buffer from the argument.
//                          The buffer is purged. If there is an existing
//                          name buffer it is assumed to be large enough.
//                          The argument is assumed to have a name buffer.
//      copyNamesTo         Copies the name buffer to the argument array's
//                          given buffer. The buffer is purged. The desti-
//                          nation buffer is assumed to be large enough.
//      allocateNameBuffer  Allocates the name buffer to the given size.
//                          Old buffer is retained if it is large enough.
//                          Contents are not copied. Returns boolTrue if
//                          successful.
//      resizeNameBuffer    Recreates the name buffer to allow the given
//                          amount of additional space. Returns boolTrue
//                          if successful.
//      addName             Adds a new name of the given size. Resizes
//                          name buffer if necessary. Returns the offset
//                          of the name in the buffer or zero, if not
//                          successful.

    void    copyNamesFrom      (const PicNamAssocArray&);
    void    copyNamesTo        (PicNamAssocArray& array, NameBuffer*) const;
    PICBOOL allocateNameBuffer (PICSIZE newSize);
    PICBOOL resizeNameBuffer   (PICSIZE increment);

    PICOFFSET addName (NAMEENTRYSIZE nameSize, const char* name);

    NameBuffer* names;   // Association indexes refer to this buffer

};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicNamAssocArray::PicNamAssocArray () :
    PicAssocArray(), names(NULLPTR) {}

inline PicNamAssocArray::PicNamAssocArray (PICSIZE initSize) :
    PicAssocArray(initSize), names(NULLPTR) {}

// --------------------------------------------------------------------------

inline PicAssoc& PicNamAssocArray::operator [] (PICPOSITION pos)
       { return ((PicAssocArray*)this)->at(pos); }
inline const PicAssoc& PicNamAssocArray::operator [] (PICPOSITION pos) const
       { return ((PicAssocArray*)this)->at(pos); }
#endif
