// --------------------------------------------------------------------------
//  NAMED MAP OBJECTS
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PICNMMAP_HXX
#define PICNMMAP_HXX

#include <pcollect.hxx>
#include <pnassarr.hxx>

/*:class *** NAMED MAP OBJECT ***********************************************

PicNmMapObject is an abstract base class for two collections of name-value
pairs: PicNmMap and PicSortedNmMap. They are both implemented by using
array-like storage.

****************************************************************************/

class PicNmMapObject : public PicNmCollect
{
    public:
// --------------------------------------------------------------------------
//:fun  ASSIGNMENT

    virtual PicObject&      operator = (const PicObject&);
            PicNmMapObject& operator = (const PicNmMapObject&);

/*
Assignment copies named collections.
*/

    PICABSOBJECTCLASS(PicNmMapObject)

// --------------------------------------------------------------------------
//:fun  COMPARISONS

    virtual PICBOOL operator == (const PicObject&)      const;
            PICBOOL isSubsetOf  (const PicNmMapObject&) const;

/*
These methods perform a deep comparison of two maps. Maps are equal if they
contain equal components at equal names in the same order. A subset map
contains equal components at equal names in any order.
*/

// --------------------------------------------------------------------------
//:fun  NULL

    virtual PICBOOL isNull  () const;
    virtual void    setNull ();

/*
Null map is considered empty. It does not allocate dynamic memory.
*/

// --------------------------------------------------------------------------

    virtual PicObject& access  (const PicIterator&);
    virtual PicObject* replace (PicIterator&, PicObject*);

    virtual PICBOOL purge (PicIterator&, PICBOOL inOrder = boolTrue) =0;
    virtual PICBOOL allocate (PICSIZE forComponents);

    virtual const char* nameAt   (const PicIterator&) const;
    virtual PICBOOL        findName (const char*, PicIterator&) const;
    virtual PICBOOL        insert   (const char* name, PicIterator&) =0;

    virtual void deleteAll ();

    virtual PICBOOL next     (PicIterator&) const;
    virtual PICBOOL previous (PicIterator&) const;

// --------------------------------------------------------------------------
//:fun  PURGE

    void purgeOnOrder ();

/*
Purges all empty locations. Because order is preserved, this method
doesn't modify the semantics of the map. It only saves space.
*/

    virtual PICSIZE size    () const;
            PICBOOL isEmpty () const;

// --------------------------------------------------------------------------
//:fun

    const char* minName () const;
    const char* maxName () const;

/*
:ret minName: The smallest name in use or NULLPTR if the map is empty.
:ret maxName: The greatest name in use or NULLPTR if the map is empty.
*/

    protected:
// ==========================================================================
//:fun  CONSTRUCTORS, COPY & ASSIGNMENT

    PicNmMapObject ();
    PicNmMapObject (PICSIZE initSize);
    PicNmMapObject (const PicNmMapObject&);
    PicNmMapObject (const PicNamAssocArray&);

    const PicNamAssocArray& assocArray () const;
          PicNamAssocArray& assocArray ();

    PICPOSITION location (const char*    name) const;
    PICPOSITION location (const BaseString& name) const;
    PICPOSITION location (const PicObject&  obj)  const;

/*
PicNmMapObject has the following protected constructors:

    1. PicNmMapObject map;
    2. PicNmMapObject map(10);
    3. PicNmMapObject map(nmMap);
    4. PicNmMapObject map(namAssocArray);

1. Creates a null map that doesn't allocate memory.
2. Creates an empty map with the given preallocated size.
3. Creates a map with components copied from the argument.
4. Creates a map out of an association array (the internal representation).
   The internal array can be accessed with the method assocArray. The
   location-methods search for an association by name or by value. If
   not found, outOfRange is returned.
*/

    PICBOOL equals (const PicNmMapObject&) const;

    private:
// ==========================================================================
//      INSTANCE VARIABLES

    PicNamAssocArray assocs;
};


/*:class *** NAMED MAP ******************************************************

PicNmMap is a collection of name-value pairs having a fixed order
independent of the alphabetic order of the names.

****************************************************************************/

class PicNmMap : public PicNmMapObject
{

    public:
// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, COPY & ASSIGNMENT

    PicNmMap ();
    PicNmMap (PICSIZE initSize);
    PicNmMap (const PicNmMap&);
    PicNmMap (const PicNmCollect& fromNmCollect);

/*
PicNmMaps can be created as follows:

    1. PicNmMap map;
    2. PicNmMap map(10);
    3. PicNmMap map(nmMap);
    4. PicNmMap map(nmCollect);

1. Creates a null map that doesn't allocate memory.
2. Creates an empty map with the given preallocated size.
3. Creates a map with components copied from the argument.
4. Creates a map with components copied from a named collection.
*/

//:fun ----------------------------------------------------------------------

   ~PicNmMap ();

/*
Destructor deletes the map and all it's components.
*/

//:fun ----------------------------------------------------------------------

    PicObject& operator = (const PicObject&);
    PicNmMap&  operator = (const PicNmMap&);

    PICOBJECTCLASS(PicNmMap)

    PICBOOL operator == (const PicObject&) const;
    PICBOOL operator == (const PicNmMap&)  const;

// --------------------------------------------------------------------------

    virtual PICBOOL purge  (PicIterator&, PICBOOL inOrder = boolTrue);
    virtual PICBOOL insert (const char* name, PicIterator&);

//:fun ----------------------------------------------------------------------

    PICBOOL putPtrAt (const char*    name, PicObject*);
    PICBOOL putPtrAt (const BaseString& name, PicObject*);

/*
Puts the given object at the given name. If there is a component with
the name, it is deleted. Returns true if the operation is successful.
*/

//:fun ----------------------------------------------------------------------

    void purge ();

/*
Purges all empty locations without preserving the order of components.
*/

// --------------------------------------------------------------------------
//:fun  SORT

    void sort (PICNMSORTER* = PicNmCollect::ascending);

/*
Map is sorted according to a compare function of the following form:

    PICSHORT compare (const BaseString& n1, const BaseString& n2)
    {
        // return 1 if n1 sorts after n2
        //        0 if n1 == n2
        //       -1 if n1 sorts before n2
    }
*/

    PicNmMap& operator *= (const PicCollection&);
    PicNmMap& operator -= (const PicCollection&);
    PicNmMap& operator &= (const PicNmCollect&);
    PicNmMap& operator |= (const PicNmCollect&);
    PicNmMap& operator += (const PicNmCollect&);

    protected:
// ==========================================================================
//  PROTECTED CONSTRUCTOR

    PicNmMap (const PicNamAssocArray&);

    private:
// ==========================================================================

    PicNamAssocArray assocs;
};


/*:class *** SORTED NAMED MAP ***********************************************

PicSortedNmMap is a collection of name-value pairs that is kept in the
alphabetic order of the names.

****************************************************************************/

class PicSortedNmMap : public PicNmMapObject
{
    public:
// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, COPY & ASSIGNMENT

    PicSortedNmMap ();
    PicSortedNmMap (PICSIZE initialSize);
    PicSortedNmMap (const PicSortedNmMap&);
    PicSortedNmMap (const PicNmCollect& nmCollect);

/*
PicSortedNmMaps can be created as follows:

    1. PicSortedNmMap map;
    2. PicSortedNmMap map(10);
    3. PicSortedNmMap map(aSortedNmMap);
    4. PicSortedNmMap map(anNmCollect);

1. Creates a null map that doesn't allocate memory.
2. Creates an empty map with the given preallocated size.
3. Creates a map with components copied from the argument.
4. Creates a map with components copied from a named collection.

By default the sorting order is PicNmCollect::ascending.
*/

//:fun ----------------------------------------------------------------------

   ~PicSortedNmMap ();

/*
Destructor deletes the map and all its components.
*/

    PicObject&      operator = (const PicObject&);
    PicSortedNmMap& operator = (const PicSortedNmMap&);

    PICOBJECTCLASS(PicSortedNmMap)

    PICBOOL operator == (const PicObject&) const;
    PICBOOL operator == (const PicSortedNmMap&) const;

// --------------------------------------------------------------------------
//:fun  ACCESSING COMPONENTS

    virtual PICBOOL findName (const char* name, PicIterator&) const;
    virtual PICBOOL insert   (const char* name, PicIterator&);
    virtual PICBOOL purge    (PicIterator&, PICBOOL inOrder = boolTrue);

/*
Insert and purge are implemented to preserve the order of names.
Searching is implemented as a linear search of ordered components.
*/

// --------------------------------------------------------------------------
//:fun  MODIFYING COMPONENTS

    PICBOOL putPtrAt (const char*    name, PicObject*);
    PICBOOL putPtrAt (const BaseString& name, PicObject*);

/*
Puts the given object at the given name. If there is a component with
the name, it is deleted. Returns true if the operation is successful.
*/

// --------------------------------------------------------------------------
//:fun  SORT

    void sort (PICNMSORTER* = PicNmCollect::ascending);

    virtual PICBOOL isSorted () const;

/*
This method sets the sorting criterion for the map. The map is sorted
according to the new order and it retains the order from that point on.

:ret isSorted: boolTrue.
*/

    PicSortedNmMap& operator *= (const PicCollection&);
    PicSortedNmMap& operator -= (const PicCollection&);
    PicSortedNmMap& operator &= (const PicNmCollect&);
    PicSortedNmMap& operator |= (const PicNmCollect&);
    PicSortedNmMap& operator += (const PicNmCollect&);

    protected:
// ==========================================================================
//  PROTECTED CONSTRUCTOR

    PicSortedNmMap (const PicNamAssocArray&);

    private:
// ==========================================================================

    PICNMSORTER* order;
};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicNmMapObject::PicNmMapObject () : PicNmCollect(), assocs() {}
inline PicNmMapObject::PicNmMapObject (PICSIZE initSize) :
    PicNmCollect(), assocs(initSize) {}
inline PicNmMapObject::PicNmMapObject (const PicNmMapObject& map) :
    PicNmCollect(), assocs(map.assocs) {}
inline PicNmMapObject::PicNmMapObject (const PicNamAssocArray& assocarr) :
    PicNmCollect(), assocs(assocarr) {}

inline PicNmMapObject& PicNmMapObject::operator = (const PicNmMapObject& map)
       { assocs = map.assocs; return *this; }
inline PICBOOL PicNmMapObject::equals (const PicNmMapObject& map) const
       { return assocs.equals(map.assocs); }
inline PICBOOL PicNmMapObject::isSubsetOf (const PicNmMapObject& map) const
       { return map.assocs.contains(this->assocs); }

inline void PicNmMapObject::purgeOnOrder () { assocs.purge(boolTrue); }
inline PICBOOL PicNmMapObject::allocate (PICSIZE size)
       { return assocs.put(size); }

inline PICBOOL PicNmMapObject::isEmpty () const { return assocs.isEmpty(); }

inline const PicNamAssocArray& PicNmMapObject::assocArray () const
       { return assocs; }
inline PicNamAssocArray& PicNmMapObject::assocArray ()
       { return assocs; }

inline PICPOSITION PicNmMapObject::location (const char* name) const
       { return assocs.locationOfName(name); }
inline PICPOSITION PicNmMapObject::location (const BaseString& name) const
       { return assocs.locationOfName(name.chars()); }
inline PICPOSITION PicNmMapObject::location (const PicObject& obj) const
       { return assocs.locationOf(obj); }

// --------------------------------------------------------------------------

inline PicObject& PicNmMap::operator = (const PicObject& obj)
       { return PicNmMapObject::operator=(obj); }
inline PicNmMap& PicNmMap::operator = (const PicNmMap& map)
       { PicNmMapObject::operator=(map); return *this; }

inline PICBOOL PicNmMap::operator == (const PicObject& obj) const
       { return PicNmMapObject::operator == (obj); }
inline PICBOOL PicNmMap::operator == (const PicNmMap& map) const
       { return PicNmMapObject::equals(map); }

inline PICBOOL PicNmMap::putPtrAt (const BaseString& name, PicObject* objPtr)
       { return putPtrAt(name.chars(), objPtr); }

inline void PicNmMap::sort (PICNMSORTER* pfSorter)
       { assocArray().sort(pfSorter); }

inline void PicNmMap::purge () { assocs.purge(boolFalse); }

inline PicNmMap& PicNmMap::operator *= (const PicCollection& cl)
       { return (PicNmMap&)PicCollection::operator *= (cl); }
inline PicNmMap& PicNmMap::operator -= (const PicCollection& cl)
       { return (PicNmMap&)PicCollection::operator -= (cl); }
inline PicNmMap& PicNmMap::operator &= (const PicNmCollect& nmcl)
       { return (PicNmMap&)PicNmCollect::operator &= (nmcl); }
inline PicNmMap& PicNmMap::operator |= (const PicNmCollect& nmcl)
       { return (PicNmMap&)PicNmCollect::operator |= (nmcl); }
inline PicNmMap& PicNmMap::operator += (const PicNmCollect& nmcl)
       { return (PicNmMap&)PicNmCollect::operator += (nmcl); }

// --------------------------------------------------------------------------

inline PicObject& PicSortedNmMap::operator = (const PicObject& obj)
       { return PicNmMapObject::operator=(obj); }
inline PicSortedNmMap& PicSortedNmMap::operator = (const PicSortedNmMap& map)
       { PicNmCollect::operator=(map); return *this; }

inline PICBOOL PicSortedNmMap::operator == (const PicObject& obj) const
       { return PicNmMapObject::operator == (obj); }
inline PICBOOL PicSortedNmMap::operator == (const PicSortedNmMap& map) const
       { return PicNmMapObject::equals(map); }

inline PICBOOL PicSortedNmMap::putPtrAt (const BaseString& name, PicObject* objPtr)
       { return putPtrAt(name.chars(), objPtr); }

inline void PicSortedNmMap::sort (PICNMSORTER* pfSorter)
       { assocArray().sort(pfSorter); order = pfSorter; }

inline PicSortedNmMap& PicSortedNmMap::operator *= (const PicCollection& cl)
       { return (PicSortedNmMap&)PicCollection::operator *= (cl); }
inline PicSortedNmMap& PicSortedNmMap::operator -= (const PicCollection& cl)
       { return (PicSortedNmMap&)PicCollection::operator -= (cl); }
inline PicSortedNmMap& PicSortedNmMap::operator &= (const PicNmCollect& nmcl)
       { return (PicSortedNmMap&)PicNmCollect::operator &= (nmcl); }
inline PicSortedNmMap& PicSortedNmMap::operator |= (const PicNmCollect& nmcl)
       { return (PicSortedNmMap&)PicNmCollect::operator |= (nmcl); }
inline PicSortedNmMap& PicSortedNmMap::operator += (const PicNmCollect& nmcl)
       { return (PicSortedNmMap&)PicNmCollect::operator += (nmcl); }

inline PICBOOL PicSortedNmMap::isSorted () const { return boolTrue; }

#endif
