// --------------------------------------------------------------------------
//  SEMAPHORE
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PSEMAPH_HXX
#define PSEMAPH_HXX

#include <bstring.hxx>

/*:class *** SEMAPHORE **************************************************

Semaphore is an object that is used to control access to limited resources
in multi-user environment. While the resource is in use (semaphore is set)
other users must wait it to become available (current user releases it by
calling clear-method).
PicSemaphore is used to control access inside a single process. 

****************************************************************************/

class PicSemaphore
{
    public:

//:fun ----------------------------------------------------------------------

    PicSemaphore ();
   ~PicSemaphore ();

/*
Destructor closes the semaphore. The last connection also destroys the
semaphore.

:err error: Semaphore creation error.
*/

    PICCLASS()

    PICERRORCODES { 
        ok,
        error
    };

// --------------------------------------------------------------------------
//:fun  SEMAPHORE ACCESS

    PICBOOL set   ();
    PICBOOL clear ();

/*
These methods change state of the semaphore. Set is used to reserve
semaphore. It's an atomic operation that first waits semaphore to become
free before setting it. Clear is used to release previously reserved
semaphore.
*/

    private:
// ==========================================================================
//  INSTANCE VARIABLES

};


/*:class *** CRITICAL SECTION ***********************************************

Critical section can be protected by a semaphore as follows:

    static PicSemaphore sem;

    {
        PicCritSec critSec (sem);
        ... // critical events
        // destructor releases the semaphore
    }

Using PicCritSec is preferable to using the semaphore directly as
the compiler takes care of releasing the lock.

****************************************************************************/

class PicCritSec
{
    public:

    PicCritSec (PicSemaphore&);
   ~PicCritSec ();

    private:

    PicSemaphore& sem;
};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicCritSec::PicCritSec (PicSemaphore& s) : sem(s) { s.set(); }
inline PicCritSec::~PicCritSec () { sem.clear(); }

#endif
