// --------------------------------------------------------------------------
//  STREAM BASE CLASS
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PSTREAM_HXX
#define PSTREAM_HXX

#include <pvirtobj.hxx>
#include <picerror.erh>
class BaseString;

// --------------------------------------------------------------------------
//      STREAM CLASS INFORMATION

#define PICSTREAMCLASS(className) \
    PICANYCLASS(className, PicVirtClass, PicStream)

#define MAKE_PICSTREAMCLASS(className,base) \
    MAKE_PICANYCLASS(className, PicVirtClass, base)


/*:class *** STREAM *********************************************************

PicStream is the base class of PIC stream hierarchy. A stream is a
communication channel for binary data. It can be written to or read
from, depending on how it was opened. Opening operations are defined
only in the derived classes.

Stream I/O classes work in cooperation with streams: they write and/or
read data of a specific format. For example, PicText handles textual
data.

Individual stream implementations may provide buffering and other 
schemes to improve performance etc.

Derived stream classes should use the following macros for class
identification:

    PICSTREAMCLASS(className)
    MAKE_PICSTREAMCLASS(className, baseName)

****************************************************************************/

class PicStream : public PicVirtObject
{
    public:

    PICSTREAMCLASS(PicStream)

    PICERRORCODES { ok,
        readError     = ERR_STREAM_READ,
        writeError    = ERR_STREAM_WRITE,
        positionError = ERR_STREAM_POSITION
    };

// --------------------------------------------------------------------------
//:fun  DESTRUCTOR

    virtual ~PicStream ();

    const PicVirtClass& id () const;

    static PicStream& null ();

/*
Derived destructors clean up the stream (flush data, close files...).
The null stream eats all output and provides no input.
*/

// --------------------------------------------------------------------------
//:fun  METHODS

    virtual PICSIZE writeFrom (PicStream&, PICSIZE bufSize = 512);

/*
This method reads all data from the given stream and writes it to the
receiver stream. Streams don't have to be of same type. Use this to copy
content of one stream to another.

:arg PicStream&: Stream from which data is read. After operation this
                 stream is empty.
:arg PICSIZE:    Size of the buffer to be used in transfer.
:ret PICSIZE:    Number of bytes actually transferred.
:err writeError: General write error. Derived classes should give more
                 descriptive error description.
:err readError:  General read error. Derived classes should give more
                 descriptive error description.
*/

//:fun ----------------------------------------------------------------------

    virtual PICSIZE write (const void* from, PICSIZE length);

/*
This method should write data from the given buffer to the stream.
The default implementation raises writeError.

:arg from:        Pointer to the data buffer.
:arg length:      Number of bytes to write. Buffer must contain at least
                  this amount bytes.
:ret PICSIZE:     Number of bytes actually written. It may differ from
                  the requested amount in error cases.
:err writeError:  General write error. Derived classes should give more
                  descriptive error description.
*/

//:fun ----------------------------------------------------------------------

    virtual PICSIZE read (void* to, PICSIZE length);

/*
This method should read a requested amount bytes from the stream and
copy them to a user-allocated buffer. The default implementation raises
readError.

:arg to:     Pointer to the buffer to hold the data. It must be large enough
             to hold at least the requested amount of bytes.
:arg length: Number of bytes to read.
:ret PICSIZE: Number of bytes actually read. It may differ from the
              requested amount, if the stream becomes empty.
:err readError: General read error. Derived classes should give more
                descriptive error description.
*/

// ---------------------------------------------------------------------

    enum PicStreamPosition { start, current, end };

//:fun ----------------------------------------------------------------------

    virtual void move   (PICLONG, PicStreamPosition = current);
            void rewind ();

    PICULONG movePointer (
        PICULONG          pointer,
        PICULONG          size,
        PICLONG           move,
        PicStreamPosition startAt);

/*
:list Method
:item move: This method should move the stream pointer forward or backward,
      relative to the given starting position, if possible. The default
      implementation moves the pointer forward from the current position by
      reading and discarding data; if any other moving is requested,
      positionError is raised.
:item rewind: Moves the pointer to the beginning of the stream.
:item movePointer: This is a helper method for implementing pointer moving
      when the stream pointer and size are known. The return value is the
      new pointer value. No errors are raised; if the limits (0..size) are
      exceeded, the pointer is set to the start or one past end.
:end

:arg PICLONG:           The amount of bytes the pointer should be moved.
                        If moving extends beyond the limits of the stream,
                        the pointer should be set to the beginning or one
                        past the end and no error raised.
:arg PicStreamPosition: The starting point of moving: PicStream::start,
                        PicStream::current or PicStream::end.
:arg pointer:           The current stream pointer.
:arg size:              The current size of the stream.
:ret PICULONG:          The new value of stream pointer (within 0..size).

:err positionError:     The stream pointer cannot be moved.
*/

//:fun ----------------------------------------------------------------------

    virtual char* readAhead (PICSIZE&);
    virtual void  moveAhead (PICSIZE);

/*
These methods access a read-ahead buffer. Every stream should be able to
be read through these methods.

readAhead returns a pointer to the read-ahead buffer, after taking care
that the buffer is not empty. If there is nothing more to be read, a
NULLPTR is returned. After the call the PICSIZE argument tells how many
unread bytes the buffer holds. Do not access more than that! Note: the
returned pointer is non-const so that you can temporarily alter the buffer,
but always remember to restore the changes! The buffer may be the stream
itself.

moveAhead moves the current position forward by the given amount. If
moving extends beyond the size of the read-ahead buffer, the stream
pointer is moved also. This method should not alter the PicStream::ended
status, but readAhead instead. If the stream pointer is moved with
PicStream::move, the read-ahead buffer should be flushed.

Important: use either read or readAhead to access a stream, but never both.
They are not necessarily synchronized.

The default implementations raise readError. When implementing buffering
for a stream, see PicBufferStream
*/

//:enum ---------------------------------------------------------------------

    enum
    {
        writeable    = 0x01,
        readable     = 0x02,
        positionable = 0x04,
        ended        = 0x08
    };

/*
This enumeration specifies the status bits of streams:

:list Status bit
:item writeable:    The stream has been prepared for writing. However, it
                    doesn't guarantee that writing will actually succeed.
:item readable:     The stream has been prepared for reading. However, it
                    doesn't guarantee that reading will actually succeed.
:item positionable: The stream pointer can be moved freely. If a stream is
                    not positionable, only moving forward from the current
                    position is possible.
:item ended:        This bit is set if an input stream has ended or if the
                    stream is not readable.
:end

You ask the status bits with the method PicStream::status.
*/

//:fun ----------------------------------------------------------------------

    virtual PICSTATUS status () const =0;

    PICBOOL endOfStream () const;

/*
The status method accesses the current status of the stream. The
endOfStream method returns whether the PicStream::ended status bit
is set.
*/

//:fun ----------------------------------------------------------------------

    virtual BaseString name () const =0;

/*
All streams should have name identifying the stream. The name may be
given by the programmer or it may be derived from the contents and the
status of the stream.

:ret BaseString: Name of the stream.
*/

    protected:
// ==========================================================================
//   PROTECTED CONSTUCTOR (to make compiler happier ;-)

    PicStream () {}

};

/*:class *** CHAIN STREAM ***************************************************

PicChainStream is used to connect several output streams together. In this
way you can write same data to several streams with one write operation.
Chain stream doesn't handle output itself but delegates writing to output
stream associated with it. Chaining is done by linking several chain
streams together. Each chain stream in the chain contains link to next
chain stream to which it passes data for writing. Reading is not defined
for a chain.

****************************************************************************/

class PicChainStream : public PicStream
{
    public:

    PICSTREAMCLASS(PicChainStream)

// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT

    PicChainStream (PicStream&);
    PicChainStream (PicChainStream&, PicStream&);

/*
The first constructor creates chain with only output stream connected to
it, the second one also creates link to given chain stream. When created
chain stream becomes the head of the chain.

:arg PicStream&: Output stream associated with the chain stream.
:arg PicChainStream&: Chain to which created object is linked.
*/

// --------------------------------------------------------------------------
//:fun  METHODS

    virtual PICSIZE write (const void* from, PICSIZE length);

/*
This method writes data to output stream associated with the chain stream.
It also passes data to next stream in the chain for writing.

:arg from:    Pointer to buffer from where data is copied to the stream.
:arg length:  Size of the buffer.
:ret PICSIZE: Number of bytes actually written to the first stream in the
              chain.
*/

//:fun ----------------------------------------------------------------------

    virtual PICSTATUS status () const;

/*
Chain stream is writeable and positionable, if any of the output streams
in the chain are. Chain stream is never readable.
*/

//:fun ----------------------------------------------------------------------

    virtual BaseString name () const;

/*
Name of a chain stream is a combination of the output stream names.

:ret BaseString: Chain stream name.
*/

    private:
// ==========================================================================

    PicChainStream* chain;
    PicStream*      stream;

    DENY_COPYING_FOR(PicChainStream)

};


/*:class *** BUFFERED STREAM ************************************************

PicBufferStream implements a read-ahead buffer. When you need such,
do the following:

    - derive from PicBufferStream
    - flush the buffer when the stream is closed
    - flush the buffer when the stream pointer is moved
    - status: stream has not ended if there are still bytes in the buffer

****************************************************************************/

class PicBufferStream : public PicStream
{
    public:

    PICSTREAMCLASS(PicBufferStream)

// --------------------------------------------------------------------------
//:fun  METHODS

             PicBufferStream ();
    virtual ~PicBufferStream ();

    virtual char* readAhead (PICSIZE&);
    virtual void     moveAhead (PICSIZE);

    PICSIZE bytesInBuffer () const;
    void    flushBuffer   ();

    virtual PICSTATUS  status () const =0;
    virtual BaseString name   () const =0;

/*
:list Method
:item constructor:   The constructor doesn't allocate the buffer; that is
                     done in the readAhead method.
:item destructor:    The destructor frees the buffer.
:item readAhead:     Implements the read-ahead buffering: allocates a buffer
                     of 512 bytes and fills it from the stream.
:item moveAhead:     Implements the moving of buffer pointer.
:item bytesInBuffer: Returns the amount of unread bytes in the buffer.
:item flushBuffer:   Sets the buffer empty.
:end
*/

    private:
// =========================================================================

    char*      buffer;         // Buffer data
    PICOFFSET  bufPtr;         // Current position in the buffer
    PICSIZE    bufSize;        // Amount of bytes in the buffer

    DENY_COPYING_FOR(PicBufferStream)

};

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline const PicVirtClass& PicStream::id () const { return virtId(); }

inline PICBOOL PicStream::endOfStream () const
       { return (status() & ended) != 0; }

inline void PicStream::rewind () { move(0, start); }

inline PicChainStream::PicChainStream (PicStream& stm) :
    PicStream(), chain(NULLPTR), stream(&stm) {}

inline PicChainStream::PicChainStream (PicChainStream& chn, PicStream& stm) :
    PicStream(), chain(&chn), stream(&stm) {}

inline PicBufferStream::PicBufferStream () :
    PicStream(), buffer(NULLPTR), bufPtr(0), bufSize(0) {}

inline PICSIZE PicBufferStream::bytesInBuffer () const
    { return bufSize - bufPtr; }

inline void PicBufferStream::flushBuffer () { bufPtr = bufSize; }

#endif
