// --------------------------------------------------------------------------
//  STRING STREAM
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PSTRSTM_HXX
#define PSTRSTM_HXX

#include <pstream.hxx>
#include <bstring.hxx>

/*:class *** INPUT STRING STREAM ********************************************

String stream is a combination of string and stream: stream features
provide a general protocol to pass string to many sources and powerful
string features make it feasible to access/modify the string.

String associated with the stream is handled by the user. Make sure that
the string is visible all the time it's associated with the stream. Same
string may have many readers/writers at the same time. All streams have
their own 'current position' where read and write operations take place.
Note that manipulating string stream as stream also affects next read and
write operation. 

There are two string streams: PicInStringStream only allows reading, while
PicStringStream allows both reading and writing.

****************************************************************************/

class PicInStringStream : public PicStream
{
    public:

    PICSTREAMCLASS(PicInStringStream)

// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT

    PicInStringStream (const BaseString&);
   ~PicInStringStream ();

/*
Constructor associates the given string to the string stream. The string
can only be read from.

:arg const BaseString&: String associated to the stream.
*/

//:fun ----------------------------------------------------------------------

    virtual PICSIZE read (void* to, PICSIZE length);

/*
This method reads length bytes from the stream at current position and
copies it to the given buffer. The current position is moved forward by
the size of buffer.
*/

//:fun ----------------------------------------------------------------------

    virtual char* readAhead (PICSIZE&);
    virtual void     moveAhead (PICSIZE);

/*
Buffered reading returns a pointer to the string itself.
*/

    virtual void move (
        PICLONG amount, 
        PicStreamPosition = PicStream::current);

//:fun ----------------------------------------------------------------------

    virtual PICSTATUS status () const;

/*
Input string stream is always readable and positionable.
*/

//:fun ----------------------------------------------------------------------

    virtual BaseString name () const;

/*
Name of a string stream displays some of the contents of the stream.
*/

// --------------------------------------------------------------------------
//:fun  STRING ACCESS

    const BaseString& string ();

    void setString (const BaseString&);

/*
These methods change/access the associated string. All direct changes
with string methods also have effect on stream operations. When the
string is changed, the stream is rewound.

:arg const BaseString&: New string associated with the stream.
:ret const BaseString&: Reference to associated stream string.
*/

    protected:
// ==========================================================================
// INSTANCE VARIABLES, PRIVATE METHODS AND FRIENDS

    BaseString    *stream;
    PICPOSITION    current; // Current position, next write/read takes
                            // place at this position. 0 = ended

    private:
// ==========================================================================

    DENY_COPYING_FOR(PicInStringStream)
};


/*:class *** STRING STREAM **************************************************

String stream is a combination of string and stream: stream features
provide a general protocol to pass string to many sources and powerful
string features make it feasible to access/modify the string.

String associated with the stream is handled by the user. Make sure that
the string is visible all the time it's associated with the stream. Same
string may have many readers/writers at the same time. All streams have
their own 'current position' where read and write operations take place.
Note that manipulating string stream as stream also affects next read and
write operation. 

There are two string streams: PicInStringStream only allows reading, while
PicStringStream allows both reading and writing.

****************************************************************************/

class PicStringStream : public PicInStringStream
{
    public:

    PICSTREAMCLASS(PicStringStream)

// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT

    PicStringStream (BaseString&);
   ~PicStringStream ();

/*
Constructor associates the given string to the string stream. The string
can be read from or written to.

:arg BaseString&: String associated to the stream.
*/

//:fun ----------------------------------------------------------------------

    virtual PICSIZE write (const void* from, PICSIZE length);

/*
This method writes length bytes from the given buffer to the stream at
current position. The current position is moved forward by the size of
buffer. If the current position is in the middle of the string, the new
data is inserted rather than overwritten.
*/

//:fun ----------------------------------------------------------------------

    virtual PICSTATUS status () const;

/*
String stream is always writeable, readable and positionable.
*/

// --------------------------------------------------------------------------
//:fun  STRING ACCESS

    BaseString& string ();

    void setString (BaseString&);

/*
These methods change/access the associated string. All direct changes
with string methods also have effect on stream operations. When the
string is changed, the stream is rewound.

:arg BaseString&: New string associated with the stream.
:ret BaseString&: Reference to associated stream string.
*/

    private:
// ==========================================================================

    DENY_COPYING_FOR(PicStringStream)
};


/*:class *** INPUT CHAR STREAM **********************************************

Character stream connects a stream to a fixed-size memory buffer, usually
a character string. User provides the buffer and the stream ensures that
its bounds are honored. If output overflows, it is simply cut off. Reading
traverses the whole buffer regardless of its contents.

There are two character streams: PicInCharStream only allows reading,
while PicCharStream allows both reading and writing.

****************************************************************************/

class PicInCharStream : public PicStream
{
    public:

    PICSTREAMCLASS(PicInCharStream)

// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT

    PicInCharStream (const char* buffer, PICSIZE length);

/*
Constructor associates the given buffer to the character stream. The
stream can only be read from.

:arg buffer: User-allocated buffer of read-only data.
:arg length: Length of the buffer (and length of the data also).
*/

//:fun ----------------------------------------------------------------------

    virtual PICSIZE read (void* to, PICSIZE length);

/*
This method reads length bytes from the stream at current position and
copies it to the given buffer. The current position is moved forward by
the size of buffer.
*/

//:fun ----------------------------------------------------------------------

    virtual char* readAhead (PICSIZE&);
    virtual void     moveAhead (PICSIZE);

/*
Buffered reading returns a pointer to the character buffer itself.
*/

    virtual void move (
        PICLONG amount, 
        PicStreamPosition = PicStream::current);

//:fun ----------------------------------------------------------------------

    virtual PICSTATUS status () const;

/*
Input character stream is always readable and positionable.
*/

//:fun ----------------------------------------------------------------------

    virtual BaseString name () const;

/*
Name of a character stream displays some of the contents of the buffer.
*/

// --------------------------------------------------------------------------
//:fun  BUFFER ACCESS

    const char* buffer     () const;
    PICSIZE        bufferSize () const;

    void setBuffer (const char*, PICSIZE length);

/*
These methods change/access the associated buffer. All direct changes
to the buffer also have effect on stream operations. When the buffer
is changed, the stream is rewound.

:arg const char*: New buffer associated with the stream.
:arg length:         Size of the new buffer.
:ret const char*: Reference to associated stream buffer.
:ret PICSIZE:        Current size of the buffer.
*/

    protected:
// ==========================================================================
//  INSTANCE VARIABLES, PRIVATE METHODS AND FRIENDS

    char*   stream;
    PICSIZE    stmSize;
    PICOFFSET  current; // Current position, next write/read takes
                        // place at this position. 0xffff = ended

    private:
// ==========================================================================

    DENY_COPYING_FOR(PicInCharStream)
};


/*:class *** CHAR STREAM ****************************************************

Character stream connects a stream to a fixed-size memory buffer, usually
a character string. User provides the buffer and the stream ensures that
its bounds are honored. If output overflows, it is simply cut off. Reading
traverses the whole buffer regardless of its contents.

There are two character streams: PicInCharStream only allows reading,
while PicCharStream allows both reading and writing.

****************************************************************************/

class PicCharStream : public PicInCharStream
{

    public:

    PICSTREAMCLASS(PicCharStream)

// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT

    PicCharStream (char* buffer, PICSIZE length);

/*
Constructor associates the given buffer to the character stream. The
stream can be read from or written to.

:arg buffer: User-allocated buffer of read-write data.
:arg length: Length of the buffer (and length of the data also).
*/

//:fun ----------------------------------------------------------------------

    virtual PICSIZE write (const void* from, PICSIZE length);

/*
This method writes length bytes from the given buffer to the stream at
current position. The current position is moved forward by the size of
buffer. Writing overwrites existing data. NOTE: terminating zero bytes
are not appended automatically. If you build zero-terminated strings
via PicCharStream, always end the writing with explicit '\0'.
*/

//:fun ----------------------------------------------------------------------

    virtual PICSTATUS status () const;

    PICBOOL full () const;

/*
Character stream is always writeable, readable and positionable.
When character stream is full, you cannot write more data.
*/

// --------------------------------------------------------------------------
//:fun  STRING ACCESS

    char* buffer ();

    void setBuffer (char*, PICSIZE length);

/*
These methods change/access the associated buffer. All direct changes
to the buffer also have effect on stream operations. When the buffer
is changed, the stream is rewound.

:arg char*: New buffer associated with the stream.
:arg length:   Size of the new buffer.
:ret char*: Reference to associated stream buffer.
*/

    private:
// ==========================================================================

    DENY_COPYING_FOR(PicCharStream)
};


// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline const BaseString& PicInStringStream::string () { return *stream; }

inline void PicInStringStream::setString (const BaseString& str)
       { stream = (BaseString*)&str; current = 1; }

inline PicStringStream::~PicStringStream () {}

inline BaseString& PicStringStream::string () { return *stream; }

inline void PicStringStream::setString (BaseString& str)
       { stream = &str; current = 1; }

// --------------------------------------------------------------------------

inline const char* PicInCharStream::buffer () const { return stream; }
inline PICSIZE PicInCharStream::bufferSize () const { return stmSize; }

inline void PicInCharStream::setBuffer (const char* buf, PICSIZE len)
       { stream = (char*)buf; stmSize = len; current = 0; }

inline PICBOOL PicCharStream::full () const { return stmSize <= current; }

inline char* PicCharStream::buffer () { return stream; }

inline void PicCharStream::setBuffer (char* buf, PICSIZE len)
       { stream = buf; stmSize = len; current = 0; }

#endif
