// --------------------------------------------------------------------------
//  TEXTUAL I/O
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PICTEXT_HXX
#define PICTEXT_HXX

#include <pstrstm.hxx>
class PicText;
class PicFmtText;
class BaseString;
class PicWrObjArray;

//:fun ----------------------------------------------------------------------

PicText&    picIn  ();
PicFmtText& picOut ();
PicFmtText& picErr ();
PicFmtText& picOut (const char* format);
PicFmtText& picErr (const char* format);

/*
These methods provide access to standard input/output devices.

:ret picIn:  Standard input.
:ret picOut: Standard output.
:ret picErr: Standard error.
*/

/*:class *** TEXT ***********************************************************

PicText operates on textual data using abstract stream for I/O operations.
Where PicStream sees data as sequence of bytes PicText views data as lines
separated by newline characters. PicText sets file stream in text mode
automatically.

****************************************************************************/

class PicText
{
    public:

    PICCLASS()

// --------------------------------------------------------------------------
//:fun  CONSTRUCTOR

    PicText (PicStream&);

/*
Constructor attaches a stream to the PicText instance. Input/output
operations are available only if the stream is prepared for input/output.
If the stream is a file stream, it is set to text mode.

:arg PicStream&: I/O stream. The stream must be prepared to handle either
                 reading or writing.
*/

// --------------------------------------------------------------------------
//:fun  INPUT

    PicText& rd   (char&);
    PicText& rd   (BaseString&, char separator, PICBOOL spc = boolTrue);
    PicText& rdln (BaseString&);
    PicText& rdc  (BaseString&, PICSIZE charCount);

/*
These methods read characters from the stream:

    - next character
    - next string using whitespace (by default) and a given character as
      a separator (separators are not included in the string)
    - next line (ending newline is not included)
    - given number of characters (or up to the next newline)

:arg char&:       Receives character from the stream. Returned character is
                  '\0' if at the end of text.
:arg BaseString&: String to receive the read characters. String is empty if
                  at the end of text or line was empty.
:arg separator:   Additional string separator.
:arg spc:         Determines whether whitespace is also used as a separator.
:arg PICSIZE:     Number of characters to read.

:ret PicText&: Reference to the receiver (allows chained calls).

:err PicStream::readError: General stream read error.
*/

//:fun ----------------------------------------------------------------------

    PICBOOL endOfText () const;

/*
This method checks if the last read-method tried to read from empty stream.

:ret endOfText: boolTrue if end of text was detected. Before the first
                read operation this returns boolFalse.
*/

// --------------------------------------------------------------------------
//:fun  OUTPUT

    PicText& wr (char);
    PicText& wr (const char*);
    PicText& wr (const BaseString&);

/*
Write text to the stream.

:arg char: Character to be written.
:arg const char*: String to be written.
:arg const BaseString&: String to be written.
:ret PicText&: Reference to the receiver.
:err PicStream::writeError: General stream write error.
*/

//:fun ----------------------------------------------------------------------

    PicText& wrln (const char*);
    PicText& wrln (const BaseString&);
    PicText& wrln ();

/*
Write text + newline to the stream.

:arg const char*: String to be written.
:arg const BaseString&: String to be written.
:ret PicText&: Reference to the receiver.
:err PicStream::writeError: General stream write error.
*/

//:fun ----------------------------------------------------------------------

    PicText& wrn (PICLONG);
    PicText& wrf (PICFLOAT);
    PicText& wrp (const void*);

/*
Write number to the stream.

:arg PICLONG: Long integer to be written.
:arg PICFLOAT: Floating-point number to be written.
:arg const void*: Address of a variable to be written. Address format is
                  segment:offset.
:ret PicText&: Reference to the receiver.
:err PicStream::writeError: General stream write error.
*/

// --------------------------------------------------------------------------
//:fun  STREAM

    const PicStream& stream () const;
          PicStream& stream ();

    void setStream (PicStream&);

/*
These methods access/change the stream associated with the receiver.

:arg PicStream&: New I/O stream. All buffers are reset before connection
                 is made.
:ret PicStream&: Associated stream. Do not change state of the stream.
*/

    private:
// ==========================================================================
//  INSTANCE VARIABLES, PRIVATE METHODS AND FRIENDS

    PicStream* theStream;       // Associated stream.

    DENY_COPYING_FOR(PicText)

};


/*:class *** TEXT FORMAT ****************************************************

PicTextFmt is a helper class to hold formatting information passed to
a type-specific writer function. It contains the following pieces of
information:

    stream      The stream where the text is to be written.

    format      Pointer to the beginning of the format string (after [).
                Be sure to support the following conventions:
                    '*' - placeholder for default image
                    '?' - placeholder for class/type name
                format must not be NULLPTR.

    indent      The desired indentation.

****************************************************************************/

class PicTextFmt
{
    public:

    PicStream&  stream;
    const char* format;
    PICSIZE     indent;

    PicTextFmt (PicStream& stream, const char* format, PICSIZE indent);

    void write (const char* data, PICSIZE size);

    PICOFFSET writeInteger (
        PICOFFSET      beginAt,
        PICULONG       value,
        const char* typeName,
        PICBOOL        negative = boolFalse);

    static PICOFFSET getInt (const char*, PICOFFSET beginAt, PICUSHORT&);

/*
:list Method
:item write:        Writes the given character data to the stream. Applies
                    indentation after each newline character.
:item writeInteger: Writes the given integer if there is a valid integer
                    format at the given offset. Moves the offset past the
                    format. In case of invalid format moves by one position.
                    If negative is specified, a minus sign is prepended.
:item getInt:       If there are decimal digits at the given offset, this
                    function reads the integer and stores it in the second
                    argument. Offset is moved past the number.
:end
*/

};

typedef PICOFFSET   PICWRFUNC    (const void*, PicTextFmt&, PICOFFSET);
typedef const void* PICWROBJFUNC (const void*, PICBOOL);


/*:class *** FORMATTED TEXT *************************************************

PicFmtText extends the output functionality of PicText to include picture
formats. It also provides keyword replacement.

The format strings contain arbitrary characters to be written to the stream
plus placeholders for picture formats:

    k@cf    k = a keyword (case-sensitive string). The keyword may be
                missing in which case the placeholder is anonymous.
                If a keyword is not supplied when using the format, the
                keyword and the following placeholders are written out
                as such.
            c = common string formatting (may be missing).
            f = class-specific formatting (required).

To output a @-sign write it twice (@@).

Common string formatting:

    <n      Left-justify to field of n characters.    @<6*   @<6|.*
    >n      Right-justify to field of n characters.   @>6*   @>6|0*
    =n      Center to field of n characters.          @=6*   @=6|-*

    <n|c  >n|c  =n|c    Pad with the character 'c' (default is space).

            If the replacement value is longer than the field,
            it is written in its full length.

    #n      Repeat n times.                                     @#4*
    t       Trim leading and trailing whitespace.               @t*
    u       Make uppercase.                                     @u*
    l       Make lowercase.                                     @l*
    p       Apply proper casing for names.                      @p*
    r       Reverse.                                            @r*
    sxxx    Take a substring. In the expressions (xxx), n is starting
            position and m is size of the substring. Plus sign implies
            counting from the beginning, minus sign counting from the end.
            Missing starting position is 1. Missing size means the rest
            of the string.
    sn+m    Take substring [n .. n + m - 1].                    @2+4*
    s+m     Take substring [1 .. m].                            @+6*
    sn+     Take substring [n .. size].                         @2+*
    sn-m    Take substring [size - n - m + 2 .. size - n + 1].  @2-4*
    s-m     Take substring [size - m + 1 .. size].              @-2*
    sn-     Take substring [1 .. size - n + 1].                 @4-*

Class-specific formats are enclosed in square brackets []. Inside the
brackets you can use placeholders for format components and literal text
enclosed in single quotes ('xxx'). To output a literal single quote write
it twice ('xxx''xxx'). Certain placeholders are commonly understood:

    *       Default image.                                @['xx'*'yy']
    ?       Class/type name.                              @['class: '?]

The following abbreviations can be used:

    *       [*]                                           @*
    ?       [?]                                           @?

Example uses of picture formats:

    text("Error CODE@*(@*) in file FILE@*(@*)")
        >>"CODE" << err.id()   << err.status()
        >>"FILE" << err.file() << err.line();
    text("Value 1@* should be of the class 2@?.") >>"1"<< value >>"2"<< obj;
    text("Thread id: @>5[h2].") << pinfo.tid;

More information about class-specific formats:

    operator<<(PicFmtText)      - numeric formats
    BaseDateTime::writeFmt      - date & time formats

****************************************************************************/

class PicFmtText : public PicText
{
    public:

    PICCLASS()

// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT

    PicFmtText (PicStream&);
   ~PicFmtText ();

/*
Constructor attaches a stream to the PicFmtText instance. Input/output
operations are available only if the stream is prepared for input/output.

The destructor doesn't close or destroy the stream.

:arg PicStream&: I/O stream. The stream must be prepared to handle either
                 reading or writing.
*/

//:fun ----------------------------------------------------------------------

    PicFmtText&    format (const char* format);
    const char* format () const;

    PicFmtText& operator () (const char* format);

    PICSIZE indent () const;
    void    indent (PICSIZE indentSize);

/*
The first format-method changes the format string (parenthesis-operator
is an abbreviation for this). The previous format, if any, is flushed to
the stream at this time. The PicFmtText is returned so that output can
follow.

The second format-method returns the current format string.

Indenting affects all output: the given amount of spaces are written
after each newline character.
*/

//:fun ----------------------------------------------------------------------

    friend PicFmtText& operator >> (PicFmtText&, const char* keyword);

/*
The operator >> starts keyword handling: all the objects to be written are
stored (by reference) from that point on. When the format is flushed, the
list of stored objects and related keywords is used to write the objects
in the desired order (defined by the format string). The format is flushed:

    * when a new format is given,
    * PicFmtText is destroyed or
    * explicitly when the special keyword PicFmtText::endFormat() is given.
*/

//:fun ----------------------------------------------------------------------

    friend PicFmtText& operator << (PicFmtText&, char);
    friend PicFmtText& operator << (PicFmtText&, const char*);
    friend PicFmtText& operator << (PicFmtText&, const BaseString&);
    friend PicFmtText& operator << (PicFmtText&, short int);
    friend PicFmtText& operator << (PicFmtText&, unsigned short int);
    friend PicFmtText& operator << (PicFmtText&, int);
    friend PicFmtText& operator << (PicFmtText&, unsigned int);
    friend PicFmtText& operator << (PicFmtText&, long int);
    friend PicFmtText& operator << (PicFmtText&, unsigned long int);
    friend PicFmtText& operator << (PicFmtText&, PICFLOAT);

/*
These operators write the given value to the stream in place of the next
placeholder in the format string. If no placeholder is found, the output
is appended (using default format).

The following formatting is available for integers, including char
interpreted as an ASCII value:

    xi      Integer with minimum of i digits (default 1). Negative sign
            is prepended as appropriate. 'x' gives the radix:

                d = decimal (the default)
                h = hexadecimal (using a..f)
                H = hexadecimal (using A..F)
                o = octal

Pointer format:

    Si      Segment part; i is any valid integer format.
    Oi      Offset part; i is any valid integer format.
    :       Literal colon.

    For example ['0x'Sh4:'0x'Oh4] gives the format '0xffff:0xffff'.
    Default pointer format (*) is [SH4:OH4].

The following formatting is available for real numbers:

    .d      Number of decimals for rounding (negative scale value).
    :d      Number of zeros left of decimal point for rounding
            (positive scale value).

    Examples:   ['$'.2]      56.099   -->  $56.10
                [:2]      45691.5     -->  45700.
                >8[.3'%']    45.12    -->   45.120%
                            167.76    -->  167.760%
                              7.7616  -->    7.762%

In addition of these the basic Pic classes each define their own operators
for formatted output.

*/

//:fun ----------------------------------------------------------------------

    PicFmtText& wrFormat (
        const void*   pobject,
        PICWRFUNC*    pfwrite,
        PICWROBJFUNC* pfwrobj = NULLPTR);

/*
This method is used when implementing the various '<<' operators. The first
argument is a pointer to the object to be written. This object must stay
alive as long as the current format is being processed. To guarantee that
you can provide a clone/clean-up function to copy the object.

The second argument is a pointer to a function that writes the object. The
function must be of the following form:

    PICOFFSET wrTYPE (
        const TYPE*       pobject,
        const PicTextFmt& format,
        PICOFFSET         beginAt);

    The arguments are the following:

        pobject   A pointer to the object to be written.
        format    The format specification. See PicTextFmt for more
                  details.
        beginAt   Index to the format string where to begin.

    The function must return the new index to the format string after the
    processed part of the format. The function is called several times
    if there are several parts in the format specification.

The third argument is a clone/clean-up function that is called:

    1. When the object is stored for later use - you can copy it to
       avoid a dangling pointer.
    2. When the whole format has been processed and the object will be
       needed no more - you can delete the copy.

The function takes the following form:

    TYPE* wrobjTYPE (TYPE* pobject, PICBOOL cleanUp);
    {
        if (cleanUp) { delete obj; return NULLPTR; }
        return new TYPE(*obj);
    }

    The arguments are the following:

        pobject   A pointer to the object to be written.
        cleanUp   False upon first call, true upon second.

    The function must return the object to be stored at the first call,
    whether the copy or the original. At the second call the return
    value is ignored.

    Clone/clean-up function is optional; it is called only when given.


The output operators (<<) usually take the following form:

    PicFmtText& operator << (PicFmtText& txt, const TYPE& object);
    {
        return txt.wrFormat(&object, (PICWRFUNC*)wrTYPE);
    }

Or with copying:

    PicFmtText& operator << (PicFmtText& txt, const TYPE& object);
    {
        return txt.wrFormat(&object,
            (PICWRFUNC*)wrTYPE, (PICWROBJFUNC*)wrobjTYPE);
    }

Important: never give the address of an object passed by value unless you
copy the object!
*/

    private:
// ==========================================================================
//  INSTANCE VARIABLES

    const char* fmt;
    PicWrObjArray* pending;
    PICSIZE        indentSize;

    DENY_COPYING_FOR(PicFmtText)

};


/*:des ----------------------------------------------------------------------

    ENDFMT

This special keyword explicitly ends the formatted output:

    fmtTxt("...") >>"kwd" << obj1 << obj2 >>"kwd2" << obj3  ENDFMT;

At the point of ENDFMT the whole format string is written to the stream
with the placeholders (@) replaced with the corresponding values. However,
if no keywords are used, the values are written immediately and ENDFMT only
writes the remaining portion of the format string.

Note: if there are temporary objects in the output expression, ENDFMT
ensures that they are written out before being destructed. For example:

    fmtTxt("String: @*.") << BaseString("string")  ENDFMT;
*/

#define ENDFMTMARK ((char*)0x0001)
#define ENDFMT >>ENDFMTMARK


/*:class *** STRING TEXT ****************************************************

PicStringText is a helper class to enable the use of the operators
of PicFmtText with BaseStrings. It combines a BaseString, an PicStringStream
and an PicFmtText. You use this class as follows:

    PicStringText string;
    string("format...") << value << value2 << value3;
    string("more text is appended...") << value4 << value5 ENDFMT;

You reset the PicStringText to begin from empty string as follows:

    string.reset("format...") << ...

****************************************************************************/

class PicStringText : public BaseString
{
    public:

    PICCLASS()

    PicStringText ();
   ~PicStringText ();

    PicFmtText& operator () (const char* format);
    PicFmtText& operator () ();
    PicFmtText& reset       (const char* format = NULLPTR);

    PicStringText& flush ();

    const PicFmtText& text () const;
          PicFmtText& text ();

    private:
// ==========================================================================
//  INSTANCE VARIABLES

    PicStringStream strStream;
    PicFmtText      fmtText;

    DENY_COPYING_FOR(PicStringText)

};


// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicText& PicText::wrln ()
       { return wr('\n'); }
inline PicText& PicText::wr(const BaseString& string)
       { theStream->write(string.chars(), string.size()); return *this; }
inline PicText& PicText::wrln(const char* string)
       { return wr(string).wr('\n'); }
inline PicText& PicText::wrln(const BaseString& string)
       { return wr(string).wr('\n'); }

inline PICBOOL PicText::endOfText() const
       { return theStream->endOfStream(); }

inline PicText& PicText::wrf (PICFLOAT real)
       { return wr(asString(real)); }
inline PicText& PicText::wrp (const void* pointer)
       { if (pointer) return wr(asString(0L)); else return wr(asString(0L)); }

inline const PicStream& PicText::stream () const
       { return *theStream; }
inline PicStream& PicText::stream ()
       { return *theStream; }
inline void PicText::setStream (PicStream& newStream)
       { theStream = &newStream; }

inline PicTextFmt::PicTextFmt (
    PicStream&     stm,
    const char* fmt,
    PICSIZE        ind)
    :
    stream (stm),
    format (fmt),
    indent (ind)
{}

inline PicFmtText& PicFmtText::format (const char* form)
{
    *this ENDFMT;
    fmt = form;
    return *this;
}

inline PicFmtText& PicFmtText::operator () (const char* form)
       { return format(form); }
inline const char* PicFmtText::format () const
       { return fmt; }

inline PICSIZE PicFmtText::indent () const
       { return indentSize; }
inline void PicFmtText::indent (PICSIZE indentSz)
       { indentSize = indentSz; }

inline PicFmtText& PicStringText::operator () (const char* format)
       { return fmtText(format); }
inline PicFmtText& PicStringText::operator () ()
       { return fmtText; }

inline const PicFmtText& PicStringText::text () const
       { return fmtText; }
inline PicFmtText& PicStringText::text ()
       { return fmtText; }

inline PicStringText& PicStringText::flush ()
       { fmtText ENDFMT; return *this; }

inline PicFmtText& picOut (const char* format)
       { return picOut()(format); }
inline PicFmtText& picErr (const char* format)
       { return picErr()(format); }

#endif
