// --------------------------------------------------------------------------
//  PicThread
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PTHREAD_HXX
#define PTHREAD_HXX

#include <picbase.hxx>
#include <psemaph.hxx>
#include <pvirtobj.hxx>

class PicThread;

 typedef int    PICTHREADID;
 typedef int    PICTHREADRESULT;
 enum PicPriority {
    normal
 };
 typedef void*  PICTHREADARG;

/*:class ====================================================================

An PicThreadPointer refers to an PicThread and you use it exactly as
you would use an PicThread*. With PicThreadPointers an PicThread
instance is, however, aware of that it is being pointed at and
therefore it is able to raise an (fatal) error if it is destructed
while there is still alive PicThreadPointers pointers pointing at it.
Strictly speaking the result is similar; the application will be
terminated but with PicThreadPointers you will know exactly why.
There is some overhead when the PicThreadPointer is constructed and
destructed but during use there is no overhead at all. 
*/

class PicThreadPointer
{
    public:
// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT
//:fun ----------------------------------------------------------------------

    PicThreadPointer(PicThread* thread);
    PicThreadPointer(const PicThreadPointer&);

    PicThreadPointer& operator = (const PicThreadPointer&);
    
    ~PicThreadPointer();

/*
The constructors, destructors and assignment of PicThreadPointer.
*/

//:fun ----------------------------------------------------------------------
    
    PicThread*  operator->();
                operator PICBOOL ();

/*
operator-> returns the thread the PicThreadPointer is pointing at. In
certain circumstances it can be NULLPTR.
operator PICBOOL returns boolTrue if the PicThreadPointer points at a
valid thread. Once it has been established it is a valid thread it is
safe to use. If the thread it is pointing at is destructed, a FATAL
error is raised.
E.g:

    PicErrorContext  context;
    PicThreadPointer tptr = PicThread::thread(1);

    if (!tptr)
    {
        // An error was raised in PicThread::thread();
        ...
    }
    else
    {
        // tptr is safe to use and pass around. If the thread it is
        // pointing at, is destructed it will be detected.
        ...
    }

*/

    private:
        
    PicThread *t;
};

/*:class ====================================================================

A process can have more than one thread. Each thread runs independently,
keeping its own register, stack, and execution state. Threads share the same
data (that is, they share the program's globally defined variables).
Although a thread can execute any part of the program, including a part
being executed by another thread, threads are typically used to execute
separate parts. This distributes the available CPU time and lets a program
carry out several tasks simultaneously--for example, loading a file and
prompting the user for input at the same time.
*/

class PicThread : public PicVirtObject
{
    public:

    PICANYCLASS(PicThread, PicVirtClass, PicThread)

    PICERRORCODES {  ok,
        alivePointers,
        alreadyRunning,
        finishedAndWaiting,
        noProcessSlots,
        notAPicThread,
        notSuspended,
        notRunning,
        runningThreadDestructed,
        threadDestructed
    };

/*
Error status codes of PicThread.
*/

//:enum ---------------------------------------------------------------------

    enum PicThreadState
    {
        created,
        idle,
        running,
        finished
    };

/*
A thread can be in four distinct states.

    created         The thread has been created but it has never been
                    started.
    idle            The thread is not running. It has run, has
                    finished and been waited for (possibly several times).
    running         The thread is executing in its thread function.
    finished        The thread has finished executing its thread function.
                    It will stay in this state until somebody waits for it.
                    Then it will switch to idle-state and can be restarted.
*/                        

// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT
//:fun ----------------------------------------------------------------------

    PicThread(PicPriority  priority = normal);
    PicThread(PicThread&);
    
    virtual ~PicThread();

    virtual PicThread& operator = (PicThread&);    

/*    
Default constructor creates a thread that has the same priority as
its parent. The priority can later be changed at any time. The copy
constructor creates a copy of the given thread only in the sence that
the priority class and level are the same. The thread has a unique id
and its 'tid' is 0 since it is not started, regardless of the state of
the given thread. 

The assignment operator behaves in the same way as the copy
constructor, i.e. the priority is copied and nothing else. If the
thread is running it continues to run (with the new priority), if it
is idle, it remains idle.

If a running thread or a thread to which there exists
PicThreadPointers is destructed, a FATAL error is raised.

:arg level                      The priority level within a particular
                                priority class.
:arg priorityClass              The priority class.

:err runningThreadDestructed    This fatal error is raised if an
                                alive and running thread is being
                                destructed.
:err alivePointers              This fatal error is raised if alive
                                pointers (PicThreadPointers) exist
                                when the thread is being destructed.
*/

// --------------------------------------------------------------------------
//      METHODS
//:fun-----------------------------------------------------------------------

    static PicThread*       self();
    static PicThreadPointer thread(PICUSHORT id);

    PICUSHORT references();

/*
Self returns a pointer to the calling thread.

Thread returns an PicThreadPointer that points to an PicThread if the
given id referred to an alive PicThread, or to a NULLPTR if the id
was invalid or if the PicThread has been destructed already. Once you
receive an PicThreadPointer, without errors being raised, it is safe
to use and pass around. A fatal error will, however, be raised if
there are alive PicThreadPointers when the corresponding PicThread is
destructed. 

References returns the number of PicThreadPointers that refer to this
thread.

Note: It the user's responsibility to ensure that there are no alive
      PicThreadPointers when the PicThread is destructed.

:err threadDestructed   The thread corresponding to the given id was
                        an PicThread but it has already been destructed.
:err notAnPicThread     The given id does not refer to an (alive or
                        dead) PicThread.
*/
        
//:fun ----------------------------------------------------------------------
        
    PICBOOL start
    (
        PICTHREADARG arg, 
        PICUSHORT stackSize = 4096
    );

    PICBOOL           wait();
    PICTHREADRESULT   result();

/*
The default size of the thread stack is 4K which should suffice for
most situations. Handle can refer to arbitrary data.
Start returns boolTrue if the thread is started successfully,
otherwise boolFalse.

If the thread has been created (but never started) or if it is idle,
wait() returns boolFalse, otherwise it waits for the thread to finish
(if it is running) and returns boolTrue. Once a thread has been
started it cannot be started again without being waited for.
 
Result() returns the value the threadFunction() returned.

    PicThread a;

    a.start();
    a.wait();

    PICTHREADRESULT result = a.result();            

      
:arg arg                A handle to arbitrary data.
:arg stackSize          The size of the stack.

:err noProcessSlots     The resources of OS are exhausted.
:err alreadyStarted     The thread has already been started and is
                        running.
*/                        

//:fun ----------------------------------------------------------------------
                
    PICBOOL suspend ();
    PICBOOL resume  ();

//:fun ----------------------------------------------------------------------

    void setPriority(const PicPriority level);
    PicPriority getPriority();

/*
Use these functions to set and get the priority of a thread.

:arg level      The priority level within a priority class.
*/
    
//:fun ----------------------------------------------------------------------
                    
    PICTHREADID         tid         ();
    PICUSHORT           id          () const;        
    PicThreadState      state       ();

    static PICTHREADID  currentTid  ();
    static PICUSHORT    currentId   ();
    static void         yield       ();    
    static void         sleep       (PICULONG milliSeconds);
    
/*
When a thread is constructed it is given a uniq id, that is never
reused during the lifetime of the running application. When a thread
is running it has a tid (thread identifier) which may be reused at a
later time. State returns the current state of the thread. The tid
and state may change at any time. CurrentTid() returns the tid  of
the calling thread. Sleep() causes the calling thread to sleep for
the specified amount of milliseconds and yield() causes the calling
thread to give up the remainder of its time-slice. Sleep(0) is
equivalent to yield().
*/

//:fun ----------------------------------------------------------------------
    
    PicThread(void* dummy);
/*
This constructor is used to automatically create thread object 
for the main thread. Don't use it yourself.
Dummy id needed to make this constructor differ from other
constructors. Nemos-problemos, this constructor is used only once.
*/

    protected:
// ==========================================================================
//      PROTECTED VARIABLES AND METHODS
//:fun ----------------------------------------------------------------------
    
    virtual PICTHREADRESULT threadFunction(PICTHREADARG arg);

    PicSemaphore        semMutex;

/*
ThreadFunction() is the actual thread function. It is called when the
thread is started. You must *not* exit the function in any other way
than by returning from it. The returned PICTHREADRESULT is made available
to other threads through the functions wait() and result().

Use semMutex to lock an instance from concurrent access. Beware of
calling functions that use the semaphore, from a function that already
has set the semaphore. You should not call another function of
PicThread while semMutex is locked.
E.g.
              
    Ok                          Not Ok (dead-lock will follow)
    --                          ------
    
    {                           {                     
        PicCriticalSection          PicCriticalSection
          section(semMutex);          section(semMutex); 
                                                      
        ...                         ...               
    }                               
                                    suspend();                              
    suspend();                  }

*/    

    private:
// ======================================================================
//      INSTANCE VARIABLES, PRIVATE METHODS AND FRIENDS

    void mark    ();
    void release ();
        
    static PICUSHORT  newId();

    friend PICTHREADRESULT threadFunc(PICTHREADARG);

    PicSemaphore    semReady;
    PICUSHORT       useCount;
    PICTHREADRESULT res;
    PicThreadState  tState;
    PicPriority     priority;
    PICUSHORT       tId;
    PICTHREADID     tTid;
   public:
    int             fildes[2]; 
    int             fildes2[2]; 
    friend class PicThreadPointer;

};

typedef PicThread::PicThreadState PicThreadState;

// --------------------------------------------------------------------------
//      INLINE DEFINITIONS
// --------------------------------------------------------------------------


inline PicThreadPointer::PicThreadPointer(PicThread* thread)
  : t(thread)
{
    if (t)
      t->mark();
}


inline PicThreadPointer::PicThreadPointer(const PicThreadPointer& p)
  : t(p.t)
{
    if (t)
      t->mark();
}


inline PicThreadPointer& PicThreadPointer::operator =
(
    const PicThreadPointer& p
)
{
    if (t)
      t->release();

    t = p.t;

    if (t)
      t->mark();

    return *this;
}


inline PicThreadPointer::~PicThreadPointer()
{
    if (t)
      t->release();
}


inline PicThread* PicThreadPointer::operator -> ()
{
    return t;
}


inline PicThreadPointer::operator PICBOOL ()
{
    return (t != NULLPTR);
}


inline void PicThread::yield()
{
    sleep(0L);
}

#endif
