// --------------------------------------------------------------------------
//  TIME OBJECT
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#ifndef PTIME_HXX
#define PTIME_HXX

#include <pobject.hxx>
#include <btime.hxx>

/*:class ********************************************************************

PicTime implements the basic time handling operations. Time values include
an hour, minute and second component. Null time is used to represent unknown
or missing value. At any moment after creation PicTime instance is either
null or represents a valid time of a day (i.e. within 00:00:00 and 23:59:59)

Time format can be selected with the PicTimeIniter class.

PicTime belongs to the kernel level 2 library. It has a counterpart BaseTime
in the level 1. Most date handling operations are defined for BaseTime and
inherited by PicTime.

****************************************************************************/

class PicTime : public PicPrimitive, public BaseTime
{
    public:

    PICOBJECTCLASS(PicTime)

// --------------------------------------------------------------------------
//:fun  CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT

    PicTime ();
    PicTime (PICHOURCOUNT plusInterval);
    PicTime (PICHOUR, PICMINUTE, PICSECOND);
    PicTime (
        const BaseString&,
        char delimiter);
    PicTime (const BaseString&);
    PicTime (const BaseTime&);
    PicTime (const PicTime&);

/*
PicTimes can be created in the following ways:

    1. PicTime a;
    2. PicTime b (now + 2);
    3. PicTime c (13, 30, 0);
    4. PicTime d ("13.30");
    5. PicTime e ("1:30:00 PM", picColon);
    6. PicTime f (aBaseTime);
    7. PicTime g (anPicTime);

1. Creates a null PicTime.
2. Creates an PicTime with the value 'after two hours' (now + 2 hours).
   This constructor uses the current system time.
3. Creates an PicTime with the value 13h 30m given as components.
4. Creates an PicTime with the value 13h 30m in default format.
5. Creates an PicTime with the value 13h 30m in a specified format.
6. Copies a BaseTime.
7. Copies another PicTime.

:err deformatError: The given string is not of the given format.
:err formatError:   The given string is not a valid time.
*/

//:fun ----------------------------------------------------------------------

    virtual PicObject& operator = (const PicObject&);
            PicTime&   operator = (const PicTime&);
            PicTime&   operator = (const BaseTime&);

/*
These methods copy the time value.

:arg PicObject&: Reference to an PicTime.
:ret PicTime&, PicObject&: Reference to the receiver.
:err typeMismatch: PicObject&-argument is not an PicTime.
*/

// --------------------------------------------------------------------------
//:fun  COMPARISONS

    virtual PICBOOL operator == (const PicObject&) const;
            PICBOOL operator != (const PicObject&) const;
    virtual PICBOOL operator <  (const PicObject&) const;

            PICBOOL operator == (const PicTime&)   const;
            PICBOOL operator != (const PicTime&)   const;
            PICBOOL operator <  (const PicTime&)   const;

            PICBOOL operator == (const BaseTime&)  const;
            PICBOOL operator != (const BaseTime&)  const;
            PICBOOL operator <  (const BaseTime&)  const;

/*
PicTimes are compared along the time axis. Later times are considered
greater.
Two null times are equal and null time is smaller than any other
non-null time.
*/

    PICBOOL is (const PicTime&)   const;
    PICBOOL is (const BaseTime&)  const;
    PICBOOL is (const PicObject&) const;

    virtual PICBOOL isNull  () const;
    virtual void    setNull ();

// --------------------------------------------------------------------------
//:fun  DEFAULT

    virtual PICBOOL isDefault  () const;
    virtual void    setDefault ();

/*
Default PicTime is null.
*/

// --------------------------------------------------------------------------
//:fun  TEXTUAL IMAGE

    virtual BaseString getImage () const;
    virtual BaseString getImage (const PicObject& format) const;

    BaseString getImage (
        PicTimeResolution,
        PicTimeStyle,
        char delimiter,
        char leader) const;

/*
These methods return the textual representation of the receiver. Default
format is used unless a specific format is given. For null time a null
string is returned.

Formatting object may contain the following components:

    {
        STYLE      = <FORMAT_HOURS_12 or FORMAT_HOURS_24>;
        RESOLUTION = <FORMAT_HOURS_ONLY or FORMAT_HOURS_TO_MINUTES or
                      FORMAT_HOURS_TO_SECONDS>;
        DELIMITER  = <char>;
        LEADER     = <char>;
        SIZE       = <PICSIZE>; // Field width
        JUSTIFY    = <FORMAT_LEFT or FORMAT_CENTER or FORMAT_RIGHT>;
        FILLER     = <char>;
    }

Constants are defined in picbase.erh.

:arg delimiter: Delimiter of the time components, e.g. ':' (13:30).
:arg leader:    Leader for one-character numbers, e.g. '0' (02:07).
*/

//:fun ----------------------------------------------------------------------

    virtual void setImage (const BaseString&);
    virtual void setImage (const PicObject&  format, const BaseString&);

    void setImage (const BaseString&, char delimiter);

/*
These methods set the value of the receiver according to the given textual
time image. A time format is needed for parsing. If no format is given, the
default is used. A formatting object may contain the following components:

    {
        STYLE      = <FORMAT_HOURS_12 or FORMAT_HOURS_24>;
        RESOLUTION = <FORMAT_HOURS_ONLY or FORMAT_HOURS_TO_MINUTES or
                      FORMAT_HOURS_TO_SECONDS>;
        DELIMITER  = <char>;
        LEADER     = <char>;
    }

Constants are defined in picbase.erh.

:err deformatError: The given string is not of the given format.
:err formatError:   The given string is not a valid time.
*/

//:fun ----------------------------------------------------------------------

    friend PicFmtText& operator << (PicFmtText&, const PicTime&);

    virtual PICOFFSET writeFmt (PicTextFmt&, PICOFFSET) const;

/*
Implementation of formatted output.

*/

// --------------------------------------------------------------------------
//:fun  INCREMENT/DECREMENT

    virtual PICBOOL increment (PICSHORT);

/*
Implementation of increment/decrement operation.
*/

// --------------------------------------------------------------------------
//:fun  UNDEFINED OPERATIONS

    virtual PICSHORT       asInt     () const;
    virtual char           asChar    () const;
    virtual PICBOOL        asBoolean () const;
    virtual const char*    asChars   () const;

/*
These raise PicObject::undefinedOperation when called.
*/

    private:
// ==========================================================================
//:fun  PRIVATE METHODS

    virtual PICBOOL setTypedValue (const PicObject& constraints);

/*
Checks the current value of the receiver for PicTime-specific type
definition constraints. The constraint object must be an indexed
dictionary; it may contain the following constraints:

    {
        NULL_IS_CURRENT = <hours>;
        // Null value is converted to the current time + <hours>.
    }

The constant is defined in picbase.erh.
*/

};


// --------------------------------------------------------------------------
//  INLINE DEFINITIONS
// --------------------------------------------------------------------------

inline PicTime& PicTime::operator = (const PicTime& time)
       { BaseTime::operator = (time); return *this; }
inline PicTime& PicTime::operator = (const BaseTime& time)
       { BaseTime::operator = (time); return *this; }

inline PICBOOL PicTime::isNull  () const { return BaseTime::isNull(); }
inline void    PicTime::setNull ()       { BaseTime::setNull(); }

inline PICBOOL PicTime::isDefault  () const { return BaseTime::isNull(); }
inline void    PicTime::setDefault ()       { BaseTime::setNull(); }

inline BaseString PicTime::getImage () const
       { return BaseTime::getImage(); }

inline BaseString PicTime::getImage (
    PicTimeResolution resolution,
    PicTimeStyle      style,
    char           delimiter,
    char           leader) const
    { return BaseTime::getImage(resolution, style, delimiter, leader); }

inline void PicTime::setImage (const BaseString& image)
       { BaseTime::setImage(image); }
inline void PicTime::setImage (const BaseString& image, char delimiter)
       { BaseTime::setImage(image, delimiter); }

inline PICBOOL PicTime::operator != (const PicObject& obj) const
       { return !(PicTime::operator == (obj)); }
inline PICBOOL PicTime::operator == (const PicTime& time) const
       { return BaseTime::operator == (time); }
inline PICBOOL PicTime::operator != (const PicTime& time) const
       { return BaseTime::operator != (time); }
inline PICBOOL PicTime::operator < (const PicTime& time) const
       { return BaseTime::operator < (time); }
inline PICBOOL PicTime::operator == (const BaseTime& time) const
       { return BaseTime::operator == (time); }
inline PICBOOL PicTime::operator != (const BaseTime& time) const
       { return BaseTime::operator != (time); }
inline PICBOOL PicTime::operator < (const BaseTime& time) const
       { return BaseTime::operator < (time); }

inline PICBOOL PicTime::is (const PicTime& time) const
       { return this == &time; }
inline PICBOOL PicTime::is (const BaseTime& time) const
       { return ((BaseTime*)this) == &time; }
inline PICBOOL PicTime::is (const PicObject& obj) const
       { return this == &obj; }

inline PicFmtText& operator << (PicFmtText& txt, const PicTime& obj)
       { return operator << (txt, (const PicObject&)obj); }

inline PICBOOL PicTime::increment (PICSHORT inc)
       { addHours(inc); return boolTrue; }

inline PICBOOL PicTime::asBoolean () const
       { undefined("asBoolean"); return boolFalse; }
inline PICSHORT PicTime::asInt () const
       { undefined("asInt"); return 0; }
inline char PicTime::asChar () const
       { undefined("asChar"); return '\0'; }
inline const char* PicTime::asChars () const
       { undefined("asChars"); return NULLPTR; }

#endif
