// --------------------------------------------------------------------------
//  BaseDate implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>
}

#include <bdatetim.hxx>
#include <bstring.hxx>
#include <bstrarr.hxx>
#include <picerror.hxx>
#include <pdefault.hxx>

MAKE_PICCLASS(BaseDate)

ERROR_MESSAGES(BaseDate)
PARAMSTR(nullInstance, "Operation on Null instance")
PARAMSTR(deformatError, "String '@*' is not of the given format")
PARAMSTR(formatError, "String '@*' is not a valid date")
PARAMSTR(illegalComponent, "Illegal date component")
PARAMSTR(dateIllegal, "Illegal date")
PARAMSTR(tooEarly, "Date @* is too early")
PARAMSTR(tooLate, "Date @* is too late")
END_PARAM_STRINGS

// --------------------------------------------------------------------------
//      VALIDATION FUNCTIONS
// --------------------------------------------------------------------------

static const PICDAY daysInMonth[2][13] =
{
    { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
    { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }
};

inline PICBOOL isLeap (PICYEAR aYear)
{
    return (aYear % 4 == 0 && (aYear % 100 != 0 || aYear % 400 == 0));
}

inline PICBOOL yearIsLegal (PICLONG aYear)
{
    return (aYear >= minYear && aYear <= maxYear);
}

inline PICBOOL monthIsLegal (PICMONTH aMonth)
{
    return (aMonth >= minMonth && aMonth <= maxMonth);
}

inline PICBOOL dayIsLegal (PICDAY aDay)
{
    return (aDay >= minDay && aDay <= maxDay);
}

inline PICBOOL weekdayIsLegal (PICDAY aDay)
{
    return (aDay >= minWeekday && aDay <= maxWeekday);
}

inline PICBOOL isLegal (PICYEAR aYear, PICMONTH aMonth, PICDAY aDay)
{
    return yearIsLegal(aYear)   &&
           monthIsLegal(aMonth) &&
           dayIsLegal(aDay)     &&
           aDay <= daysInMonth[isLeap(aYear)][aMonth];
}

inline PICBOOL dayNumIsLegal (PICLONG dayNum)
{
    // Lower limit: 1600 January  1.0  == day number 0
    // Upper limit: 2500 December 31.0 == day number 329083

    return (dayNum >= 0 && dayNum <= 329083L);
}

static PICMONTH getMonthFromName (
    const BaseString &aMonth,
    const char    *monthNames[])
{
    for (PICMONTH month = minMonth; month <= maxMonth; month++)
        if (strcmp(monthNames[month],aMonth.chars()) == 0)
            return month;
    return 0;
}

// --------------------------------------------------------------------------
//      DEFAULT FORMATS
// --------------------------------------------------------------------------

static PicDateResolution resolution    = defaultDateResolution;
static PicDateStyle      style         = defaultDateStyle;
static char              delimiter     = defaultDateDelimiter;
static char              leader        = defaultDateLeader;
static PicYearMWStyle    yearMWStyle   = defaultYearMWStyle;

PicDateStyle BaseDate::defaultStyle ()
{
    return ::style;
}

PicDateResolution BaseDate::defaultResolution ()
{
    return ::resolution;
}

char BaseDate::defaultDelimiter ()
{
    return ::delimiter;
}

char BaseDate::defaultLeader ()
{
    return ::leader;
}

PicYearMWStyle BaseDate::defaultYearMWStyle ()
{
    return ::yearMWStyle;
}

// --------------------------------------------------------------------------
//      Month and weekday names
// --------------------------------------------------------------------------

const PICSIZE weekdaySize =  8;
const PICSIZE monthSize   = 13;

BaseString BaseDate::monthName (PICMONTH aMonth)
{
    if (monthIsLegal(aMonth))
        return ::picLongMonths()[aMonth];

    ERAISE(BaseDate, illegalComponent);
    PicError::cumulate("Month out of range: " & asString((short)aMonth));

    return BaseString();
}

// --------------------------------------------------------------------------

BaseString BaseDate::shortMonthName (PICMONTH aMonth)
{
    if (monthIsLegal(aMonth))
        return ::picShortMonths()[aMonth];

    ERAISE(BaseDate, illegalComponent);
    PicError::cumulate("Month out of range: " & asString((short)aMonth));

    return BaseString();
}

// --------------------------------------------------------------------------

BaseString BaseDate::weekdayName (PICWEEKDAY weekday)
{
    if (weekdayIsLegal((PICDAY)weekday))
        return ::picWeekdays()[weekday];

    ERAISE(BaseDate, illegalComponent);
    PicError::cumulate("Weekday out of range: " & asString(weekday));

    return BaseString();
}

// --------------------------------------------------------------------------

BaseString BaseDate::shortWeekdayName (PICWEEKDAY weekday)
{
    if (weekdayIsLegal((PICDAY)weekday))
        return ::picShortWeekdays()[weekday];

    ERAISE(BaseDate, illegalComponent);
    PicError::cumulate("Weekday out of range: " & asString(weekday));

    return BaseString();
}

// --------------------------------------------------------------------------

PICDAYCOUNT BaseDate::daysInYear (PICYEAR aYear)
{
    if (yearIsLegal(aYear))
        return daysInNormalYear + (isLeap(aYear) ? 1 : 0);

    ERAISE(BaseDate, illegalComponent);
    PicError::cumulate("Year out of range: " & asString(aYear));

    return 0L;
}

// --------------------------------------------------------------------------
//      INITIALIZATION
// --------------------------------------------------------------------------

static void initFromArray (
    const char         **charArray,
    const BaseStringArray  *stringArray)
{
    PICINDEX size = stringArray->size();

    for (PICPOSITION pos = 1; pos <= size; pos++)
        charArray[pos] = stringArray->at(pos).chars();
}

PicDateIniter::PicDateIniter (
    PicDateStyle            style,
    PicDateResolution       resolution,
    char                 delimiter,
    char                 leader,
    const BaseStringArray*  longMonthNames,
    const BaseStringArray*  shortMonthNames,
    const BaseStringArray*  dayNames,
    const BaseStringArray*  shortDayNames,
    PicYearMWStyle          ymwStyle)
    :
    dStyle (::style), dResol(::resolution), dDelim(::delimiter),
    dLeader(::leader), ymwStyle(::yearMWStyle)
{
    memcpy(longNames,  ::picLongMonths(),    monthSize * sizeof(char*));
    memcpy(shortNames, ::picShortMonths(),   monthSize * sizeof(char*));
    memcpy(days,       ::picWeekdays(),      weekdaySize * sizeof(char*));
    memcpy(shortDays,  ::picShortWeekdays(), weekdaySize * sizeof(char*));

    ::style         = style;
    ::resolution    = resolution;
    ::delimiter     = delimiter;
    ::leader        = leader;
    ::yearMWStyle   = ymwStyle;

    if (longMonthNames && longMonthNames->size() == maxMonth)
        initFromArray(::picLongMonths(), longMonthNames);
    if (shortMonthNames && shortMonthNames->size() == maxMonth)
        initFromArray(::picShortMonths(), shortMonthNames);
    if (dayNames && dayNames->size() == maxWeekday)
        initFromArray(::picWeekdays(), dayNames);
    if (shortDayNames && shortDayNames->size() == maxWeekday)
        initFromArray(::picShortWeekdays(), shortDayNames);
}

PicDateIniter::~PicDateIniter ()
{
    ::resolution    = dResol;
    ::style         = dStyle;
    ::delimiter     = dDelim;
    ::leader        = dLeader;
    ::yearMWStyle   = ymwStyle;

    memcpy(::picLongMonths(),    longNames,  monthSize * sizeof(char*));
    memcpy(::picShortMonths(),   shortNames, monthSize * sizeof(char*));
    memcpy(::picWeekdays(),      days,       weekdaySize * sizeof(char*));
    memcpy(::picShortWeekdays(), shortDays,  weekdaySize * sizeof(char*));
}

// --------------------------------------------------------------------------
//      BASIC DATE OPERATIONS
// --------------------------------------------------------------------------

static void getCurrentDate (PICYEAR &aYear, PICMONTH &aMonth, PICDAY &aDay)
{
    tm    *today;
    time_t timer;

    aYear  = 1900;
    aMonth = 1;

    time(&timer);
    today = localtime(&timer);

    aYear  += today->tm_year;
    aMonth += today->tm_mon;
    aDay    = today->tm_mday;
}

static PICLONG calcDayNum (PICYEAR year, PICMONTH month, PICDAY day)
{
    // Calculate the number of days since 1600 January 1.0.

    // Algorithm adapted from 
    //    Astronomical Algorithms by Jean Meeus, Willmann-Bell, Inc. 1991,
    //    Chapter 7, 'Julian Day'

    if (month < 3)
    {
        year--;
        month += 12;
    }

    int cent = year / 100;

    return long(365.25 * (year + 4716)) + long(30.6001 * (month + 1))
           + day + (2 - cent + cent / 4) - 2306972L;
}

static void calcDate (
    PICLONG dayNum,
    PICYEAR& year, PICMONTH& month, PICDAY& day)
{
    // Calculate date from number of days since 1600 January 1.0.

    // Algorithm adapted from 
    //    Astronomical Algorithms by Jean Meeus, Willmann-Bell, Inc. 1991,
    //    Chapter 7, 'Julian Day'

    long jd = dayNum + 2305448L;  // Note: actually JD + 0.5
    long b  = long((jd - 1867216.25) / 36524.25);
         b += jd + 1 - b / 4 + 1524;
    long c  = long((b - 122.1) / 365.25);
    long d  = long(365.25 * c);
    long e  = long((b - d) / 30.6001);

    day   = PICDAY(b - d - long(30.6001 * e));
    month = PICMONTH((e < 14) ? e - 1 : e - 13);
    year  = PICYEAR((month > 2) ? c - 4716 : c - 4715);
}

// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR & ASSIGNMENT
// --------------------------------------------------------------------------

BaseDate::BaseDate ()
{
    setNull();
}

BaseDate::BaseDate (
    PicRelativeDate relativeDate,
    PICLONG         unitInterval)
{
    setNull();
    setRelativeDate(relativeDate,unitInterval);
}

BaseDate::BaseDate (
    PICYEAR  aYear,
    PICMONTH aMonth,
    PICDAY   aDay)
{
    setNull();
    setDate(aYear,aMonth,aDay);
}

BaseDate::BaseDate (
    const BaseString& date,
    PicDateStyle      aStyle,
    char           aDelimiter)
{
    setNull();
    setImage(date,aStyle,aDelimiter);
}

BaseDate::BaseDate (
    const BaseString& date,
    PicDateStyle      aStyle)
{
    setNull();
    setImage(date,aStyle,defaultDelimiter());
}

BaseDate::BaseDate (
    const BaseString& date)
{
    setNull();
    setImage(date,defaultStyle(),defaultDelimiter());
}

BaseDate::BaseDate (const BaseDate& right)
{
    *this = right;
}

BaseDate& BaseDate::operator = (const BaseDate &right)
{
    theYear  = right.theYear;
    theMonth = right.theMonth;
    theDay   = right.theDay;

    return *this;
}

// --------------------------------------------------------------------------
//      SET DATE
// --------------------------------------------------------------------------

void BaseDate::setDate (PICYEAR aYear, PICMONTH aMonth, PICDAY aDay)
{
    if (isLegal(aYear, aMonth, aDay))
    {
        theYear  = aYear;
        theMonth = aMonth;
        theDay   = aDay;
    }
    else
    {
        ERAISE(BaseDate, dateIllegal);

        BaseString msg;
        msg << "Illegal date: " << aYear << "y "
            << (PICSHORT)aMonth << "m " << (PICSHORT)aDay << "d";
        PicError::cumulate(msg);
    }
}

void BaseDate::setDate (
    PICYEAR           aYear,
    const BaseString& monthName,
    PICDAY            aDay)
{
    PICMONTH aMonth = getMonthFromName(monthName,::picLongMonths());

    if (!aMonth)
        aMonth = getMonthFromName(monthName,::picShortMonths());

    setDate(aYear,aMonth,aDay);
}

void BaseDate::setRelativeDate (
    PicRelativeDate relativeDate,
    PICLONG         unitInterval)
{
    PICYEAR  aYear;
    PICMONTH aMonth;
    PICDAY   aDay;
    
    getCurrentDate(aYear,aMonth,aDay);

    BaseDate aDate(aYear,aMonth,aDay);

    switch (relativeDate)
    {
        case today:
            aDate.addDays(unitInterval);
            break;

        case thisWeek:
            aDate.addWeeks(unitInterval);
            break;

        case thisMonth:
            aDate.addMonths(unitInterval);
            break;

        case thisYear:
            aDate.addYears(unitInterval);
            break;
    }
    
    if (!PicError::isDetected(classId()))
        *this = aDate;
}

// --------------------------------------------------------------------------
//      INTERVALS
// --------------------------------------------------------------------------

PICDAYCOUNT BaseDate::daysBetween (const BaseDate &right) const
{
    if (isNull() || right.isNull())
    {
        ERAISE(BaseDate, nullInstance);
        return 0;
    }

    return calcDayNum(theYear, theMonth, theDay) -
           calcDayNum(right.theYear, right.theMonth, right.theDay);
}

PICMONTHCOUNT BaseDate::monthsBetween (const BaseDate &right) const
{
    if (isNull() || right.isNull())
    {
        ERAISE(BaseDate, nullInstance);
        return 0;
    }

    PICSHORT        sign;
    const BaseDate *bigger, *smaller;

    if (*this > right)
    {
        sign    = 1;
        bigger  = this;
        smaller = &right;
    }
    else
    {
        sign    = -1;
        bigger  = &right;
        smaller = this;
    }

    PICMONTHCOUNT months = bigger->theMonth - smaller->theMonth;

    months += maxMonth * (bigger->theYear - smaller->theYear);

    if (smaller->theDay > bigger->theDay)
        months--;

    return sign * months;
}

PICYEARCOUNT BaseDate::yearsBetween (const BaseDate &right) const
{
    if (isNull() || right.isNull())
    {
        ERAISE(BaseDate, nullInstance);
        return 0;
    }

    PICSHORT        sign;
    const BaseDate *bigger, *smaller;

    if (*this >= right)
    {
        sign    = 1;
        bigger  = this;
        smaller = &right;
    }
    else
    {
        sign    = -1;
        bigger  = &right;
        smaller = this;
    }

    PICYEARCOUNT years = bigger->year() - smaller->year();

    if (smaller->dayNumber() > bigger->dayNumber())
        years--;

    return sign * years;
}

void BaseDate::addDays (PICDAYCOUNT days)
{
    if (isNull())
    {
        ERAISE(BaseDate, nullInstance);
        return;
    }

    PICLONG dayNum = calcDayNum(theYear, theMonth, theDay) + days;

    if (!dayNumIsLegal(dayNum))
    {
        ERAISE(BaseDate, dateIllegal);
        return;
    }
    calcDate(dayNum, theYear, theMonth, theDay);
}

void BaseDate::addMonths (PICMONTHCOUNT monthsToAdd)
{
    if (isNull())
    {
        ERAISE(BaseDate, nullInstance);
        return;
    }

    PICYEARCOUNT  aYear  = theYear;
    PICMONTHCOUNT aMonth = theMonth;

    // Warning: this might not work in another enviroment. The
    // behaviour of % is not standardized when one of the
    // operands is negative.

    aYear  += monthsToAdd / maxMonth;
    aMonth += monthsToAdd % maxMonth;

    if (aMonth > maxMonth)
    {
        aMonth -= maxMonth;
        aYear++;
    }
    else if (aMonth < minMonth)
    {
        aMonth += maxMonth;
        aYear--;
    }

    if (yearIsLegal(aYear))
    {
        theYear  = PICYEAR(aYear);
        theMonth = PICMONTH(aMonth);

        PICDAY days = daysInMonth[isLeap(theYear)][theMonth];

        if (theDay > days)
            theDay = days;
    }
    else
    {
        ERAISE(BaseDate, dateIllegal);
        PicError::cumulate("Year out of range: " & asString(aYear));
    }
}

void BaseDate::addYears (PICYEARCOUNT yearsToAdd)
{
    if (isNull())
    {
        ERAISE(BaseDate, nullInstance);
        return;
    }

    PICYEARCOUNT aYear = theYear + yearsToAdd;

    if (yearIsLegal(aYear))
    {
        theYear = PICYEAR(aYear);

        PICDAY days = daysInMonth[isLeap(theYear)][theMonth];

        if (theDay > days)
            theDay = days;
    }
    else
    {
        ERAISE(BaseDate, dateIllegal);
        PicError::cumulate("Year out of range: " & asString(aYear));        
    }
}

// --------------------------------------------------------------------------
//      COMPARISONS
// --------------------------------------------------------------------------

PICBOOL BaseDate::operator == (const BaseDate &right) const
{
    if (isNull() || right.isNull())
        return isNull() && right.isNull();

    return theYear  == right.theYear  &&
           theMonth == right.theMonth &&
           theDay   == right.theDay;
}

PICBOOL BaseDate::operator > (const BaseDate &right) const
{
    if (isNull())
        return boolFalse;
    else if (right.isNull())
        return boolTrue;
    else if (theYear > right.theYear)
        return boolTrue;
    else if (theYear < right.theYear)
        return boolFalse;
    else if (theMonth > right.theMonth)
        return boolTrue;
    else
        return theMonth == right.theMonth && theDay > right.theDay;
}

// --------------------------------------------------------------------------
//      DATE ATTRIBUTES
// --------------------------------------------------------------------------

PICWEEKDAY BaseDate::weekday () const
{
    if (isNull())
        return 0;

    PICLONG dayNum = calcDayNum(theYear, theMonth, theDay);

    // The day number zero, 1600 January 1.0, was Saturday, so:

    return PICWEEKDAY((dayNum + 5) % 7) + 1;
}

PICWEEKNUMBER BaseDate::weekNumber () const
{
    if (isNull())
        return 0;

    BaseDate      firstDay(theYear, minMonth, minDay);
    PICWEEKNUMBER numOfWeek;
    PICWEEKDAY    dayOfWeek = firstDay.weekday();
    int           offset;

    if (dayOfWeek <= 4)                // Number of first week is 1
        offset = dayOfWeek - 1 + 6;
    else                               // Number of first week is 52 or 53
        offset = dayOfWeek - 1 - 1;

    numOfWeek = PICWEEKNUMBER((dayNumber() + offset) / maxWeekday);

    if (numOfWeek == 0)                // Can be 52 or 53
    {
        firstDay.addDays(-1);
        numOfWeek = firstDay.weekNumber();
    }
    else if (numOfWeek == 53)          // Can be 53 or 1
    {
        firstDay.addYears(1);

        dayOfWeek = firstDay.weekday();
        if (dayOfWeek <= 4)
            numOfWeek = 1;
        else
            numOfWeek = 53;
    }

    return numOfWeek;
}

PICBOOL BaseDate::setWeek (PICYEAR year, PICWEEKNUMBER week)
{
    BaseDate last ((PICYEAR)year, (PICMONTH)12, (PICDAY)28);
    if (week >= 1 && week <= last.weekNumber())
    {
        BaseDate first ((PICYEAR)year, (PICMONTH)1, (PICDAY)1);
        PICSHORT daysToWeek;
        PICSHORT day1OfWeek1 = first.weekday();

        if (day1OfWeek1 <= 4)              // Number of first week is 1
            day1OfWeek1 = 2 - day1OfWeek1;
        else                               // Number of first week is 52 or 53
            day1OfWeek1 = 9 - day1OfWeek1;

        daysToWeek = (week - 1) * maxWeekday + day1OfWeek1 - 1;

        first.addDays(daysToWeek);
        *this = first;
        return boolTrue;
    }
    return boolFalse;
}

PICDAYCOUNT BaseDate::dayNumber () const
{
    if (isNull())
        return 0;

    // Algorithm adapted from 
    //    Astronomical Algorithms by Jean Meeus, Willmann-Bell, Inc. 1991,
    //    Chapter 7, 'Julian Day'

    return (275 * theMonth) / 9 -
           (isLeap(theYear) ? 1 : 2) * ((theMonth + 9) / 12) + theDay - 30;
}

PICBOOL BaseDate::isLeapYear () const
{
    return isNull() ? boolFalse : isLeap(theYear);
}

PICDAY BaseDate::lastDayOfMonth () const
{
    return isNull() ? 0 : daysInMonth[isLeap(theYear)][theMonth];
}

// --------------------------------------------------------------------------
//      CHARACTER CONVERSION (FORMATTING)
// --------------------------------------------------------------------------

static const char* imageFmts[5][3] =
{
//    yearsOnly  yearsToMonths    yearsToDays
//    ---------  -------------    -----------
    { "@[yyyy]", "@[yyyy|+m]",    "@[yyyy|+m|+d]"     }, // yearFirst
    { "@[yyyy]", "@[+m|yyyy]",    "@[+m|+d|yyyy]"     }, // monthFirst
    { "@[yyyy]", "@[+m|yyyy]",    "@[+d|+m|yyyy]"     }, // dayFirst
    { "@[yyyy]", "@[month yyyy]", "@[month +d, yyyy]" }, // monthByName
    { "@[yyyy]", "@[mmm yyyy]",   "@[mmm +d, yyyy]"   }  // monthByShortName
};

BaseString BaseDate::getImage () const
{
    DEBUG_REQUIRE(::style > 0 && ::style < 6);
    DEBUG_REQUIRE(::resolution > 0 && ::resolution < 4);

    PicStringText image;
    image(imageFmts[::style - 1][::resolution - 1]) << *this;
    return image.flush();
}

BaseString BaseDate::getImage (
    PicDateResolution aResolution,
    PicDateStyle      aStyle,
    char           aDelimiter,
    char           aLeader) const
{
    PicDateResolution oldResolution = ::resolution;
    PicDateStyle      oldStyle      = ::style;
    char           oldDelimiter  = ::delimiter;
    char           oldLeader     = ::leader;

    ::resolution = aResolution;
    ::style      = aStyle;
    ::delimiter  = aDelimiter;
    ::leader     = aLeader;

    BaseString image = getImage();

    ::resolution = oldResolution;
    ::style      = oldStyle;
    ::delimiter  = oldDelimiter;
    ::leader     = oldLeader;

    return image;
}

void BaseDate::setImage (const BaseString& dateImage)
{
    BaseString image = dateImage;
    image.trim();

    if (image.isEmpty())
        setNull();
    else if (image == "*")
        setRelativeDate(today, 0);
    else
    {
        BaseDateTime dt;
        PICOFFSET i = dt.deformat(image.chars(), 0, ::style, 1);

        if (i == 0 || i < image.size())
        {
            ERAISEFMT(BaseDate, deformatError) << image;
        }
        else if (i == 0xffff)
        {
            ERAISEFMT(BaseDate, formatError) << image;
        }
        else
            *this = dt.date;
    }
}

void BaseDate::setImage (
    const BaseString& aDate,
    PicDateStyle      aStyle,
    char           aDelimiter)
{
    PicDateStyle      oldStyle      = ::style;
    char           oldDelimiter  = ::delimiter;

    ::style      = aStyle;
    ::delimiter  = aDelimiter;

    setImage(aDate);

    ::style      = oldStyle;
    ::delimiter  = oldDelimiter;
}

// --------------------------------------------------------------------------
//      FORMATTED OUTPUT
// --------------------------------------------------------------------------

static void wrYear1 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char year[8];
    sprintf(year, "%d", dt.date.year());
    if (dt.date.year() > 1949 && dt.date.year() < 2050)
        fmt.write(&year[2], 2);
    else
        fmt.write(year, 4);
}

static void wrYear2 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char year[8];
    sprintf(year, "%d", dt.date.year());
    fmt.write(year, 4);
}

static void wrMon1 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char mon[4];
    sprintf(mon, "%d", dt.date.month());
    fmt.write(mon, strlen(mon));
}

static void wrMon2 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char mon[4];
    mon[0] = '0';
    fmt.write(mon + (mon[2] != 0), 2);
}

static void wrMon (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char mon[4];
    if (mon[0] = ::leader)
    {
        sprintf(mon+1, "%d", dt.date.month());
        fmt.write(mon + (mon[2] != 0), 2);
    }
    else
        wrMon1(dt, fmt);
}

static void wrMon3 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    const char* name = ::picShortMonths()[dt.date.month()];
    fmt.write(name, strlen(name));
}

static void wrMon4 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    BaseString name = ::picShortMonths()[dt.date.month()];
    name.makeUppercase();
    fmt.write(name.chars(), name.size());
}

static void wrMon5 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    const char* name = ::picLongMonths()[dt.date.month()];
    fmt.write(name, strlen(name));
}

static void wrMon6 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    BaseString name = ::picLongMonths()[dt.date.month()];
    name.makeUppercase();
    fmt.write(name.chars(), name.size());
}

static void wrDay1 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char day[4];
    sprintf(day, "%d", dt.date.day());
    fmt.write(day, strlen(day));
}

static void wrDay2 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char day[4];
    day[0] = '0';
    sprintf(day+1, "%d", dt.date.day());
    fmt.write(day + (day[2] != 0), 2);
}

static void wrDay (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char day[4];
    if (day[0] = ::leader)
    {
        sprintf(day+1, "%d", dt.date.day());
        fmt.write(day + (day[2] != 0), 2);
    }
    else
        wrDay1(dt, fmt);
}

static void wrWD1 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char wd[4];
    sprintf(wd, "%d", dt.date.weekday());
    fmt.write(wd, 1);
}

static void wrWD2 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    const char* name = ::picShortWeekdays()[dt.date.weekday()];
    fmt.write(name, strlen(name));
}

static void wrWD3 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    BaseString name = ::picShortWeekdays()[dt.date.weekday()];
    name.makeUppercase();
    fmt.write(name.chars(), name.size());
}

static void wrWD4 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    const char* name = ::picWeekdays()[dt.date.weekday()];
    fmt.write(name, strlen(name));
}

static void wrWD5 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    BaseString name = ::picWeekdays()[dt.date.weekday()];
    name.makeUppercase();
    fmt.write(name.chars(), name.size());
}

static void wrWN1 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char wn[4];
    sprintf(wn, "%d", dt.date.weekNumber());
    fmt.write(wn, strlen(wn));
}

static void wrWN2 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char wn[4];
    wn[0] = '0';
    sprintf(wn+1, "%d", dt.date.weekNumber());
    fmt.write(wn + (wn[2] != 0), 2);
}

static void wrAMPM1 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    fmt.write(dt.time.hour() < 12 ? "am" : "pm", 2);
}

static void wrAMPM2 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    fmt.write(dt.time.hour() < 12 ? "AM" : "PM", 2);
}

static void wrHourH1 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char hour[4];
    sprintf(hour, "%d", dt.time.hour());
    fmt.write(hour, strlen(hour));
}

static void wrHourH2 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char hour[4];
    hour[0] = '0';
    sprintf(hour+1, "%d", dt.time.hour());
    fmt.write(hour + (hour[2] != 0), 2);
}

static void wrHourH (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char hour[4];
    if (hour[0] = BaseTime::defaultLeader())
    {
        sprintf(hour+1, "%d", dt.time.hour());
        fmt.write(hour + (hour[2] != 0), 2);
    }
    else
        wrHourH1(dt, fmt);
}

static void wrHourA1 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char hour[4];
    sprintf(hour, "%d", dt.time.hour()%12);
    fmt.write(hour, strlen(hour));
}

static void wrHourA2 (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char hour[4];
    hour[0] = '0';
    sprintf(hour+1, "%d", dt.time.hour()%12);
    fmt.write(hour + (hour[2] != 0), 2);
}

static void wrHourA (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char hour[4];
    if (hour[0] = BaseTime::defaultLeader())
    {
        sprintf(hour+1, "%d", dt.time.hour()%12);
        fmt.write(hour + (hour[2] != 0), 2);
    }
    else
        wrHourA1(dt, fmt);
}

static void wrMin (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char minute[4];
    minute[0] = '0';
    sprintf(minute+1, "%d", dt.time.minute());
    fmt.write(minute + (minute[2] != 0), 2);
}

static void wrSec (const BaseDateTime& dt, PicTextFmt& fmt)
{
    char second[4];
    second[0] = '0';
    sprintf(second+1, "%d", dt.time.second());
    fmt.write(second + (second[2] != 0), 2);
}

// --------------------------------------------------------------------------

struct WrDateComp {
    PICUSHORT      size;
    const char* tag;
    void (*wrf)(const BaseDateTime&, PicTextFmt&);
};

static WrDateComp wrDateComp[] =
{
    // --- Time components ---

    { 2, "AP",      wrAMPM2  }, // AM/PM
    { 2, "ap",      wrAMPM1  }, // am/pm

    { 2, "aa",      wrHourA2 }, // Hour (00 .. 11)
    { 1, "a",       wrHourA1 }, // Hour (0 .. 11)
    { 2, "hh",      wrHourH2 }, // Hour (00 .. 23)
    { 1, "h",       wrHourH1 }, // Hour (0 .. 23)
    { 2, "+a",      wrHourA  }, // Hour (use default leader)
    { 2, "+h",      wrHourH  }, // Hour (use default leader)

    { 2, "nn",      wrMin    }, // Minute (00 .. 59)

    { 2, "ss",      wrSec    }, // Second (00 .. 59)

    // --- Date components ---

    { 4, "yyyy",    wrYear2  }, // Year (1600 .. 2500)
    { 2, "yy",      wrYear1  }, // Year (00 .. 99, i.e. 1950 .. 2049)

    { 5, "MONTH",   wrMon6   }, // Month (JANUARY .. DECEMBER)
    { 5, "month",   wrMon5   }, // Month (January .. December)
    { 3, "MMM",     wrMon4   }, // Month (JAN .. DEC)
    { 3, "mmm",     wrMon3   }, // Month (Jan .. Dec)
    { 2, "mm",      wrMon2   }, // Month (01 .. 12)
    { 1, "m",       wrMon1   }, // Month (1 .. 12)
    { 2, "+m",      wrMon    }, // Month (use default leader)

    { 2, "dd",      wrDay2   }, // Day (01 .. 31)
    { 1, "d",       wrDay1   }, // Day (1 .. 31)
    { 2, "+d",      wrDay    }, // Day (use default leader)

    { 7, "WEEKDAY", wrWD5    }, // Weekday (MONDAY .. SUNDAY)
    { 7, "weekday", wrWD4    }, // Weekday (Monday .. Sunday)
    { 3, "WWW",     wrWD3    }, // Weekday (MON .. SUN)
    { 3, "www",     wrWD2    }, // Weekday (Mon .. Sun)
    { 1, "w",       wrWD1    }, // Weekday (1 .. 7)

    { 2, "##",      wrWN2    }, // Week number (01 .. 53)
    { 1, "#",       wrWN1    }, // Week number (1 .. 53)
};

const PICSIZE cWrDateComp = sizeof(wrDateComp) / sizeof(WrDateComp);

// --------------------------------------------------------------------------

PICOFFSET BaseDateTime::writeFmt (
    PicTextFmt&    fmt,
    PICOFFSET      i,
    const char* typeName) const
{
    switch (fmt.format[i])
    {
        case '*':
        {
            BaseString image = "";
            if (!date.isNull())
                image.append(date.getImage());
            if (!date.isNull() && !time.isNull())
                image.append(' ');
            if (!time.isNull())
                image.append(time.getImage());
            fmt.write(image.chars(), image.size());
            break;
        }

        case 'D':
        {
            BaseString image = date.getImage();
            fmt.write(image.chars(), image.size());
            break;
        }

        case 'T':
        {
            BaseString image = time.getImage();
            fmt.write(image.chars(), image.size());
            break;
        }

        case '?':
        {
            fmt.write(typeName, strlen(typeName));
            break;
        }

        case ' ':
        case '.':
        case ',':
        case ':':
        case ';':
        case '-':
        case '/':
        case '\\':
            fmt.write(fmt.format + i, 1);
            break;

        case '|':
        {
            char delim = BaseDate::defaultDelimiter();
            fmt.write(&delim, 1);
            break;
        }

        case '~':
        {
            char delim = BaseTime::defaultDelimiter();
            fmt.write(&delim, 1);
            break;
        }

        default:
        {
            for (PICUSHORT j = 0; j < cWrDateComp; j++)
            {
                WrDateComp& comp = wrDateComp[j];
                if (strncmp(fmt.format + i, comp.tag, comp.size) == 0)
                {
                    comp.wrf(*this, fmt);
                    return i + comp.size;
                }
            }
            break;
        }
    }

    return i + 1;
}

// --------------------------------------------------------------------------

static PICOFFSET wrDate (const BaseDate* obj, PicTextFmt& fmt, PICOFFSET i)
{
    static const char typeName[] = "BaseDate";

    BaseDateTime datetime = *obj;

    if (obj->isNull())
    {
        PicTextFmt nullFmt (PicStream::null(), fmt.format, 0);
        return datetime.writeFmt(nullFmt, i, typeName);
    }

    return datetime.writeFmt(fmt, i, typeName);
}

static BaseDate* wrobjDate (BaseDate* obj, PICBOOL cleanUp)
{
    if (cleanUp)
    {
        delete obj;
        return NULLPTR;
    }
    return new BaseDate(*obj);
}

PicFmtText& operator << (PicFmtText& txt, const BaseDate& obj)
{
    return txt.wrFormat (
        &obj, (PICWRFUNC*)wrDate, (PICWROBJFUNC*)wrobjDate);
}

// --------------------------------------------------------------------------
//      DEFORMATTING PROCEDURE
// --------------------------------------------------------------------------

static PICOFFSET numToken (const char* str, PICOFFSET i, PICSHORT& num)
{
    if (isdigit(str[i]))
    {
        num = str[i] - '0';
        while (isdigit(str[++i]))     // Read decimal digits
        {
            num *= 10;                // and form an integer
            num += str[i] - '0';
        }
    }
    return i;
}

// --------------------------------------------------------------------------

enum TimeCompType   // Possible DateTime components
{
    YR  = 0x0001,
    MON = 0x0002,
    DAY = 0x0004,
    HRS = 0x0010,
    MIN = 0x0020,
    SEC = 0x0040
};

const PICSHORT DAT = YR | MON | DAY;
const PICSHORT MS  = MIN | SEC;
const PICSHORT TIM = MS | HRS;

struct TimeCompLimit { PICSHORT type; PICSHORT minv; PICSHORT maxv; };

static const TimeCompLimit limits[] =  // Limit values for each component
{
    { YR,                        minYear, maxYear },
    { YR,                        60,      99      },
    { YR | MS,                   32,      59      },
    { YR | MS | DAY,             25,      31      },
    { YR | MS | DAY | HRS,       13,      24      },
    { YR | MS | DAY | HRS | MON, 1,       12      },
    { YR | MS | HRS,             0,       0       }
};

const PICSIZE cLimits = sizeof(limits) / sizeof(TimeCompLimit);

struct TimeComp { PICSHORT type; PICSHORT value; };

// --------------------------------------------------------------------------

enum DateOrder { DMY, YMD, MDY, YDM }; // Possible date component orders

static const DateOrder defOrder [4][4] =
{
    { DMY, YMD, MDY, YDM },  // Search from left to right --->
    { YMD, MDY, DMY, YDM },  // (first is the default date order)
    { MDY, YMD, DMY, YDM },
    { YDM, DMY, YMD, MDY }
};

static const TimeCompType d2models1[4][2] = // [comp][order]
{
    { DAY, MON },    // DMY    // Possible orders of two date components
    { MON, DAY },    // YMD
    { MON, DAY },    // MDY
    { DAY, MON }     // YDM
};

static const TimeCompType d2models2[4][2] = // [comp][order]
{
    { MON, YR  },    // DMY    // Possible orders of two date components
    { YR,  MON },    // YMD    // (secondary set)
    { MON, YR  },    // MDY
    { YR,  MON }     // YDM
};

static const TimeCompType d3models[4][3] = // [comp][order]
{
    { DAY, MON, YR  },    // DMY  // Possible orders of three date components
    { YR,  MON, DAY },    // YMD
    { MON, DAY, YR  },    // MDY
    { YR,  DAY, MON }     // YDM
};

static PICBOOL checkDate (TimeComp items[6], PICSIZE dsize, DateOrder dOrder)
{
    for (PICOFFSET iord = 0; iord < 4; iord++)
    {
        DateOrder order = defOrder[dOrder][iord];
        TimeCompType c[3];
        PICOFFSET i;

        if (dsize == 3) for (i = 0; i < 3; i++)
        {
            if (d3models[order][i] & items[i].type)
                c[i] = d3models[order][i];
            else
                goto NextOrder;
        }
        else for (i = 0; i < 2; i++)
        {
            if (d2models1[order][i] & items[i].type)
                c[i] = d2models1[order][i];
            else for (i = 0; i < 2; i++)
            {
                if (d2models2[order][i] & items[i].type)
                    c[i] = d2models2[order][i];
                else
                    goto NextOrder;
            }
        }

        if (iord > 1)  // 3rd and 4th date order: more unambiguity required
        {
            PICSHORT unambig = (items[0].type == c[0]) +
                               (items[1].type == c[1]) +
                               (dsize == 3 ? items[2].type == c[2] : 1);
            if (unambig < 2)
                return boolFalse;
        }

        items[0].type = c[0];
        items[1].type = c[1];
        if (dsize == 3)
            items[2].type = c[2];
        return boolTrue;

        NextOrder: ;
    }
    return boolFalse;
}

// --------------------------------------------------------------------------

struct TimeModel { PICSHORT size, dsize, tsize; };

static const TimeModel tmodels[] =  // Possible interpretations of date/time
{                                   // image in guessing preference order
    { 2, 0, 2 }, // h.n
    { 2, 2, 0 }, // d.m, m.y
    { 2, 1, 1 }, // d h
    { 3, 0, 3 }, // h.n.s
    { 3, 3, 0 }, // d.m.y
    { 3, 2, 1 }, // d.m h, m.y h
    { 3, 1, 2 }, // d h.n
    { 4, 2, 2 }, // d.m h.n
    { 4, 3, 1 }, // d.m.y h
    { 4, 1, 3 }, // d h.n.s
    { 5, 3, 2 }, // d.m.y h.n
    { 5, 2, 3 }, // d.m h.m.s
    { 6, 3, 3 }  // d.m.y h.n.s
};

const cTModels = sizeof(tmodels) / sizeof(TimeModel);

static PICBOOL checkTime (
    TimeComp      items[6],
    const char separ[7],
    PICSIZE       size,
    DateOrder     dOrder)
{
    for (PICOFFSET i = 0; i < cTModels; i++)
    {
        const TimeModel& tm = tmodels[i];
        if (tm.size == size)
        {
            // Check DATE/TIME separator (should be blank)

            if (tm.dsize && tm.tsize && separ[tm.dsize] != ' ')
                continue;

            // Check TIME components (always HRS MIN SEC)

            if (tm.tsize > 0 && !(items[tm.dsize].type & HRS))
                continue;
            if (tm.tsize > 1 && !(items[tm.dsize + 1].type & MIN))
                continue;
            if (tm.tsize > 2 && !(items[tm.dsize + 2].type & SEC))
                continue;

            // Check DATE components (that's more complicated...)

            if (tm.dsize == 0 ||
                (tm.dsize == 1 && (items[0].type & DAY)) ||
                checkDate(items, tm.dsize, dOrder))
            {
                // All's well, fix the time components

                if (tm.tsize > 0) items[tm.dsize].type     = HRS;
                if (tm.tsize > 1) items[tm.dsize + 1].type = MIN;
                if (tm.tsize > 2) items[tm.dsize + 2].type = SEC;

                return boolTrue;
            }
        }
    }
    return boolFalse;
}

// --------------------------------------------------------------------------

PICOFFSET BaseDateTime::deformat (
    const char* str,
    PICOFFSET      i,
    PicDateStyle   dStyle,
    PICSHORT       fDateTime)
{
    TimeComp items[6];
    char  separ[7];
    char  ampm  = 0;
    PICSHORT count = 0;

    while (1)
    {
        while (isspace(str[i])) // Read leading spaces
            i++;

        if (str[i] && !isalnum(str[i])) // Memorize separator
        {
            separ[count] = str[i];
            while (isspace(str[++i]))
                ;

            if (str[i] && !isalnum(str[i]))
                return 0; // **** multiple separator characters
        }
        else
            separ[count] = ' ';

        if (isalpha(str[i]))    // Alphanumeric --> AM/PM ?
        {
            PICSHORT am = strncmp(str + i, "AM", 2);
            PICSHORT pm = strncmp(str + i, "PM", 2);

            if (am == 0 || pm == 0)
            {
                if (ampm || fDateTime == 1)
                    return 0; // **** second AM/PM or date only
                ampm = (am == 0 ? 'A' : 'P');
                i += 2;
                continue; // Don't increment count after AM/PM
            }
        }

        if (count >= (fDateTime == 3 ? 6 : 3) || !str[i])
            goto Out;

        if (isalpha(str[i]))    // Alphanumeric --> month name ?
        {
            PICMONTH mon = 0;

            while (++mon < 13)
            {
                if (strncmp(str + i, ::picLongMonths()[mon],
                         strlen(::picLongMonths()[mon])) == 0)
                {
                    items[count].type  = MON;
                    items[count].value = mon;
                    i += strlen(::picLongMonths()[mon]);
                    goto Next;
                }
                else if (strncmp(str + i, ::picShortMonths()[mon],
                    strlen(::picShortMonths()[mon])) == 0)
                {
                    items[count].type  = MON;
                    items[count].value = mon;
                    i += strlen(::picShortMonths()[mon]);
                    goto Next;
                }
            }
            return 0; // *** no such month name
        }
        else // Numeric item
        {
            i = numToken(str, i, items[count].value);
            if (isalpha(str[i]))
                return 0; // *** letters after number

            for (PICSHORT j = cLimits; --j >= 0;) // check limits
            {
                if (items[count].value <= limits[j].maxv &&
                    items[count].value >= limits[j].minv)
                {
                    items[count].type = limits[j].type;
                    goto Next;
                }
            }
            return 0; // *** number out of range
        }
      Next:
        count++;
    }
  Out:

    if (count == 0)
        return 0; // **** only a single AM/PM flag

    PICSHORT j;

    switch (fDateTime)
    {
        case 1: for (j = 0; j < count; j++)
        {
            items[j].type &= ~TIM;
            if (items[j].type == 0)
                return 0; // *** only date items expected
        }
        break;

        case 2: for (j = 0; j < count; j++)
        {
            items[j].type &= ~DAT;
            if (items[j].type == 0)
                return 0; // *** only time items expected
        }
        break;

        case 3: if (::delimiter != BaseTime::defaultDelimiter())
        for (j = 0; j < count; j++)
        {                                                  // Delimiter may
            if (separ[j]     == ::delimiter ||             // give a clue
                separ[j + 1] == ::delimiter)               // for DateTime
            {                                              // components
                if ((items[j].type & DAT) == 0)
                    return 0; // *** should be date item
                items[j].type &= ~TIM;
            }
            else if (separ[j]     == BaseTime::defaultDelimiter() ||
                     separ[j + 1] == BaseTime::defaultDelimiter())
            {
                if ((items[j].type & TIM) == 0)
                    return 0; // *** should be time item
                items[j].type &= ~DAT;
            }
        }
    }

    if (count == 1)
    {
        if (items[0].type == MON || items[0].type == YR)
            ;
        else if ((fDateTime & 1) && (items[0].type & DAY))
            items[0].type = DAY;
        else if (items[0].type & HRS)
            items[0].type = HRS;
        else
            return 0; // *** illegal single-item image
    }
    else
    {
        DateOrder dOrder;
        switch (dStyle)
        {
            case BaseDate::yearFirst:
                dOrder = YMD; break;
            case BaseDate::dayFirst:
                dOrder = DMY; break;
            default:
            // case BaseDate::monthFirst:
            // case BaseDate::monthByName:
            // case BaseDate::monthByShortName:
                dOrder = MDY; break;
        }

        if (!checkTime(items, separ, count, dOrder))
            return 0;  // *** no such date/time image
    }

    PICSHORT maxComp = YR;

    date.setRelativeDate(BaseDate::today, 0);
    time.setRelativeTime(::now);

    for (j = 0; j < count; j++)
    {
        if (maxComp < items[j].type)
            maxComp = items[j].type;

        switch (items[j].type)
        {
            case YR:
                date.theYear = (PICYEAR)items[j].value;
                if (date.theYear < 50)
                    date.theYear += 2000;
                else if (date.theYear < 100)
                    date.theYear += 1900;
                break;
            case MON:
                date.theMonth = (PICMONTH)items[j].value;
                break;
            case DAY:
                date.theDay = (PICDAY)items[j].value;
                break;
            case HRS:
                time.theHour = (PICHOUR)items[j].value;
                if (ampm && time.theHour > 12)
                    return 0xffff; // *** AM/PM error
                if (ampm == 'p')
                    time.theHour += 12;
                if (time.theHour == 24)
                    time.theHour = 0;
                break;
            case MIN:
                time.theMinute = (PICMINUTE)items[j].value;
                break;
            case SEC:
                time.theSecond = (PICSECOND)items[j].value;
                break;
        }
    }

    switch (maxComp)
    {
        case YR:  date.theMonth  = 1; // Fall through
        case MON: date.theDay    = 1; // Fall through
        case DAY: time.theHour   = 0; // Fall through
        case HRS: time.theMinute = 0; // Fall through
        case MIN: time.theSecond = 0; // Fall through
        case SEC: ;
    }

    if (!isLegal(date.theYear, date.theMonth, date.theDay))
        return 0xffff; // **** impossible date

    return i;  // Success
}

// --------------------------------------------------------------------------

const char** picShortMonths()
{
    static const char *buf[monthSize] =
        {
            "",
            PIC_SHORTMONTH1,
            PIC_SHORTMONTH2,
            PIC_SHORTMONTH3,
            PIC_SHORTMONTH4,
            PIC_SHORTMONTH5,
            PIC_SHORTMONTH6,
            PIC_SHORTMONTH7,
            PIC_SHORTMONTH8,
            PIC_SHORTMONTH9,
            PIC_SHORTMONTH10,
            PIC_SHORTMONTH11,
            PIC_SHORTMONTH12
        };
    return buf;
}

const char** picLongMonths()
{
    static const char *buf[monthSize] =
        {
            "",
            PIC_LONGMONTH1,
            PIC_LONGMONTH2,
            PIC_LONGMONTH3,
            PIC_LONGMONTH4,
            PIC_LONGMONTH5,
            PIC_LONGMONTH6,
            PIC_LONGMONTH7,
            PIC_LONGMONTH8,
            PIC_LONGMONTH9,
            PIC_LONGMONTH10,
            PIC_LONGMONTH11,
            PIC_LONGMONTH12
        };
    return buf;
}

const char** picWeekdays()
{
    static const char *buf[weekdaySize] =
        {
            "",
            PIC_WEEKDAY1,
            PIC_WEEKDAY2,
            PIC_WEEKDAY3,
            PIC_WEEKDAY4,
            PIC_WEEKDAY5,
            PIC_WEEKDAY6,
            PIC_WEEKDAY7
        };
    return buf;
}

const char** picShortWeekdays()
{
    static const char *buf[weekdaySize] =
        {
            "",
            PIC_SHORT_WEEKDAY1,
            PIC_SHORT_WEEKDAY2,
            PIC_SHORT_WEEKDAY3,
            PIC_SHORT_WEEKDAY4,
            PIC_SHORT_WEEKDAY5,
            PIC_SHORT_WEEKDAY6,
            PIC_SHORT_WEEKDAY7
        };
    return buf;
}

