// --------------------------------------------------------------------------
//  REAL NUMBER base class implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <float.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
}

#include <breal.hxx>
#include <bstring.hxx>
#include <picerror.hxx>
#include <picbase.erh>
#include <pdefault.hxx>

MAKE_PICCLASS(BaseReal)

ERROR_MESSAGES(BaseReal)
PARAMSTR(nullInstance, "Operation on Null instance.")
PARAMSTR(deformatError, "String '@*' is not of the given format.")
PARAMSTR(formatError, "String '@*' is not a valid real number.")
PARAMSTR(underflow, "Underflow")
PARAMSTR(overflow, "Overflow: @* ")
PARAMSTR(divideByZero, "Divide by zero")
END_PARAM_STRINGS

// --------------------------------------------------------------------------
//      Notes on the implementation
//
//      Floating point Reals are of form {imp=<value>,scale=floatReal,...}
//      Fixed point Reals are of form {imp = <value>, scale = <scale>,...}
//      where the actual value is <value> * 10**<scale>.
//
// --------------------------------------------------------------------------
//      STATIC (LOCAL) CLASS DATA
// 
//    nullDouble     Null value for double.

const  PICUSHORT maxNrOfDigits  = 16;
const  PICFLOAT  nullDouble     = DBL_MAX;
const  PICFLOAT  maxIntegerPart = 999999999999999.0;

static char   decimalPoint   = defaultDecimalPoint;
//  static char   buf[maxNrOfDigits + 10];

// --------------------------------------------------------------------------
//      INITIALIZATION

PicRealIniter::PicRealIniter (char newDecPoint)
  : prevDecPoint(decimalPoint)
{
    decimalPoint = newDecPoint;
}

PicRealIniter::~PicRealIniter ()
{
    decimalPoint = prevDecPoint;
}

// --------------------------------------------------------------------------
//      LOCAL INLINE FUNCTIONS
// --------------------------------------------------------------------------
//
//      BaseReal::flags has the following format:
//        bit      15  - null if clear, non-null if set
//        bits 0 - 14  - precision + maxDigit + 1 (so floatReal == 0)
//
//      getPrec        Extracts precision from BaseReal::flags field
//      set*Flags      Forms the BaseReal::flags field out of the info
//      isFloatNumber  Returns whether BaseReal is a floating point number
//      isNullNumber   Returns whether BaseReal is null
//
//      scaleDiff  Counts a scale difference from a scale to another.
//                 If you have a value A in scale F, you get it in scale G
//                 by multiplying A by 10 ** scaleDiff (F, G).
//                 In other words:
//                 {f, s}  ==  {f * 10 ** scaleDiff (s, s2), s2}
//
// --------------------------------------------------------------------------

const PICUSHORT nullFloatFlags = 0;

inline PICDIGIT getPrec (const PICUSHORT flags)
{
    return (flags & 0x7fff) - maxDigit - 1;
}

inline void setFlags (
    PICUSHORT& flags, const PICDIGIT prec, const PICBOOL isNull)
{
    flags = (prec + maxDigit + 1) | (isNull ? 0 : 0x8000);
}

inline void setPrecFlags (PICUSHORT& flags, const PICDIGIT prec)
{
    flags = (prec + maxDigit + 1) | (flags & 0x8000);
}

inline void setNullFlags (PICUSHORT& flags)
{
    flags &= 0x7fff;
}

inline void setNonNullFlags (PICUSHORT& flags)
{
    flags |= 0x8000;
}

inline PICBOOL isFloatNumber (const PICUSHORT flags)
{
    return (flags & 0x7fff) == 0;
}

inline PICBOOL isNullNumber (const PICUSHORT flags)
{
    return (flags & 0x8000) == 0;
}

inline PICDIGIT scaleDiff (const PICDIGIT from, const PICDIGIT to)
{
    return ((from == floatReal) ? 0 : from) - ((to == floatReal) ? 0 : to);
}

// --------------------------------------------------------------------------
//      SCALE DOUBLE
//
//      Scales a double.  The scaledValue value is an integer, except when
//      toScale == floatReal. The function checks for possible overflows.
//      In case of overflow it returns boolFalse and scaledValue is left
//      intact.
//
// --------------------------------------------------------------------------

static PICBOOL scaleDouble (
    PICFLOAT  value,
    PICFLOAT& scaledValue,
    PICDIGIT  fromScale,
    PICDIGIT  toScale,
    PICBOOL   rounding = boolTrue)
{
    if (toScale == floatReal)
    {
        scaledValue = value * pow(10.0, scaleDiff(fromScale, toScale));
        return boolTrue;
    }

    PICFLOAT integerPart, fractPart, real = value < 0.0 ? -value : value;

    real      *= pow(10.0, scaleDiff(fromScale, toScale));
    fractPart  = modf(real, &integerPart);

    if (rounding && (fractPart >= 0.5))
        integerPart += 1.0;
    if (integerPart > maxIntegerPart)
    {
        ERAISEFMT(BaseReal, overflow) << value;
        return boolFalse;
    }

    scaledValue = ((value < 0.0) ? -integerPart : integerPart);
    return boolTrue;
}

// --------------------------------------------------------------------------
//      evaluateImage evaluates a string image.
//
//      Arguments:
//        image       The string image.
//        value       The double value of the image. Scaled with flags:prec.
//        flags       The precision & nullity of the value. Precision is
//                    determined by decimal point position and by the number
//                    of significant digits in the image. Nullity flag is
//                    set if the image represents null real.
//
//      Return value: True if the image represents a valid PICFLOAT or null
//                    real. In case of error arguments are not affected.
//
// --------------------------------------------------------------------------

static PICBOOL evaluateImage (
    BaseString image,
    PICFLOAT&  value,
    PICUSHORT& flags)
{
    image.trim();

    if (image.isEmpty())
    {
        value = nullDouble;
        flags = nullFloatFlags;
        return boolTrue;
    }

    // We check the syntax using a 7 state DFA.
    // All states are needed, because we want to get the number of
    // significant digits. (Otherwise 5 would have been enough.)

    char     aChar;
    PICSIZE     len        = image.size ();
    PICPOSITION pos        = firstPos, decPoint = outOfRange;
    PICUSHORT   nrOfDigits = 0;

    {
        // state_1: // initial state: nothing scanned
        aChar = image.at(pos);
        if ((aChar == '-') || (aChar == '+'))
        {
            goto state_2;
        }
        if (aChar == '0')
            goto state_3;
        if (PicChar::isDigit(aChar))
        {
            nrOfDigits++;
            goto state_4;
        }
        if (aChar == decimalPoint)
        {
            decPoint = pos;
            goto state_6;
        }
        goto state_error;

        state_2: // scanned: "-"
        if (++pos > len)
            goto state_error;
        aChar = image.at(pos);
        if (aChar == '0')
            goto state_3;
        if (PicChar::isDigit(aChar))
        {
            nrOfDigits++;
            goto state_4;
        }
        if (aChar == decimalPoint)
        {
            decPoint = pos;
            goto state_6;
        }
        goto state_error;

        state_3: // scanned: "-000" or "000"
        if (++pos > len)
            goto state_ok;
        aChar = image.at(pos);
        if (aChar == '0')
            goto state_3;
        if (PicChar::isDigit(aChar))
        {
            nrOfDigits++;
            goto state_4;
        }
        if (aChar == decimalPoint)
        {
            decPoint = pos;
            goto state_7;
        }
        goto state_error;

        state_4: // scanned "-000ddd" or "000ddd"
        if (++pos > len)
            goto state_ok;
        aChar = image.at(pos);
        if (PicChar::isDigit(aChar))
        {
            nrOfDigits++;
            goto state_4;
        }
        if (aChar == decimalPoint)
        {
            decPoint = pos;
            goto state_5;
        }
        goto state_error;

        state_5: // scanned "-000ddd.ddd"
        if (++pos > len)
            goto state_ok;
        aChar = image.at(pos);
        if (PicChar::isDigit(aChar))
        {
            nrOfDigits++;
            goto state_5;
        }
        goto state_error;

        state_6: // scanned "-." or "."
        if (++pos > len)
            goto state_error;
        aChar = image.at(pos);
        if (aChar == '0')
            goto state_7;
        if (PicChar::isDigit(aChar))
        {
            nrOfDigits++;
            goto state_5;
        }
        goto state_error;

        state_7: // scanned "-000.000" or "000.000"
        if (++pos > len)
            goto state_ok;
        aChar = image.at(pos);
        if (aChar == '0')
            goto state_7;
        if (PicChar::isDigit(aChar))
        {
            nrOfDigits++;
            goto state_5;
        }
        goto state_error;

        state_error:
        ERAISEFMT(BaseReal, deformatError) << image;
        return boolFalse;

        state_ok:;
    }

    // NOW: Syntax is ok.
    //      nrOfDigits   is the number of significant digits
    //      decPoint     is the position of decimal point
    //      sign         is true iff any sign exists (+ or -)
    // We continue by removing the decimal point and truncating
    // the image to at most 100 characters (function atof does
    // not accept more).

    PICDIGIT imageScale;

    if (decPoint == outOfRange)
        imageScale = 0;
    else
    {
        imageScale = decPoint - len;
        image.deleteAt(decPoint, 1);
        len--;
    }
    if (len > 99)
    {
        image.deleteAfter(99);
        imageScale += len - 99;
    }
    value = atof(image.chars());
    if (nrOfDigits > maxNrOfDigits)
    {
        setFlags(flags, floatReal, boolFalse);
        value *= pow(10.0, imageScale);
    }
    else
        setFlags(flags, imageScale, boolFalse);

    return boolTrue;
}

// --------------------------------------------------------------------------
//      METHODS
// --------------------------------------------------------------------------
//      CONSTRUCTORS & ASSIGNMENT

BaseReal::BaseReal () : imp(nullDouble), flags(nullFloatFlags) {}

BaseReal::BaseReal (PICDIGIT digit)
  : imp(nullDouble)
{
    setFlags(flags, digit, boolTrue);
}

BaseReal::BaseReal (PICFLOAT value, PICDIGIT digit)
  : imp(value)
{
    setFlags(flags, digit, boolFalse);
    if (digit != floatReal)
    {
        if (!scaleDouble(value, imp, floatReal, digit))
            setNullFlags(flags);
    }
}

BaseReal::BaseReal (const BaseString& image)
  : imp(nullDouble), flags(nullFloatFlags)
{
    evaluateImage(image, imp, flags);
}

BaseReal::BaseReal (const BaseString& image, PICDIGIT digit)
  : imp(nullDouble)
{
    setFlags(flags, digit, boolTrue);
    setImage(image);
}

BaseReal::BaseReal (const BaseReal& real) : imp(real.imp), flags(real.flags) {}

BaseReal& BaseReal::operator = (PICFLOAT value)
{
    if (scaleDouble(value, imp, floatReal, getPrec(flags)))
        setNonNullFlags(flags);
    return *this;
}

BaseReal& BaseReal::assignValue (const BaseReal& value)
{
    if (isNullNumber(value.flags))
        setNull();
    else
    {
        if (scaleDouble(value.imp, imp, getPrec(value.flags), getPrec(flags)))
            setNonNullFlags(flags);
    }
    return *this;
}

BaseReal& BaseReal::assignValue (PICLONG value)
{
    if (scaleDouble((PICFLOAT)value, imp, 0, getPrec(flags)))
        setNonNullFlags(flags);
    return *this;
}

// --------------------------------------------------------------------------
//      CONVERSION

PICFLOAT BaseReal::asFloat () const
{
    if (isNullNumber(flags))
    {
        ERAISE(BaseReal, nullInstance);
        return 0.0;
    }
    else if (isFloatNumber(flags))
        return imp;
    else
        return imp * pow(10.0, getPrec(flags));
}

// --------------------------------------------------------------------------
//      NULL HANDLING

PICBOOL BaseReal::isNull () const
{
    return isNullNumber(flags);
}

void BaseReal::setNull ()
{
    imp = nullDouble;
    setNullFlags(flags);
}

// --------------------------------------------------------------------------
//      PRECISION

PICDIGIT BaseReal::getPrecision() const
{
    return getPrec(flags);
}

void BaseReal::setPrecision (PICDIGIT newPrec)
{
    if (isNullNumber(flags) || scaleDouble(imp, imp, getPrec(flags), newPrec))
        setPrecFlags(flags, newPrec);
}

PICBOOL BaseReal::isFloat() const
{
    return isFloatNumber(flags);
}

PICBOOL BaseReal::isFixed() const
{
    return !isFloatNumber(flags);
}

// --------------------------------------------------------------------------
//      ARITHMETIC OPERATIONS

BaseReal operator + (const BaseReal& value1, const BaseReal& value2)
{
    if (value1.getPrecision() <= value2.getPrecision())
//          { BaseReal tmp(value1); return tmp += value2; }
        return BaseReal(value1) += value2;  // Doesn't compile -- bug?
    else
        { BaseReal tmp(value2); return tmp += value1; }
}

BaseReal operator - (const BaseReal& value1, const BaseReal& value2)
{
    if (value1.getPrecision() <= value2.getPrecision())
        { BaseReal tmp(value1); return tmp -= value2; }
    else
        { BaseReal tmp(value2); tmp.negate(); return tmp += value1; }
}

BaseReal operator * (const BaseReal& value1, const BaseReal& value2)
{
    if (value1.getPrecision() <= value2.getPrecision())
        { BaseReal tmp(value1); return tmp *= value2; }
    else
        { BaseReal tmp(value2); return tmp *= value1; }
}

BaseReal operator / (const BaseReal& value1, const BaseReal& value2)
{
    BaseReal tmp(value1);
    if (value1.getPrecision() > value2.getPrecision())
        tmp.setPrecision(value2.getPrecision());
    return tmp /= value2;
}

// --------------------------------------------------------------------------
//      ASSIGNMENT + ARITHMETIC OPERATIONS

BaseReal& BaseReal::operator += (const BaseReal& value)
{
    if (isNullNumber(flags) || isNullNumber(value.flags))
        ERAISE(BaseReal, nullInstance);
    else
    {
        PICDIGIT scale = getPrec(flags);
        scaleDouble (imp + value.imp *
            pow(10.0, scaleDiff(getPrec(value.flags), scale)),
            imp, scale, scale);
    }
    return *this;
}

BaseReal& BaseReal::operator -= (const BaseReal& value)
{
    if (isNullNumber(flags) || isNullNumber(value.flags))
        ERAISE(BaseReal, nullInstance);
    else
    {
        PICDIGIT scale = getPrec(flags);
        scaleDouble (imp - value.imp *
            pow(10.0, scaleDiff(getPrec(value.flags), scale)),
            imp, scale, scale);
    }
    return *this;
}

BaseReal& BaseReal::operator *= (const BaseReal& value)
{
    if (isNullNumber(flags) || isNullNumber(value.flags))
        ERAISE(BaseReal, nullInstance);
    else
    {
        scaleDouble (
            imp * value.asFloat(), imp, getPrec(flags), getPrec(flags));
    }
    return *this;
}

BaseReal& BaseReal::operator /= (const BaseReal& value)
{
    if (isNullNumber(flags) || isNullNumber(value.flags))
        ERAISE(BaseReal, nullInstance);
    else if (value.isZero())
        ERAISE(BaseReal, divideByZero);
    else
    {
        scaleDouble (
            imp / value.asFloat(), imp, getPrec(flags), getPrec(flags));
    }
    return *this;
}

BaseReal& BaseReal::operator += (PICLONG value)
{
    if (isNullNumber(flags))
        ERAISE(BaseReal, nullInstance);
    else
    {
        PICDIGIT scale = getPrec(flags);
        scaleDouble (imp + (PICFLOAT)value *
            pow(10.0, scaleDiff(floatReal, scale)),
            imp, scale, scale);
    }
    return *this;
}

BaseReal& BaseReal::operator -= (PICLONG value)
{
    if (isNullNumber(flags))
        ERAISE(BaseReal, nullInstance);
    else
    {
        PICDIGIT scale = getPrec(flags);
        scaleDouble (imp - (PICFLOAT)value *
            pow(10.0, scaleDiff(floatReal, scale)),
            imp, scale, scale);
    }
    return *this;
}

BaseReal& BaseReal::operator *= (PICLONG value)
{
    if (isNullNumber(flags))
        ERAISE(BaseReal, nullInstance);
    else
    {
        scaleDouble (
            imp * (PICFLOAT)value, imp, getPrec(flags), getPrec(flags));
    }
    return *this;
}

BaseReal& BaseReal::operator /= (PICLONG value)
{
    if (isNullNumber(flags))
        ERAISE(BaseReal, nullInstance);
    else if (value == 0L)
        ERAISE(BaseReal, divideByZero);
    else
    {
        scaleDouble (
            imp / (PICFLOAT)value, imp, getPrec(flags), getPrec(flags));
    }
    return *this;
}

// --------------------------------------------------------------------------
//      ABSOLUTE VALUE

void BaseReal::makeAbs ()
{
    if (isNullNumber(flags))
        ERAISE(BaseReal, nullInstance);
    else
        imp = fabs(imp);
}

// --------------------------------------------------------------------------
//      NEGATE

void BaseReal::negate ()
{
    if (isNullNumber(flags))
        ERAISE(BaseReal, nullInstance);
    else
        imp = -imp;
}

// --------------------------------------------------------------------------
//      COMPARISONS

PICBOOL operator == (const BaseReal& value1, const BaseReal& value2)
{
    if (isNullNumber(value1.flags) || isNullNumber(value2.flags))
        return isNullNumber(value1.flags) && isNullNumber(value2.flags);

    PICDIGIT delta = scaleDiff(getPrec(value1.flags),getPrec(value2.flags));

    if (delta > 0)
        return (value1.imp * pow(10.0, delta)) == value2.imp;
    else
        return value1.imp == (value2.imp * pow(10.0, -delta));
}

PICBOOL operator > (const BaseReal& value1, const BaseReal& value2)
{
    if (isNullNumber(value1.flags))
        return boolFalse;
    else if (isNullNumber(value2.flags))
        return boolTrue;
    else
    {
        PICDIGIT delta;

        delta = scaleDiff(getPrec(value1.flags),getPrec(value2.flags));

        if (delta > 0)
            return (value1.imp * pow(10.0, delta)) > value2.imp;
        else
            return value1.imp > (value2.imp * pow(10.0, -delta));
    }
}

PICBOOL operator >= (const BaseReal& value1, const BaseReal& value2)
{
    if (isNullNumber(value1.flags))
        return isNullNumber(value2.flags);
    else if (isNullNumber(value2.flags))
        return boolTrue;
    else
    {
        PICDIGIT delta;

        delta = scaleDiff(getPrec(value1.flags), getPrec(value2.flags));

        if (delta > 0)
            return (value1.imp * pow(10.0, delta)) >= value2.imp;
        else
            return value1.imp >= (value2.imp * pow(10.0, -delta));
    }
}

PICBOOL BaseReal::isZero () const
{
    return isNullNumber(flags) ? boolFalse : imp == 0;
}

PICBOOL BaseReal::isPositive () const
{
    return isNullNumber(flags) ? boolFalse : imp > 0;
}

PICBOOL BaseReal::isNegative () const
{
    return isNullNumber(flags) ? boolFalse : imp < 0;
}

// --------------------------------------------------------------------------
//      ROUND & TRUNCATE

void BaseReal::round (PICDIGIT newPrec)
{
    if (isNullNumber(flags))
        ERAISE(BaseReal, nullInstance);
    else
    {
        if (scaleDouble (imp, imp, getPrec(flags), newPrec))
            setPrecFlags(flags, newPrec);
    }
}

void BaseReal::truncate (PICDIGIT newPrec)
{
    if (isNullNumber(flags))
        ERAISE(BaseReal, nullInstance);
    else
    {
        if (scaleDouble (imp, imp, getPrec(flags), newPrec, boolFalse))
            setPrecFlags(flags, newPrec);
    }
}

// --------------------------------------------------------------------------
//      ROOF & FLOOR etc.

PICLONG BaseReal::roof () const
{
    if (isNullNumber(flags))
    {
        ERAISE(BaseReal, nullInstance);
        return 0L;
    }

    PICFLOAT roofValue = ::ceil(asFloat());

    if ((roofValue < (PICFLOAT)LONG_MIN) || (roofValue > (PICFLOAT)LONG_MAX))
    {
        ERAISEFMT(BaseReal, overflow) << roofValue;
        return 0L;
    }
    return (PICLONG)roofValue;
}

PICLONG BaseReal::floor () const
{
    if (isNullNumber(flags))
    {
        ERAISE(BaseReal, nullInstance);
        return 0L;
    }

    PICFLOAT floorVal = ::floor(asFloat());

    if ((floorVal < (PICFLOAT)LONG_MIN) || (floorVal > (PICFLOAT)LONG_MAX))
    {
        ERAISEFMT(BaseReal, overflow) << floorVal;
        return 0L;
    }
    return (PICLONG)floorVal;
}

PICLONG BaseReal::integerPart () const
{
    PICFLOAT intPart;
    
    if (isNullNumber(flags))
    {
        ERAISE(BaseReal, nullInstance);
        return 0L;
    }

    modf(asFloat(), &intPart);
    if ((intPart < (PICFLOAT)LONG_MIN) || (intPart > (PICFLOAT)LONG_MAX))
    {
        ERAISEFMT(BaseReal, overflow) << intPart;
    }
    return (PICLONG)intPart;
}

BaseReal BaseReal::fraction () const
{
    PICFLOAT integerPart, fractPart;

    if (isNullNumber(flags))
        return BaseReal(getPrec(flags));

    fractPart = modf(asFloat(), &integerPart);
    return BaseReal(fractPart, getPrec(flags));
}

// --------------------------------------------------------------------------
//      FORMATTING and DEFORMATTING
//
//      Formatting is done by the following simple way:
//      1. Find digits using stdlib function ecvt or fcvt.
//      2. Insert decimal point and append zeroes, if needed.
//      3. Prepend sign.
//      Sign and point are int's because ecvt requires it.

BaseString BaseReal::getImage () const
{
    return getImage(getPrec(flags));
}

BaseString BaseReal::getImage (PICDIGIT digit, char decPoint) const
{
    if (isNullNumber(flags))
        return BaseString();

    int        sign, point;
    PICDIGIT   scale;
    BaseString textImage;

    if (digit == floatReal)
    {
        // take image with 16 significant numbers

        textImage = ecvt(asFloat(), 16, &point, &sign);
        scale = point - 16;
    }
    else
    {
        // take image with 0 decimals

        PICFLOAT value;

        if (!scaleDouble(this->imp, value, getPrec(flags), digit))
            return BaseString();

        textImage = fcvt(value, 0, &point, &sign);

        // The following correction is needed because fcvt returns empty
        // string if value == 0.0:
        if (textImage.size() == 0)
            textImage.append('0', 1);

        scale = digit;
        point = textImage.size() + digit;
    }

    // NOW: scale == scale of the last digit in textImage
    //      point == number of digits to the left of the decimal point
    //      sign  == boolTrue if value was negative

    if (scale > 0)
        textImage.append('0', digit);
    else if (scale < 0)
    {
        if (point < 1)
        {
            textImage.prepend('0', 1 - point);
            point = 1;
        }
        point++;

        if (decPoint == '\0')
            textImage.insert(point, decimalPoint); // Parameterized dec.point
        else
            textImage.insert(point, decPoint); // Given dec.point

        if (digit == floatReal) // Remove trailing zeros
        {
            while (textImage.last() == '0')
                textImage.deleteAfter(textImage.size() - 1);
        }
    }
    if (sign)
        textImage.prepend('-');
    return textImage;
}

// --------------------------------------------------------------------------

void BaseReal::setImage (const BaseString& value)
{
    PICUSHORT imageFlags;
    PICFLOAT  newVal;

    if (evaluateImage(value,newVal,imageFlags))
    {
        if (isNullNumber(imageFlags))
            setNullFlags(flags);
        else if (scaleDouble(newVal,imp,getPrec(imageFlags),getPrec(flags)))
            setNonNullFlags(flags);
    }
}

// --------------------------------------------------------------------------
//      Formatted output
// --------------------------------------------------------------------------

PICOFFSET BaseReal::writeFmt (
    PicTextFmt&     fmt,
    PICOFFSET       i,
    const char*  typeName) const
{
    const PICDIGIT minDigit = -16;
    const PICDIGIT maxDigit =  16;

    PICUSHORT ushort;
    PICDIGIT  digit = getPrecision();

    switch (fmt.format[i])
    {
        case '*':
            i++;
            break;

        case '?':
            fmt.write(typeName, strlen(typeName));
            return i + 1;

        case '.':
            i = fmt.getInt(fmt.format, i + 1, ushort);
            digit = -ushort;
            if (digit < minDigit)
                digit = minDigit;
            break;

        case ':':
            i = fmt.getInt(fmt.format, i + 1, ushort);
            digit = ushort;
            if (digit > maxDigit)
                digit = maxDigit;
            break;

        default:
            return i + 1;
    }

    BaseString image = getImage(digit);
    fmt.write(image.chars(), image.size());
    return i;
}

static PICOFFSET wrFloat (const PICFLOAT* obj, PicTextFmt& fmt, PICOFFSET i)
{
    BaseReal real = *obj;
    return real.writeFmt(fmt, i, "PICFLOAT");
}

static PICFLOAT* wrobjPICFLOAT (PICFLOAT* obj, PICBOOL cleanUp)
{
    if (cleanUp)
    {
        delete obj;
        return NULLPTR;
    }
    PICFLOAT* pf = new PICFLOAT;
    *pf = *obj;
    return pf;
}

PicFmtText& operator << (PicFmtText& txt, PICFLOAT obj)
{
    return txt.wrFormat (
        &obj, (PICWRFUNC*)wrFloat, (PICWROBJFUNC*)wrobjPICFLOAT);
}

static PICOFFSET wrReal (const BaseReal* obj, PicTextFmt& fmt, PICOFFSET i)
{
    return obj->writeFmt(fmt, i, "BaseReal");
}

static BaseReal* wrobjBaseReal (BaseReal* obj, PICBOOL cleanUp)
{
    if (cleanUp)
    {
        delete obj;
        return NULLPTR;
    }
    return new BaseReal(*obj);
}

PicFmtText& operator << (PicFmtText& txt, const BaseReal& obj)
{
    return txt.wrFormat (
        &obj, (PICWRFUNC*)wrReal, (PICWROBJFUNC*)wrobjBaseReal);
}

// --------------------------------------------------------------------------
//      PICFLOAT operations
// --------------------------------------------------------------------------

BaseString asString (PICFLOAT value)
{
    PicStringText text;
    text() << value;
    return text;
}

BaseString& operator << (BaseString& toStr, PICFLOAT value)
{
    toStr.append(asString(value));
    return toStr;
}

PicText& PicText::wrn (PICLONG number)
{
    return wr(asString(number));
}

