// --------------------------------------------------------------------------
//  ARRAY OBJECT implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#include <parray.hxx>
#include <piterimp.hxx>
#include <psemaph.hxx>
#include <picerror.hxx>
#include <picbase.erh>
#include <pstring.hxx>

#define RET_CALL_GUARD(className, func, defval) \
    if (TYPE_GUARD(className::classId(), *this)) \
        return ((className*)this)->func; \
    return defval

PICSIZE PicObject::size () const
{
    if (id() >= PicString::classId())
        return ((PicString*)this)->size();
    RET_CALL_GUARD(PicCollection, size(), 0);
}

// --------------------------------------------------------------------------

MAKE_PICABSOBJECTCLASS(PicArrayObject, PicSeqIxCollect)
MAKE_PICOBJECTCLASS(PicArray, PicArrayObject)
MAKE_PICOBJECTCLASS(PicSortedArray, PicArrayObject)

// --------------------------------------------------------------------------
//      PicArrayObject
// --------------------------------------------------------------------------

// --------------------------------------------------------------------------
//      ASSIGNMENT

PicObject& PicObject::operator = (const PicObject &obj)
{
    return(*this);
}

PicObject& PicArrayObject::operator = (const PicObject& obj)
{
    if (obj.id() >= id())
        return *this = (const PicArrayObject&)obj;
    else
        return PicSeqIxCollect::operator=(obj);
}

// --------------------------------------------------------------------------
//      COMPARISONS

PICBOOL PicArrayObject::operator == (const PicObject& obj) const
{
    if (id() == obj.id())
        return equals((PicArrayObject&)obj);

    return boolFalse;
}

// --------------------------------------------------------------------------
//      NULL

PICBOOL PicArrayObject::isNull  () const { return array.isNull(); }
void    PicArrayObject::setNull ()       { array.setNull(); }

// --------------------------------------------------------------------------
//      ACCESSING COMPONENTS

PicObject& PicArrayObject::access (const PicIterator& here)
{
    PICPOSITION accessPos = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(array.contains(accessPos));
    return *array[accessPos];
}

PICINDEX PicArrayObject::indexAt (const PicIterator& here) const
{
    PICPOSITION indexPos = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(array.contains(indexPos));
    return (PICINDEX)indexPos;
}

PICBOOL PicArrayObject::findIndex (PICINDEX index, PicIterator& here) const
{
    SET_ITERATE_INDEX(here, index);
    return array.contains((PICPOSITION)index);
}

// --------------------------------------------------------------------------

PICBOOL PicArrayObject::insert (PICINDEX index, PicIterator& here)
{
    index = (PICINDEX)array.insert (
        index != outOfIndex ? (PICPOSITION)index : 0);
    SET_ITERATE_INDEX(here, index);
    return index != 0;
}

PICBOOL PicArrayObject::put (PICINDEX index, PicIterator& here)
{
    array.put((PICPOSITION)index);
    SET_ITERATE_INDEX(here, index);
    return index != 0;
}

// --------------------------------------------------------------------------

PICBOOL PicArrayObject::next (PicIterator& i) const
{
    PICPOSITION current = (PICPOSITION)GET_ITERATE_INDEX(i);
    current = array.next(current);
    SET_ITERATE_INDEX(i, current);
    return current != 0;
}

PICBOOL PicArrayObject::previous (PicIterator& i) const
{
    PICPOSITION current = (PICPOSITION)GET_ITERATE_INDEX(i);
    current = array.previous(current);
    SET_ITERATE_INDEX(i, current);
    return current != 0;
}

// --------------------------------------------------------------------------
//      DELETING

void PicArrayObject::deleteAll ()
{
    array.deleteAll();
}

void PicArrayObject::removeAll ()
{
    array.removeAll();
}

// --------------------------------------------------------------------------
//      ASKING ATTRIBUTES

PICSIZE PicArrayObject::size () const { return array.size(); }


// --------------------------------------------------------------------------
//      PicArray
// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR, COPY & ASSIGNMENT

PicArray::PicArray () : PicArrayObject() {}

PicArray::PicArray (const PicArray& arr) : PicArrayObject(arr) {}

PicArray::PicArray (const PicObjectPtrArray& objarr) :
    PicArrayObject(objarr) {}

PicArray::PicArray (const PicCollection& fromCollection) :
    PicArrayObject()
{
    *this = fromCollection;
}

PicArray::~PicArray () {}

// --------------------------------------------------------------------------
//      MODIFYING COMPONENTS

PicObject* PicArray::replace (PicIterator& here, PicObject* pobj)
{
    PICPOSITION replaceIndex = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(objectArray().inRange(replaceIndex));
    return objectArray().replace(replaceIndex, pobj);
}

PICBOOL PicArray::purge (PicIterator& here, PICBOOL inOrder)
{
    PICPOSITION purgeIndex = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(objectArray().inRange(purgeIndex));
    objectArray().purge(purgeIndex, inOrder);
    return previous(here);
}


// --------------------------------------------------------------------------
//      PicSortedArray
// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR, COPY & ASSIGNMENT

PicSortedArray::PicSortedArray () :
    PicArrayObject(), order(PicSeqIxCollect::ascending) {}

PicSortedArray::PicSortedArray (const PicSortedArray& arr) :
    PicArrayObject(arr), order(PicSeqIxCollect::ascending) {}

PicSortedArray::PicSortedArray (const PicObjectPtrArray& objarr) :
    PicArrayObject(objarr), order(PicSeqIxCollect::ascending) {}

PicSortedArray::PicSortedArray (const PicCollection& fromCollection) :
    PicArrayObject(), order(PicSeqIxCollect::ascending)
{
    *this = fromCollection;
}

PicSortedArray::~PicSortedArray () {}

// --------------------------------------------------------------------------
//      MODIFYING COMPONENTS

PicObject* PicSortedArray::replace (PicIterator& here, PicObject* pobj)
{
    PICPOSITION replaceIndex = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(objectArray().inRange(replaceIndex));

    PICPOSITION idx = objectArray().lastPos() + 1;

    if (replaceIndex >= idx)
        objectArray().put(replaceIndex = idx);

    if (pobj)
    {
        while (idx = objectArray().previous(replaceIndex))
        {
            switch ((*order)(*pobj, *objectArray()[idx]))
            {
                case 0:
                    delete pobj;
                    pobj = NULLPTR;
                    goto Out;
                case 1:
                    goto Next;
                case -1:
                    objectArray().swap(idx, replaceIndex);
                    replaceIndex = idx;
                    // Fall through
            }
        }

      Next:
        while (idx = objectArray().next(replaceIndex))
        {
            switch ((*order)(*objectArray()[idx], *pobj))
            {
                case 0:
                    delete pobj;
                    pobj = NULLPTR;
                    goto Out;
                case 1:
                    goto Next2;
                case -1:
                    objectArray().swap(idx, replaceIndex);
                    replaceIndex = idx;
                    // Fall through
            }
        }
      Next2:
        idx = replaceIndex;
    }

  Out:
    PicObject* old = objectArray().replace(replaceIndex, NULLPTR);

    if (!pobj)
        objectArray().purge(replaceIndex, boolTrue);
    else
        objectArray()[replaceIndex] = pobj;

    SET_ITERATE_INDEX(here, idx);
    return old;
}

PICBOOL PicSortedArray::insert (PICINDEX index, PicIterator& here)
{
    index = (PICINDEX)objectArray().insert();
    SET_ITERATE_INDEX(here, index);
    return index != 0;
}

PICBOOL PicSortedArray::put (PICINDEX index, PicIterator& here)
{
    if (index > (PICINDEX)objectArray().lastPos())
        index = (PICINDEX)objectArray().lastPos() + 1;
    SET_ITERATE_INDEX(here, index);
    return objectArray().put(index);
}

PICBOOL PicSortedArray::purge (PicIterator& here, PICBOOL)
{
    PICPOSITION purgeIndex = (PICPOSITION)GET_ITERATE_INDEX(here);
    DEBUG_REQUIRE(objectArray().inRange(purgeIndex));
    objectArray().purge(purgeIndex, boolTrue); // Must retain order
    return previous(here);
}

PICBOOL PicSortedArray::addPtr (PicObject* pobj)
{
    PicIterator here;
    if (insert(outOfIndex, here))
    {
        replace(here, pobj);
        return boolTrue;
    }
    return boolFalse;
}

// --------------------------------------------------------------------------
//      SORTING

static PICOBJSORTER* pfSorter;

static int sortCompare (const void* elem1, const void* elem2)
{
    if (!*(PicObject**)elem1 || !*(PicObject**)elem2)
        return (!*(PicObject**)elem1 - !*(PicObject**)elem2);
    return (*pfSorter)(**(PicObject**)elem1, **(PicObject**)elem2);
}

static PicSemaphore semArray;

void PicArray::sort (PICOBJSORTER* pfCompare)
{
    PicCritSec sec (semArray);
    ::pfSorter = pfCompare;
    objectArray().sort(::sortCompare);
}

void PicSortedArray::sort (PICOBJSORTER* pfCompare)
{
    PicCritSec sec (semArray);
    ::pfSorter = pfCompare;
    objectArray().sort(::sortCompare);
    order = pfCompare;
}

PICBOOL PicSortedArray::findObject (
    const PicObject& obj,
    PicIterator&     here) const
{
    PICPOSITION pos = 0;

    if (!isEmpty())
    {
        PPicObject pobj = (PicObject*)&obj;

        PicCritSec sec (semArray);
        ::pfSorter = order;
        pos = objectArray().findSorted(&pobj, ::sortCompare);

        SET_ITERATE_INDEX(here, pos);
    }
    return pos != 0;
}
