// --------------------------------------------------------------------------
//  PIC DATE implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#include <pdate.hxx>
#include <bdatetim.hxx>
#include <pstring.hxx>
#include <picerror.hxx>
#include <picbase.erh>
#include <pcollect.hxx>
#include <piterate.hxx>

MAKE_PICOBJECTCLASS(PicDate,PicPrimitive)

// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR, COPY & ASSIGNMENT

PicDate::PicDate () : PicPrimitive(), BaseDate() {}

PicDate::PicDate (PicRelativeDate relative, PICLONG unitInterval) :
    PicPrimitive(), BaseDate(relative, unitInterval) {}

PicDate::PicDate (PICYEAR year, PICMONTH month, PICDAY day) :
    PicPrimitive(), BaseDate(year, month, day) {}

PicDate::PicDate (
    const BaseString& image,
    PicDateStyle      style,
    char           delimiter)
    :
    PicPrimitive(), BaseDate(image, style, delimiter) {}

PicDate::PicDate (
    const BaseString& image,
    PicDateStyle      style)
    :
    PicPrimitive(), BaseDate(image, style, defaultDelimiter()) {}

PicDate::PicDate (
    const BaseString& image)
    :
    PicPrimitive(), BaseDate(image, defaultStyle(), defaultDelimiter()) {}

PicDate::PicDate (const BaseDate& date) :
    PicPrimitive(), BaseDate(date) {}

PicDate::PicDate (const PicDate& date) :
    PicPrimitive(), BaseDate(date) {}

PicObject& PicDate::operator = (const PicObject& obj)
{
    if (TYPE_GUARD(classId(), obj))
        BaseDate::operator = ((const PicDate&)obj);
    return *this;
}

// --------------------------------------------------------------------------
//      COMPARISONS

PICBOOL PicDate::operator == (const PicObject& obj) const
{
    if (id() == obj.id())
        return BaseDate::operator == ((const PicDate&)obj);
    return boolFalse;
}

PICBOOL PicDate::operator < (const PicObject& obj) const
{
    if (id() == obj.id())
        return BaseDate::operator < ((const PicDate&)obj);
    ERAISEFMT(PicObject, uncomparable)
        >>"OBJ1" << *this
        >>"OBJ2" << obj;
    return boolFalse;
}

// --------------------------------------------------------------------------
//      TEXTUAL IMAGE

static void getFormat (
    const PicObject&   format,
    PicDateResolution& resolution,
    PicDateStyle&      style,
    char&           delimiter,
    char&           leader)
{
    PicIterator iter;
    PICBOOL     pictFmt = format.IS_A(PicString);

    if (!pictFmt && format.find(DELIMITER,iter))
        delimiter = format.current(iter).asChar();
    else
        delimiter = BaseDate::defaultDelimiter();

    if (!pictFmt && format.find(LEADER,iter))
        leader = format.current(iter).asChar();
    else
        leader = BaseDate::defaultLeader();

    if (!pictFmt && format.find(STYLE,iter))
        switch (format.current(iter).asInt())
        {
            case FORMAT_YEAR_FIRST:
                style = BaseDate::yearFirst;
                break;
            case FORMAT_MONTH_FIRST:
                style = BaseDate::monthFirst;
                break;
            case FORMAT_DAY_FIRST:
                style = BaseDate::dayFirst;
                break;
            case FORMAT_MONTH_BY_NAME:
                style = BaseDate::monthByName;
                break;
            case FORMAT_MONTH_BY_SHORT_NAME:
                style = BaseDate::monthByShortName;
                break;
            default:
                ERAISE(BaseDate, formatError);
                style = BaseDate::defaultStyle();
                break;
        }
    else
        style = BaseDate::defaultStyle();

    if (!pictFmt && format.find(RESOLUTION,iter))
        switch (format.current(iter).asInt())
        {
            case FORMAT_YEARS_ONLY:
                resolution = BaseDate::yearsOnly;
                break;
            case FORMAT_YEARS_TO_MONTHS:
                resolution = BaseDate::yearsToMonths;
                break;
            case FORMAT_YEARS_TO_DAYS:
                resolution = BaseDate::yearsToDays;
                break;
            default:
                ERAISE(BaseDate, formatError);
                resolution = BaseDate::defaultResolution();
                break;
        }
    else
        resolution = BaseDate::defaultResolution();
}

BaseString PicDate::getImage (const PicObject& format) const
{
    BaseString image;

    if (!pictureFmt(format, image))
    {
        PicDateResolution resolution;
        PicDateStyle      style;
        char           delimiter;
        char           leader;

        getFormat(format, resolution, style, delimiter, leader);
        image = PicString::format (
            image = BaseDate::getImage(resolution, style, delimiter, leader),
            format);
    }
    return image;
}

void PicDate::setImage (const PicObject& format, const BaseString& image)
{
    PicDateResolution resolution;
    PicDateStyle      style;
    char           delimiter;
    char           leader;

    getFormat(format, resolution, style, delimiter, leader);
    BaseDate::setImage(image, style, delimiter);
}

// --------------------------------------------------------------------------
//      FORMATTED OUTPUT

PICOFFSET PicDate::writeFmt (PicTextFmt& fmt, PICOFFSET i) const
{
    static const char* typeName = id().className();

    BaseDateTime datetime = *(BaseDate*)this;

    if (isNull() && fmt.format[i] != '?')
    {
        PicTextFmt nullFmt (PicStream::null(), fmt.format, 0);
        return datetime.writeFmt(nullFmt, i, typeName);
    }

    return datetime.writeFmt(fmt, i, typeName);
}

// --------------------------------------------------------------------------
//      SET TYPED VALUE

PICBOOL PicDate::setTypedValue (const PicObject& constraints)
{
    PicIterator here;
    const PicIxCollect& cntr = AS(PicIxCollect, constraints);

    if (isNull() && cntr.find(NULL_IS_CURRENT, here))
        setRelativeDate(today, cntr.at(here).asInt());

    if (cntr.find(NOT_BEFORE, here))
    {
        PICSHORT days = cntr.at(here).asInt();
        BaseDate limit (today, +days);
        if (*this < limit)
        {
            ERAISEFMT(BaseDate, tooEarly) << limit;
            return boolFalse;
        }
    }
    if (cntr.find(NOT_AFTER, here))
    {
        PICSHORT days = cntr.at(here).asInt();
        BaseDate limit (today, +days);
        if (*this > limit)
        {
            ERAISEFMT(BaseDate, tooLate) << limit;
            return boolFalse;
        }
    }

    return boolTrue;
}
