// --------------------------------------------------------------------------
//  PicDateTime implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#include <pdatetim.hxx>
#include <pstring.hxx>
#include <picerror.hxx>
#include <picbase.erh>
#include <pcollect.hxx>
#include <piterate.hxx>

MAKE_PICOBJECTCLASS(PicDateTime,PicPrimitive)

// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR, COPY & ASSIGNMENT

PicDateTime::PicDateTime () : dt() {}

PicDateTime::PicDateTime (const BaseDate& date) : dt(date) {}

PicDateTime::PicDateTime (const BaseTime& time) : dt(time) {}

PicDateTime::PicDateTime (const BaseDate& d, const BaseTime& t) : dt(d,t) {}

PicDateTime::PicDateTime (
    PicRelativeDate erd,
    PICLONG         dateInterval,
    PICHOURCOUNT    timeInterval)
    :
    dt(erd, dateInterval, timeInterval)
{}

PicDateTime::PicDateTime (
    PICYEAR year, PICMONTH month,   PICDAY day,
    PICHOUR hour, PICMINUTE minute, PICSECOND second)
    :
    dt(year, month, day, hour, minute, second)
{}

PicDateTime::PicDateTime (const BaseDateTime& copy) :
    PicPrimitive(), dt(copy) {}

PicDateTime::PicDateTime (const PicDateTime& copy) :
    PicPrimitive(), dt(copy.dt) {}

PicObject& PicDateTime::operator = (const PicObject& obj)
{
    if (TYPE_GUARD(classId(), obj))
        *this = ((const PicDateTime&)obj);
    return *this;
}

// --------------------------------------------------------------------------
//      COMPARISONS

PICBOOL PicDateTime::operator == (const PicObject& obj) const
{
    if (id() == obj.id())
        return *this == ((const PicDateTime&)obj);
    return boolFalse;
}

PICBOOL PicDateTime::operator < (const PicObject& obj) const
{
    if (id() == obj.id())
        return *this < ((const PicDateTime&)obj);
    ERAISEFMT(PicObject, uncomparable)
        >>"OBJ1" << *this
        >>"OBJ2" << obj;
    return boolFalse;
}

// --------------------------------------------------------------------------
//      FORMATTING
// --------------------------------------------------------------------------

BaseString PicDateTime::getImage (const PicObject& format) const
{
    BaseString image;

    pictureFmt(format, image);
    return image;
}

void PicDateTime::setImage (const PicObject& format, const BaseString& image)
{
    dt.setImage(image);
}

PICOFFSET PicDateTime::writeFmt (PicTextFmt& fmt, PICOFFSET i) const
{
    static const char* typeName = id().className();

    if (isNull() && fmt.format[i] != '?')
    {
        PicTextFmt nullFmt (PicStream::null(), fmt.format, 0);
        return dt.writeFmt(nullFmt, i, typeName);
    }

    return dt.writeFmt(fmt, i, typeName);
}

// --------------------------------------------------------------------------
//      SET TYPED VALUE

PICBOOL PicDateTime::setTypedValue (const PicObject& constraints)
{
    PicIterator here;
    const PicIxCollect& cntr = AS(PicIxCollect, constraints);

    if (isNull() && cntr.find(NULL_IS_CURRENT, here))
    {
        dt.date.setRelativeDate(BaseDate::today, cntr.at(here).asInt());
        dt.time.setRelativeTime(0);
    }

    if (cntr.find(NOT_BEFORE, here))
    {
        PICSHORT days = cntr.at(here).asInt();
        BaseDate limit (BaseDate::today, +days);
        if (dt.date < limit)
        {
            ERAISEFMT(BaseDate, tooEarly) << limit;
            return boolFalse;
        }
    }
    if (cntr.find(NOT_AFTER, here))
    {
        PICSHORT days = cntr.at(here).asInt();
        BaseDate limit (BaseDate::today, +days);
        if (dt.date > limit)
        {
            ERAISEFMT(BaseDate, tooLate) << limit;
            return boolFalse;
        }
    }

    return boolTrue;
}
