// --------------------------------------------------------------------------
//  ERROR LOG implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>

extern "C" {
#include <time.h>
#include <stdio.h>
}

#include <perrlog.hxx>
#include <psemaph.hxx>
#include <pthread.hxx>

// --------------------------------------------------------------------------
//      CLASS DATA
//
//      - reference to error manager parameters
//      - count of error log users
//      - pointer to circular log writer
//      - log open status
//      - sequential log data

static PicErrMgrParam* param = &PicErrorManager::param();

static PICUSHORT userCount = 0;

// FWRITECIRCLOG* picWriteToCircularLog = NULLPTR;
PICBOOL picWriteToCircularLog = boolFalse;


// --------------------------------------------------------------------------

static void closeCircularLog(void)
{

}

static void writeToCircularLog(
    const char *className, 
    PICUSHORT status, 
    const char *logText,
    PICUSHORT line, 
    const char *fileName)
{

}

// --------------------------------------------------------------------------


PICBOOL PicErrorLog::opened = boolFalse;

static PICBOOL seqLogOpen = boolFalse;

static int seqLog = 0; // file descriptor

// --------------------------------------------------------------------------
//      SEQUENTIAL LOG

static PicSemaphore semErrLog;

static void closeSeqLog ()
{
    if (seqLogOpen)
    {
        close(seqLog);
        seqLogOpen = boolFalse;
    }
}

static BaseString reserveString;
static PICBOOL writeToSeqLogImp (BaseString&);

PICBOOL writeToSeqLog (const char* logText)
{
    PicCritSec sec (semErrLog);
    reserveString = logText;
    return writeToSeqLogImp(reserveString);
}

PICBOOL writeToSeqLog (BaseString& logText)
{
    PicCritSec sec (semErrLog);
    return writeToSeqLogImp(logText);
}

static PICBOOL writeToSeqLogImp (BaseString& logText)
{
    PICUINT usRet = 0;
    PICULONG usBytesWritten;

    if (!seqLogOpen)
    {
//    seqLog = open(param->elogFileName, O_WRONLY | O_CREAT | O_APPEND, 0660);
        seqLog = open(param->elogFileName, O_WRONLY | O_CREAT | O_TRUNC, 0660);
        if (seqLog < 0)
        {
            usRet = errno;
            static const char* errLog2 = "ERROR$.LOG";

            if (param->elogFileName == errLog2)
            {
                ERAISEFMT_FATAL(PicErrorManager, cannotOpenLog)
                    << errLog2 << usRet;
                return boolFalse;
            }
            else
            {
                PicStringText err;
                err(PicErrorManager::errors(PicErrorManager::cannotOpenLog))
                    << param->elogFileName << usRet;

                param->elogFileName = errLog2;
                if (!writeToSeqLogImp(err.flush()))
                    return boolFalse;
            }
        }
        else // Open ok, now append
        {
            seqLogOpen = boolTrue;
        }
    }

//  ----------------------------
//  Stamp before log text:
//
//           1         2
//  012345679012345678901234567
//  5009[01] 09/24/93 13:33:03
//  ----------------------------

    int pid;
    PICUSHORT tid;
    pid = getpid();
    tid = PicThread::self()->currentId();

    char*      pchTime;
    time_t     ltime;
    char       infoBuf[64];
    PicCharStream infoStm  (infoBuf, sizeof(infoBuf));
    PicFmtText    infoText (infoStm);

    time(&ltime);
    pchTime = asctime(localtime(&ltime));
    pchTime[19] = '\0';
    infoText("@[H4][@[H2]] @*  ")
        << pid << tid << pchTime + 4 << '\0' ENDFMT;

    logText.prepend(infoBuf);

    while (logText.last() == '\n')
        logText.deleteAfter(logText.size() - 1);
    logText.replace("\n", "\r\n         ");
    logText.append("\r\n");

    errno = 0;
    usRet = (write(seqLog, logText.chars(), logText.size())==0) ? 0 : errno;
    if (usRet)
    {
        ERAISEFMT_FATAL(PicErrorManager, logWriteFailed)
            << param->elogFileName << usRet;
    }
    return usRet == 0;
}


// --------------------------------------------------------------------------
//      LOOP GUARD
//
//      loopGuard prevents recursive calls to error log functions.

static PICBOOL loopGuard = boolFalse;

static PICBOOL enterGuard ()
{
    if (loopGuard)
        return boolFalse;

    return (loopGuard = boolTrue);
}

inline void exitGuard () { loopGuard = boolFalse; }

// --------------------------------------------------------------------------
//      REGISTER USERS OF ERROR LOG

void PicErrorLog::login ()
{
    PicCritSec sec (semErrLog);

    userCount++;

    if (opened)
    {
        opened = boolFalse;
        closeSeqLog();
        if (picWriteToCircularLog)
            closeCircularLog();
    }
}

void PicErrorLog::logout ()
{
    PicCritSec sec (semErrLog);

    if (userCount)
        userCount--;

    if (userCount == 0 && opened)
    {
        opened = boolFalse;
        closeSeqLog();
        if (picWriteToCircularLog)
            closeCircularLog();
    }
}

// --------------------------------------------------------------------------

inline PICBOOL stampLegal (PicErrorLogStamp stamp)
{
    return (stamp >= PicErrorLog::baseStamp &&
            stamp <= PicErrorLog::internalStamp);
}

// --------------------------------------------------------------------------
//      LOGGING

void PicErrorLog::write (
    const PicError*  message,
    PicErrorLogStamp aStamp)
{
    if (!message || !enterGuard())
        return;

    PicStringText    logText;
    PicErrorManager* mgr = PicErrorManager::get();
    const char*   stamp = stampLegal(aStamp) ?
        param->elogStamps[aStamp] : "TEXT@*";

    if (param->elogCircular && picWriteToCircularLog)
        message->description(logText.text(), stamp);
    else
    {
        BaseString wholeStamp = param->elogStamps[baseStamp];
        wholeStamp.append(stamp);
        message->description(logText.text(), wholeStamp.chars());
    }

    if (!mgr || !mgr->logErrorHook(logText))
    {
        if (param->elogCircular && picWriteToCircularLog)
        {
            writeToCircularLog (
                message->id().className(),
                message->status(),
                logText.chars(),
                message->line(),
                message->file());
        }
        else
            writeToSeqLog(logText);

        opened = boolTrue;
    }

    exitGuard();
}

void PicError::log (BaseString& logText)
{
    if (!enterGuard())
        return;

    PicErrorManager* mgr = PicErrorManager::get();

    if (!mgr || !mgr->logErrorHook(logText))
    {
        if (param->elogCircular && picWriteToCircularLog)
        {
            writeToCircularLog("", 0, logText.chars(), 0, "");
        }
        else
            writeToSeqLog(logText);

        PicErrorLog::opened = boolTrue;
    }
    exitGuard();
}

void PicErrorLog::writeMemoryError (
    const char*  logText,
    PICUSHORT       lineNumber,
    const char*  fileName)
{
    if (!enterGuard())
        return;

    if (param->elogCircular && picWriteToCircularLog)
    {
        writeToCircularLog (
            "Out of memory",
            ERR_OUT_OF_MEMORY,
            logText,
            lineNumber,
            fileName);
    }
    else
    {
        char       message [128];
        PicCharStream msgStm  (message, sizeof(message));
        PicFmtText    msgText (msgStm);

        if (fileName)
        {
            msgText("Out of memory in @*(@*):\n@*\n")
                << fileName << lineNumber << logText << '\0' ENDFMT;
        }
        else
            msgText("Out of memory: @*\n") << logText << '\0' ENDFMT;

        writeToSeqLog(message);
    }

    opened = boolTrue;
    exitGuard();
}


