// --------------------------------------------------------------------------
//  PicErrorManager & PicErrorContext implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <stdio.h>
}

#include <perrmgr.hxx>
#include <picerror.hxx>
#include <perrlog.hxx>
#include <pthread.hxx>

MAKE_PICERRORMGRCLASS(PicErrorManager)
MAKE_PICCLASS(PicNotifier)
MAKE_PICCLASS(PicErrorContext)
MAKE_PICCLASS(PicTestContext)

ERROR_MESSAGES(PicErrorManager)
PARAMSTR(duplicateManager, "Duplicate error manager created.")
PARAMSTR(cannotOpenLog, "Can not open log.")
PARAMSTR(logWriteFailed, "Can not write to log.")
PARAMSTR(error, "Internal error")
END_PARAM_STRINGS

// --------------------------------------------------------------------------
//      The default notifier
// --------------------------------------------------------------------------

class PicDefaultNotifier : public PicNotifier
{
    public:

    virtual void fatalExit (
        PICUSHORT   lineNumber,
        const char* fileName)
    {
        printf("\nFatal error in %s(%i)\n", fileName, lineNumber);
        if (PicError::isDetected())
            printf("%s\n", PicError::current().description().chars());
    }

    virtual void memoryExit (
        const char* message,
        PICUSHORT   lineNumber,
        const char* fileName)
    {
        printf("\nOut of memory in %s(%i):\n%s\n", 
                fileName, lineNumber, message);
    }

    virtual void status (const BaseString& msg)
    {
        printf("%s\n", msg.chars());
    }

    virtual void progress (const BaseString& msg, PICUSHORT readyness)
    {
        printf("%s, %i ready.\r", msg.chars(), readyness);
        if (readyness >= 100)
            printf("\n");
    }

};

static PicNotifier& defaultNotifier ()
{
    static PicDefaultNotifier notifier;
    return notifier;
}

// --------------------------------------------------------------------------
//      IMPLEMENTATION FOR PicErrorManager
// --------------------------------------------------------------------------
//      A pointer to the current error manager is retained here (error
//      managers themselves are automatic objects in the stack of the
//      current thread).

PicErrMgrParam& PicErrorManager::param ()
{
    static PicErrMgrParam defParams;
    static PICBOOL inited = boolFalse;

    if(!inited){
        inited = boolTrue;

        defParams.elogCircular = boolFalse;
        defParams.elogFilter = 
            PicErrorLog::raised | PicErrorLog::held | PicErrorLog::discarded;
        defParams.elogSuspend = 0;
        defParams.elogFileName = "ERROR.LOG";
        defParams.elogRecLength = 220;
        defParams.elogFileSize = 100;
        defParams.elogStamps[0] = "CLASS@* error STATUS@* in FILE@*(LINE@*)";
        defParams.elogStamps[1] = "Error cleared:    TEXT@*";
        defParams.elogStamps[2] = "Error discarded:  TEXT@*";
        defParams.elogStamps[3] = "Error held:       TEXT@*";
        defParams.elogStamps[4] = "Error raised:     TEXT@*";
        defParams.elogStamps[5] = "-------- Error manager reset --------";
        defParams.elogStamps[6] = "Fatal error:      TEXT@*";
        defParams.elogStamps[7] = 
            "Previous error was internal to error manager (FILE@* on line LINE@*).";
        defParams.emgrRaiseTemplate = "FILE@*(LINE@*) : CLASS@* : TEXT@*";
        defParams.emgrHoldTemplate = "-- FILE@*(LINE@*) : CLASS@* : TEXT@*";
        defParams.emgrNoDescription = "-- No description available --";
    }

    return defParams;
}

static PicSemaphore errManMutex;

const  PICUSHORT maxErrorManagers = 20;
static PicErrorManager* picErrorManagers[maxErrorManagers];
static PICUSHORT        picErrorManagerOwners[maxErrorManagers];

// Since we can get here from PicThread, we must ensure
// that we don't call any functions of PicThread that use
// its semMutex (this could deadlock the current thread)

PicErrorManager* PicErrorManager::get ()
{
    PicThread* curr = PicThread::self();
    if(!curr)
        return NULLPTR;

    for(PICUSHORT i=0; i<maxErrorManagers; i++){
        if(picErrorManagerOwners[i] == curr->id())
            return picErrorManagers[i];
    }
    return NULLPTR;
}

PicErrorManager::PicErrorManager () :
    pNotifier(&defaultNotifier())
{
    PicCritSec critSec(errManMutex);
    PicThread* curr = PicThread::self();
    if(!curr){
        FATAL_EXIT();
        return;
    }

    for(PICUSHORT i = 0; i < maxErrorManagers; i++){
        if(picErrorManagerOwners[i] == curr->id()){
            ERAISE(PicErrorManager, duplicateManager);
            return;
        }
        if(picErrorManagerOwners[i] == 0){
            picErrorManagerOwners[i] = curr->id();
            picErrorManagers[i]   = this;
            currentContext        = this;
            memoryError           = boolFalse;
            return;
        }
    }
    ERAISE(PicErrorManager, error);
}

PicErrorManager::~PicErrorManager ()
{
    PicCritSec critSec(errManMutex);
    PicThread* curr = PicThread::self();
    if(!curr)
        return;
    if (currentError())
    {
        PicError::cumulate (
            "*** Still one error left in program. "
            "Please, debug the error away!");
        FATAL_EXIT();
    }
    for(PICUSHORT i = 0; i < maxErrorManagers; i++){
        if(picErrorManagerOwners[i] == curr->id()){
            picErrorManagerOwners[i] = 0;
            picErrorManagers[i] = NULLPTR;
        }
    }
}

// --------------------------------------------------------------------------
//      IMPLEMENTATION FOR PicErrorContext
// --------------------------------------------------------------------------
//      Error context creation has effect only, if the error manager
//      object exists. Note the order of execution, when the error
//      manager is created:
//          1) construct PicErrorContext (no active error manager)
//          2) construct PicErrorManager (active error manager is set)
//      and destroyed:
//          1) destroy PicErrorManager (deactivate error manager)
//          2) destroy PicErrorContext (no active error manager)

PicErrorContext::PicErrorContext()
{
    error = NULLPTR;

    PicErrorManager* mgr = PicErrorManager::get();
    if (mgr)
        mgr->createContext(this);
    else
    {
        previous = NULLPTR;
        // To ensure sensible behaviour in single-context state:
        PicError::clearCurrent();
    }
}

PicErrorContext::~PicErrorContext()
{
    PicErrorManager* mgr = PicErrorManager::get();
    if (mgr)
        mgr->deleteContext();
}

// --------------------------------------------------------------------------
//      PicTestContext
// --------------------------------------------------------------------------

PicTestContext::PicTestContext () :
    PicErrorContext()
{
    PicErrorManager::param().elogSuspend++;
}

PicTestContext::~PicTestContext ()
{
    PicError::clearCurrent();
    if (PicErrorManager::param().elogSuspend > 0)
        PicErrorManager::param().elogSuspend--;
}

