// --------------------------------------------------------------------------
//  ERROR MANAGER LPICL1 - implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#include <perrmgr1.hxx>
#include <perrlog.hxx>
#include <picbase.erh>

MAKE_PICERRORMGRCLASS(PicL1ErrorManager)

static PicErrMgrParam* prm = &PicErrorManager::param();

// --------------------------------------------------------------------------
//      CONSTRUCTOR AND DESTRUCTOR

PicL1ErrorManager::PicL1ErrorManager () : PicErrorManager ()
{
    PicErrorLog::login();
}

PicL1ErrorManager::~PicL1ErrorManager ()
{
    reset();
    PicErrorLog::logout();
}

// --------------------------------------------------------------------------
//      CLEARING ERRORS

void PicL1ErrorManager::clearCurrent ()
{
    if (outOfMemory())
    {
        // The application may be able to deallocate some memory itself, so:
        memoryError = boolFalse;

        PicStringText clearStamp;
        clearStamp(prm->elogStamps[PicErrorLog::clearStamp])
            >>"TEXT" << "Out-of-memory error.";
        PicErrorLog::write(clearStamp);
    }
    else if (isRaised())
    {
        if (prm->elogFilter & PicErrorLog::cleared && !prm->elogSuspend)
        {
            PicErrorContext e;

            PicErrorLog::write (
                currentContext->previous->error, PicErrorLog::clearStamp);

            if (CHECK_INTERNAL_ERROR())
                return;
        }
        delete currentContext->error;
        currentContext->error = NULLPTR;
    }
}

void PicL1ErrorManager::reset ()
{
    // Discard errors in the context stack
    // and logs the reset action.

    PicErrorContext* context = currentContext;

    while (context)
    {
        if (context->error)
        {
            if (prm->elogFilter & PicErrorLog::discarded && !prm->elogSuspend)
            {
                PicErrorLog::write (
                    context->error, PicErrorLog::discardStamp);
            }
            delete context->error;
            context->error = NULLPTR;
        }
        context = context->previous;
    }

    if (prm->elogFilter && PicErrorLog::open())
    {
        BaseString resetStamp = prm->elogStamps[PicErrorLog::resetStamp];
        PicErrorLog::write(resetStamp);
        if (currentContext->error)
        {
            // Error when writing resetStamp...
            delete currentContext->error;
            currentContext->error = NULLPTR;
        }
    }
}

// --------------------------------------------------------------------------
//      RAISING ERRORS

void PicL1ErrorManager::raise (const PicErrorText& message, PICBOOL hold)
{
    PicError* error = NULLPTR;

    {
        PicErrorContext e;
        error = new PicError (message, this);
        CHECK_INTERNAL_ERROR();
    }

    // There is always current context

    if (currentContext->error)   // The context was not clear
    {
        if (hold)
            error->appendError(currentContext->error);
        else
        {
            if (prm->elogFilter & PicErrorLog::discarded && !prm->elogSuspend)
            {
                PicErrorContext e;
                PicErrorLog::write (
                    currentContext->previous->error, PicErrorLog::discardStamp);
                CHECK_INTERNAL_ERROR();
            }
            delete currentContext->error;
        }
    }

    currentContext->error = error;

    if (prm->elogFilter & (hold ? PicErrorLog::held : PicErrorLog::raised)
        && !prm->elogSuspend
        )
    {
        PicErrorContext e;
        PicErrorLog::write (
            error, hold ? PicErrorLog::holdStamp : PicErrorLog::raiseStamp);
        CHECK_INTERNAL_ERROR();
    }
}

void PicL1ErrorManager::cumulate (const BaseString& cumulationText)
{
    if (isRaised())
    {
        PicErrorContext e;

        currentContext->previous->error->appendText(cumulationText);

        if (prm->elogFilter & (PicErrorLog::raised | PicErrorLog::held)
            && !prm->elogSuspend
            )
        {
            BaseString nonConstStr = cumulationText;
            PicErrorLog::write(nonConstStr);
        }

        CHECK_INTERNAL_ERROR();
    }
}

void PicL1ErrorManager::raiseMemoryError (const char* explanation)
{
    if (prm->elogFilter)
        PicErrorLog::writeMemoryError(explanation);

    memoryError = boolTrue;
}

PICBOOL PicL1ErrorManager::checkMemoryError (
    PICBOOL     wantToContinue,
    const char* cumulationText,
    PICUSHORT   lineNumber,
    const char* fileName)
{
    if (outOfMemory())
    {
        if (prm->elogFilter)
            PicErrorLog::writeMemoryError (
                cumulationText, lineNumber, fileName);

        if (wantToContinue)
            return boolTrue;
        else
            PicError::exitOnMemoryError (
                cumulationText, lineNumber, fileName);
    }
    return boolFalse;
}

// --------------------------------------------------------------------------
//      READING ERROR INFO FROM OBJECTS/RESOURCES

BaseString PicL1ErrorManager::getObjErrorText (
    const PicObject&, const PicClass&, PICSTATUS) const
{
    return BaseString("");
}

// --------------------------------------------------------------------------
//      HANDLING ERROR CONTEXTS

PicError* PicL1ErrorManager::removeError ()
{
    if (currentContext->error)
    {
        PicError* error = currentContext->error;
        currentContext->error = NULLPTR;

        if (prm->elogFilter & PicErrorLog::cleared && !prm->elogSuspend)
        {
            PicErrorContext e;

            PicErrorLog::write(error, PicErrorLog::clearStamp);

            CHECK_INTERNAL_ERROR();
        }
        return error;
    }
    return NULLPTR;
}

// --------------------------------------------------------------------------
//      PRIVATE METHODS

void PicL1ErrorManager::transfer (PicErrorContext* current)
{
    if (current->previous)
    {
        if (current->previous->error)
        {
            if (prm->elogFilter & PicErrorLog::discarded && !prm->elogSuspend)
            {
                PicErrorContext e;

                PicErrorLog::write (
                    current->previous->error,
                    PicErrorLog::discardStamp);

                CHECK_INTERNAL_ERROR();
            }
            delete current->previous->error;
        }
        current->previous->error = current->error;
    }
}

PICBOOL PicL1ErrorManager::checkInternalError (
    PICUSHORT   lineNumber,
    const char* fileName)
{
    if (isDetected())
    {
        if (checkMemoryError(boolTrue, "Allocation error in error manager",
                             lineNumber, fileName))
        {
            return boolTrue;
        }

        PicStringText internalStamp;
        internalStamp(prm->elogStamps[PicErrorLog::internalStamp])
            >>"LINE" << lineNumber
            >>"FILE" << fileName;

        PicErrorLog::write(internalStamp);
        clearCurrent();
    }
    return boolFalse;
}
