// --------------------------------------------------------------------------
//  FILE STREAM implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/file.h>
#include <dirent.h>
#include <stdio.h>

extern "C" {
#include <string.h>
}

#include <pfilestm.hxx>
#include <pfile.hxx>
#include <psemaph.hxx>
#include <picerror.hxx>
#include <picbase.erh>

MAKE_PICSTREAMCLASS(PicFileStream,PicBufferStream)

#define FILE_RAISE(status)  ERAISEFMT_LINE_FILE (PicFile::errors, \
    PicFile::classId, status, __LINE__, __FILE__)

// --------------------------------------------------------------------------
//      SELF is used in some const methods to cast this-pointer from
//      const PicFileStream* to PicFileStream.
//      In those cases natural constant abstraction overrides normal
//      C++ meaning of const method.

#define SELF ((PicFileStream*)this)

// --------------------------------------------------------------------------
//      CONSTANTS AND STATIC VARIABLES
//
//      * standardFlags is always added to file mode. They define file to
//        be sequential.

const PICUINT standardFlags = 0;

const PICHFILE fInput  = 0;
const PICHFILE fOutput = 1;
const PICHFILE fError  = 2;

// --------------------------------------------------------------------------
//      STATIC (LOCAL) METHODS

static PICBOOL isStandard (PICHFILE handle)
{
  return (handle == fInput || handle == fOutput || handle == fError);
}

static void errorInfo(
          PICUINT     err,
    const BaseString& name,
          PICUSHORT&  status,
          BaseString& info)
{
    switch(err)
    {
        case EACCES:
        case EROFS:
        case ETXTBSY:
            status = PicFile::accessError;
            info   = name;
            break;
        case ENOENT:
            status = PicFile::fileNotFound;
            info   = name;
            break;
        case EBADF:
	    status = PicFile::eBadFileNbr; // fall through
        case EISDIR:
	    status = PicFile::eIsDir; // fall through
        case EINVAL:
	    status = PicFile::eInvalidArg; // fall through
        case ENAMETOOLONG:
	    status = PicFile::nameError;
            info   = name;
            break;
        case ENOTDIR:
            status = PicFile::pathError;
            info   = name;
            break;
        default:
            status = PicFile::error;
            info = "Status: ";
            info.append(asString(err));
            break;
    }
}

// --------------------------------------------------------------------------
//      Open flags are given in openFlags-array. If you update
//      PicOpenMode make sure to change this too.
//      Operations are :
//          - forReading   : open, must exist
//          - forWriting   : truncate if exists, create if doesn't
//          - forAppending : open if exist, create if doesn't

static PICUINT openFlags [3] =
{
    O_RDONLY, O_WRONLY | O_CREAT | O_TRUNC, O_WRONLY | O_CREAT
//    O_RDONLY, O_WRONLY | O_CREAT | O_TRUNC, O_WRONLY | O_CREAT | O_APPEND
};

static PICUINT fileModeFlags (PicOpenMode mode)
{
    PICUINT fMode = standardFlags;

    switch (mode)
    {
        case PicFile::forReading:   break;
        case PicFile::forWriting:   fMode |= 0660; break;
        case PicFile::forAppending: fMode |= 0660; break;
    }

    return fMode;
}

void PicFileStream::fOpen (
    PicOpenMode  mode,
    PICUINT      attr)
{
    if (isOpen())
    {
        ERAISEFMT(PicFile, fileAlreadyOpen) << theFile.fullName();
        return;
    }

    PICUINT err = 0, fMode = fileModeFlags(mode);

    if((fHandle = ::open(theFile.fullName().chars(), openFlags[mode], fMode))<0)
    {
        err = errno;
        PICUSHORT  status;
        BaseString info;

        errorInfo(err, theFile.fullName(), status, info);

        FILE_RAISE(status) << info;
        endOfFile = boolTrue;
        fHandle   = picHNullFile;
    }
    else
    {
        endOfFile = boolFalse;
        if (mode == PicFile::forAppending)
            move(0, PicStream::end);
    }
}

// --------------------------------------------------------------------------
//      METHODS

// Second argument is because PicFileStream("file.nam") should
// invoke PicFileStream(PicFile("file.nam")) instead of this ctor.

PicFileStream::PicFileStream (PICHFILE handle, PICHFILE) :
    PicBufferStream(), 
    theFile        (), 
    fHandle        (handle), 
    endOfFile      (0)
{}

PicFileStream::PicFileStream (const PicFile& file) :
    PicBufferStream (),
    theFile         (file),
    fHandle         (picHNullFile),
    endOfFile       (0)
{}

PicFileStream::PicFileStream (
    const PicFile& file,
    PicOpenMode    mode,
    PICUINT        attributes)
    :
    PicBufferStream (),
    theFile         (file),
    fHandle         (picHNullFile),
    endOfFile       (0),
    saveOpenFlags   (mode)
{
    fOpen(mode, attributes);
}

PicFileStream::~PicFileStream ()
{
    if (isOpen() && !isStandard(fHandle))
        close();
}

PicFileStream& PicFileStream::in ()
{
    static PicFileStream picInput(fInput, 0);
    return picInput;
}

PicFileStream& PicFileStream::out ()
{
    static PicFileStream picOutput(fOutput, 0);
    return picOutput;
}

PicFileStream& PicFileStream::err ()
{
    static PicFileStream picError(fError, 0);
    return picError;
}

PicText& picIn()
{
    static PicText input(PicFileStream::in());
    return input;
}

PicFmtText& picOut()
{
    static PicFmtText output(PicFileStream::out());
    return output;
}

PicFmtText& picErr()
{
    static PicFmtText errOut(PicFileStream::err());
    return errOut;
}

// --------------------------------------------------------------------------


PICSIZE PicFileStream::write (const void* from, PICSIZE length)
{
    PICUINT err = 0;
    int written;

    written = ::write(fHandle, (char *)from, length);
    if(written == -1){
        written = 0;
        err = errno;
    }

    if (err)
    {
        PICUSHORT  status;
        BaseString info;

        errorInfo(err, theFile.fullName(), status, info);
        FILE_RAISE(status) << info;
        endOfFile = boolTrue;
        return 0;
    }
    return written;
}

PICSIZE PicFileStream::read (void* to, PICSIZE length)
{
    PICUINT err = 0;
    int bytes;

    bytes = ::read(fHandle, (char *)to, length);
    if(bytes == -1){
        bytes = 0;
        err = errno;
    }

    if (err)
    {
        PICUSHORT  status;
        BaseString info;

        errorInfo (err, theFile.fullName(), status, info);
        FILE_RAISE(status) << info;
        endOfFile = boolTrue;
        return 0;
    }
    if (bytes == 0)
        endOfFile = boolTrue;
    return bytes;
}

PICSTATUS PicFileStream::status () const
{
    PICSTATUS result = isStandard(fHandle) ? 0 : positionable;

    if (endOfFile && bytesInBuffer() == 0)
        result |= ended;
    if (isOpen())
    {
        if (openMode() == PicFile::forWriting ||
            openMode() == PicFile::forAppending)
            result |= writeable;
        else if (openMode() == PicFile::forReading)
            result |= readable;
    }
    return result;
}

BaseString PicFileStream::name () const
{
    return theFile.fullName();
}

void PicFileStream::changeFile (const PicFile& file)
{
    if (isOpen())
        close();
    theFile = file;
}

void PicFileStream::close ()
{
    PICUINT err = 0;

    if (!isOpen())
        ERAISEFMT(PicFile, fileNotOpened) << theFile.fullName();
    else if (err = (::close(fHandle)==0) ? 0 : errno)
    {
        PICUSHORT  status;
        BaseString info;

        errorInfo (err,theFile.fullName(),status,info);
        FILE_RAISE(status) << info;
    }
    flushBuffer();
    endOfFile = boolFalse;
    fHandle   = picHNullFile;
}

void PicFileStream::flush ()
{
    if (isOpen() && !isStandard(fHandle))
    {
        PICUINT ret = 0;
        ret = 0;  // nop

        if (ret)
        {
            PICUSHORT  status;
            BaseString info;

            errorInfo(ret, theFile.fullName(), status, info);
            FILE_RAISE(status) << info;
        }
    }
}

PicOpenMode PicFileStream::openMode () const
{
    if (fHandle == fInput)
        return PicFile::forReading;
    else if (fHandle == fOutput || fHandle == fError)
        return PicFile::forWriting;
    else if (fHandle == picHNullFile)
    {
        ERAISEFMT(PicFile, fileNotOpened) << theFile.fullName();
        return PicFile::forReading;
    }
    else
    {
        return saveOpenFlags;
    }
}

void PicFileStream::move (PICLONG distance, PicStreamPosition origin)
{
    if (isOpen())
    {
        // This is indexed with PicStreamPosition:
        static int fMethod[3] = { L_SET, L_INCR, L_XTND };

        int osRet = 
          (lseek(fHandle, distance, fMethod[origin])==-1 ? errno : 0);
        if (osRet)
        {
            PICUSHORT  status;
            BaseString info;

            errorInfo(osRet,theFile.fullName(),status,info);
            FILE_RAISE(status) << info;
        }
        else
        {
            endOfFile = boolFalse;
            flushBuffer();
        }
    }
    else
        ERAISEFMT(PicFile, fileNotOpened) << theFile.fullName();
}

