// --------------------------------------------------------------------------
//  INTEGER OBJECT implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <limits.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
}

#include <pinteger.hxx>
#include <picerror.hxx>
#include <pstring.hxx>
#include <breal.hxx>
#include <picbase.erh>
#include <piterate.hxx>

const PICSIZE   maxIntegerLength = 34;
const PICLONG   nullLong         = (PICLONG)LONG_MIN;
const PICSHORT  nullInt          = (PICSHORT)SHRT_MIN;

MAKE_PICOBJECTCLASS(PicInteger, PicPrimitive)

ERROR_MESSAGES(PicInteger)
PARAMSTR(nullInstance, "Operation on Null instance")
PARAMSTR(deformatError, "String '@*' is not of the given format")
PARAMSTR(formatError, "String '@*' is not a valid integer")
PARAMSTR(underflow, "Underflow: @*")
PARAMSTR(overflow, "Overflow: @* ")
PARAMSTR(divideByZero, "Divide by zero")
PARAMSTR(uncomparable, "Integer OBJ1@* is uncomparable with OBJ2@*")
END_PARAM_STRINGS

// --------------------------------------------------------------------------
//      SAFE ADD AND MULTIPLY
//
//      Adding to integers can overflow only if both have the same sign and
//      |a| > MAX_LONG - |b|. Implementation expects that unsigned long can
//      hold the result of any two long number.
//      Multiply can overflow if |a| > MAX_LONG / |b|.
//      Note that a-b == a+(-b) and divide can't overflow.
//
//      These return boolean value indicating if the value is non-null (
//      boolTrue -> non-null, boolFalse -> null). Also result is changed
//      only for non-null values.

static PICBOOL safeAdd (PICLONG right, PICLONG left, PICLONG& result)
{
    PICULONG value;
    PICBOOL  hasSign = boolFalse;

    if (left > 0 && right > 0)
        value = PICULONG(left) + PICULONG(right);
    else if (left < 0 && right < 0)
    {
        value   = PICULONG(-left) + PICULONG(-right);
        hasSign = boolTrue;
    }
    else
    {
        result = left + right;
        return boolTrue;
    }

    if (value > PICULONG(LONG_MAX))
    {
        ERAISEFMT(PicInteger, overflow)
            << (asString(left) & " + " & asString(right));
        return boolFalse;
    }

    if (hasSign)
        result = -PICLONG(value);
    else
        result = PICLONG(value);
    return boolTrue;
}

static PICBOOL safeMultiply (PICLONG left, PICLONG right, PICLONG& result)
{
    PICLONG value;
    PICLONG comp = left >= 0 ? left : -left;

    if (right == 0)
    {
        result = 0;
        return boolTrue;
    }
    else if (right > 0)
        value = LONG_MAX / right;
    else
        value = LONG_MIN / right;

    if (comp > value)
    {
        ERAISEFMT(PicInteger, overflow)
            << (asString(left) & " * " & asString(right));
        return boolFalse;
    }
    else
    {
        result = left * right;
        return boolTrue;
    }
}

// --------------------------------------------------------------------------
//
//      unformattedValue doesn't accept any grouping, sign position...
//      Image must not contain any blank characters at the beginning
//      or at the end (setImage takes care of them).
//
//      Returns : boolTrue if given image is valid integer, boolFalse
//                otherwise. If the image is ok then the integer value
//                is returned in val.
//      Errors  : deformatError

static PICBOOL unformattedValue (const BaseString& image, PICLONG& val)
{
    PICSIZE     len   = image.size();
    PICPOSITION pos   = firstPos;
    char        aChar = image.first();

    //  If first character is SIGN (-) or PLUS (+) it's
    //  accepted but not if it's also the only character.

    if (((aChar == '-') || (aChar == '+')) && (pos != len))
        pos++;

    // In this loop only digits are allowed. If string 
    // passes this loop it's valid Integer.

    for (; pos <= len; pos++)
        if (!PicChar::isDigit(image.at(pos)))
        {
            ERAISEFMT(PicInteger, deformatError) << image;
            return boolFalse;
        }

    // Conversion to PICFLOAT makes it possible to check overflow
    // conditions.

    PICFLOAT value = atof(image.chars());

    if (value <= minLongInt || value > maxLongInt)
    {
        ERAISEFMT(PicInteger, overflow) << image;
        return boolFalse;
    }

    val = (PICLONG)value;
    return boolTrue;
}

// --------------------------------------------------------------------------
//      CONSTRUCTORS, COPY & ASSIGNMENT

PicInteger::PicInteger () :
    PicPrimitive(), imp(nullLong), isNullInteger(boolTrue) {}

PicInteger::PicInteger (PICLONG val) :
    PicPrimitive(), imp(val), isNullInteger(boolFalse) {}

PicInteger::PicInteger (const PicInteger& val) :
    PicPrimitive(), imp(val.imp), isNullInteger(val.isNullInteger) {}

PicInteger::PicInteger (const BaseString& image) :
    PicPrimitive(), imp(nullLong), isNullInteger(boolTrue)
{
    setImage(image);
}

PicObject& PicInteger::operator = (const PicObject& obj)
{
    if (TYPE_GUARD(classId(), obj))
    {
        imp = ((const PicInteger&)obj).imp;
        isNullInteger = ((const PicInteger&)obj).isNullInteger;
    }
    return *this;
}

void PicInteger::setNull ()
{
    imp = nullLong;
    isNullInteger = boolTrue;
}

// --------------------------------------------------------------------------
//      CONVERSIONS

PICLONG PicInteger::asLong () const
{
    if (isNullInteger)
    {
        ERAISE(PicInteger, nullInstance);
        return nullLong;
    }
    return imp;
}

PICSHORT PicInteger::asInt () const
{
    if (isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else if (imp >= (PICLONG)SHRT_MIN && (imp <= (PICLONG)SHRT_MAX))
        return (PICSHORT)imp;
    else
    {
        ERAISEFMT(PicInteger, overflow) << imp;
    }
    return nullInt;
}

char PicInteger::asChar () const
{
    if (isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else if (imp >= 0 && imp <= 255)
        return (char)imp;
    else
    {
        ERAISEFMT(PicInteger, overflow) << imp;
    }
    return (char)UCHAR_MAX;
}

// --------------------------------------------------------------------------
//      COMPARISONS

PICBOOL PicInteger::operator == (const PicObject& obj) const
{
    if (id() == obj.id())
        return *this == (const PicInteger&)obj;
    return boolFalse;
}

PICBOOL PicInteger::operator < (const PicObject& obj) const
{
    if (classId() == obj.id())
        return *this < (const PicInteger&)obj;

    ERAISEFMT(PicInteger, uncomparable)
        >>"OBJ1" << *this
        >>"OBJ2" << obj;
    return boolFalse;
}

PICBOOL operator == (const PicInteger& value1, const PicInteger& value2)
{
    if (value1.isNullInteger || value2.isNullInteger)
        return value1.isNullInteger && value2.isNullInteger;
    return value1.imp == value2.imp;
}

PICBOOL operator < (const PicInteger& value1, const PicInteger& value2)
{
    if (value1.isNullInteger)
        return !value2.isNullInteger;
    else if (value2.isNullInteger)
        return boolFalse;
    else
        return value1.imp < value2.imp;
}

PICBOOL operator <= (const PicInteger& value1, const PicInteger& value2)
{
    if (value1.isNullInteger)
        return boolTrue;
    else if (value2.isNullInteger)
        return boolFalse;
    else
        return value1.imp <= value2.imp;
}

// --------------------------------------------------------------------------
//      ASSIGNMENT + ARITHMETIC OPERATIONS

PicInteger& PicInteger::operator += (const PicInteger& value)
{
    if (isNullInteger || value.isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else
        safeAdd(imp, value.imp, imp);
    return *this;
}

PicInteger& PicInteger::operator -= (const PicInteger& value)
{
    if (isNullInteger || value.isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else
        safeAdd(imp, -value.imp, imp);
    return *this;
}

PicInteger& PicInteger::operator *= (const PicInteger& value)
{
    if (isNullInteger || value.isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else
        safeMultiply(imp, value.imp, imp);
    return *this;
}

PicInteger& PicInteger::operator /= (const PicInteger& value)
{
    if (isNullInteger || value.isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else if (value.imp == 0L)
        ERAISE(PicInteger, divideByZero);
    else
        imp /= value.imp;
    return *this;
}

PicInteger& PicInteger::operator %= (const PicInteger& value)
{
    if (isNullInteger || value.isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else if (value.imp == 0L)
        ERAISE(PicInteger, divideByZero);
    else
        imp %= value.imp;
    return *this;
}

PicInteger& PicInteger::operator &= (const PicInteger& value)
{
    if (isNullInteger || value.isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else
        imp &= value.imp;
    return *this;
}

PicInteger& PicInteger::operator |= (const PicInteger& value)
{
    if (isNullInteger || value.isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else
        imp |= value.imp;
    return *this;
}

PicInteger& PicInteger::operator ^= (const PicInteger& value)
{
    if (isNullInteger || value.isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else
        imp ^= value.imp;
    return *this;
}

// --------------------------------------------------------------------------
//      PROTECTED OPERATIONS

void PicInteger::multiplyTo (BaseReal& real) const
{
    if (isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else
        real *= imp;
}

void PicInteger::divideFrom (BaseReal& real) const
{
    if (isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else
        real /= imp;
}

// --------------------------------------------------------------------------
//      INCREMENT & DECREMENT

PicInteger& PicInteger::operator ++ ()
{
    if (isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else if (imp < LONG_MAX)
        ++imp;
    else
    {
        ERAISEFMT(PicInteger, overflow) << (asString(imp) & " + 1");
    }
    return *this;
}

PicInteger& PicInteger::operator -- ()
{
    if (isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else if (imp > LONG_MIN)
        --imp;
    else
    {
        ERAISEFMT(PicInteger, underflow) << (asString(imp) & " - 1");
    }
    return *this;
}

// --------------------------------------------------------------------------
//      ROUND & TRUNCATE

PicInteger& PicInteger::round (PICDIGIT digit)
{
    if (isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else if (digit != 0)
    {
        PICFLOAT precision = pow(10, digit);

        if (imp > 0)
            safeAdd(imp, 5 * (PICLONG(precision) / 10), imp);
        else
            safeAdd(imp, -(5 * (PICLONG(precision) / 10)), imp);

        imp = PICLONG(imp / precision) * precision;
    }
    return *this;
}

PicInteger& PicInteger::truncate (PICDIGIT digit)
{
    if (isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else if (digit != 0)
    {
        PICFLOAT precision = pow(10, digit);
        imp = PICLONG(imp / precision) * precision;
    }
    return *this;
}

// --------------------------------------------------------------------------
//      ABSOLUTE VALUE

PicInteger& PicInteger::makeAbs ()
{
    if (isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else
        imp = labs(imp);
    return *this;
}

// --------------------------------------------------------------------------
//      NEGATE

PicInteger& PicInteger::negate ()
{
    if (isNullInteger)
        ERAISE(PicInteger, nullInstance);
    else
        imp = -imp;
    return *this;
}

// --------------------------------------------------------------------------
//      ODD and PICN

PICBOOL PicInteger::odd () const
{
    return isNullInteger ? boolFalse : (imp % 2) != 0;
}

PICBOOL PicInteger::even () const
{
    return isNullInteger ? boolFalse : (imp % 2) == 0;
}

// --------------------------------------------------------------------------
//      TEXTUAL IMAGE
//
//      Formatting is done is the following simple way.
//      1. Find image of absolute value
//      2. Prepend the sign if needed.

BaseString PicInteger::getImage (PICUSHORT radix) const
{
    char    buffer [maxIntegerLength];
    BaseString result;

    if (isNullInteger)
        return result;

    if(sprintf(buffer, "%ld", labs(imp))){
        result = buffer;
    }
    //    result = ltoa(labs(imp), buffer, radix);

    if (imp < 0 && !result.isEmpty())
        result.prepend('-', 1);

    return result;
}

BaseString PicInteger::getImage (const PicObject& format) const
{
    BaseString image;

    if (!pictureFmt(format, image))
    {
        PicIterator here;
        if (format.find(RADIX, here))
        {
            PicString::format (
                image = getImage(format.current(here).asInt()), format);
        }
        else
            PicString::format(image = getImage(), format);
    }
    return image;
}

void PicInteger::setImage (const BaseString& imageStr)
{
    BaseString image(imageStr);

    image.trim();

    if (image.isEmpty())
        setNull();
    else
    {
        PICLONG newValue;

        if (unformattedValue(image,newValue))
            *this = newValue;
    }
}

// --------------------------------------------------------------------------
//      FORMATTED OUTPUT

PICOFFSET PicInteger::writeFmt (PicTextFmt& fmt, PICOFFSET i) const
{
    static const char* typeName = id().className();

    if (isNull() && fmt.format[i] != '?')
    {
        PicTextFmt nullFmt (PicStream::null(), fmt.format, 0);
        return nullFmt.writeInteger(i, labs(imp), typeName, imp < 0);
    }

    return fmt.writeInteger(i, labs(imp), typeName, imp < 0);
}

