// --------------------------------------------------------------------------
//  OBJECT PTR ARRAY implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

// Edit this include:
// ------------------
#include <pobjarr.hxx>
#include <bstring.hxx>

// --------------------------------------------------------------------------
//      STATIC PARAMETER FUNCTIONS
//
//  Modify these functions to suit the component class. Default
//  implementations are suitable for most Pic classes. You may replace
//  ITEM_CLASS with the real component class name for the sake of
//  readability. For an example, see BSTRARR.CXX.
//
// --------------------------------------------------------------------------
//  Create a new component at the address pointed to by 'to'.
//  The component should be created as null-valued.

static void createItem (PPicObject* to)
{
    *to = NULLPTR;
}

// --------------------------------------------------------------------------
//  Create a new component at the address pointed to by 'to'.
//  Copy the value of the argument 'from' (may be NULL).

static void createItemCopy (PPicObject* to, const PPicObject* from)
{
    PicObject* pobj = *from;
    if (pobj)
        *to = pobj->copy();
    else
        *to = NULLPTR;
}

// --------------------------------------------------------------------------
//  Return boolTrue if the given component is in null state.
//  Such components are regarded as non-existing, so locations
//  occupied by them are empty.

static PICBOOL isItemNull (const PPicObject* item)
{
    return *item == NULLPTR;
}

// --------------------------------------------------------------------------
//  Set the given component in null state. After that it must not
//  allocate dynamic memory.

static void setItemNull (PPicObject* item)
{
    delete *item;
    *item = NULLPTR;
}

// --------------------------------------------------------------------------
//  Compare two components for ascending order. Return
//   0  if they are equal
//   1  if 'i1' sorts after 'i2'
//  -1  if 'i1' sorts before 'i2'
//  Make sure that null components are sorted after all others.

static int compareAsc (PPicObject* i1, PPicObject* i2)
{
    if (!*i1 || !*i2)
        return (!*i1 - !*i2);
    if (**i1 == **i2)
        return 0;
    return **i1 < **i2 ? -1 : 1;
}

// --------------------------------------------------------------------------
//  Compare two components for descending order. Return
//   0  if they are equal
//  -1  if 'i1' sorts after 'i2'
//   1  if 'i1' sorts before 'i2'
//  Make sure that null components are sorted after all others.

static int compareDesc (PPicObject* i1, PPicObject* i2)
{
    if (!*i1 || !*i2)
        return (!*i1 - !*i2);
    if (**i1 == **i2)
        return 0;
    return **i1 < **i2 ? 1 : -1;
}

// --------------------------------------------------------------------------
//      TYPE INFORMATION
// --------------------------------------------------------------------------

static PicGenArrayInfo typeInfo =
{
    sizeof(PPicObject),

    (PROC_SETITEM*)      createItem,
    (PROC_SETITEM*)      NULLPTR, // or NULLPTR to init as zeros
    (PROC_COPYITEM*)     createItemCopy,
    (PROC_SETITEM*)      setItemNull,
    (PROC_TESTITEM*)     isItemNull,
    (PROC_COMPAREITEMS*) compareAsc,
    (PROC_COMPAREITEMS*) compareDesc
};

#include <genarray.cgn>

// --------------------------------------------------------------------------
//      COMPARISON
// --------------------------------------------------------------------------

PICBOOL PicObjectPtrArray::operator == (const PicObjectPtrArray& array) const
{
    if (lastPos() != array.lastPos())
        return boolFalse;

    for (PICPOSITION ix = 1; ix <= ub; ix++)
    {
        const PicObject *me = at(ix), *he = array.at(ix);
        if ((!me && he) || (me && !he) || (me && *me != *he))
            return boolFalse;
    }
    return boolTrue;
}
