// --------------------------------------------------------------------------
//  PIC OBJECT implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <string.h>
}
#include <picbase.erh>
#include <picerror.hxx>
#include <pobject.hxx>

MAKE_PICCLASS(PicObjectClass)
MAKE_PICABSOBJECTCLASS(PicObject, PicVirtObject)
MAKE_PICABSOBJECTCLASS(PicPrimitive, PicObject)

ERROR_MESSAGES(PicObject)
PARAMSTR(staticDeleted, "Static PicObject deleted")
PARAMSTR(uncomparable, "Objects OBJ1@* and OBJ2@* are uncomparable")
END_PARAM_STRINGS

// --------------------------------------------------------------------------
//      PicObject
// --------------------------------------------------------------------------

PicObject::~PicObject() {}

// --------------------------------------------------------------------------

#define CALL_GUARD(className, func) \
    (TYPE_GUARD(className::classId(), *this)) ? ((className*)this)->func

PICSHORT PicObject::asInt () const
{
    return CALL_GUARD(PicPrimitive, asInt()) : 0;
}

PICBOOL PicObject::asBoolean () const
{
    return CALL_GUARD(PicPrimitive, asBoolean()) : boolFalse;
}

char PicObject::asChar () const
{
    return CALL_GUARD(PicPrimitive, asChar()) : '\0';
}

const char* PicObject::asChars () const
{
    return CALL_GUARD(PicPrimitive, asChars()) : NULLPTR;
}

// --------------------------------------------------------------------------
//      PicPrimitive
// --------------------------------------------------------------------------

PICBOOL PicPrimitive::setTypedValue (const PicObject&) { return boolTrue; }

// --------------------------------------------------------------------------
//      FORMATTED OUTPUT
// --------------------------------------------------------------------------

PICOFFSET PicObject::writeFmt (PicTextFmt& fmt, PICOFFSET i) const
{
    switch (fmt.format[i])
    {
        case '*':
        {
            BaseString image = getImage();
            fmt.write(image.chars(), image.size());
            break;
        }

        case '?':
        {
            const char* name = id().className();
            fmt.write(name, strlen(name));
            break;
        }
    }

    return i + 1;
}

static PICOFFSET wrObject (const PicObject* obj, PicTextFmt& fmt, PICOFFSET i)
{
    return obj->writeFmt(fmt, i);
}

static PicObject* wrobjPicObject (PicObject* obj, PICBOOL cleanUp)
{
    if (cleanUp)
    {
        delete obj;
        return NULLPTR;
    }
    return obj->copy();
}

PicFmtText& operator << (PicFmtText& txt, const PicObject& obj)
{
    return txt.wrFormat (
        &obj, (PICWRFUNC*)wrObject, (PICWROBJFUNC*)wrobjPicObject);
}

PICBOOL PicObject::pictureFmt (
    const PicObject& format,
    BaseString&      image) const
{
//      if (format.IS_A(PicString::classId()))
//      {
//          PicStringStream stm  (image);
//          PicFmtText      text (stm);
//  
//          text(AS(PicString,format).asChars()) << *this;
//          return boolTrue;
//      }
    return boolFalse;
}

// --------------------------------------------------------------------------
//      GLOBAL NULL OBJECTS
// --------------------------------------------------------------------------

class PicNullObject : public PicObject
{
    public:

    PICANYOBJECTCLASS(PicNullObject)

    virtual PicObject* copy () const { return NULLPTR; }
    virtual PicObject& operator = (const PicObject&) 
       { undefined ("assignment = "); return *this; }
    virtual PICBOOL operator == (const PicObject&) const { return boolFalse; }
    virtual PICBOOL operator < (const PicObject& obj) const 
        { 
            ERAISEFMT(PicObject, uncomparable)
                >>"OBJ1" << *this
                >>"OBJ2" << obj;
            return boolTrue; 
        }
    virtual PICBOOL isNull  () const { return boolTrue; }
    virtual void    setNull () {}
    virtual PICBOOL isDefault  () const { return boolTrue; }
    virtual void    setDefault () {}
    virtual BaseString getImage () const { return ""; }
    virtual BaseString getImage (const PicObject& format) const { return ""; }
    virtual void setImage (const BaseString& image) {}
    virtual void setImage (
        const PicObject&  format,
        const BaseString& image) {}
};

const PicObjectClass& PicNullObject::classId() 
{
  static PicVirtClassInfo info = { 
        "PicNullObject",
        sizeof(PicNullObject), 
        (PICFVIRTCLASSID*)PicObject::classId 
  };
  static PicObjectClass id (&info, NULLPTR);
  return id; 
}
MAKE_PICCLASSID(PicNullObject)

PicObject& PicObject::null()
{
    static PicNullObject  nullObject;
    return nullObject;
}

class PicNullPrimitive : public PicPrimitive
{
    public:

    PICANYOBJECTCLASS(PicNullPrimitive)

    virtual PicObject* copy () const { return NULLPTR; }
    virtual PicObject& operator = (const PicObject&)  
       { undefined ("assignment = "); return *this; }
    virtual PICBOOL operator == (const PicObject&) const { return boolFalse; }
    virtual PICBOOL operator < (const PicObject& obj) const 
        { 
            ERAISEFMT(PicObject, uncomparable)
                >>"OBJ1" << *this
                >>"OBJ2" << obj;
            return boolTrue; 
        }
    virtual PICBOOL isNull  () const { return boolTrue; }
    virtual void    setNull () {}
    virtual PICBOOL isDefault  () const { return boolTrue; }
    virtual void    setDefault () {}
    virtual BaseString getImage () const { return ""; }
    virtual BaseString getImage (const PicObject& format) const { return ""; }
    virtual void setImage (const BaseString& image) {}
    virtual void setImage (
        const PicObject&  format,
        const BaseString& image) {}
    virtual PICSHORT       asInt     () const 
        { undefined ("asInt"); return 0; }
    virtual PICBOOL        asBoolean () const
        { undefined ("asBoolean"); return boolFalse; }
    virtual char        asChar    () const 
        { undefined ("asChar"); return '\0'; }
    virtual const char* asChars   () const 
        { undefined ("asChars"); return NULLPTR; }
};

const PicObjectClass& PicNullPrimitive::classId() 
{
  static PicVirtClassInfo info = { 
        "PicNullPrimitive",
        sizeof(PicNullPrimitive), 
        (PICFVIRTCLASSID*)PicPrimitive::classId 
  };
  static PicObjectClass id (&info, NULLPTR);
  return id; 
}
MAKE_PICCLASSID(PicNullPrimitive)

PicPrimitive& PicPrimitive::null()
{
    static PicNullPrimitive nullPrimit;
    return nullPrimit;
}
