// --------------------------------------------------------------------------
//  STREAM BASE CLASS implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#include <palloc.hxx>
#include <pstream.hxx>
#include <picerror.hxx>

// --------------------------------------------------------------------------
//      PicNullStream
// --------------------------------------------------------------------------

class PicNullStream : public PicStream
{
    public:

    PICSTREAMCLASS(PicNullStream)

             PicNullStream () {}
    virtual ~PicNullStream () {}

    virtual PICSIZE write (const void*, PICSIZE) { return 0; }
    virtual PICSIZE read  (      void*, PICSIZE) { return 0; }
    virtual PICBOOL skip  (PICLONG)              { return boolTrue; }

    virtual PICSTATUS status () const
    {
        return writeable | readable | positionable | ended;
    }

    virtual BaseString name () const { return "NullStream"; }

};

PicStream& PicStream::null ()
{
    static PicNullStream null;
    return null;
}

// --------------------------------------------------------------------------

MAKE_PICSTREAMCLASS(PicStream,PicVirtObject)
MAKE_PICSTREAMCLASS(PicNullStream,PicStream)
MAKE_PICSTREAMCLASS(PicChainStream,PicStream)
MAKE_PICSTREAMCLASS(PicBufferStream,PicStream)

ERROR_MESSAGES(PicStream)
PARAMSTR(readError, "Can't read with class @*")
PARAMSTR(writeError, "Can't write with class @*")
PARAMSTR(positionError, "Only forward movement possible with class @*")
END_PARAM_STRINGS

// --------------------------------------------------------------------------
//      PicStream
// --------------------------------------------------------------------------

PicStream::~PicStream () {}

PICSIZE PicStream::writeFrom (PicStream& from, PICSIZE bufSize)
{
    if (!(from.status() & readable && this->status() & writeable))
        return 0;

    void    *buffer = picAllocate(bufSize, boolFalse);
    PICSIZE  sz, total = 0;

    if (!MEMORY_ERROR("PicStream::writeFrom"))
    {
        while (sz = from.read(buffer, bufSize))
            total += write(buffer, sz);
        picFree(buffer);
    }
    return total;
}

PICSIZE PicStream::write (const void*, PICSIZE)
{
    ERAISEFMT(PicStream, writeError) << id().className();
    return 0;
}

PICSIZE PicStream::read (void*, PICSIZE)
{
    ERAISEFMT(PicStream, readError) << id().className();
    return 0;
}

char* PicStream::readAhead (PICSIZE&)
{
    ERAISEFMT(PicStream, readError) << id().className();
    return NULLPTR;
}

void PicStream::moveAhead (PICSIZE)
{
    ERAISEFMT(PicStream, readError) << id().className();
}

void PicStream::move (PICLONG amount, PicStreamPosition startAt)
{
    if (amount > 0 && startAt == current)
    {
        char* buffer = (char*)picAllocate(512);
        if (!MEMORY_ERROR("PicStream::move"))
        {
            PICLONG nbuf = amount / 512;
            PICSIZE rest = (PICSIZE)(amount % 512);

            while (nbuf-- > 0)
                if (read(buffer, 512) == 0)
                {
                    rest = 0;
                    break;
                }
            if (rest > 0)
                read(buffer, rest);

            picFree(buffer);
        }
    }
    else if (amount != 0)
        ERAISEFMT(PicStream, positionError) << id().className();
}

PICULONG PicStream::movePointer (
    PICULONG          pointer,
    PICULONG          size,
    PICLONG           move,
    PicStreamPosition startAt)
{
    if (move < 0)
    {
        switch (startAt)
        {
            case start:
                pointer = 0;
                break;
            case end:
                pointer = size;
                // Fall through
            case current:
                if (pointer < -move)
                    pointer = 0;
                else
                    pointer += move;
                break;
        }
    }
    else
    {
        switch (startAt)
        {
            case end:
                pointer = size;
                break;
            case start:
                pointer = 0;
                // Fall through
            case current:
                if (size - pointer < move)
                    pointer = size;
                else
                    pointer += move;
                break;
        }
    }

    return pointer;
}

// --------------------------------------------------------------------------
//      PicChainStream
// --------------------------------------------------------------------------

PICSIZE PicChainStream::write (const void* from, PICSIZE length)
{
    PICSIZE sz = stream->write(from, length);

    if (chain)
        chain->write(from, length);

    return sz;
}

PICSTATUS PicChainStream::status () const
{
    PICSTATUS result = stream->status();
    if (chain)
        result |= chain->status();
    return result & ~readable | ended;
}

BaseString PicChainStream::name () const
{
    if (chain)
        return stream->name() & "->" & chain->name();
    else
        return stream->name();
}

// --------------------------------------------------------------------------
//      PicBufferStream
// --------------------------------------------------------------------------

PicBufferStream::~PicBufferStream ()
{
    picFree(buffer);
}

char* PicBufferStream::readAhead (PICSIZE& amount)
{
    const PICSIZE maxBufSize = 512;

    if (bufPtr == bufSize)
    {
        if (!buffer)
        {
            buffer = (char*)picAllocate(maxBufSize);
            if (MEMORY_ERROR("PicBufferStream::readAhead"))
                return NULLPTR;
        }
        bufSize = read(buffer, maxBufSize);
        if ((bufPtr = 0) == bufSize)
            return NULLPTR;
    }

    amount = bytesInBuffer();
    return buffer + bufPtr;
}

void PicBufferStream::moveAhead (PICSIZE amount)
{
    PICSIZE left = bytesInBuffer();

    if (amount > left)
    {
        flushBuffer();
        move(amount - left);
    }
    else
        bufPtr += amount;
}
