// --------------------------------------------------------------------------
//  TEXTUAL I/O implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

extern "C" {
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
}

#include <ptext.hxx>
#include <psemaph.hxx>

MAKE_PICCLASS(PicText)
MAKE_PICCLASS(PicFmtText)
MAKE_PICCLASS(PicStringText)

// --------------------------------------------------------------------------
//      PicText
// --------------------------------------------------------------------------

#include <pfilestm.hxx>

PicText::PicText (PicStream& stream) : theStream(&stream)
{
//      if (stream.id() == PicFileStream::classId())
//          ((PicFileStream&)stream).textMode();
}

PicText& PicText::rd (char& character)
{
    PICSIZE  size;
    char* current = theStream->readAhead(size);

    if (current)
    {
        character = *current;
        theStream->moveAhead(1);
    }
    else
        character = '\0';

    return *this;
}

PicText& PicText::rdln (BaseString& string)
{
    PICSIZE  size;
    char* current;
    char* pnl;

    string.setEmpty();

    while (current = theStream->readAhead(size))
    {
        if (pnl = (char*)memchr(current, '\n', size))
        {
            if (pnl > current)
                string.append(current, pnl - current);
            theStream->moveAhead(pnl - current + 1);
            return *this;
        }
        else
            string.append(current, size);

        theStream->moveAhead(size);
    }
    return *this;
}

PicText& PicText::rd (BaseString& string, char separator, PICBOOL spc)
{
    PICSIZE  size;
    char* current;
    char* start;
    char* end;

    string.setEmpty();

    while (start = current = theStream->readAhead(size))
    {
        for (end = current + size; current < end; current++)
            if (!(spc && PicChar::isSpace(*current)) &&
                 *current != separator)
                goto FoundNonSeparator;
        theStream->moveAhead(size);
    }
    return *this;

  FoundNonSeparator:
    for (char* begin = current; ; current++)
    {
        if (current == end)
        {
            if (current > begin)
                string.append(begin, current - begin);
            theStream->moveAhead(size);
            if (start = begin = current = theStream->readAhead(size))
                end = start + size;
            else
                break;
        }
        if ((spc && PicChar::isSpace(*current)) || *current == separator)
        {
            if (current > begin)
                string.append(begin, current - begin);
            theStream->moveAhead(current - start + 1);
            break;
        }
    }
    return *this;
}

PicText& PicText::rdc (BaseString& string, PICSIZE charCount)
{
    PICSIZE  size;
    char* current = NULLPTR; // To avoid compiler warning
    char* end;

    string.setEmpty();

    while (charCount && (current = theStream->readAhead(size)))
    {
        if ((end = (char*)memchr(current, '\n', size)) &&
            charCount > end - current)
        {
            if (end > current)
                string.append(current, end - current);
            theStream->moveAhead(end - current + 1);
            break;
        }
        else if (charCount <= size)
        {
            string.append(current, charCount);
            theStream->moveAhead(charCount);
            break;
        }
        else
        {
            string.append(current, size);
            charCount -= size;
            theStream->moveAhead(size);
        }
    }
    return *this;
}

// --------------------------------------------------------------------------

PicText& PicText::wr (char character)
{
    theStream->write(&character, sizeof(char));
    return *this;
}

PicText& PicText::wr (const char* string)
{
    if (string)
        theStream->write(string, strlen(string));
    return *this;
}

// --------------------------------------------------------------------------
//      FORMATTED OUTPUT
// --------------------------------------------------------------------------

PicFmtText::PicFmtText (PicStream& stream) :
    PicText(stream), fmt(NULLPTR), pending(NULLPTR), indentSize(0) {}

PicFmtText::~PicFmtText ()
{
    if (fmt)
        *this ENDFMT;
}

// --------------------------------------------------------------------------

PicStringText::PicStringText () :
    BaseString(), strStream(*this), fmtText(strStream) {}

PicStringText::~PicStringText () {}

PicFmtText& PicStringText::reset (const char* format)
{
    setEmpty();
    strStream.rewind();
    return fmtText(format);
}


/* --------------------------------------------------------------------------
    writeObject
   --------------------------------------------------------------------------

   This function is called recursively to implement string formatting
   codes. The output of each recursive call goes to a temporary string
   stream to which contents the formatting is applied.

   Also this function implements the usual abbreviations * and ? 
   and literal strings within ''.

   ----------------------------------------------------------------------- */

static PICSIZE writeObject (
    PicStream&     stream,
    PICWRFUNC*     pfwrite,
    const void*    pobject,
    const char* format,
    PICSIZE        indent)
{
    if (!format)
    {
        PicTextFmt txtfmt (stream, "*]", indent);
        (*pfwrite)(pobject, txtfmt, 0);
        return 0;
    }
    else switch (*format)
    {
        case '*':  // Default image
        {
            PicTextFmt txtfmt (stream, "*]", indent);
            (*pfwrite)(pobject, txtfmt, 0);
            return 1;
        }

        case '?':  // Class/type name
        {
            PicTextFmt txtfmt (stream, "?]", indent);
            (*pfwrite)(pobject, txtfmt, 0);
            return 1;
        }

        case '[':  // Class-specific format
        {
            PicTextFmt fmt (stream, format, indent);
            PICOFFSET i = 1;

            do switch (fmt.format[i])
            {
                case '\'':
                {
                    PICOFFSET start = ++i;

                  FindApostrophe:
                    while (fmt.format[i] != '\'' && fmt.format[i] != '\0')
                        i++;
                    if (i > start)
                        fmt.write(fmt.format + start, i - start);
                    if (fmt.format[++i] == '\'')
                    {
                        start = i++;
                        goto FindApostrophe;
                    }
                    break;
                }

                case ']':
                    return i + 1;

                case '\0':
                    return i;

                default:
                    i = (*pfwrite)(pobject, fmt, i);
                    break;

            } while (i);

            return 0;
        }
    }

    // String formatting is implemented here.

    BaseString      tempStr;
    PicStringStream tempStm (tempStr);

    switch (*format)
    {
        case '<':  // Justify left
        case '>':  // Justify right
        case '=':  // Center
        {
            PICUSHORT fldWidth = 0;
            char   padChar  = ' ';

            PICOFFSET i = PicTextFmt::getInt(format, 1, fldWidth);

            if (format[i] == '|')
            {
                padChar = format[++i];
                i++;
            }

            i += writeObject(tempStm, pfwrite, pobject, format + i, indent);

            if (fldWidth > tempStr.size())
            {
                PICSIZE    padSize = fldWidth - tempStr.size();
                BaseString padStr (padChar, padSize);

                if (*format == '>')      // Right-justify
                {
                    stream.write(padStr.chars(), padSize);
                    stream.write(tempStr.chars(), tempStr.size());
                }
                else if (*format == '<') // Left-justify
                {
                    stream.write(tempStr.chars(), tempStr.size());
                    stream.write(padStr.chars(), padSize);
                }
                else                     // Center
                {
                    stream.write(padStr.chars(), padSize / 2);
                    stream.write(tempStr.chars(), tempStr.size());
                    stream.write(padStr.chars(), padSize / 2 + padSize % 2);
                }
            }
            else
                stream.write(tempStr.chars(), tempStr.size());

            return i;
        }

        case '#':  // Repeat
        {
            PICUSHORT count = 1;
            PICOFFSET i     = PicTextFmt::getInt(format, 1, count);

            i += writeObject(tempStm, pfwrite, pobject, format + i, indent);

            for (PICUSHORT j = 1; j <= count; j++)
                stream.write(tempStr.chars(), tempStr.size());

            return i;
        }

        case 's':  // Substring
        {
            char   direct = '+';
            PICUSHORT start  = 0;
            PICUSHORT size   = 0xffff;
            PICOFFSET i      = 1;

            if (isdigit(format[1]))
            {
                i = PicTextFmt::getInt(format, i, start);
                start--;
            }

            if (format[i] == '+' || format[i] == '-')
            {
                direct = format[i++];
                if (isdigit(format[i]))
                    i = PicTextFmt::getInt(format, i, size);
            }

            i += writeObject(tempStm, pfwrite, pobject, format + i, indent);

            if (start > tempStr.size())
                start = tempStr.size();
            if (size > tempStr.size() - start)
                size = tempStr.size() - start;

            if (direct == '-')
                start = tempStr.size() - start - size;

            stream.write(tempStr.chars() + start, size);
            return i;
        }

        case 't':  // Trim
        case 'u':  // Uppercase
        case 'l':  // Lowercase
        case 'p':  // Proper
        case 'r':  // Reverse
        {
            PICOFFSET i = 1;

            i += writeObject(tempStm, pfwrite, pobject, format + 1, indent);

            switch (*format)
            {
                case 't': tempStr.trim();          break;
                case 'u': tempStr.makeUppercase(); break;
                case 'l': tempStr.makeLowercase(); break;
                case 'p': tempStr.makeProper();    break;
                case 'r': tempStr.reverse();       break;
            }

            stream.write(tempStr.chars(), tempStr.size());
            return i;
        }

        default:   // Unknown formatting code
        {
            static char errstr[] = "[@ERROR:'x']";
            errstr[9] = *format;
            stream.write(errstr, 12);
            return 1;
        }
    }
}

/* --------------------------------------------------------------------------
        List of pending write objects
   --------------------------------------------------------------------------
        A fixed maximum amount of write objects can be stored in a list.
        The list is read when the format is written to the stream.
        Write objects are of the type PicWrObj and the list is of the
        type PicWrObjArray. The list structure is as follows:

            current   - the current write object when building the list

            pfwrite   - pointer to a function to write the object
            pfwrobj   - pointer to a function to copy/delete the object
            pobject   - pointer to the object
            keyword   - keyword of the object (NULLPTR for anonymous)

            pfwrite
            pfwrobj
            pobject
            keyword

            ...

   ----------------------------------------------------------------------- */

const PICSIZE maxWrObj = 20;

struct PicWrObj
{
    PICWRFUNC*     pfwrite;
    PICWROBJFUNC*  pfwrobj;
    const void*    pobject;
    const char* keyword;

    void set (const void* pobj, PICWRFUNC* pfwr, PICWROBJFUNC* pfobj)
    {
        pfwrite = pfwr;
        pfwrobj = pfobj;
        pobject = pfobj ? (*pfobj)(pobj, boolFalse) : pobj;
    }

    PICSIZE write (PicStream& stream, const char* format, PICSIZE indent)
    {
        return writeObject(stream, pfwrite, pobject, format, indent);
    }

    void cleanUp ()
    {
        if (pfwrobj) (*pfwrobj)(pobject, boolTrue);
    }

};

struct PicWrObjArray
{
    PicWrObj* current;
    PicWrObj  wrobj[maxWrObj];

    void cleanUp ()
    {
        current = wrobj + maxWrObj;
        do (--current)->cleanUp(); while (current > wrobj);
        memset(wrobj, 0, sizeof(wrobj));
    }

    PicWrObjArray ()
    {
        current = wrobj;
        memset(wrobj, 0, sizeof(wrobj));
    }

   ~PicWrObjArray ()
    {
        cleanUp();
    }

    void add (const void* pobj, PICWRFUNC* pfwr, PICWROBJFUNC* pfobj)
    {
        current->set(pobj, pfwr, pfobj);
        if (current < &wrobj[maxWrObj - 1])
            current++;
    }
};

struct PicWrObjArrayMgr
{
    PicWrObjArray* list;

    PicWrObjArrayMgr () : list(NULLPTR) {}
   ~PicWrObjArrayMgr ()
    {
        delete list;
        list = NULLPTR;
    }
};


// --------------------------------------------------------------------------
//  Operator << is implemented here.
// --------------------------------------------------------------------------

PicFmtText& PicFmtText::wrFormat (
    const void*   pobject,
    PICWRFUNC*    pfwrite,
    PICWROBJFUNC* pfwrobj)
{
    if (pending)
    {
        // Keyword handling is active:
        // add the object to the pending list.

        pending->add(pobject, pfwrite, pfwrobj);
        return *this;
    }

    // Otherwise, format and write the object immediately:

    PicTextFmt tfmt (stream(), NULLPTR, indentSize);
    const char* pch = fmt;

    if (pch)
    {
        FindAtSign:
        while (*pch != '@' && *pch != '\0')
            pch++;
        if (pch > fmt)
            tfmt.write(fmt, pch - fmt);
        if (*pch == '\0')
            pch = NULLPTR;
        else if (*++pch == '@')
        {
            fmt = pch++;      // '@@' was found; write literal '@'
            goto FindAtSign;
        }
    }

    fmt += writeObject(stream(), pfwrite, pobject, fmt = pch, indentSize);

    return *this;
}


// --------------------------------------------------------------------------
//  Operator >> : memorize a keyword or process ENDFMT.
// --------------------------------------------------------------------------

static PicSemaphore  opsemPicFmtText;

PicFmtText& operator >> (PicFmtText& txt, const char* keyword)
{
    // One global pointer is reserved to store a pending list for later
    // reuse. Because formatted output is usually flushed before the
    // next one begins, a great deal of allocation can be avoided with
    // this simple optimization.

    static PicWrObjArrayMgr savePending;

    // When the first keyword is specified in the output, the pending
    // list is activated. From that point on every write object is
    // stored in the list and they are written to the stream when
    // the format is flushed by ENDFMT.

    if (keyword != ENDFMTMARK)
    {
        if (!txt.pending)
        {
            PicCritSec sec (opsemPicFmtText);

            if (savePending.list)
            {
                // Reuse previous list:
                txt.pending = savePending.list;
                savePending.list = NULLPTR;
            }
            else
                txt.pending = new PicWrObjArray();
        }

        txt.pending->current->keyword = keyword;
        return txt;
    }

    // ENDFMT was given; flush the whole format string now.

    PicTextFmt tfmt (txt.stream(), NULLPTR, txt.indentSize);
    const char* pch = txt.fmt;
    PicWrObj* end   = txt.pending ? txt.pending->current : NULLPTR;
    PicWrObj* wrobj = end;

    while (pch)
    {
        while (*pch != '@' && *pch != '\0')
            pch++;

        if (*pch == '\0') // End of format; write out the tail
        {
            if (pch > txt.fmt)
                tfmt.write(txt.fmt, pch - txt.fmt);
            txt.fmt = pch = NULLPTR;
            if (txt.pending)
            {
                PicCritSec sec (opsemPicFmtText);

                if (savePending.list)
                    delete txt.pending;
                else
                    // Save for later reuse:
                    (savePending.list = txt.pending)->cleanUp();
                txt.pending = NULLPTR;
            }
        }
        else if (pch[1] == '@')   // '@@' was found; write literal '@'
        {
            tfmt.write(txt.fmt, pch - txt.fmt + 1);
            pch = txt.fmt = pch + 2;
        }
        else if (txt.pending)  // Write next pending object in place of '@'
        {
            PICSIZE sizkwd = 0;

            if (++wrobj >= end || wrobj->keyword) // Find the next keyword
            {
                wrobj = txt.pending->wrobj;

                while ((wrobj->keyword == NULLPTR ||
                        pch - txt.fmt < (sizkwd = strlen(wrobj->keyword)) ||
                        strncmp(wrobj->keyword, pch - sizkwd, sizkwd) != 0) &&
                       ++wrobj < end)
                    ;
            }

            if (wrobj < end) // Write the next pending object
            {
                if (pch - sizkwd > txt.fmt)
                    tfmt.write(txt.fmt, pch - txt.fmt - sizkwd);
                txt.fmt  = pch + 1;
                txt.fmt += wrobj->write(txt.stream(), txt.fmt, txt.indentSize);
                pch      = txt.fmt;
            }
            else
                pch++; // Keyword not found; skip over the '@'
        }
        else
            pch++; // No pending objects; skip over the '@'
    }

    return txt;
}


// --------------------------------------------------------------------------
//  Helper methods of PicTextFmt
// --------------------------------------------------------------------------

void PicTextFmt::write (const char* data, PICSIZE size)
{
    if (indent)  // Indenting is implemented here.
    {
        BaseString indentString (' ', indent);
        PICUSHORT  i0 = 0, i = 0;

        while (i < size)
        {
            if (data[i++] == '\n')
            {
                stream.write(data + i0, i - i0);
                stream.write(indentString.chars(), indent);
                i0 = i;
            }
        }
        if (i > i0)
            stream.write(data + i0, i - i0);
    }
    else
        stream.write(data, size);
}

PICOFFSET PicTextFmt::getInt (
    const char* format,
    PICOFFSET      i,
    PICUSHORT&     num)
{
    if (isdigit(format[i]))
    {
        num = format[i] - '0';
        while(isdigit(format[++i]))
        {
            num *= 10;
            num += format[i] - '0';
        }
    }
    return i;
}

PICOFFSET PicTextFmt::writeInteger (
    PICOFFSET   i,
    PICULONG    obj,
    const char* typeName,
    PICBOOL     neg)
{
    const PICUSHORT maxMinLen = 33;

    PICUSHORT radix  = 10;
    PICUSHORT minlen = 1;
    PICBOOL   ucase  = boolFalse;
    char      buf[maxMinLen + 1];
    PICSIZE   size;

    switch (format[i])
    {
        case '?':
            write(typeName, strlen(typeName));
            return i + 1;

        case '*': // Fall through
        case 'd': radix = 10; i++; break;
        case 'H': ucase = boolTrue; // Fall through
        case 'h': radix = 16; i++; break;
        case 'o': radix = 8;  i++; break;

        default:  if (!isdigit(format[i]))
                      return i + 1;
    }

    i = getInt(format, i, minlen);
    if (minlen > maxMinLen)
        minlen = maxMinLen;

//      ltoa(obj, buf + 1, radix);
    switch(radix){
        case 8:  sprintf(buf +1, "%lo", obj); 
                 break;
        case 16: if(ucase) sprintf(buf +1, "%lX", obj); 
                 else sprintf(buf +1, "%xld", obj); 
                 break;
        default: sprintf(buf +1, "%ld", obj); 
                 break;
    }
    buf[0] = '-';

    if ((size = strlen(buf + !neg)) < minlen)
    {
        memmove(buf + 1 + minlen - size, buf + 1, size);
        memset(buf + 1, '0', minlen - size);
        write(buf + !neg, minlen);
    }
    else
        write(buf + !neg, size);

    return i;
}


// --------------------------------------------------------------------------
//  Output operators of fundamental types.
// --------------------------------------------------------------------------

static PICOFFSET wrChar (const void* obj, PicTextFmt& fmt, PICOFFSET i)
{
    switch (fmt.format[i])
    {
        case '*':
            fmt.write((char*)(PICLONG*)&obj, 1);
            return i + 1;

        default:
            return fmt.writeInteger(i, (PICULONG)obj, "char");
    }
}

PicFmtText& operator << (PicFmtText& txt, char obj)
{
    return txt.wrFormat((void*)(PICULONG)obj, wrChar);
}

// --------------------------------------------------------------------------

static PICOFFSET wrPChar (const char* obj, PicTextFmt& fmt, PICOFFSET i)
{
    switch (fmt.format[i])
    {
        case '*':
            if (obj)
                fmt.write(obj, strlen(obj));
            break;

        case '?':
            fmt.write("char*", 8);
            break;
    }
    return i + 1;
}

PicFmtText& operator << (PicFmtText& txt, const char* obj)
{
    return txt.wrFormat(obj, (PICWRFUNC*)wrPChar);
}

static PICOFFSET wrStr (const BaseString* obj, PicTextFmt& fmt, PICOFFSET i)
{
    switch (fmt.format[i])
    {
        case '*':
            if (!obj->isNull())
                fmt.write(obj->chars(), obj->size());
            break;

        case '?':
            fmt.write("BaseString", 10);
            break;
    }
    return i + 1;
}

static BaseString* wrobjBaseString (
    BaseString* obj,
    PICBOOL     cleanUp)
{
    if (cleanUp)
    {
        delete obj;
        return NULLPTR;
    }
    return new BaseString(*obj);
}

PicFmtText& operator << (PicFmtText& txt, const BaseString& obj)
{
    return txt.wrFormat (
        &obj, (PICWRFUNC*)wrStr, (PICWROBJFUNC*)wrobjBaseString);
}

// --------------------------------------------------------------------------

static PICOFFSET wrSInt (PICLONG obj, PicTextFmt& fmt, PICOFFSET i)
{
    return fmt.writeInteger(i, labs(obj), "short int", obj < 0);
}

PicFmtText& operator << (PicFmtText& txt, short int obj)
{
    return txt.wrFormat((void*)(PICLONG)obj, (PICWRFUNC*)wrSInt);
}

static PICOFFSET wrUSInt (PICULONG obj, PicTextFmt& fmt, PICOFFSET i)
{
    return fmt.writeInteger(i, obj, "unsigned short int", boolFalse);
}

PicFmtText& operator << (PicFmtText& txt, unsigned short int obj)
{
    return txt.wrFormat((void*)(PICULONG)obj, (PICWRFUNC*)wrUSInt);
}

static PICOFFSET wrInt (PICLONG obj, PicTextFmt& fmt, PICOFFSET i)
{
    return fmt.writeInteger(i, labs(obj), "int", obj < 0);
}

PicFmtText& operator << (PicFmtText& txt, int obj)
{
    return txt.wrFormat((void*)(PICLONG)obj, (PICWRFUNC*)wrInt);
}

static PICOFFSET wrUInt (PICULONG obj, PicTextFmt& fmt, PICOFFSET i)
{
    return fmt.writeInteger(i, obj, "unsigned int", boolFalse);
}

PicFmtText& operator << (PicFmtText& txt, unsigned int obj)
{
    return txt.wrFormat((void*)(PICULONG)obj, (PICWRFUNC*)wrUInt);
}

static PICOFFSET wrLInt (PICLONG obj, PicTextFmt& fmt, PICOFFSET i)
{
    return fmt.writeInteger(i, labs(obj), "long int", obj < 0);
}

PicFmtText& operator << (PicFmtText& txt, long int obj)
{
    return txt.wrFormat((void*)obj, (PICWRFUNC*)wrLInt);
}

static PICOFFSET wrULInt (PICULONG obj, PicTextFmt& fmt, PICOFFSET i)
{
    return fmt.writeInteger(i, obj, "unsigned long int", boolFalse);
}

PicFmtText& operator << (PicFmtText& txt, unsigned long int obj)
{
    return txt.wrFormat((void*)obj, (PICWRFUNC*)wrULInt);
}


