// --------------------------------------------------------------------------
//  PicThread implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#include <unistd.h>
#include <sys/wait.h>
#include <stdlib.h>

extern "C"
{
#include <stdlib.h>
}

#include <palloc.hxx>
#include <pthread.hxx>
#include <picerror.hxx>
#include <ptext.hxx>
#include <palloc.hxx>

MAKE_PICANYCLASS(PicThread, PicVirtClass, PicVirtObject)

ERROR_MESSAGES(PicThread)
PARAMSTR(alivePointers, "Thread being destructed has alive pointers.")
PARAMSTR(alreadyRunning, "Thread is already running.")
PARAMSTR(finishedAndWaiting, "Thread has finished running and is waiting.")
PARAMSTR(noProcessSlots, "Could not launch thread.")
PARAMSTR(notAPicThread, "Not a PicThread")
PARAMSTR(notRunning, "Thread not running.")
PARAMSTR(notSuspended, "Thread not suspended.")
PARAMSTR(notRunning, "Thread not running.")
PARAMSTR(runningThreadDestructed,"Running thread destructed.")
PARAMSTR(threadDestructed, "Thread has been destructed")
END_PARAM_STRINGS

static int priorityToOsValue(PicPriority p)
{
    return 1;
}

static PicPriority osValueToPriority(int p)
{
    return normal;
}

struct ThreadNode
{
    ThreadNode(PicThread *t) : thread(t) {}

    PicThread*  thread;
};


struct ThreadData
{
    ThreadData(PicThread *t, PICTHREADARG a) : thread(t), arg(a) {}
    
    PicThread*   thread;
    PICTHREADARG arg;
};


PICTHREADRESULT threadFunc(PICTHREADARG arg);

#define MAXTHREADS  100

static ThreadNode*  threads[MAXTHREADS];
static PICUSHORT    threadBufferSize    = 0;
static PICUSHORT    maxThread           = 1;

static PicThread *mainThread = new PicThread((void *)0);

PicThread::PicThread(PicPriority p)
  : tId         (newId()),
    tTid        (0),
    tState      (created),
    useCount    (0),
    res         (0)
{
    pipe(fildes);
    pipe(fildes2);
    setPriority(p);
    // The allocation of space for the array is taken care of in newId().
    threads[tId] = new ThreadNode(this);
}


PicThread::PicThread(PicThread& thread)
  : tId         (newId()),
    tTid        (0),
    tState      (created),
    useCount    (0),
    res         (0)
{
    pipe(fildes);
    pipe(fildes2);
    *this = thread;
    threads[tId] = new ThreadNode(this);
}


// This is the constructor for the very first thread. It should not be used
// for any other purpose.

PicThread::PicThread(void*)
  : tId         (1),
    tTid        (0),
    tState      (running),
    useCount    (0),
    res         (0)
{
    fildes[0] = fildes[1] = 0;
    tTid = getpid();
    threads[tId] = new ThreadNode(this);
}

   
PicThread::~PicThread()
{
}


PicThread& PicThread::operator = (PicThread& thread)
{
    PicPriority p = thread.getPriority();

    setPriority(p);
    return *this;
}


static PicSemaphore selfCrsem;

PicThread* PicThread::self()
{
    PicCritSec crSec(selfCrsem); 

    PicThread *t = NULLPTR;
    PICUSHORT i = currentId();

    if (i != 0){
        t = threads[i]->thread;
    }

    return t;
}


PicThreadPointer PicThread::thread(PICUSHORT id)
{
    PicThread *t = NULLPTR;

    if (id >= 1 && id < maxThread)
    {
        t = threads[id]->thread;

        if (!t)
            ERAISE(PicThread, threadDestructed);
    }
    else
        ERAISE(PicThread, notAPicThread);

    return PicThreadPointer(t);
}


PICUSHORT PicThread::references()
{
    PicCritSec section(semMutex);
    
    return useCount;
}


PICBOOL PicThread::start(PICTHREADARG arg, PICUSHORT stackSize)
{
    PicCritSec section(semMutex);
    switch (tState)
    {
        case created:
        case idle:
            break;

        case running:
            ERAISE(PicThread, alreadyRunning);
            return boolFalse;

        case finished:
            ERAISE(PicThread, finishedAndWaiting);
            return boolFalse;
    }
    
    PicCritSec sectReady(semReady);
    ThreadData *data = new ThreadData(this, arg);

    tTid = fork();
    if(tTid != -1){
      tState = running;
      if(tTid == 0){  // I'm child
	tTid = currentTid();
        threadFunc(data);
	exit(0);
      }
    }else{
      ERAISE(PicThread, noProcessSlots);
      delete data;
    }
    return (tTid != -1);
}


PICBOOL PicThread::wait() 
{
    int status;
    
    if(tState != created){
      waitpid(tTid, &status, 0);
      tState = idle;
    }
    return boolTrue;
}

PICTHREADRESULT PicThread::result() 
{
    PicCritSec section(semMutex);
    
    return res;
}

PICBOOL PicThread::suspend()
{
    PICBOOL success = boolTrue;
    PICTHREADID t = tid();

    if (!t)
    {
        ERAISE(PicThread, notRunning);
        success = boolFalse;
    }
    else
    {
        undefined("Thread suspension");
        success = boolFalse;
    }
    return success;
}


PICBOOL PicThread::resume()
{
    PICBOOL success = boolTrue;
    PICTHREADID t = tid();

    if (!t)
    {
        ERAISE(PicThread, notRunning);
        success = boolFalse;
    }
    else
    {
    }
    return success;
}

void PicThread::setPriority(const PicPriority p)
{
    PicCritSec section(semMutex);
    priority = p;
}


PicPriority PicThread::getPriority() 
{
    PicCritSec section(semMutex);
    
    if (tTid){
    }
    return priority;
}

PICTHREADID PicThread::tid()
{
    PicCritSec section(semMutex);
      
    return tTid;
}

PICUSHORT PicThread::id() const
{
    return tId;
}


PicThreadState PicThread::state()
{
    PicCritSec section(semMutex);

    if(currentTid() != tid()){
      int status;

      if(tState == running){  // check if really
	int ret = waitpid(tTid, &status, WNOHANG);
	if(ret>0){
	  tState = finished;
	}
      }
    }
    return tState;
}

static PicSemaphore currentIdCrsem;

PICUSHORT PicThread::currentId()
{
    PicCritSec crSec(currentIdCrsem); 

    PICTHREADID t = currentTid();
    PICINDEX i;

    for (i = 1; i <= maxThread; i++){
      if (threads[i]->thread){
        if(threads[i]->thread->tTid == t){
          break;
        }
      }
    }

    if (i == maxThread + 1)
        i = 0;

    return i;
}


PICTHREADID PicThread::currentTid()
{
    return getpid();
}


void PicThread::sleep(PICULONG milliseconds)
{
    usleep(milliseconds * 1000);
}


PICTHREADRESULT PicThread::threadFunction(PICTHREADARG arg)
{
    return 0; // by default do nothing
}


PICTHREADRESULT threadFunc(PICTHREADARG arg)
{
    ThreadData *data = (ThreadData*) arg;

    // when we get this, our launcher has exited PicThread::start :
    data->thread->semReady.set();
    PICTHREADRESULT res = data->thread->threadFunction(data->arg);

    data->thread->semMutex.set();
    data->thread->tState = PicThread::finished;
    data->thread->tTid = 0;
    data->thread->res = res;
    data->thread->semMutex.clear();

    // we're done:
    data->thread->semReady.clear();

    delete data;
    return res;
}

ThreadNode** allocateNodes()
{
    threadBufferSize = 2;

    PICSIZE size = (PICUSHORT) (threadBufferSize * sizeof(ThreadNode*));

    return (ThreadNode **) malloc(size);
}

static PicSemaphore newIdSem;

PICUSHORT PicThread::newId()
{
    PicCritSec critSec(newIdSem);
    
    PICUSHORT i = ++maxThread;

    if (maxThread+1 >= threadBufferSize)
    {
        threadBufferSize = 2 * maxThread;
	//        threads = (ThreadNode **) realloc(threads,
	//                      (PICUSHORT) (threadBufferSize * sizeof(ThreadNode*)));
    }       

    return i;
}


void PicThread::mark()
{
    PicCritSec section(semMutex);
    
    useCount++;
}


void PicThread::release()
{
    PicCritSec section(semMutex);
    
    useCount--;
}

