// --------------------------------------------------------------------------
//  PIC TIME implementation
//  Part of : PicLib
// --------------------------------------------------------------------------
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// --------------------------------------------------------------------------

#include <ptime.hxx>
#include <bdatetim.hxx>
#include <pstring.hxx>
#include <picerror.hxx>
#include <picbase.erh>
#include <pcollect.hxx>
#include <piterate.hxx>

MAKE_PICOBJECTCLASS(PicTime,PicPrimitive)

// --------------------------------------------------------------------------
//      CONSTRUCTORS, DESTRUCTOR, COPY & ASSIGNMENT

PicTime::PicTime () : PicPrimitive(), BaseTime() {}

PicTime::PicTime (PICHOURCOUNT plusInterval) :
    PicPrimitive(), BaseTime(plusInterval) {}

PicTime::PicTime (PICHOUR hour, PICMINUTE minute, PICSECOND second) :
    PicPrimitive(), BaseTime(hour, minute, second) {}

PicTime::PicTime (const BaseString& image, char delimiter) :
    PicPrimitive(), BaseTime(image, delimiter) {}

PicTime::PicTime (const BaseString& image) :
    PicPrimitive(), BaseTime(image, defaultDelimiter()) {}

PicTime::PicTime (const BaseTime& time) :
    PicPrimitive(), BaseTime(time) {}

PicTime::PicTime (const PicTime& time) :
    PicPrimitive(), BaseTime(time) {}

PicObject& PicTime::operator = (const PicObject& obj)
{
    if (TYPE_GUARD(classId(), obj))
        BaseTime::operator = ((const PicTime&)obj);
    return *this;
}

// --------------------------------------------------------------------------
//      COMPARISONS

PICBOOL PicTime::operator == (const PicObject& obj) const
{
    if (id() == obj.id())
        return BaseTime::operator == ((const PicTime&)obj);
    return boolFalse;
}

PICBOOL PicTime::operator < (const PicObject& obj) const
{
    if (id() == obj.id())
        return BaseTime::operator < ((const PicTime&)obj);
    ERAISEFMT(PicObject, uncomparable)
        >>"OBJ1" << *this
        >>"OBJ2" << obj;
    return boolFalse;
}

// --------------------------------------------------------------------------
//      TEXTUAL IMAGE

static void getFormat (
    const PicObject&   format,
    PicTimeResolution& resolution,
    PicTimeStyle&      style,
    char&           delimiter,
    char&           leader)
{
    PicIterator iter;
    PICBOOL     pictFmt = format.IS_A(PicString);

    if (!pictFmt && format.find(DELIMITER,iter))
        delimiter = format.current(iter).asChar();
    else
        delimiter = BaseTime::defaultDelimiter();

    if (!pictFmt && format.find(LEADER,iter))
        leader = format.current(iter).asChar();
    else
        leader = BaseTime::defaultLeader();

    if (!pictFmt && format.find(RESOLUTION,iter))
        switch (format.current(iter).asInt())
        {
            case FORMAT_HOURS_ONLY:
                resolution = BaseTime::hoursOnly;
                break;
            case FORMAT_HOURS_TO_MINUTES:
                resolution = BaseTime::hoursToMinutes;
                break;
            case FORMAT_HOURS_TO_SECONDS:
                resolution = BaseTime::hoursToSeconds;
                break;
            default:
                ERAISE(BaseTime, formatError);
                resolution = BaseTime::defaultResolution();
                break;
        }
    else
        resolution = BaseTime::defaultResolution();

    if (!pictFmt && format.find(STYLE,iter))
        switch (format.current(iter).asInt())
        {
            case FORMAT_HOURS_12:
                style = BaseTime::hours12;
                break;
            case FORMAT_HOURS_24:
                style = BaseTime::hours24;
                break;
            default:
                ERAISE(BaseTime, formatError);
                style = BaseTime::defaultStyle();
                break;
        }
    else
        style = BaseTime::defaultStyle();
}

BaseString PicTime::getImage (const PicObject& format) const
{
    BaseString image;

    if (!pictureFmt(format, image))
    {
        PicTimeResolution resolution;
        PicTimeStyle      style;
        char           delimiter;
        char           leader;

        getFormat(format, resolution, style, delimiter, leader);
        image = PicString::format (
            image = BaseTime::getImage(resolution, style, delimiter, leader),
            format);
    }
    return image;
}

void PicTime::setImage (const PicObject& format, const BaseString& image)
{
    PicTimeResolution resolution;
    PicTimeStyle      style;
    char           delimiter;
    char           leader;

    getFormat(format, resolution, style, delimiter, leader);
    BaseTime::setImage(image, delimiter);
}

// --------------------------------------------------------------------------
//      FORMATTED OUTPUT

PICOFFSET PicTime::writeFmt (PicTextFmt& fmt, PICOFFSET i) const
{
    static const char* typeName = id().className();

    BaseDateTime datetime = *(BaseTime*)this;

    if (isNull() && fmt.format[i] != '?')
    {
        PicTextFmt nullFmt (PicStream::null(), fmt.format, 0);
        return datetime.writeFmt(nullFmt, i, typeName);
    }

    return datetime.writeFmt(fmt, i, typeName);
}

// --------------------------------------------------------------------------
//      SET TYPED VALUE

PICBOOL PicTime::setTypedValue (const PicObject& constraints)
{
    if (isNull())
    {
        PicIterator here;
        const PicIxCollect& cntr = AS(PicIxCollect, constraints);

        if (cntr.find(NULL_IS_CURRENT, here))
            setRelativeTime(cntr.at(here).asInt());
    }
    return boolTrue;
}
