// **********************************************************************
//  This file contains the code-part of PicGUI class
//  Part of : PicGUI
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#include <stdlib.h>
#include <string.h>
#include <iostream.h>
#include <strstream.h>
#include "mmain.hxx"
#include "PicGUI.hxx"
#include "psemaph.hxx"
#include "merrors.hxx"
#include <PICClient.h>
#include <tcl.h>

const PicGUI::funcPointer PicGUI::transmitFuncs[] = {
	{"acceptRights", 0, acceptRights},
	{"addAlgorithm",3, addAlgorithm},
	{"addUser", 2, addUser},
	{"batchInsert", 2, batchInsert},
	{"browseResults", 1, browseResults},
	{"cancelQuery", 0, cancelQuery},
	{"changePassword", 3, changePassword},
	{"changeSearchType", 1, changeSearchType},
	{"changeUserDbAuthorization", 3, changeUserDbAuthorization},
	{"changeUserServerAuthorization", 3, changeUserServerAuthorization},
	{"closeQueryImage", 1, closeQueryImage},
	{"connect", 3, connect},
	{"createDatabase", 3, createDatabase},
	{"deleteImage", 1, deleteImage},
	{"disconnect", 0, disconnect},
	{"getAlgorithmInfo", 1, getAlgorithmInfo},
	{"getAlgorithmParams", 1, getAlgorithmParams},
	{"getAllDbNames", 0, getAllDbNames},
	{"getAllUsers", 0, getAllUsers},
	{"getClientOptions", 0, getClientOptions},
	{"getDatabaseAlgNames", 1, getDatabaseAlgNames},
	{"getDatabaseNames", 0, getDatabaseNames},
	{"getImage", 3, getImage},
	{"getServerAddress", 0, getServerAddress},
	{"getServerAdvert", 0, getServerAdvert},
	{"getServerOptions", 0, getServerOptions},
	{"getSignature", 2, getSignature},
	{"getThumbnail", 5, getThumbnail},
	{"getUserDbAuthorization", 2, getUserDbAuthorization},
	{"getUserServerAuthorization", 1, getUserServerAuthorization},
	{"guiStarted", 0, guiStarted},
	{"insertImage", 3, insertImage},
	{"openQueryImage", 4, openQueryImage},
	{"query", -1, query},
	{"quit", 1, quit},
	{"removeAlgorithm", 1, removeAlgorithm},
	{"removeDatabase", 2, removeDatabase},
	{"removeUser", 1, removeUser},
	{"replaceAlgorithm", -1, replaceAlgorithm},
	{"saveQueryImage", 3, saveQueryImage},
	{"scaleImage", 4, scaleImage},
	{"setAlgorithmParams", 2, setAlgorithmParams},
	{"setClientOptions", 1, setClientOptions},
	{"setDatabaseAlgorithms", -1, setDatabaseAlgorithms},
	{"setImageAttributes", 2, setImageAttributes},
	{"setNumOfDbFrames", 1, setNumOfDbFrames},
	{"setServerOptions", 4, setServerOptions}
};

// A dummy constructor.
PicGUI::PicGUI()
{}

// A dummy destructor.
PicGUI::~PicGUI()
{}


//========================================================================
//  The function to initialize the Tcl/Tk interpreter
//========================================================================

int Tcl_AppInit(Tcl_Interp *interp) {
	if ((Tcl_Init(interp) == TCL_ERROR) || (Tk_Init(interp) == TCL_ERROR))
		return TCL_ERROR;
  
	//
	// Register the 'transmit' command with the Tcl interpreter.
	//
	Tcl_CreateCommand(interp, "transmit", (Tcl_CmdProc *)PicGUI::transmitCmd,
		(ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);


  interpreter = interp;

  return TCL_OK;
}


//========================================================================
//  This method initially starts the Tcl/Tk GUI
//========================================================================

PICBOOL PicGUI::start(PICINT argc, char *argv[]) {
	Tk_Main(argc, argv, (Tcl_AppInitProc *) Tcl_AppInit);
  
	return boolTrue;
}


//========================================================================
//  This method receives a command from GUI and sends it to the main 
//  program for processing
//========================================================================

PICINT PicGUI::transmitCmd(ClientData clientdata, Tcl_Interp *interp,
	PICINT argc, char *argv[]) {
	static unsigned int numTransmitFuncs = sizeof(transmitFuncs)/sizeof(funcPointer);
	ostrstream debug;

	if(!interp)
		return TCL_ERROR;
  
	interpreter = interp;
  
	if(clientdata) {
		interp->result = "Invalid data, no client data expected.";
		return TCL_ERROR;
	}
  
	if(argc < 2) {
		interp->result = "Wrong # argc, at least 2 arguments needed.";
		return TCL_ERROR;
	}

	PicMain::guiEnteringC();

	//
	// Check which (if any!) operation is applicable.
	//
	debug << "transmitCmd searching for [" << argv[1] << "] ... ";

	PICBOOL done = boolFalse;
	for (unsigned int i = 0; (i < numTransmitFuncs) && !done; i++) {
		if (!strcmp(transmitFuncs[i].name, argv[1])) {
			debug << "found at " << i << "/" << numTransmitFuncs-1 << "." << endl << ends;
			DEBUG_GUI(debug.str());

			//
			// Check that the number of arguments is correct (-1 signifies a
			// variable number of arguments.
			//
			if ( (transmitFuncs[i].argc != -1) && (transmitFuncs[i].argc != argc-2) ) {
				interp->result = "Wrong number of args to transmit.";
				return TCL_ERROR;
			}

			//
			// If the command has a variable number of arguments,
			// replace argv[1] (the command's name) with the number
			// of arguments.
			//
			if (transmitFuncs[i].argc == -1) {
				PicInteger argv2 = argc;
				
				argv[1] = (char *)(argv2.getImage()).chars();
			}

			//
			// Call the requested method - this _needs_ to be done after debug
			// message printing (the #defined call above) or Tcl doesn't get
			// the return value a previous call to setResult() intended.
			//
			(*transmitFuncs[i].fptr)(argv);
			
			done = boolTrue;
		}
	}

	if (!done) {
		debug << "no match found from a set of " << numTransmitFuncs <<" methods." << endl << ends;
		DEBUG_GUI(debug.str());
		interp->result = "No such transmit command.";
		return TCL_ERROR;
	}

	PicMain::guiLeavingC();

	return TCL_OK;
}


//========================================================================
//  This function prints a message in the info text box
//========================================================================

PICBOOL PicGUI::infoPrint(BaseString mesg) {
	PICBOOL ret = boolTrue;

	if(!interpreter){
		ret = boolFalse;
	}else{
		Tcl_VarEval(interpreter, "printInfo {", (char *)mesg.chars(), "}",
			(char *) NULL);
	}
 
	return ret;
}


//========================================================================
//  This function sets the results of a query to the main window
//========================================================================

PICBOOL PicGUI::setQueryResults(const PicIxCollect &pics) {
	Tk_PhotoImageBlock *imageBlock;
	Tk_PhotoHandle pHandle;
	PicIterator iter;
	BaseStringArray results;

	ITERATE(pics, iter) {
		const PicPicture &pic = AS(PicPicture, pics.at(iter));
		PICUSHORT i = pics.indexAt(iter);
		ostrstream imageInfos;

		//
		// The number of the actual result image frame the data belongs to
		// (i.e. a number from 1 to PicParams.dbImageFramesOnScreen.)
		//
		imageInfos << i << " ";

		BaseString id, index, match, name;

		if (!pic.isNull()) {
			id = asString(pic.id);
			index = asString(pic.matchRanking);
			match = pic.matchRate;
			name = pic.name;

			if(DEBUG_GUI_ON)
    	    	cout << "Query result " << i << ": " << id.chars() << 
        			", " << index.chars() << ", " << match.chars() << 
        			", " << name.chars() << endl;

			imageInfos << id.chars() << " ";

			//
			// Do not set index & match, if index is 0 (means insert case).
			//
			if(atoi(index.chars())) {
				imageInfos << index.chars() << " " << match.chars() << " ";
			}
			else {
				imageInfos << "\"\"" << " " << "\"\"" << " ";
			}

			imageInfos << "\"" << name.chars() << "\"" << ends;
			results.add(imageInfos.str());

			BaseString imageName = "imgResult" & asString(i);
#ifdef TCLTK_80
			pHandle = Tk_FindPhoto(interpreter, (char *)imageName.chars());
#else
			pHandle = Tk_FindPhoto((char *) imageName.chars());
#endif
			imageBlock = &(pic.raw->rawImage);
			
			Tk_PhotoSetSize( pHandle, imageBlock->width, imageBlock->height);
			Tk_PhotoPutBlock(pHandle, imageBlock, 0, 0,
				imageBlock->width, imageBlock->height);
		}
	}

	setActions("setQueryResults", results);

	return boolTrue;
}

//========================================================================
//  This function the current query progress the GUI
//========================================================================

PICBOOL PicGUI::QueryProgressReport(PICULONG completed, PICULONG total) {
	ostrstream tmp;
	tmp << completed << " images scanned" << ends;
	BaseString report = tmp.str();

	DEBUG_GUI("GUI received query progress report.\n");

	setActions("queryProgressReport", report);

	return boolTrue;
}





/*** BEGIN v2 relics. 				***/
/*** some relics have been nuked /TI 10.3.99 	***/

//
// The methods below are relics from version 2.
//

// =============================================================  
// This function sets the percentage of advancement in current query 
// =============================================================  

/*
PICBOOL PicGUI::sendQueryAdvance(PICUSHORT percentage)
{
	cout << "*** THIS IS OFF!! in SendQueryAdvance(): " << percentage << endl;
	return boolTrue;
  
}
*/

// =============================================================  
// This function sets the new segmentation algorithms to GUI 
// =============================================================  

PICBOOL PicGUI::setSegmentationAlgorithms(BaseStringArray &names)
{
  DEBUG_GUI("GUI received new segmentation algorithms\n");
  
  PICPOSITION pos = names.next(0);
  
  while(pos)
  {
    Tcl_SetVar(interpreter, "segalgs", (char *) names[pos].chars(), \
	       TCL_LEAVE_ERR_MSG | TCL_APPEND_VALUE); 
    Tcl_SetVar(interpreter, "segalgs", " ", \
	       TCL_LEAVE_ERR_MSG | TCL_APPEND_VALUE);   
    pos = names.next(pos);
  }

  Tcl_SetVar(interpreter, "newaction", (char *) "segalgs ", \
	     TCL_LEAVE_ERR_MSG | TCL_APPEND_VALUE);

  return boolTrue;
}

//  * * * * * * * NEW FUNCTIONS ADDED TO NETPIC * * * * * * * *
/*** These are operational. I don't know if they are really needed,	***/
/*** but they are used.	/TI 10.3.99					***/
/*** BEGIN KNOWN USED METHODS						***/

PICBOOL PicGUI::FailedToCompleteRequest()
{
  INFO_PRINT("Failed to complete the request.\n");  
  return boolTrue;
}

PICBOOL PicGUI::NoMoreQueryResults()
{
  INFO_PRINT("No more query results.\n.\n");
  return boolTrue;
}

PICBOOL PicGUI::NoQueryMatches()
{
  BaseStringArray results;
  
  setActions("cancelQuery", results);
  
  INFO_PRINT("No query matches.\n");
  return boolTrue;
}

PICBOOL PicGUI::OperationComplete()
{
  INFO_PRINT("Operation complete.\n");  
  return boolTrue;
}


PICBOOL PicGUI::PasswordIncorrect()
{
  INFO_PRINT("Password incorrect.\n");
  return boolTrue;
}

/*** 	END KNOWN USED METHODS	/TI 10.3.99				***/


// =============================================================  
// Get the segment algorithm params 
// =============================================================  

void PicGUI::tr_segalgParams_get(char *algname)
{

}  

// =============================================================  
// User has changed the segment algorithm params 
// =============================================================  

void PicGUI::tr_segalgParams_set(char *algname, char *algparams)
{

}


/*** NEW IN V3 ***/

//========================================================================
//  Public auxiliary methods.
//========================================================================

//
//___ Gives server options to the GUI ____________________
//
// NOTE! There is also a private method called setServerOptions() but it
//		is used to inform the server about changed options.
//
void PicGUI::setServerOptions(BaseStringArray options) {
	DEBUG_GUI("GUI received server options' names\n");

	//
	// 'options' consists of the maximum number of clients, disconnection
	// timeout, server's http address and advertisement.
	//
	setActions("getServerOptions", options);
}       

//
//___ Gives the search types to the GUI ____________________
//
PICBOOL PicGUI::setSearchTypes(BaseStringArray &searchTypes) {
	setActions("setSearchTypes", searchTypes);

	return(boolTrue);
}


//========================================================================
//  Private auxiliary methods.
//========================================================================

//
//___ Append a new action to the 'actions' list ____________________
//
// Appends the operation name given by the parameter 'action' (and its
// arguments, if 'args' != NULL) to the 'actions' list. If 'setResult'
// equals 'boolTrue' then no action is set and the arguments are sent
// back to Tcl as transmit's return value.
//
void PicGUI::setActions(char *action, const BaseStringArray &args) {
	ostrstream result;

	result << action << " {" << BaseStringArrayToList(args) << "}" << ends;

	Tcl_SetVar(interpreter, "actions", result.str(), TCL_GLOBAL_ONLY |
		TCL_LEAVE_ERR_MSG | TCL_APPEND_VALUE | TCL_LIST_ELEMENT);

	interpreter->result = "1";
}

//
//___ Makes a new query image out of a PicPicture ____________________
//
void PicGUI::newQueryImage(PicPicture &newPicture, char *destImageName,
	char *listIndex) {
	Tk_PhotoHandle pHandle;
	Tk_PhotoImageBlock *imageBlock;
	BaseStringArray imageAttrs;

	if (newPicture.raw != NULL) {
		imageBlock = &(newPicture.raw->rawImage);

		imageAttrs.add(listIndex);					// List index.
		imageAttrs.add(asString(newPicture.id));
		imageAttrs.add(newPicture.name.isEmpty() ? "" : newPicture.name.chars());
		imageAttrs.add(newPicture.filename.chars());
		imageAttrs.add(asString(imageBlock->width));
		imageAttrs.add(asString(imageBlock->height));
		imageAttrs.add(newPicture.par1.chars());

		//
		// Get a handle on the query image specified by argv[3] and
		// copy the new image to it.
		//
#ifdef TCLTK_80
		pHandle = Tk_FindPhoto(interpreter,destImageName);
#else
		pHandle = Tk_FindPhoto(destImageName);
#endif

		Tk_PhotoSetSize(pHandle, imageBlock->width, imageBlock->height);
		Tk_PhotoPutBlock(pHandle, imageBlock, 0, 0,
			imageBlock->width, imageBlock->height);

		setActions("newQueryImage", imageAttrs);
	}
}

//
//___ Convert a PicArray to a BaseStringArray ____________________
//
BaseStringArray PicGUI::arrayToBaseStringArray(PicArray &array) {
	BaseStringArray result;
	PicIterator iter;
		
	ITERATE(array, iter) {
		result.add( (array.at(iter)).getImage() );
	}
	return result;
}

//
//___ Convert a BaseStringArray to a Tcl list ____________________
//
char *PicGUI::BaseStringArrayToList(const BaseStringArray &array) {
	ostrstream result;

	if (!array.isNull())
		for (PICPOSITION pos = array.next(0); pos; pos = array.next(pos))
			result << "{" << array[pos].chars() << "} ";

	result << ends;		// Null termination.

	return result.str();
}





//========================================================================
//  Methods to handle 'transmit' requests.
//========================================================================

//
//___ Make changes in user rights permanent ____________________
//
// No parameters.
//
void PicGUI::acceptRights(char *argv[]) {
	if (!PICClient::AcceptRights()) {
		INFO_PRINT("Error while setting new users' rights.\n");
	}
	else {
		INFO_PRINT("New users' rights set.\n");
	}
}

//
//___ Add a new user ____________________
//
// argv[2] == username
// argv[3] == password
//
void PicGUI::addUser(char *argv[]) {
	ostrstream result;

	if (!PICClient::AddUser(argv[2], argv[3])) {
		result << "Error while adding user ";
	}
	else {
		result << "Added user ";
	}
	result << argv[2] << "." << endl << ends;
	INFO_PRINT(result.str());
}

//
//___ Insert a set of images into a database ____________________
//
// argv[2] == filename (full path + name)
// argv[3] == abort running batch (0 or 1)
//
void PicGUI::batchInsert(char *argv[]) {
	if (!PicMain::batchInsert(argv[2], atoi(argv[3]))) {
		if (atoi(argv[3])) {
			DEBUG_GUI("Batch insertion aborted.\n");
		}
		else {
			setResult("running");
		}
	}
}

//
//___ Get a new batch of query result images after scrolling ___________________
//
// argv[2] == number of the first image to show
//
void PicGUI::browseResults(char *argv[]) {
	if (!PicMain::browseQueryResults(atoi(argv[2])))   
		DEBUG_GUI("No finished query available.\n");
}

//
//___ Attempt to stop an ongoing query ____________________
//
// No Parameters.
//
void PicGUI::cancelQuery(char *argv[]) {
	if (!PICClient::StopQuery()) {
		INFO_PRINT("Error while aborting query.\n");
	}
	else {
		INFO_PRINT("Query aborted.\n");
		setActions("cancelQuery");
	}
}

//
//___ Change a user's password ____________________
//
// argv[2] == username
// argv[3] == new password
// argv[4] == old password ( == "" when the superuser is changing someone else's password.)
//
void PicGUI::changePassword(char *argv[]) {
	char *oldPass = strcmp(argv[4], "") ? argv[4] : (char *)NULL;
	ostrstream result;

	if ( !PICClient::ChangePassword(argv[2], argv[3], oldPass) ) {
		result << "Error while changing password for ";
	}
	else {
		result << "Password changed for ";
	}
	result << argv[2] << "." << endl << ends;
	INFO_PRINT(result.str());
}

//
//___ The seach type was changed ____________________
//
// argv[2] == search type name
//
void PicGUI::changeSearchType(char *argv[]) {
	BaseStringArray algNames;

	DEBUG_GUI("Search type about to be changed\n");

	if(!PicMain::changeSearchType(argv[2], algNames)) {
		DEBUG_GUI("Searchtype changing failed!\n");
	}
	else {
		setActions("setSearchTypeAlgNames", algNames);
	}
}

//
//___ Change a user's database authorization ____________________
//
// argv[2] == username
// argv[3] == database name
// argv[4] == database authorization level (0, 1, 2 or 3)
//
void PicGUI::changeUserDbAuthorization(char *argv[]) {
	if (!PICClient::changeUserDbAuthorization(argv[2], argv[3], atoi(argv[4])))
		DEBUG_GUI("Failed to change user database authorization: db id not found.\n")
}

//
//___ Change a user's server authorization ____________________
//
// argv[2] == username
// argv[3] == server authorization level (0, 1 or  2)
// argv[4] == can change password (0 if can't, 1 if can)
//
void PicGUI::changeUserServerAuthorization(char *argv[]) {
	if (!PICClient::changeUserServerAuthorization(argv[2], atoi(argv[3]), atoi(argv[4])))
		DEBUG_GUI("Failed to change user server authorization: user not found.\n")
}

//
//___ Close active query image ____________________
//
// argv[2] == image list index
//
void PicGUI::closeQueryImage(char *argv[]) {
	DEBUG_GUI("Sending 'close image'.\n");

	if(!PicMain::exImageClosed(atoi(argv[2]))) {
		DEBUG_GUI("Failed to close image.\n")
	}
	else {
		setActions("closeQueryImage");
	}
}

//
//___ Attempt to establish connection ____________________
//
// argv[2] == host name
// argv[3] == username
// argv[4] == password
//
void PicGUI::connect(char *argv[]) {
	BaseStringArray initData;

	switch (PICClient::Connect(argv[2], argv[3], argv[4])) {
		case AUTHORIZED:
			initData = PICClient::GetInitData();
			//
			// initData[0]      = 0, 1 or 2 (server authorization)
			// initData[1]      = 0 or 1 (can change password)
			// initData[2]   	= algorithm name 1
			// initData[3]		= algorithm parameter window 1
			//    ...
			// initData[n]   	= algorithm name n
			// initData[n+1]	= algorithm parameter window n+1
			//

			INFO_PRINT("Connection established.\n")
			DEBUG_GUI("GUI received init data.\n");

			//
			// Add user & servername to the BaseStringArray after 'can
			// change password' and inform the GUI of a new action.
			//
			initData.addAt(3, argv[3]);
			initData.addAt(4, argv[2]);

			setActions("connect", initData);
			break;

		case UNAUTHORIZED:
			INFO_PRINT("Connection failure: wrong password or no permission.\n"); 
			break;

		case SERVERBUSY:
			INFO_PRINT("Connection failure: server is busy.\n");
			break;

		case SOCKETREADERROR:
			INFO_PRINT("Connection failure: socket read error.\n");
			break;

		case INITIALIZATIONERROR:
			INFO_PRINT("Connection failure: socket initialization error.\n");
			break;

		default:
			INFO_PRINT("Connection failure: server is down or does not exist.\n");
	}

}

//___ Create a new database ____________________
//
// argv[2] == database name
// argv[3] == database algorithms
// argv[4] == database description
//
// NOTE! This method still uses a version of PICClient::CreateDatabase()
// 		implemented by PicSearch v2. Thus is does _not_ support setting
//		the database's algorithms or description with just one call.
//
void PicGUI::createDatabase(char *argv[]) {
	ostrstream result;

	if (!PICClient::CreateDatabase(argv[2], argv[2])) {
		result << "Error while creating database ";
	}
	else
		result << "Created database ";

	result << argv[2] << "." << endl << ends;
	INFO_PRINT(result.str());
}

//
//___ Delete a database image ____________________
//
// argv[2] == id of the image to delete
//
void PicGUI::deleteImage(char *argv[]) {
	DEBUG_GUI("Deleting a databasepicture\n");
    
	if(!PicMain::deleteImageFromDb(atoi(argv[2]), 0))
		DEBUG_GUI("Deleting aborted!\n");
}

//
//___ Disconnect ____________________
//
// No parameters.
//  
void PicGUI::disconnect(char *argv[]) {
	if (!PICClient::Disconnect())
    	INFO_PRINT("Error while disconnecting.\n")
	else
		setActions("disconnect");
}   

//
//___ Get an algorithm's description ____________________
//
// argv[2] == algorithm name
//
void PicGUI::getAlgorithmInfo(char *argv[]) {
	setResult(PICClient::GetAlgorithmInfo(argv[2]));
}

//
//___ Get the current parameters of an algorithm ____________________
//
// argv[2] == algorithm name
//
void PicGUI::getAlgorithmParams(char *argv[]) 
{	
	PicArray algParams;
	PICDbAlgorithm *thisAlg;
  
    DEBUG_GUI("[GUI] getAlgorithmParams.\n");

	if((thisAlg=PicMain::getSearchAlg(argv[2])))
		setResult(arrayToBaseStringArray(thisAlg->params));

}

//
//___ Get the names of all the server's databases ____________________
//
// No parameters.
//
void PicGUI::getAllDbNames(char *argv[]) {
	setResult(PICClient::GetAllDbNames());
}

//
//___ Get the names of all the users ____________________
//
// No parameters.
//
void PicGUI::getAllUsers(char *argv[]) {
	setResult(PICClient::GetDbUsers());
}

//
//___ Get client options ____________________
//
// No parameters.
//  
void PicGUI::getClientOptions(char *argv[]) {
	PicArray params;
  
	if (PicMain::getParams(params)) {
		setResult(arrayToBaseStringArray(params));
	}
	else 
		DEBUG_GUI("GUI did not receive client options.\n");
}

//
//___ Get the algorithm names of a database ____________________
//
// argv[2] == database name
//
void PicGUI::getDatabaseAlgNames(char *argv[]) {
	setResult(PICClient::GetDbAlgorithms(argv[2]));
	cout << "Leftthis!\n";
}

//
//___ Get names of databases the user has access to ____________________
//
// No parameters.
//
void PicGUI::getDatabaseNames(char *argv[]) {
	setResult(PICClient::GetDbNames());
}

//
//___ Make the selected image a new query image ____________________
//
// argv[2] == image id
// argv[3] == destination image name
// argv[4] == image list index
//
void PicGUI::getImage(char *argv[]) {
	PicPicture *result;
  
	DEBUG_GUI("Copying a database picture to a new query picture.\n");

	if (!PicMain::newExampleImage(atoi(argv[4])))
		DEBUG_GUI("Could not create a new example image.\n");

	result=PICClient::ReadImageFromDb(atoi(argv[2]), atoi(argv[4]));
	if(!result)
	{
		DEBUG_GUI("Image reading failed...\n");
		return;
	}
		
	newQueryImage(*result, argv[3], argv[4]);
}

//
//___ Get server address ____________________
//
// No parameters.
//
void PicGUI::getServerAddress(char *argv[]) {
	setActions("getServerAddress", PICClient::GetAddress());
}

//
//___ Get server advertisement ____________________
//
// No parameters.
//
void PicGUI::getServerAdvert(char *argv[]) {
	setActions("getServerAdvert", PICClient::GetServerAdvertisement());
}

//
//___ Get server options ____________________
//
// No parameters.
//  
void PicGUI::getServerOptions(char *argv[]) {
	if(!PICClient::GetServerOptions()) {
		INFO_PRINT("Error while getting server options.\n");
	}
	else {
		INFO_PRINT("Got server options.\n");
	}
}

//
//___ Make the selected thumbnail a new query image ____________________
//
// argv[2] == name of thumbnail image
// argv[3] == name of destination query image
// argv[4] == image list index
// argv[5] == image id
// argv[6] == image name
//
void PicGUI::getThumbnail(char *argv[]) {
	Tk_PhotoImageBlock imageBlock;
	PicPicture *result;
  
	if (!PicMain::newExampleImage(atoi(argv[4])))
	{
		DEBUG_GUI("Could not create a new example image.\n");
		return;
	}

	result=PicMain::getExamplePic(atoi(argv[4]));
	if(!result)
	{
		DEBUG_GUI("Image reading failed...\n");
		return;
	}

#ifdef TCLTK_80
	Tk_PhotoGetImage(Tk_FindPhoto(interpreter, argv[2]), &imageBlock);
#else
	Tk_PhotoGetImage(Tk_FindPhoto(argv[2]), &imageBlock);
#endif
	result->getBlock(&imageBlock); 
	
	newQueryImage(*result, argv[3], argv[4]);
}

//
//___ Get user's rights to a specified database ____________________
//
// argv[2] == username
// argv[3] == database name
//
void PicGUI::getUserDbAuthorization(char *argv[]) {
	setResult(PICClient::getUserDbAuthorization(argv[2], argv[3]));
}

//
//___ Get user's server authorization ____________________
//
// argv[3] == username
//
void PicGUI::getUserServerAuthorization(char *argv[]) {
	setResult(PICClient::getUserServerAuthorization(argv[2]));
}

//
//___ Get an image's signature for a certain algorithm ____________________
//
// argv[2] == algorithm name
// argv[3] == image list index
//
void PicGUI::getSignature(char *argv[]) {
	setResult(PicMain::getSignatureImage(argv[2], atoi(argv[3])));
}

//
//___ The GUI has started ____________________
//
// No parameters.
//  
void PicGUI::guiStarted(char *argv[]) {
	PicMain::guiHasStarted();
}

//
//___ Insert an image to the database ____________________
//
// argv[2] == image name
// argv[3] == image list index
// argv[4] == update an existing image (== 0 or 1)
//
// NOTE! PicMain::insertImageToDb() seems to ignore the 'updating' flag,
//		so no replacing is possible.
// 
void PicGUI::insertImage(char *argv[]) {
	PicPicture newImage;
	Tk_PhotoImageBlock imageBlock;
  
	DEBUG_GUI("About to insert the query image into the database.\n");

#ifdef TCLTK_80
	Tk_PhotoGetImage(Tk_FindPhoto(interpreter, argv[2]), &imageBlock);
#else
	Tk_PhotoGetImage(Tk_FindPhoto(argv[2]), &imageBlock);
#endif
	newImage.getBlock(&imageBlock); 

	if (!PicMain::insertImageToDb(newImage, atoi(argv[3]), atoi(argv[4]))) {
		DEBUG_GUI("Inserting aborted!\n");
	}

	//
	// Update the id of the inserted query image.
	//
	setResult(asString(newImage.id));
}

//
//___ Load new query image from disk ____________________
//
// argv[2] == filename (full path + name)
// argv[3] == destination query image name
// argv[4] == image list index
// argv[5] == update flag
//
void PicGUI::openQueryImage(char *argv[]) {
    PicPicture newPicture;

    if (!PicMain::newExampleImage(atoi(argv[4])))
        DEBUG_GUI("Could not create a new example image.\n");

    PicMain::loadImage(argv[2], newPicture, atoi(argv[4]), atoi(argv[5]));

//	printf("JLSIZE: %d\n", newPicture.origData.size());
	
	if (newPicture.isNull()) {
        DEBUG_GUI("Image loading failed, GUI received a null picture\n")
    }
    else
        newQueryImage(newPicture, argv[3], argv[4]);
}

// new /TI 13.4.99
//
//___ Add new algorithm to system____________________
//
// argv[2] == library file name (full path + name)
// argv[3] == tk script file name (full path + name) [not required!]
// argv[4] == algorithm type
//
void PicGUI::addAlgorithm(char *argv[]) 
{
	cout << "==> addAlg gets lib:" << argv[2] << ", tk: " << argv[3];
	cout << " and type: " << argv[4] << endl;

	PICClient::AddAlgorithm(argv[2], argv[3], atoi(argv[4]));

	DEBUG_MSG("[GUI]: addAlgorithm message sent.\n");

	// Should we wait for ACK?
}

// new /TI 26.4.99
//
//___ Remove an algorithm from system____________________
//
// argv[2] == name of the algorithm to be removed
//
void PicGUI::removeAlgorithm(char *argv[]) 
{
	cout << "==> removeAlg gets: " << argv[2] << " through." << endl;
  
	PICClient::RemoveAlgorithm(argv[2]);
	
	DEBUG_MSG("[GUI]: removeAlgorithm message sent.\n");

}

// new /TI 27.4.99
//
//___ Replace an existing algorithm with new implementation ____________________
//
// argv[2] == library file name (full path + name)
// argv[3] == tk script file name (full path + name) [not required!]
// argv[4] == algorithm type
// argv[5] == algorithm to replace
// argv[6] == 1=recalcula sigs, 0=use old sigs
//
void PicGUI::replaceAlgorithm(char *argv[]) 
{
	cout << "==> replAlgiin tulee " << argv[2] << "->libbi, ";
	cout << argv[3] << "->tk, " << argv[4] << "->type,";
	cout << argv[5] << "->to repl name" << argv[6] << "->recalc (1=doit!)";
	cout << endl;   

	PICClient::ModifyAlgorithm(argv[2], argv[3], argv[5],
								atoi(argv[4]), atoi(argv[6]));
	
	DEBUG_MSG("[GUI]: modifyAlgorithm message sent.\n");
}

// 
//___ Start a query ____________________
//
// argv[2] == algorithm name
// argv[3] == index number of the picture
//
//

void PicGUI::query(char *argv[]) 
{
	DEBUG_GUI("GUI: Query starting\n");

	PicPicture *queryPic=PicMain::getExamplePic(atoi(argv[3]));

//	if (DEBUG_GUI_ON) {
	cout << "Query image attributes: " << queryPic->id << ", " <<
	  queryPic->filename.chars() << ", " << queryPic->name.chars() <<
	  ", " << queryPic->par1.chars() << ", " << queryPic->par2.chars() <<
	  endl;
//	}

	if(!PICClient::StartQuery(argv[2], *queryPic))
		DEBUG_GUI("Query failed!\n");
}

//
//___ Attempt end program execution ____________________
//
// No parameters.
//
void PicGUI::quit(char *argv[]) {
    DEBUG_GUI("Trying to exit\n");

    if (PicMain::exitProgram() == boolFalse)
		INFO_PRINT("PicSearch is not allowed to exit right now.\n")
	else
		Tcl_Eval(interpreter, "exit 0");
}

//
//___ Remove a database ____________________
//
// argv[2] == database name
// argv[3] == mode ("list" == remove from list, "disk" == remove from disk)
//
void PicGUI::removeDatabase(char *argv[]) {
	BaseString name = argv[2];
	ostrstream result;

	if (!strcmp(argv[3], "list")) {
	        result << "Database removal property removed. ";
		cout << "[CLNT] removeDatabase: not supported!" << endl;
	}
	else if (!strcmp(argv[3], "disk")) {
		if (!PICClient::DeleteDatabase(name)) {
			result << "Error while deleting database ";
		}
		else {
			result << "Deleted database ";
		}
	}
	result << argv[2] << "." << endl << ends;
	INFO_PRINT(result.str());
}

//
//___ Remove a user ____________________
//
// argv[2] == username
//
void PicGUI::removeUser(char *argv[]) {
	ostrstream result;

	if (!PICClient::RemoveUser(argv[2])) {
		result << "Error while removing user ";
	}
	else {
		result << "Removed user ";
	}
	result << argv[2] << "." << endl << ends;
	INFO_PRINT(result.str());
}

//
//___ Save the active query image to disk ____________________
//
// argv[2] == filename (full path + name)
// argv[3] == image list index
// argv[4] == overwrite (0 or 1)
//
void PicGUI::saveQueryImage(char *argv[]) {
	PICULONG listIndex = atoi(argv[3]);
  
	if (!PicMain::saveImage(argv[2], listIndex, atoi(argv[4]))) {
		if (atoi(argv[4])) {
			DEBUG_GUI("Image saving failed!\n");
			cout << "overwrite failed!" << endl;
		}
		else {
			cout << "exists!" << endl;
			setResult("fileExists");
		}
	}
}

//
//___ Shrink a given image if need be ____________________
//
// argv[2] == source image name
// argv[3] == destination image name
// argv[4] == destination width
// argv[5] == destination height
//
void PicGUI::scaleImage(char *argv[]) {
	Tk_PhotoImageBlock sourceImageBlock, *destImageBlock;
	
#ifdef TCLTK_80
	Tk_PhotoHandle sourcePHandle = Tk_FindPhoto(interpreter, argv[2]);
	Tk_PhotoHandle destPHandle = Tk_FindPhoto(interpreter, argv[3]);
#else
	Tk_PhotoHandle sourcePHandle = Tk_FindPhoto(argv[2]);
	Tk_PhotoHandle destPHandle = Tk_FindPhoto(argv[3]);
#endif
		
	int width = atoi(argv[4]), height = atoi(argv[5]);

	Tk_PhotoGetImage(sourcePHandle, &sourceImageBlock);

	if ((sourceImageBlock.width > width) || (sourceImageBlock.height > height)) {
		DEBUG_GUI("Starting scaling of query image.\n");

		PicPicture sourcePic;
		sourcePic.getBlock(&sourceImageBlock);
		
		destImageBlock = &(PicTools::scaleImage(*sourcePic.raw, width, height, 1))->rawImage;

		Tk_PhotoSetSize(destPHandle, destImageBlock->width, destImageBlock->height);
		Tk_PhotoPutBlock(destPHandle, destImageBlock, 0, 0,
			destImageBlock->width, destImageBlock->height);
	}
	else {
		DEBUG_GUI("No scaling needed, copying image.\n");

		Tk_PhotoSetSize(destPHandle, sourceImageBlock.width, sourceImageBlock.height);
		Tk_PhotoPutBlock(destPHandle, &sourceImageBlock, 0, 0,
			sourceImageBlock.width, sourceImageBlock.height);
	}
}

//
//___ Set an alogorithm's parameters ____________________
//
// argv[2] == algorithm name
// argv[3] == algorithm's parameters as one string
//
void PicGUI::setAlgorithmParams(char *argv[]) {
	PicArray algParams;
	PICDbAlgorithm *thisAlg;

	DEBUG_GUI("[GUI] setAlgorithmParams()\n");
  
	if (!PicTools::listToCollect(argv[3], algParams))
		DEBUG_GUI("Internal error while converting algorithm parameters.\n");
	
	if((thisAlg=PicMain::getSearchAlg(argv[2])))
		thisAlg->params=algParams;
}

//
//___ Store changed client options ____________________
//
// argv[2] == a Tcl list containing the client options (see source/cmain/
// 				inc/mshared.hxx for the element of this list.)
//
void PicGUI::setClientOptions(char *argv[]) {
	PicArray options, optionsCheck;
  
	DEBUG_GUI("Setting client options.\n");
  
	if (!PicTools::listToCollect(argv[2], options, ' ')) {
		DEBUG_GUI("Internal error while converting options list, aborting.\n");
		return;
	}

	optionsCheck = options;

	if (PicMain::setParams(options)) {
		if (options != optionsCheck) {
			//
			// PicMain didn't accept some of the options, so reset them.
			//
			DEBUG_GUI("Some options were refused, resetting options.\n");

			setActions("resetClientOptions", arrayToBaseStringArray(options));
		}
	}
	else 
		DEBUG_GUI("GUI was not able to set general options.\n");
}

/*  */
//
//___ Set algorithms in use with a database ____________________
//
// argv[2] == database name
// argv[3] == list of algorithms activated in db
//
void PicGUI::setDatabaseAlgorithms(char *argv[])
{
	BaseString paramString(argv[3]);
	PicArray algArray;

	cout << "setDatAlgos saa db:n " << argv[2] << endl;
	cout << "ja sille agot: " << argv[3] << endl;

	DEBUG_GUI("Setting database info.\n");
	
	if (!PicTools::listToCollect(paramString, algArray, ' ')) {
		DEBUG_GUI("Internal error while processing attributes\n");
		return;
	}

	if (!PICClient::setDatabaseAlgorithms(argv[2], algArray))
		DEBUG_GUI("Failed to set database info.\n");
}

//
//___ Tell PicMain about changes in query image attributes ____________________
//
// argv[2] == image list index
// argv[3] == new attributes (name desc1 desc2 width height sizeChanged)
//
void PicGUI::setImageAttributes(char *argv[]) {
	BaseString paramString(argv[3]);
	PicArray paramArray;
	PICULONG listIndex = atoi(argv[2]);
  
	DEBUG_GUI("Image attributes about to be changed.\n");
  
	cout << argv[2] << ": " << argv[3] << endl;

	if (!PicTools::listToCollect(paramString, paramArray, ' ')) {
		DEBUG_GUI("Internal error while processing attributes\n");
	}
	else {
		if (!PicMain::setImageAttributes(paramArray, listIndex)) {
			DEBUG_GUI("Image attributes setting failed\n");
		}
		else {
			BaseStringArray newParams = arrayToBaseStringArray(paramArray);
			
			setActions("setImageAttributes", newParams);
		}
	}
}

//
//___ Sets the number of thumbnails received in one go ____________________
//
// argv[2] == number of database frames
//
void PicGUI::setNumOfDbFrames(char *argv[]) {
    DEBUG_GUI("Trying to change the number of result image frames.\n");
    
    PicMain::setNumOfDbFrames(atoi(argv[2]));
}

//
//___ Tell the server about changes in its options ____________________
//
// argv[2] == maximum number of connections 
// argv[3] == disconnect timeout in minutes
// argv[4] == server's http address
// argv[5] == server advertisement text
//
// NOTE! There is also a public method called setServerOptions() but it
//		is used to set the Tcl variables containing the server options.
//
void PicGUI::setServerOptions(char *argv[]) {
  DEBUG_GUI("Setting server options\n");

	if (!PICClient::SetServerOptions(argv[2], argv[3], argv[4], argv[5])) {
		DEBUG_GUI("Error while setting server options.\n");
	}
	else {
		DEBUG_GUI("Server options set.\n");
	}
}


