// **********************************************************************
//
// 	PICHost.h
//
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//
// General Public License for more details.
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************  

  
#ifndef PICHOST_H_
#define PICHOST_H_ 

#include <unistd.h>
#include <signal.h>
#include <PICNet.h>
#include <PICServer.h>

#include <cfgread.h>
#include <mess.h>

#define AUTHORIZED   0
#define UNAUTHORIZED 1
#define SERVERBUSY   2
#define PORTNUMBER 10000

/**
  * PICHost is the main part of the whole PICSearch server software.
  * Its main purpose is to initialize and start all processes needed
  * to receive and process all requests sent by client programs.
  * PICHost is the mother of PICDB process and all PICServer processes.
  *
  * @version 0.1
  * @author Petri Kanerva
  */
                 
          
class PICHost
{	
private:
  int maxClient;        // Maximum number of clients
  int numClient;        // Numbers of clients right now
  pid_t pidArray[50];   // Pids of servers.
  pid_t pidSuper;       // Pid of the superuser server;
  int process;          // Pid of the server, which will be removed or added.
  bool superIn;         // Is superuser in or not
  char* userName;       // Username
  char* passWord;       // Users password
  ServerAuth authLevel; // Authorization level
  char changePassword;  // Can the user change his password
  PICNet net;           // PICNet object

  paramConfigFile params; // host config
  userConfigFile users;   // user info
  DBConfigFile DBs;       // database info
  xdrHandler *xdrH;       // xdr handler.

  static void SignalHandler(int);

  /** Reads the request. Extract the delivered packet
    * and initializes userName and passWord fields
    *
    * @return False on error. 
    */
  bool ReadRequest();

  /** Packs the servers reply to the packet and send it to the
    * Client.
    */
  void SendAcknowledgement(unsigned int replyValue);        
         
  /** Initializes the server options and the server options table if needed.
    *
    * @return true if initialization is succesful, otherways return false
    */
  bool PICHost::InitializeServerOptionsTable();

  /** Initializes a value in the server options table. The value is read
    * from the database and if it is not found, the default value given as
    * a parameter is written to the database. This function ensures that the
    * option exists in the database.
    * 
    * @return true if the option was initialized succesfully. Otherways false
    */
  bool InitializeServerOption( char* optionName, char* def );

  /** Checks the authorization of the username.
    *
    * @return Clients rights.
    */
  bool GetAuthorization();
    
  /** Initializes the user table so that logging in is possible.
    *
    * @return true if initialization is succesful, otherways return false
    */
  bool InitializeUserTable();

  /** Adds user to the user table and create anon
    * access to all databases for the user.
    */
  bool AddUser(string name,string pass,ServerAuth auth,bool changePW);

  /** Fetches authorization level of the anonymous user in the specified
    * database. If the database id is zero (server database), the return
    * value is the server authorization level of the anonymous user.
    *
    * @param db_id Database ID.
    *
    * @return Authorization level for the anonymous user.
   **/
   long AnonAccess(long db_id); 

  /** Forks a dedicated server for the client.
    * 
    * @param address Address of the client.
    * @param rights Clients rights.
    */
  void CreateServer();  

  /** Adds a server.
    */
  void AddServer();

  /** Removes a server.
    *
    * @param Pid of the terminated child process.
    */
  void RemoveServer(int childPid);
  
  /** Fetchs is there a free space in the server.
    * 
    * @return True if number of clients < maximum number of clients.
    */
  bool IsFreeSpace();
  
  /** Checks if the user is the superuser.
    *
    * @return True if it is, otherwise False.
    */
  bool IsSuperuser();

  /** Checks if the user is permited to access the server.
    *
    * @return True if the user has authorization level of 'access'
    *  or 'superuser', otherwise False.
    */
  bool IsAccess();

  /** Checks if the user is the first superuser.
    *
    * @return True if it is, otherwise False.
    */
  bool IsFirstSuperuser();

  /** Changes superuser rights to normal usr rights
    */
  void DropRights();

public:
  /** The constructor.
    *
    * @param the base path for the database.
    */
  PICHost(void);
  /** Main loop of the PICHost
    */
  void Main();
  
  /** The destructor.
    */
  ~PICHost();
};    
   
#endif // PICHOST_H_  
