#ifndef NEWDB_H
#define NEWDB_H

#include<cfgread.h>
#include<gdbm.h>
#include<message.h> /* for algorithmID_t and such. */
#include<stdlib.h>
#include<time.h>

/* The only possible use for this template is in direct & immediate connection
 * with the dbHandler below. It is very strictly bound by the memory
 * allocation "strategy" (see datum conversion) and any use other than the
 * intended one would easily result in bugs and weird code. */
 
template<class T> class simpleDatum {
public:
  typedef T value_type;
  T value;
  bool valid;				//only relevant after fetch used.

  simpleDatum(void) : valid(false) { }
  simpleDatum(T value_) : value(value_), valid(false) { }
  simpleDatum(datum d) {		//conversion from a datum
    if(d.dptr) { value=*(T *)d.dptr; valid=true; }
    else { value=T(); valid=false; }
    free(d.dptr); //assume this malloced earlier.
  }
  operator datum() {		//conversion to a datum
    datum d; 
    d.dptr=(char *)&value; d.dsize=sizeof(value);
    return d;
  }
  operator T() {			//conversion to T.
    return value;
  }
};

enum lockType {
  noLock,
  readLock,
  writeLock
};

/* the database handler class. The operations are simply delegated to GDBM.
 * The current implementation gives away all memory malloced by GDBM so that
 * it becomes the responsibility of whoever calls fetch. */

class dbHandler {
  GDBM_FILE dbFile;
  const string fileName;
  lockType lockStatus;
  bool lock(lockType type);
  void unlock(void);
public:
  dbHandler(const string fName);
  ~dbHandler(void);

  //delegate all operations to GDBM.
  datum fetch(datum k);
  bool store(datum k, datum d);
  bool remove(datum k);
  datum firstKey(void);
  datum nextKey(datum k);
};

#define PARAMSTRING_MAX 255

/* the following structures describe images and signatures as
 * they are stored in the DB. They need to be continuous in memory
 * because of GDBM. */

struct picture {
  char fileName[_POSIX_PATH_MAX+1];
  char paramStr1[PARAMSTRING_MAX+1];
  char paramStr2[PARAMSTRING_MAX+1];
  unsigned int thumbSize;
  char thumbNail[0];		
};

struct signature {
  unsigned int sigSize;
  char sig[0];
};

struct image {
  int size;
  char *img;
};

//used to refer to picture and signature databases
typedef simpleDatum<pictureID_t> picKey;

/* there is now a separate DB to hold the next available key value for a new
 * DB, algoritm and picture, separately for pictures in each DB. The ID DB
 * is referenced by a key that holds a database ID which is 0 for "server 
 * database" (assumes you want to add a new DB/algorithm) and non-0 for
 * other DBs (assumes you want to add a new picture). No IDs are EVER reused
 * in the current system, but 32-bit ints don't run out just like that.. */

typedef simpleDatum<databaseID_t> IDDBKey; //used to refer to ID database.
typedef simpleDatum<unsigned int> IDDBVal; //contained in the ID database.

class imageDB {
  dbHandler picDB;
  string directory;
  map<algorithmID_t,dbHandler *,less<algorithmID_t> > sigDBs;
  bool picKeyValid;
  bool sigKeyValid;
 public:

  imageDB(string dir);
  bool addSigDB(algorithmID_t alg);
  bool removeSigDB(algorithmID_t alg);
  
  string sigDBPathName(algorithmID_t alg);
  string imagePathName(string fileName);
  bool	saveImage(string fileName,image img);
  image loadImage(string fileName);
  
  picture *fetchPic(pictureID_t ID);
  bool storePic(pictureID_t ID,const char *fileName, const char *paramStr1,
		const char *paramStr2,unsigned int thumbSize,
		const char *thumbNail);
  bool removePic(pictureID_t ID);
  pictureID_t firstPicKey(void);
  pictureID_t nextPicKey(pictureID_t oldKey);
  bool picKeyFound(void) { return picKeyValid; }
  
  signature *fetchSig(algorithmID_t alg,pictureID_t pic);
  bool storeSig(algorithmID_t alg,pictureID_t pic,
		unsigned int sigSize,const char *sig);
  bool removeSig(algorithmID_t alg,pictureID_t pic);
  pictureID_t firstSigKey(algorithmID_t alg);
  pictureID_t nextSigKey(algorithmID_t alg,pictureID_t oldKey);
  bool sigKeyFound(void) { return sigKeyValid; }
};

typedef map<databaseID_t,imageDB *,less<databaseID_t> > imageDBMap;

class newPICDB {
  imageDBMap imgDBs;
  dbHandler keyDB;
  unsigned int getNewID(databaseID_t DB);
  void setNewID(databaseID_t DB, unsigned int value);
 public:
  DBConfigFile DBs;
  userConfigFile users;
  algoConfigFile algs;
  paramConfigFile params;

  bool addImgDB(databaseID_t DBID,const string& dir);
  bool removeImgDB(databaseID_t ID);
  imageDB *getImgDB(databaseID_t ID);

  string DBPathName(string dbName);

  //these services provided here instead of having them in the cfgFiles
  //because doing them right requires a locking scheme, which the cfgFile
  //does not provide.

  algorithmID_t newAlgID(void);
  databaseID_t newDBID(void);
  pictureID_t newPicID(databaseID_t DB);

  newPICDB(void);
};

#endif





