// **********************************************************************
//
//      PICServer.h
//
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//
// General Public License for more details.
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef PICSERVER_H_
#define PICSERVER_H_

#include <message.h>

#include <PICDebug.h>
#include <PICNet.h>
#include <mmain.hxx>
#include <mess.h>
#include <newdb.h>

// Added because of PICCache

/**
 * The server object.
 * This class takes care of one client's needs by translating its queries
 * to the database and reporting results back to the client while always
 * checking the client's authorization when appropriate. The host constructs
 * a server object at startup and creates copies of it (via fork) for each
 * authorized client.
 *
 * @version 1.0
 * @author Juha Riihimki
 * @version 1.0.1
 * @revised by SP
 **/

class PICServer {
private:
  //xdr handler.
  xdrHandler *xdrH;
  //the database. sorry about the mutable, It's a pretty ugly hack.
  mutable newPICDB DB;
  /// Reference to an interface object for database communication.
  PICNet &net;
  /// Pointer to the client's user name string.
  char *user_name;
  /// Client's server authorization level.
  ServerAuth server_auth;
  /// Flag indicating whether or not client can change its password (0/1).
  char can_change_password;
  /// Object holding the results from the current (last) query
  PicQueryArray current_query;
  /** Flag variable indicating whether or not last query was successfully
    * completed. This is also false when no queries have yet been made.
    **/
  int query_ok;
  /// ID of the database where the previous (current) query was done.
  long current_query_db_id;
  /// Server's timeout in seconds (read from the server database)
  long timeout;
  // Variable to control whether PICCache is disabled or not
  bool cacheIsDisabled;
  // Array for PICCache
  long *cacheElementsArray;
  // Elements in PICCache array
  long cacheElements;
  /// Private copy constructor to prevent assignments of PICServer objects.
  PICServer(PICServer &s);
  // Method for checking the array for matches 
  bool isInPICCache(long, long);
  /**
   * Checks for adequate authorization for an operation.
   *
   * @param req_auth The authorization required for an operation.
   * @param db_id    Database ID to check from.
   * @return True if user is authorized for the operation, false otherwise.
   **/
  int CheckAuth(DatabaseAuth req_auth,databaseID_t DBID) const;
  /**
   * Decides what to do with a specified message.
   *
   * @return False if server should shut down, true otherwise. 
   **/
  int xdrTranslate(void);
  /**
    * Constructs the initialization message (number 00). Function returns
    * after the message has been sent to the client.
    *
    * @return True if initialization succeeded, false otherwise.
    **/
  int Initialize(void) const;
  /**
   * Sends a failure message to the client. The message contains only the
   * type field (byte) MSG_ANS_FailedToCompleteRequest.
   **/
  void SendAck(ackValue val) const;
  void SendFail(void) const;
  /**
   * Sends a successful completion message to the client. The message
   * contains only the type field (byte) MSG_ANS_OperationComplete.
   **/
  void SendComplete(void) const;
  /**
    * Starts a query. All message header information before the actual
    * query parameter data in the message must have been read before
    * calling this function (i.e. the PICMessage object's read pointer
    * must point at the start of the actual query parameter data).
    *
    * @param msg StartQuery message from the client containing the
    *            query parameters (client request 02).
    **/
  void Query(query_msg &msg);
  /**
    * Stops the current query in progress. Releases all data structures
    * allocated by the query. Subsequent calls to SendMoreNails() will
    * fail until the next Query() call.
    **/
  void StopQuery(void);
  /**
    * Sends the next packet of result thumbnails. A query must be in
    * progress in order to have a chance of success. If no query is
    * currently in progress, a failure message is sent to the client.
    *
    * @param msg Reference to a message object holding client request 04.
    **/
  void SendMoreNails(moreThumbs_msg &msg);
  /**
    * Sends a collection of thumbnails to the client using the message
    * Query Results. A previous query must exist or nothing gets sent.
    *
    * @param start Number of the first result thumbnail to send.
    * @param count Number of thumbnails to send.
    **/
  void SendNails(unsigned int start,unsigned int count);
  /**
    * Insert a picture to a database. The parameter is a message object
    * of client request 05. The current read pointer of the message object
    * must point at the message body (after the type byte).
    *
    * @param msg Reference to a message object holding client request 05.
    **/
  void InsertPicture(insertPic_msg &msg);
  /**
    * Send a requested picture to the client. The parameter is a message
    * object of client request 06. The current read pointer of the message
    * object must point at the message body (after the type byte).
    *
    * @param msg Reference to a message object holding client request 06.
    **/
  void SendPicture(getPic_msg &msg);
  /**
    * Delete a specified picture from a database. The parameter is a message
    * object of client request 07. The current read pointer of the message
    * object must point at the message body (after the type byte). This method
    * does not, despite of the filename, physically delete the picture file
    * but only removes the entry for the picture from a database.
    *
    * @param msg Reference to a message object holding client request 07.
    **/
  void DeletePicture(deletePic_msg &msg);
  /**
    * Adds and/or removes algorithms from a database. The parameter is a
    * message object of client request 09. The current read pointer of the
    * message object must point at the message body (after the type byte).
    *
    * @param msg Reference to a message object holding client request 09.
    **/
  void AdjustAlgorithms(modifyAlgs_msg &msg);
  /**
    * Creates a new database. The parameter is a message object of client
    * request 15. The current read pointer of the message object must point
    * at the message body (after the type byte). This method creates the
    * directory structure for a new database and then performs the same
    * tasks as the AddDB() method. If the target base directory already
    * exists (even if empty), this operation will fail.
    *
    * @param msg Reference to a message object holding client request 15.
    **/
  void CreateDB(createDB_msg &msg);
  /**
    * Deletes a database. The parameter is a message object of client request
    * 18. The current read pointer of the message object must point at the
    * message body (after the type byte). This method not only removes the
    * database entry from the server's configuration but also physically
    * deletes the database files. Picture files are left intact (deleting
    * the picture files might create inconsistency in other, still present
    * databases since the databases may share pictures).
    *
    * @param msg Reference to a message object holding client request 18.
    **/
  void DeleteDB(deleteDB_msg &msg);
  /**
    * Sends the user database (information about all users) to the client. 
    **/
  void SendUsers(void);
  /**
    * Changes user authorization levels for n users. The parameter is a
    * message object of client request 11. The current read pointer of the
    * message object must point at the message body (after the type byte).
    *
    * @param msg Reference to a message object holding client request 11.
    **/
  void ChangeUserRights(changeRights_msg &msg);
  /**
    * Adds a new user to the server. The parameter is a message object of
    * client request 12. The current read pointer of the message object
    * must point at the message body (after the type byte).
    *
    * @param msg Reference to a message object holding client request 12.
    **/
  void AddUser(addUser_msg &msg);
  /**
    * Removes an user from the server. The parameter is a message object
    * of client request 13. The current read pointer of the message object
    * must point at the message body (after the type byte).
    *
    * @param msg Reference to a message object holding client request 13.
    **/
  void RemoveUser(deleteUser_msg &msg);
  /**
    * Changes the password for an user. The parameter is a message object
    * of client request 14. The current read pointer of the message object
    * must point at the message body (after the type byte).
    *
    * @param msg Reference to a message object holding client request 14.
    **/
  void ChangePassword(changePW_msg &msg);
  /**
    * Sends server options to the client. The message sent to the client
    * is a message object of server reply 33.
    **/
  void SendOptions(void);
  /**
    * Sets server options according to the message. The parameter is a
    * message object of client request 20. The current read pointer of the
    * message object must point at the message body (after the type byte).
    *
    * @param msg Reference to a message object holding client request 20.
    **/
  void SetOptions(setServOpt_msg &msg);
  /**
   * Sends an empty init message if the client has super user access level.
   *
   * @param msg Reference to a message object already holding the init
   *            message type byte.
   **/
  /*inline void SendCrippledInit(void) const {
    if (server_auth==SuperUser) {
      PICMessage msg(13);
      msg.WriteChar(MSG_SRV_Init);
      msg.WriteInt32(0);
      msg.WriteInt32(0);
      msg.WriteInt32(0);
      net.WritePacket(&msg);
      DEBUG_MSG("[SRV] crippled init message sent\n");
    }
    }*/
  /**
   * Fetches authorization level of the anonymous user in the specified
   * database. If the database id is zero (server database), the return
   * value is the server authorization level of the anonymous user.
   *
   * @param db_id Database ID.
   * @return Authorization level for the anonymous user. 
   **/
  long AnonAccess(databaseID_t db_id) const;
  /**
   * Locks the specified database. If an error occurs, an error message
   * is outputted to cerr and the error is ignored.
   *
   * @param db_id ID of the database to be locked.
   * @param func_name Name of the function to be printed in the possible
   *                  error message.
   **/
  /*inline void LockDB(long db_id,char *func_name) const {
    ASSERT(func_name);
    PICDBOperation lock_op(LOCK,db_id,0,0);
    if (db.ExecuteOperation(lock_op)!=DB_SUCCESS)
      ERROR_MSG("[SRV] "<<func_name<<"(): db failed to lock ("<< \
		lock_op.GetError()<<") - ignored\n");
		}*/
  /**
   * Unlocks the specified database. If an error occurs, an error message
   * is outputted to cerr and the error is ignored.
   *
   * @param db_id ID of the database to be locked.
   * @param func_name Name of the function to be printed in the possible
   *                  error message.
   **/
  /*inline void UnlockDB(long db_id,char *func_name) const {
    ASSERT(func_name);
    PICDBOperation unlock_op(UNLOCK,db_id,0,0);
    if (db.ExecuteOperation(unlock_op)!=DB_SUCCESS)
      ERROR_MSG("[SRV] "<<func_name<<"(): db failed to unlock ("<< \
		unlock_op.GetError()<<") - ignored\n");
		}*/
  /**
    * Sends a query progress report message to the client. The specified
    * number is coded into the reply message.
    *
    * @param percent Desired number to be sent to the client (32bit).
    **/
  void SendQueryProgress(long num) const;
  /**
    * Reads server parameters from the server database. This currently
    * sets up the server timeout, server advertisement (information) text
    * and WWW address string.
    *
    * @return True if server parameters were successfully initialized, false
    *         otherwise.
    **/
  int InitOptions(void);
  /**
    * Tries to find a free filename based on the specified file name.
    * The filename string may change during this function call.
    *
    * @param name_ptr Pointer to a pointer to a filename string (with
    * full path) which is to be tried out.
    * @return True if a suitable filename was generated, false otherwise.
    **/
  int GenerateFileName(char **name_ptr) const;
  
  bool addAlgorithm(addAlg_msg &msg);
  bool modifyAlgorithms(modifyAlg_msg &msg);
  bool removeAlgorithm(removeAlg_msg &msg);

public:
  /**
   * The constructor.
   * Initializes the common information shared between all the server
   * processes (all the processes are copies (forked by the host) of
   * one PICServer object).
   *
   * @param d A reference to the interface object for DB communication.
   **/
  inline PICServer(PICNet &n,xdrHandler *x) : xdrH(x), DB(), net(n) { }
  /**
   * The destructor.
   * Performs all the needed uninitialization when the object is deleted.
   **/
  inline ~PICServer(void) { 
    cout << endl << "in PICServer destructor!" << endl << endl;
  }
  /**
   * The server process main loop.
   * This method is called by the host process to change to server
   * process mode. This function is responsible for the communication
   * between the server and a client. When this method returns, this
   * server session between a client and the server has ended.
   *
   * @param user Pointer to client's user name string.
   * @param auth Client's server authorization level.
   * @param chgpass True if client can change its password, false otherwise.
   **/
  void Main(char *user,ServerAuth auth,char chgpass);
};

#endif // PICSERVER_H_

