// **********************************************************************
//
//      Database.cpp
//
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//
// General Public License for more details.
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#include <limits.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h> // remove()
#include <sys/stat.h> // for mkdir() & stat()
#include <iostream.h>
#include <PICDebug.h>
#include <PICNet.h>
#include <PICServer.h>
#include <mmain.hxx>

void PICServer::AdjustAlgorithms(modifyAlgs_msg &msg) {
  const string DBName=msg.targetDB;
  unsigned int algNum=msg.usedAlgs.usedAlgs_len;
  algorithmID_t *algs=msg.usedAlgs.usedAlgs_val;

  if(!DB.DBs.exists(DBName)) {
    DEBUG_MSG("[SRV] AdjustAlgorithms: db doesn't exist.\n");
    SendAck(ack_failed);
  }

  //no go through all possible algorithms.
  for(cfgAlgIterator ai=DB.algs.begin();ai!=DB.algs.end();++ai) {
    algorithmID_t algID=ai.ID();
    if(!DB.DBs[DBName].algs[algID]) { //this algorithm not active in DB.

      //se if it needs to be set active (per the request).
      unsigned int i;
      for(i=0;i<algNum;i++)
	if(algs[i]==algID)
	  break; //alg found in the list.
      if(i==algNum) //not found, so nothing needs to be changed.
	continue;

      //a new algorithm needs to be added to the DB.
      DEBUG_MSG("Adding algorithm with ID " << algID << endl);

      imageDB *imgDB=DB.getImgDB(DB.DBs[DBName].ID);
      if(imgDB==NULL) {
	DEBUG_MSG("[SRV] AdjustAlgorithms: failed to get imgDB\n");
	continue; //go try the next one.
      }

      //add a signature database for the new alg.
      imgDB->addSigDB(algID);

      DB.DBs[DBName].algs[algID]=true; //set alg active.

      //iterate all pictures in the image database
      for(pictureID_t picID=imgDB->firstPicKey();imgDB->picKeyFound();
	  picID=imgDB->nextPicKey(picID)) {
	picture *thePic=imgDB->fetchPic(picID);
	if(thePic==NULL) {
	  DEBUG_MSG("thumb for pic " << picID << " not fetched\n");
	  continue; //try the rest, anyways.
	}

	PicPicture pic; //get the picture.
	pic.loadFromFile(imgDB->imagePathName(thePic->fileName).c_str());
	pic.id=picID;
	pic.name=thePic->fileName;
	pic.par1=thePic->paramStr1;
	pic.par2=thePic->paramStr2;

	//FIXME: using calcAllSignatures since I don't know how to do just
	//one..
	if(!PicMain::calcAllSignatures(pic)) {
	  DEBUG_MSG("failed to calc sigs!\n");
	  continue; //so skip this one.
	}

	//iterate all signatures and insert the relevant one.
	PicSignature *sigtmp;
	pic.signatures.listRewind();
	while((sigtmp=(PicSignature *)pic.signatures.listNext())) {
	  // Algorithm id is sig->creator_id
	  if((algorithmID_t)sigtmp->creator_id!=algID)
	    continue; //not there yet..

	  DEBUG_MSG("correct sig found!\n");

	  unsigned int sigSize=sigtmp->signature_size;
	  char *sig=sigtmp->signature; 

	  if (!imgDB->storeSig(algID,picID,sigSize,sig)) {
	    ERROR_MSG("[SRV] InsertPicture: storeSig: db failed to comply\n");
	    continue;
	    return;
	  }
	}
	free(thePic); //free the GDBM mem. I'll probably lose some here..
      }
    } else { //the algorithm is active in the DB.      

      //see if in needs to be set inactive (per the request).
      unsigned int i;
      for(i=0;i<algNum;i++)
	if(algs[i]==algID)
	  break; //alg found in the list.
      if(i!=algNum) //found, nothing needs to be changed.
	continue;

      //an algorithm needs to be removed from the DB.
      DEBUG_MSG("Removing algorithm with ID " << algID << endl);

      imageDB *imgDB=DB.getImgDB(DB.DBs[DBName].ID);
      if(imgDB==NULL) {
	DEBUG_MSG("[SRV] AdjustAlgorithms: failed to get imgDB\n");
	continue; //go try the next one.
      }

      if(!imgDB->removeSigDB(algID)) {
	DEBUG_MSG("[SRV] AdjustAlgorithms: failed to remove sigDB\n");
	continue; //go try the next one.
      }

      DB.DBs[DBName].algs[algID].remove(); //remove the algorithm.
      //FIXME: delete the physical sigDB.
    }
  }
  SendComplete();
}

void PICServer::CreateDB(createDB_msg &msg) {
  DEBUG_MSG("[SRV] creating db '"<< msg.DBName <<"' at '"<< msg.directory <<
	    "' with "<< msg.algs.algs_len <<" algorithms\n");

  if(DB.DBs.exists(msg.DBName)) {
    DEBUG_MSG("[SRV] db already exists.\n");
    SendAck(ack_failed);
  }

  const string DBPath=DB.DBPathName(msg.directory);
  if (mkdir(DBPath.c_str(),0700)) {
    DEBUG_MSG("[SRV] failed to create db directory\n");
    SendAck(ack_failed);
    return;
  }
  
  const string imageDirPath=DBPath + "/images";
  if (mkdir(imageDirPath.c_str(),0700)) {
    DEBUG_MSG("[SRV] failed to create image directory\n");
    SendAck(ack_failed);
    return;
  }

  const databaseID_t newDBID=DB.newDBID();

  DB.DBs[msg.DBName].dir=msg.directory;
  DB.DBs[msg.DBName].ID=newDBID;

  DEBUG_MSG("[SRV] new db added with id " << newDBID << endl);

  DB.users[user_name].dbAuth[msg.DBName]=DB_WriteAccess;

  //FIXME: this part of the .cfg design is not quite up to par.
  for(unsigned int i=0;i<msg.algs.algs_len;++i) {
    const algorithmID_t algID=msg.algs.algs_val[i];
    
    if(DB.algs[algID].active) {  //this algorithm active..
      DB.DBs[msg.DBName].algs[algID]=true; //set alg present in DB.
    } 
    else { //alg not active! FIXME: make a special case for replaced algs.
      DEBUG_MSG("[SRV] CreateDB: invalid algorithm "<< algID << 
		" received" << endl);
    }
  }

  if(!DB.addImgDB(newDBID,DBPath)) {
    DEBUG_MSG("[SRV] failed to add image DB handler\n");
    SendAck(ack_failed);
    return;
  }

  newDB_msg reply;
  
  reply.DB.ID=newDBID;
  reply.DB.authLevel=DB_WriteAccess;
  reply.DB.name=msg.DBName;

  //this is a bit complicated, sorry..
  reply.DB.algs.algs_len=DB.algs.size(); //get mem for maximum possible.
  reply.DB.algs.algs_val=new algorithmID_t[reply.DB.algs.algs_len];

  int algc=0;
  for(cfgAlgIterator ai=DB.algs.begin(); ai!=DB.algs.end();++ai) {
    if(DB.DBs[msg.DBName].algs[ai.ID()]) //algorithm active.
      reply.DB.algs.algs_val[algc++]=ai.ID(); //add it.
  }
  reply.DB.algs.algs_len=algc; //set actual number found.

  xdrH->send(&reply);
  //FIXME: leaks mem.
}

//delete does a quick & dirty cleanup.
void PICServer::DeleteDB(deleteDB_msg &msg) {
  DEBUG_MSG("[SRV] deleting db " << msg.DBName << endl);

  if(!DB.DBs.exists(msg.DBName)) {
    ERROR_MSG("[SRV] '"<<user_name<<"' tried to delete a non-existent "<<
	      "database "<<msg.DBName<<endl);
    SendAck(ack_failed);
    return;
  }

  DB.removeImgDB(DB.DBs[msg.DBName].ID); //remove the image database.

  //and then, go for the files.
  string DBPath="database/" + (string)DB.DBs[msg.DBName].dir;
  string command="rm -R " + DBPath; //ho ho.
  system(command.c_str());

  if(!DB.DBs.remove(msg.DBName)) {
    ERROR_MSG("[SRV] unable to erase " << msg.DBName << endl);
    SendAck(ack_failed);
    return;
  }

  //go through all users and remove the DB auth.
  for(cfgIterator i=DB.users.begin();i!=DB.users.end();++i) 
    DB.users[i.name()].dbAuth[msg.DBName].remove();

  SendAck(ack_OK);
}




