// **********************************************************************
//
//      Server.cpp
//
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//
// General Public License for more details.
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#include <limits.h>
#include <stddef.h>
#include <signal.h> // signal handling functions
#include <string.h>
#include <unistd.h> // standard Unix functions
#include <iostream.h>
#include <PICDebug.h>
#include <PICServer.h>
#include <PICNet.h>
#include <strstream.h> //v3 for parameters
#include <stdio.h>     //v3 for parameters
#include <mmain.hxx> // for the V1 stuff...

// this is used by the signal handler function to access the net object
// this could also be defined INSIDE the PICServer class as static member...

//FIXME: this is one of those ugly-as-hell solutions.
static PICNet *net_static;
static xdrHandler *xdr_static;

static void ServerSign(int s) {
  switch (s) {
  case SIGUSR1:
    // this signal is raised by the net object on a read error. this typically
    // means either that the network connection has been closed by the client
    // or that malformed data has been received. on any occasion, the server
    // just closes the connection (if applicable) and dies
    DEBUG_MSG("[SRV] unexpected network communication error while reading "<< \
	      "from socket\n");
    net_static->CloseServerSocket();
    exit(0);
    break;
  case SIGUSR2:
    // this signal is raised by the host object if it desires to kill this
    // server object. we need to simply just close the network connection
    // and commit suicide
    DEBUG_MSG("[SRV] received termination request from host\n");
    net_static->CloseServerSocket();
    exit(0);
  case SIGPIPE:
    // this signal is raised by the kernel when it detects that the network
    // connection has been lost while sending data to the client. on this
    // signal the server will simply close the network connection (if
    // applicable) and die
    DEBUG_MSG("[SRV] unexpected network communication error while writing "<< \
	      "to socket\n");
    net_static->CloseServerSocket();
    exit(0);
    break;
  case SIGALRM:
    // this signal is the server's timeout trap. it will be raised by the
    // kernel when the server should timeout. when this signal is received
    // the server will send a timeout message to the client, close the
    // network connection and die
    DEBUG_MSG("[SRV] client timeout\n");
    {
      ack_msg timeout;

      timeout.value=ack_timeout;
      xdr_static->send(&timeout);
      net_static->CloseServerSocket();
    }
    exit(0);
    break;
  default:
    // unexpected signals will be dumped to the debug log (this should never
    // happen but...)
    DEBUG_MSG("[SRV] unexpected signal "<<s<<" received - ignored\n");
    break;
  }
}

void PICServer::Main(char *user,ServerAuth auth,char chgpass) {
  // the server starts up here... first init few variables

  user_name=user;
  server_auth=auth;
  can_change_password=chgpass;
  net_static=&net;
  xdr_static=xdrH;
  query_ok=0;
  current_query_db_id=-1;
  // Clear the PICCache information
  cacheIsDisabled = false;
  cacheElements = 0;
  cacheElementsArray = NULL;

  // set up the signal handler...
  signal(SIGPIPE,ServerSign); // error in socket write
  signal(SIGALRM,ServerSign); // timeout trap
  signal(SIGUSR1,ServerSign); // error in socket read (raised by PICNet)
  signal(SIGUSR2,ServerSign); // termination request from host
  signal(SIGINT,SIG_IGN);     // we don't want host's SIGINT handler
  signal(SIGCHLD,SIG_IGN);    // we don't want host's SIGCHLD handler

  // the main loop. before entering the acutal loop, we still have some
  // initializing to do - we need to read the server parameters from
  // the server database; InitOptions() will do this. if it fails, we
  // can't start up the server at all
  if (InitOptions()) {
    // now we need to initialize the old code from version 1. we achieve this
    // by making a dummy instance of the PicMain class
    PicMain main_obj;
    // we're ready to send the initialization message to the client. try to
    // construct it and if it fails, let the user in only if he/she has
    // superuser authorization level
    if (Initialize()||server_auth==SuperUser) {
      // the infinite main loop which can be breaked only from inside
      for (int cont=1; cont;) {
	DEBUG_MSG("[SRV] waiting for request\n");
	// first we set up a timeout alarm clock
	signal(SIGALRM,ServerSign); // double check the signal handler ;)

	//FIXME: make all important stuff here active again.
	//alarm(timeout);
	// wait for a message from the client
	//PICMessage msg(net.ReadPacketLen());
	// kill the timeout alarm for the duration of the processing of
	// the request
	//alarm(0);
	// read in the message which has been received from the client
	// net.ReadPacket(&msg);
	// translate & execute the request. xdtTranslate() will return
	// a non-zero return value when disconnect is desired
	cont=xdrTranslate();
      }
    }
  }
  // on disconnect, just violently close down the network connection and exit
  DEBUG_MSG("[SRV] disconnecting\n");
  net.CloseServerSocket();
}

int PICServer::InitOptions(void) {
  timeout=DB.params["host"].timeout;
  DEBUG_MSG("[SRV] timeout set to " << timeout << endl);
  return 1;
}

long PICServer::AnonAccess(databaseID_t db_id) const {
  for(cfgIterator i=DB.DBs.begin();i!=DB.DBs.end();++i) {
    if(DB.DBs[i.name()].ID==db_id) {
      return DB.users["anonymous"].dbAuth[i.name()];
    }
  }
  return 0;
}

int PICServer::Initialize(void) const {
  DEBUG_MSG("[SRV] constructing init message\n");
  init_msg init;
  string userName(user_name);

  // now just fill in the data to the message and free memory as we go
  init.authLevel=DB.users[userName].authLevel;
  init.changePW =DB.users[userName].changePass;
  
  init.DBs.DBs_len=DB.DBs.size();
  init.DBs.DBs_val=new DBInfo[init.DBs.DBs_len];
  
  unsigned int DBc=0;
  for(cfgIterator i=DB.DBs.begin();i!=DB.DBs.end();++i) {
    DBInfo &DBi=init.DBs.DBs_val[DBc];
    const string& dbName=i.name();

    DBi.ID=DB.DBs[dbName].ID;
    cout << "DBi.ID: " << DBi.ID << endl;
    DBi.authLevel=DB.users[userName].dbAuth[dbName];
    cout << "DBi.authLevel: " << (int)DBi.authLevel << endl;
    DBi.name=strdup(dbName.c_str());

    DBi.algs.algs_len=DB.algs.size();
    DBi.algs.algs_val=new algorithmID_t[DBi.algs.algs_len]; //max possible size

    unsigned int algc=0;
    for(cfgAlgIterator ai=DB.algs.begin();ai!=DB.algs.end();++ai) {
	  if(DB.DBs[dbName].algs[ai.ID()]) //algorithm present
        DBi.algs.algs_val[algc++]=ai.ID(); //send.
    }
    DBi.algs.algs_len=algc; //set real num.

    DBc++;
  }

  if(DBc!=init.DBs.DBs_len) { //this _really_ shouldn't happen.
    DEBUG_MSG("NUMBER OF DATABASES MISMATCH!\n");
  }

  init.algs.algs_len=DB.algs.size();
  init.algs.algs_val=new AlgInfo[init.algs.algs_len];

  PicMain::globData.searchAlgs.listClean();

  int algc=0;
  for(cfgAlgIterator ai=DB.algs.begin();ai!=DB.algs.end();++ai) {
    AlgInfo& alg=init.algs.algs_val[algc];
    const algorithmID_t algID=ai.ID();

    alg.ID=algID;

    PicSearchAlgorithm *algtmp;
    const string location="database/alg" + itos(algID) + ".so";
	
    algtmp=new PicSearchAlgorithm((char *)location.c_str(), algID);
    if(algtmp && algtmp->algID())			  // Alg seems ok
    {
		PicMain::globData.searchAlgs.listAdd((void *)algtmp);

        if(DB.algs[algID].paramWin)
        {
            const string tkLocation="database/alg" + itos(algID) + ".tk";

            FILE *fp=fopen(tkLocation.c_str(), "r");
            if(fp)
            {
                int fileSize;

                fseek(fp, 0, SEEK_END);
                fileSize=ftell(fp);
                fseek(fp, 0, SEEK_SET);

                alg.algParamWin=new char[fileSize+1];
                fread(alg.algParamWin, 1, fileSize, fp);
                alg.algParamWin[fileSize]=0;

                fclose(fp);
            }
            else
                alg.algParamWin="";
        }
        else
            alg.algParamWin="";

        alg.ID=algID;
        alg.name=strdup(algtmp->algName());
        alg.description=strdup(algtmp->algInfo());
    }
	else					// Do a lamo-kludge
	{
        const string failName=itos(algID) + "_Invalid!";
		
		alg.algParamWin="";
		alg.ID=0;
		alg.name=strdup(failName.c_str());
		alg.description=strdup("Server initialize failed for this alg!\n");
	}

    algc++;
  }

  const string info=DB.params["host"].info;
  const string http=DB.params["host"].http;

  init.serverAd=strdup(info.c_str());
  init.serverHTTP=strdup(http.c_str());

  xdrH->send(&init);
  DEBUG_MSG("[SRV] init message sent\n");

  //FIXME: this one leaks a helluva lotta mem.

  return 1;
}

//FIXME: make 'em use SendAck instead.
void PICServer::SendFail(void) const {
  SendAck(ack_failed);
  LOG_MSG("[SRV] failure message sent\n");
}

void PICServer::SendAck(ackValue val) const {
  ack_msg ack;

  ack.value=val;
  xdrH->send(&ack);
  LOG_MSG("[SRV] ack sent\n");
}

//FIXME: make 'em use SendAck instead.
void PICServer::SendComplete(void) const {
  SendAck(ack_OK);
  LOG_MSG("[SRV] success message sent\n");
}

int PICServer::CheckAuth(DatabaseAuth req_auth,databaseID_t DBID) const {
  for(cfgIterator i=DB.DBs.begin();i!=DB.DBs.end();++i) {
    if(DB.DBs[i.name()].ID==DBID) {
      return DB.users[user_name].dbAuth[i.name()]>=req_auth;
    }
  }
  return 0;
}




