// **********************************************************************
//  PicQueryArray and PicQueryObj class definitions
//  Part of : PicCore
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef MQUERY_HXX
#define MQUERY_HXX

#include <parray.hxx>
#include <btime.hxx>

/*:class ********************************************************************

PicQueryObj represents one image in a result set of a query. 
It contains a imageId + matchRate pair. PicQueryObj:s can
be sorted according to matchRate and are typically put into 
PicQueryArray (the higher the matchRate, the *smaller* is the object
in sort order, so ascending sort order will put it close to the
head of the array).

****************************************************************************/

class PicQueryObj : public PicObject
{
 public:
  PICOBJECTCLASS(PicQueryObj)

  PicQueryObj();
  PicQueryObj(PICULONG imageId, PICUSHORT matchRate);

  PICULONG  imageId;
  PICUSHORT matchRate;

  virtual PicObject& operator = (const PicObject&);

  virtual PICBOOL operator == (const PicObject&) const;
  virtual PICBOOL operator <  (const PicObject&) const;
  virtual PICBOOL isNull() const;
  virtual void    setNull();
  virtual PICBOOL isDefault() const;
  virtual void    setDefault();
  virtual BaseString getImage() const;
  virtual BaseString getImage(const PicObject& format) const;
  virtual void setImage(const BaseString& image);
  virtual void setImage(const PicObject& format, const BaseString& image);
};

/*:class ********************************************************************

PicQueryArray represents a result set of a query. It contains 
a collection of PicQueryObj:s.

****************************************************************************/

class PicQueryArray : public PicArray
{
 public:
  PICOBJECTCLASS(PicQueryArray)

  PicQueryArray ();
  PicQueryArray (const PicQueryArray&);
 ~PicQueryArray ();

  PicObject& operator = (const PicObject&);

  virtual void    setNull();
//:fun
  virtual PICBOOL add(const PicObject& obj);
/*
Use this function to add new elements to the array.
*/

//:fun
  void crop(int maxSize);
/*
Sorts the array and discards elements from the end 
until size is not greater than maxSize.
*/

//:fun
  void updateTotalTime();
/*
Updates the total execution time of this query.
*/

//:var
  PICULONG scannedPics;
  BaseTime startTime;
  BaseTime totalTime;
/*
Statistics about this query:
  - scannedPics is the number of images so far scanned
    (crop may modify this)
  - startTime is used to hold the time when the query started 
    (reset at PicEnhancedThread::checkCommand after a query 
    suspension is released)
  - totalTime is used to hold the total execution time of 
    this query
*/

};

// **********************************************************************

inline PICBOOL PicQueryObj::isDefault() const 
  { return isNull(); }
inline void PicQueryObj::setDefault()
  { setNull(); }
inline BaseString PicQueryObj::getImage(const PicObject& format) const
  { if (&format) return getImage(); else return getImage(); }
inline void PicQueryObj::setImage(const PicObject& format, 
                                    const BaseString& image)
  { if (&format) setImage(image); else setImage(image); }

#endif
