// **********************************************************************
//  Shared data members
//  Part of : PicCore
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef MSHARED_HXX
#define MSHARED_HXX

#include <parray.hxx>
#include <PicList.hxx>

class PicSearchType;
class PicSearchAlgorithm;
class PicSegmentAlgorithm;
class PicImageFormat;

/*:class ********************************************************************

PicParams contains the general parameters of the system.

****************************************************************************/

#define PICPARAMS_SHOW_ABORT_QUERY  1
#define PICPARAMS_ASK_CHANGE_CONF   2
#define PICPARAMS_DB_INS_PHYS       3
#define PICPARAMS_GUI_DEBUG         4
#define PICPARAMS_CORE_DEBUG        5
#define PICPARAMS_ALG_DEBUG         6
#define PICPARAMS_DB_DEBUG          7
#define PICPARAMS_DEF_WIDTH         8
#define PICPARAMS_DEF_HEIGHT        9

class PicParams
{
 public:
  PicParams();

//:fun
  PICBOOL fromColl(PicIxCollect &coll);
  PICBOOL toColl(PicIxCollect &coll);
/*
These functions set/get the following parameters 
from/to a collection of PicStrings (in this order):
  showAbortedQuery, askChangeConfirm, dbInsPhysically,
  picGUIDebug, picCoreDebug, picAlgDebug, picDbDebug,
  defWidth and defHeight.
Validity of defWidth and defHeight are checked.
*/

//:var
  PICBOOL showAbortedQuery;
/*
If a query gets aborted (by a newly launched query) do we 
show the results anyway? 
*/
//:var
  PICBOOL askChangeConfirm;
/*
If an image has been edited but the changes not exported from
the image editor window, do we ask a confirmation?
(Implemented by PicGUI).
*/
//:var
  PICBOOL dbInsPhysically;
/*
Should we insert image data to database or just the filenames
("links").
*/
//:var
  PICBOOL picGUIDebug;
  PICBOOL picCoreDebug;
  PICBOOL picAlgDebug;
  PICBOOL picDbDebug;
/*
Flags controlling the debug modes. Subsystems print many debugging
messages when their flag is on.
*/

//:var
  PICUSHORT getDefWidth();
  PICUSHORT setDefWidth(PICUSHORT newWidth);
  PICUSHORT getDefHeight();
  PICUSHORT setDefHeight(PICUSHORT newHeight);
/*
Get and set default dimensions of example images.
Set methods have range checking and return the actual value in effect.
*/

//:var
  PICUSHORT threadAbortTimeout;
  PICUSHORT dbImageFramesOnScreen;
  PICUSHORT stampWidth;
  PICUSHORT stampHeight;
  PicImageFormat *formatForScaledImages;
/*
These parameters are mainly for internal use. They are not set
or accessed by the user or written into database.
 - threadAbortTimeout indicates the amount of seconds to wait for 
   a thread to respond in PicMain::abortThread and suspendThread
 - dbImageFramesOnScreen is the number of frames for database images
   on the screen. This may be changed by PicGUI calling
   PicMain::setNumOfDbFrames
 - stampWidth is the maximum width for thumbnail representations of
   database images
 - stampHeight is the maximum height for thumbnail representations of
   database images
 - formatForScaledImages is a pointer to the PicImageFormat to be
   used for thumbnail representations of database images 
   (if NULLPTR, then raw RGB values are used)
   Used by PicMainDbUser::insertBasicData and updateBasicData.
*/

 private:

  PICUSHORT defWidth;
  PICUSHORT defHeight;

};

/*:class ********************************************************************

PicMainSharedData contains various globally accessible data objects.

****************************************************************************/

class PicMainSharedData
{
 public:

//:fun
  PicImageFormat*        getImageFormat(PICULONG id);
  PicSearchType*         getSearchType(BaseString name);
  PicSearchAlgorithm*    getSearchAlg(PICULONG id);
  PicSegmentAlgorithm*   getSegmentAlg(BaseString name);
  PicSegmentAlgorithm*   getSegmentAlg(PICULONG id);
/*
These functions can be used to retrieve a pointer to algorithms,
search types and image formats. Retrieval can be done by id,
name or both.
*/

//:var
  PicArray imageFormats;
  PicArray searchTypes;
  PicList searchAlgs;
  PicArray imageProcessAlgs;
  PicArray segmentationAlgs;
/*
These collections hold "staticly" the prevalent algorithms and
image format capabilities of the system. The objects inside them
are created at program start-up.

imageFormats        -- collection of PicImageFormat:s
searchTypes         -- collection of PicSearchType:s
searchAlgs          -- collection of PicSearchAlgorithm:s
segmentationAlgs    -- collection of PicSegmentAlgorithm:s
*/

//:var
  BaseString currentSearchType;
/*
Name of the current search type.
*/

//:var
  PicParams params;
/*
General parameters for the system.
*/

};

#endif

