// **********************************************************************
//  Defintions of threads for PicCore
//  Part of : PicCore
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#ifndef MTHREADS_HXX
#define MTHREADS_HXX

#include <pthread.hxx>
#include <parray.hxx>
#include <pfile.hxx>
#include <btime.hxx>
#include <PicSearchAlgorithm.hxx>
#include <mquery.hxx>

/*:class ********************************************************************

PicEnhancedThread adds a clean way to suspend/abort a thread when 
it's running. This is achieved through a public command variable
that can be set by anyone. The actual thread implementation can check
occasionally the status of the variable with checkCommand function 
and exit/suspend action if so requested.

****************************************************************************/

class PicEnhancedThread : public PicThread
{
 public:

  PICVIRTCLASS(PicEnhancedThread)
  PicEnhancedThread(PicPriority priority = normal);

//:var
  enum PicThreadCmd
  {
     go,
     suspend,
     isSuspended,
     abort
  }command;
/*
Set the value of this variable to control thread's activity:
  go          -- free to run 
  suspend     -- should suspend all action
  isSuspended -- thread sets command from suspend ==> isSuspended 
  abort       -- should halt
*/

//:fun
  PICBOOL checkCommand();
/*
As long as command==suspend, calls PicThread::sleep(200).
If command==abort returns boolTrue, otherwise boolFalse.
*/

//:var
  BaseString name;
/*
Name for "the function the thread is for".
*/

};

/*:class ********************************************************************

Query thread implements the database query. The main thread provides
query image(s) and the search algorithm before launching this thread.

****************************************************************************/

class PicQueryThread : public PicEnhancedThread
{
 public:

  PICVIRTCLASS(PicQueryThread)
  PICERRORCODES {  ok,
    noExampleImages,
    noSearchAlg,
    noQueryAvailable
  };

  PicQueryThread(PicPriority priority = normal);

//:var
  PicSeqIxCollect    *exImages;
  PicSearchAlgorithm *searchAlg;
/*
These variables are set by main thread before launching the query.
ExImages contains the PicPicture:s used in the query (thread 
deletes this before exiting). SearchAlg points to the search 
algorithm to be used in the query.
*/

//:fun
  PICUSHORT  donePercentage();
/*
Returns the percentage of images scanned so far.
*/
//:fun
  BaseString queryTime();
/*
Returns as string the time used in executing the current query.
*/

//:var
  PicQueryArray currentQuery;
  PicQueryArray abortedQuery;
  PicArray      oldQueries;
/*
Array currentQuery holds the result set of currently running query.
Array abortedQuery holds the result set of the last aborted query.
Array oldQueries holds a number of finished query result sets 
(that is PicQueryArray:s). Currently only one is used.
*/

 protected:

  virtual PICTHREADRESULT threadFunction(PICTHREADARG arg);
  virtual PICBOOL         queryLoop(PicSignature *exSign,
                                    PicSignature *testSign);

  PicSemaphore  timeMutex;
  PicSemaphore  oldQueriesMutex;
  PicSemaphore  abortedQueryMutex;
};

/*:class ********************************************************************

Mass insert thread implements the batch insertion of images into 
database. It reads the pictures from files specified by a special 
list file or a simple file pattern (e.g. "*.jpg").
A list file should contain filenames or masks (one per line).
Each line can optionally have strings for name, par1 and par2 
attributes for the image(s). E.g.
    myface.jpg # Me # Apr 1997 # people
    /images/chr96/ *.jpg # Christmas # Dec 1997

****************************************************************************/

class PicMassInsertThread : public PicEnhancedThread
{
 public:

  PICVIRTCLASS(PicMassInsertThread)
  PICERRORCODES {  ok,
    noFileName,
    fileNotExist
  };

  PicMassInsertThread(PicPriority priority = normal);

//:var
  PicFile listFile;
/*
Main thread sets listFile before starting the thread. If it has 
suffix "lst" (case-insensitive) it is considered a list file
and will be opened to read the actual filenames of the images.
Otherwise it is taken to mean the actual file(s) to be inserted
into database.
*/

 protected:

  virtual PICTHREADRESULT threadFunction(PICTHREADARG arg);

  virtual PICBOOL doInits();
  virtual PICBOOL printStats();
  virtual PICBOOL extractParams(BaseString line, BaseString &fileName, 
                                PicPicture &pic);
  virtual PICBOOL insertPic(PicPicture &pic);

  PICULONG   insertedPics;
  PICULONG   failedPics;
  PICULONG   notFound;
  PicStream  *inputStream;
  BaseString tempBuf;  

};

/*:class ********************************************************************

Database update thread is launched at start-up by main thread when 
the system is detected to have more algorithms than are present
according to the database. This situation happens when a new 
algorithm has been programmed and the system starts up after that 
for the first time. It may also happen when the database has been 
restored from a back-up. 
NOTE: Using a new algorithm in queries before the database has been 
updated will not work properly.

****************************************************************************/

class PicDbUpdateThread : public PicEnhancedThread
{
 public:

  PICVIRTCLASS(PicDbUpdateThread)

  PicDbUpdateThread(PicPriority priority = normal);

//:var
  PicArray newSearchAlgorithms;
  PicArray newSegmAlgorithms;
/*
Main thread inserts into newSearchAlgorithms instances of those 
PicSearchAlgorithm:s, which are new and need to process the
images in the database. Likewise newSegmAlgorithms contains 
new PicSegmentAlgorithm:s.
*/

 protected:

  virtual PICTHREADRESULT threadFunction(PICTHREADARG arg);
  PICBOOL removeOldEntries();
  PICBOOL insertAlgNames();
  PICBOOL doSegments(PicPicture &pic);
  PICBOOL doSignatures(PicPicture &pic);

};

#endif

