// **********************************************************************
//  This file contains various generally useful functions
//  Part of : PicCore
// **********************************************************************
// Copyright (C) 1997 PicSearch Team
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//
// General Public License for more details. 
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// For any comments please contact: klemstro@cs.helsinki.fi
//
// **********************************************************************

#include <string.h>
#include <unistd.h>

#include <breal.hxx>
#include <palloc.hxx>
#include <pfilestm.hxx>
#include <picerror.hxx>
#include <pstring.hxx>

#include <mtools.hxx>
#include <mquery.hxx>
#include <mthreads.hxx>
#include <PicPicture.hxx>

// **********************************************************************

PicRawPicture* PicTools::scaleImage(const PicRawPicture &raw,
                                    PICUSHORT newWidth,
                                    PICUSHORT newHeight,
                                    PICBOOL   keepAspectRatio)
{
  PicErrorContext ecxt;
  // Here for now
  if(newWidth==0 || newHeight==0 || 
     raw.getWidth() == 0 || raw.getHeight() == 0 ||
     raw.getDataPtr() == NULLPTR)
    return NULLPTR;

  PICULONG  trueWidth = newWidth > 1024 ? 1024 : newWidth;
  PICULONG  trueHeight = newHeight > 768 ? 768 : newHeight;
  PicRawPicture *newPic = new PicRGBPicture();

  if(newPic){
    if(keepAspectRatio){
      BaseReal oldAR = (PICFLOAT)(raw.getWidth());
      oldAR /= raw.getHeight();
      BaseReal newAR = (PICFLOAT)(trueWidth);
      newAR /= trueHeight;
      if(oldAR < newAR){
        trueWidth = ((PICFLOAT)trueHeight * oldAR).integerPart();
      }
      if(oldAR > newAR){
        trueHeight = ((PICFLOAT)trueWidth / oldAR).integerPart();
      }
    }
    trueWidth = trueWidth < 1 ? 1 : trueWidth;
    trueHeight = trueHeight < 1 ? 1 : trueHeight;

    PICULONG size=trueWidth*trueHeight*3;
    unsigned char *source = raw.getDataPtr();
    unsigned char *dest = (unsigned char *)picAllocate(size+1);

    newPic->setDataPtr(dest);
    if(dest){
      newPic->setWidth(trueWidth);
      newPic->setHeight(trueHeight);

      PICULONG x, y=0;
      PICULONG ox, oy;
      BaseReal  cy = (PICFLOAT)(raw.getHeight());
      cy /= trueHeight;
      BaseReal  cx = (PICFLOAT)(raw.getWidth());
      cx /= trueWidth;

      for(y = 0; y < trueHeight; y++){
        oy = (cy * (PICFLOAT)y).integerPart();
        if(oy >= raw.getHeight())
          oy = raw.getHeight() - 1;
        for(x = 0; x < trueWidth; x++){
          ox = (cx * (PICFLOAT)x).integerPart();
          if(ox >= raw.getWidth())
            ox = raw.getWidth() - 1;
          memcpy(&dest[(x + y*trueWidth)*3], 
                 &source[(ox + oy*raw.getWidth())*3], 3);
        }
      }
    }else{
      delete newPic;
      newPic = NULLPTR;
    }
  }
  return newPic;
}

// **********************************************************************

PICBOOL PicTools::listToCollect(BaseString from,
                                PicIxCollect &to,
                                char delimiter)
{
  PicErrorContext ecxt;

  PicInStringStream readBuf(from);
  PicText           input(readBuf);
  BaseString        val, readVal;
  PICINDEX          ix = 1;
  PICBOOL           seeking = boolFalse;
  to.setNull();

  while(!input.endOfText())
  {
    input.rd(readVal, delimiter, boolFalse);
    val.append(readVal);
    if(val.first() == '{'){
      seeking = boolTrue;
      val.deleteBefore(2);
    }
    if(val.last() == '}'){
      seeking = boolFalse;
      val.deleteAfter(val.size()-1);
    }
    val.append(" ");
    if(!seeking || input.endOfText()){
      val.deleteAfter(val.size()-1);
      to.putAt(ix, PicString(val));
      ix++;
      val.setEmpty();
    }
  }
  return !PicError::isDetected();
}

// **********************************************************************

PICBOOL PicTools::collectToList(BaseString &to,
                                PicIxCollect &from,
                                char delimiter)
{
  PicErrorContext ecxt;
  PicIterator     iter;
  PICBOOL         firstRound = boolTrue;

  to = "";

  ITERATE(from, iter)
  {
    PicObject &curr = from.at(iter);
    BaseString toAdd = curr.getImage();

    if(toAdd.isEmpty() || toAdd.positionOf(' ')>0){
      toAdd.prepend("{");
      toAdd.append("}");
    }
    if(firstRound){
      firstRound = boolFalse;
    }else{
      toAdd.prepend(delimiter);
    }
    to.append(toAdd);
  }
  return !PicError::isDetected();
}

// **********************************************************************

#include <pthread.hxx>

PICBOOL PicTools::sendMasterPrint(BaseString mesg)
{
  PicThread *me = PicThread::self();
  if(!me || me->id()==1)
    return boolFalse;
  BaseString sendBuf("INFO:" & asString(mesg.size()) & ":" & mesg);

  return 
    write(me->fildes[1], sendBuf.chars(), sendBuf.size()) == (ssize_t)sendBuf.size();
}


PICBOOL PicTools::sendMasterQueryResults(const PicQueryArray &pics)
{
  PicThread *me = PicThread::self();
  if(!me || me->id()==1)
    return boolFalse;

  PicIterator iter;
  BaseString  mesg;

  ITERATE(pics, iter)
  {
    const PicQueryObj &qobj = AS(PicQueryObj, pics.at(iter));
    if(!qobj.isNull()){
      mesg.append(asString(qobj.imageId));
      mesg.append("#");
      mesg.append(asString(qobj.matchRate));
      mesg.append("#");
    }
  }
  BaseString sendBuf("QUER:" & asString(mesg.size()) & ":" & mesg);
  return write(me->fildes[1], sendBuf.chars(), sendBuf.size()) == (ssize_t)sendBuf.size();
}

PICBOOL PicTools::sendThreadCommand(PicEnhancedThread &t)
{
  BaseString sendBuf;

  switch(t.command)
  {
  case PicEnhancedThread::go :
    sendBuf = "TRUN:";
    break;
  case PicEnhancedThread::abort :
    sendBuf = "TEND:";
    break;
  case PicEnhancedThread::suspend :
    sendBuf = "TSUS:";
    break;
  }
  if(sendBuf.isEmpty())
    return boolFalse;
  return write(t.fildes2[1], sendBuf.chars(), sendBuf.size()) == (ssize_t)sendBuf.size();
}

#include <sys/time.h>
#include <picerror.hxx>

PICBOOL PicTools::listenToMaster(PicEnhancedThread &thread)
{
  #define RECVBUFSIZE 500

  PicErrorContext errCxt;
  static char recvBuf[RECVBUFSIZE+1];
  int recv;
  struct timeval tv;
  fd_set rset, wset, xset;

  FD_ZERO (&rset);
  FD_SET (thread.fildes2[0], &rset);
  FD_ZERO (&wset);
  FD_ZERO (&xset);
  tv.tv_sec = 0;
  tv.tv_usec = 0;

  if(select(thread.fildes2[0]+1, &rset, &wset, &xset, &tv) > 0){
    recv = read(thread.fildes2[0], recvBuf, RECVBUFSIZE);
    recvBuf[recv] = NULLPTR;
    BaseString mesg(recvBuf);

    while(mesg.size() > 4){
      BaseString command = mesg.between(1,4);
      mesg.deleteBefore(6);

      if(command == "TRUN"){
	thread.command = PicEnhancedThread::go;
      }
      if(command == "TEND"){
	thread.command = PicEnhancedThread::abort;
      }
      if(command == "TSUS"){
	thread.command = PicEnhancedThread::suspend;
      }
    }
  }

	return(boolTrue);
}

// **********************************************************************

MAKE_PICOBJECTCLASS(PicBin, PicObject)

PicBin::PicBin()
  : data(NULLPTR), dataSize(0)
{
}

PicBin::~PicBin()
{
  setNull();
}

PicBin::PicBin(const PicBin &other)
  : data(NULLPTR), dataSize(0)
{
  *this = (const PicObject &)other;
}

PicObject& PicBin::operator = (const PicObject& obj)
{
  if(obj.IS_A(PicBin)){
    const PicBin &other = AS(PicBin, obj);

    setNull();
    dataSize = other.dataSize;
    data = (char *)picAllocate(dataSize);
    if(!data){
      dataSize = 0;
    }else{
      memcpy(data, other.data, dataSize);
    }
  }
  return *this;
}

PicObject& PicBin::operator = (BaseString other)
{
  setNull();
  if(!other.isEmpty()){
    dataSize = other.size();
    data = (char *)picAllocate(dataSize);
    if(!data){
      dataSize = 0;
    }else{
      memcpy(data, other.chars(), dataSize);
    }
  }
  return *this;
}

PicObject& PicBin::set(void *from, PICULONG newSize)
{
  setNull();
  if(from && newSize>0){
    dataSize = newSize;
    data = (char *)picAllocate(dataSize);
    if(!data){
      dataSize = 0;
    }else{
      memcpy(data, from, dataSize);
    }
  }
  return *this;
}

PICBOOL PicBin::operator == (const PicObject& other) const
{
  if(this == &other)
    return boolTrue;
  if(other.IS_A(PicBin)){
    if(dataSize == ((const PicBin&)other).dataSize){
      return (memcmp(data, ((const PicBin&)other).data, dataSize)==0);
    }
  }
  return boolFalse;
}

PICBOOL PicBin::operator <  (const PicObject&) const
{
  undefined("PicBin::operator <");
  return boolFalse;
}

PICBOOL PicBin::isNull() const
{
  return (dataSize==0);
}

void PicBin::setNull()
{
  if(data){
    picFree(data);
  }
  data = NULLPTR;
  dataSize = 0;
}

BaseString PicBin::getImage() const
{
  return "Bin (size=" & asString(dataSize) & ")";
}

void PicBin::setImage(const BaseString& image)
{
  *this = image;
}


const void *PicBin::get() const
{
  return data;
}

PICULONG PicBin::size() const
{
  return dataSize;
}

PICBOOL PicBin::readFromFile(const PicFile &file)
{
  PicErrorContext errCxt;
  PICBOOL ret = boolFalse;
  
  setNull();
  if(!file.exists())
    return boolFalse;

  PICULONG      fSize = file.itemCount();
  PicFileStream stream(file, PicFile::forReading);
  if(!PicError::isDetected()){
    data = (char *)picAllocate(fSize);
    if(data){
      dataSize = fSize;
      if(stream.read(data, fSize) == fSize){
        ret = boolTrue;
      }else{
        setNull();
      }
    }
  }
  return ret;
}

PICBOOL PicBin::writeToFile(const PicFile &file) const
{
  PicErrorContext errCxt;
  PICBOOL ret = boolFalse;
  
  PicFileStream stream(file, PicFile::forWriting);
  if(!PicError::isDetected()){
    if(dataSize>0){
      if(stream.write(data, dataSize) == dataSize){
        ret = boolTrue;
      }
    }else{
      ret = boolTrue;
    }
  }
  return ret;
}

// **********************************************************************

PICBOOL PicTools::windowGrab(PicPicture &pic)
{
  return boolFalse;
}

// **********************************************************************

